/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeDebug.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the debugging function that are used for debugging the 
 * all the modules of the distributed pipes.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 28/September/98
 *
 ***********************************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>

#include "DistPipeDebug.h"


int g_nDebugLevel = 0;
static char     g_pszDebugMessageBuffer[10 * 1024];


void InitializeDistPipeDebug()
{
  char* l_pszEnvirnoment;
  char* l_pszTmpPtrtoEnv;
  char* l_pszToken;



  g_nDebugLevel = 0;
  
  l_pszEnvirnoment = getenv("DIST_PIPE_DEBUG");
  if(l_pszEnvirnoment)
    {
      l_pszTmpPtrtoEnv = (char *)strdup(l_pszEnvirnoment);
    }
  else
    return;
  fprintf(stdout,
	  "\n-------------------------------------------------------\n");
  fprintf(stdout,
	  "DIST_PIPE_DEBUG found. started with the following :-\n\n");
  l_pszToken = (char *)strtok(l_pszTmpPtrtoEnv,":");
  while(l_pszToken)
    {
      if (strcmp(l_pszToken,"socket") == 0)
	{
	  g_nDebugLevel |= SOCKET_DEBUG;
	  fprintf(stdout,"Debug level : socket\n");
	}
      if (strcmp(l_pszToken,"host") == 0)
	{
	  g_nDebugLevel |= HOST_DEBUG;
	  fprintf(stdout,"Debug level : host\n");
	}
      if (strcmp(l_pszToken,"lctosc") == 0)
	{
	  g_nDebugLevel |= LCTOSC_DEBUG;
	  fprintf(stdout,"Debug level : Messages LC -> SC\n");
	}
      if (strcmp(l_pszToken,"lctoup") == 0)
	{
	  g_nDebugLevel |= LCTOUSER_DEBUG;
	  fprintf(stdout,"Debug level : Messages LC -> UP\n");
	}
      if (strcmp(l_pszToken,"uptolc") == 0)
	{
	  g_nDebugLevel |= USERTOLC_DEBUG;
	  fprintf(stdout,"Debug level : Messages UP -> LC\n");
	}
      if (strcmp(l_pszToken,"sctolc") == 0)
	{
	  g_nDebugLevel |= SCTOLC_DEBUG;
	  fprintf(stdout,"Debug level : Messages SC -> LC\n");
	}
      if (strcmp(l_pszToken,"general") == 0)
	{
	  g_nDebugLevel |= GENERAL_DEBUG;
	  fprintf(stdout,"Debug level : General\n");
	}
      if (strcmp(l_pszToken,"mem") == 0)
	{
	  g_nDebugLevel |= MEM_DEBUG;
	  fprintf(stdout,"Debug level : Memory \n");
	}
      if (strcmp(l_pszToken,"xdr") == 0)
	{
	  g_nDebugLevel |= XDR_DEBUG;
	  fprintf(stdout,"Debug level : XDR Routines \n");
	}
      if (strcmp(l_pszToken,"steadystate") == 0)
	{
	  g_nDebugLevel |= STEADY_STATE_DEBUG;
	  fprintf(stdout,"Debug level : Steady State Problem \n");
	}
 
      l_pszToken = (char *)strtok(NULL,":");
    }
  (void)free(l_pszTmpPtrtoEnv);
  fprintf(stdout,
	  "\n-------------------------------------------------------\n\n");
}


void DistPipeDebug(int a_nDebugBit, ...)
{
  va_list l_stArguments;
  char* l_pszFormat;
  
  va_start(l_stArguments, a_nDebugBit);
  if (! (a_nDebugBit & g_nDebugLevel))
    {
      va_end(l_stArguments);
      return;
    }
  
  l_pszFormat = (char *)va_arg(l_stArguments, char *);
  if (!l_pszFormat)
    return;

  (void)vsprintf(g_pszDebugMessageBuffer,l_pszFormat, l_stArguments);
  va_end(l_stArguments);

  printf("%s", g_pszDebugMessageBuffer);
  fflush(stdout);
}


