/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeHost.c
 *
 * Purpose 
 * ~~~~~~~
 * This containts the function that are used to obtain the host information and a 
 * Link list structure used to store the host details.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 24/September/98
 *
 ***********************************************************************************************/


#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <sys/un.h>
#include <pwd.h>


#include <netinet/in.h>
#include <arpa/inet.h>

#include "DistPipeHost.h"
#include "DistPipeDebug.h"
#include "DistPipeDefines.h"
#include "DistPipeError.h"
#include "DistPipeNetLevel.h"
#include "DistPipeMesg.h"

extern int g_nDebugLevel;

char* g_pszProgramName;



/************************************************************************************************************
 * Description : 
 *               This Function is used to intialize the host details which may be a TCP or UNIX stream host
 * Arguments   :
 *               a_pstHostDetails - Pointer to the Host structure
 *               a_nWhichAddress - Intialize the TCP / UNIX Stream address structure
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int InitializeHost(a_pstHostDetails, a_nWhichAddress)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nWhichAddress;
{
  

  struct hostent* l_pstHostEntry;
  int nHostAddressIndex;
  struct in_addr* l_pstInAddress;

  uid_t l_uidtUserID;
  struct passwd* l_pstPasswordEntry;

  DISTPIPES_ZERO(a_pstHostDetails, sizeof(HOST_DESCRIPTOR));

  /* 
   * Allcoate for
   *  Host Name, Login Name, Path of the executable, 
   *  socket address 
   */
  
  DISTPIPES_CALLOC(a_pstHostDetails->m_pszHostName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_pszHostLogin, char, LOGINNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_pszDaemonPath, char, PATH_LENGTH);

  if (TCP_AND_UNIXSTREAM_HOST == a_nWhichAddress)
    {
      DISTPIPES_CALLOC(a_pstHostDetails->m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
      DISTPIPES_CALLOC(a_pstHostDetails->m_pstUNIXStreamHostAddress, struct sockaddr_un, sizeof(struct sockaddr_un));
    }
  if (UNIXSTREAM_HOST == a_nWhichAddress)
    {
      DISTPIPES_CALLOC(a_pstHostDetails->m_pstUNIXStreamHostAddress, struct sockaddr_un, sizeof(struct sockaddr_un));
    }
  if (TCP_HOST == a_nWhichAddress)
    {
      DISTPIPES_CALLOC(a_pstHostDetails->m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
    }
  
  gethostname(a_pstHostDetails->m_pszHostName, HOSTNAME_LENGTH);
   

  /* 
   * Must get the architecture and set it
   */
  a_pstHostDetails->m_nHostArchitecture = 1;


  a_pstHostDetails->m_nHostProcessID = getpid();


  l_uidtUserID = getuid();
  l_pstPasswordEntry = getpwuid(l_uidtUserID);

  if (NULL == l_pstPasswordEntry->pw_name)
    {
      strcpy(a_pstHostDetails->m_pszHostLogin, "      ");
    }
  else
    {
      strcpy(a_pstHostDetails->m_pszHostLogin, l_pstPasswordEntry->pw_name);
    }


  getcwd(a_pstHostDetails->m_pszDaemonPath, PATH_LENGTH);


  /* 
   * Set the memberes in the strtucture variable
   */
  
  l_pstHostEntry = gethostbyname(a_pstHostDetails->m_pszHostName);

  if (NULL == l_pstHostEntry)
    {
      DistPipeLogSystemError("gethostbyname");
      return HOST_GETHOSTBYNAME_ERROR;
    }
 
  /*
   * Must understand this better
   */
  if ((TCP_AND_UNIXSTREAM_HOST == a_nWhichAddress) || (TCP_HOST == a_nWhichAddress))
    {
      a_pstHostDetails->m_pstTCPHostAddress->sin_family = AF_INET;
      /*   
       *   l_pstInAddress = (struct in_addr*) l_pstHostEntry->m_h_addr_list[0];
       *   a_pstHostDetails->m_pstTCPHostAddress->sin_addr.s_addr = inet_addr(inet_ntoa(*l_pstInAddress));
       */
    }
  if ((TCP_AND_UNIXSTREAM_HOST == a_nWhichAddress) || (UNIXSTREAM_HOST == a_nWhichAddress))
    {
      a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_family = AF_UNIX;
    }
  

  /*
   * DEBUG Information only if the env is set
   */

  DistPipeDebug(HOST_DEBUG, "(InitializeNode)Host Name -> %s\n", a_pstHostDetails->m_pszHostName);
  if ((g_nDebugLevel | HOST_DEBUG) == HOST_DEBUG)
    {
      nHostAddressIndex = 0;
      while ((l_pstInAddress = (struct in_addr*) l_pstHostEntry->h_addr_list[nHostAddressIndex]) != NULL)
	{ 
	  DistPipeDebug(HOST_DEBUG, "(InitializeNode)Host Address[%d] -> %s \n", nHostAddressIndex, inet_ntoa(*l_pstInAddress));
	  nHostAddressIndex++;
	}
    }
  
  DistPipeDebug(HOST_DEBUG, "(InitializeNode)Host Architecture -> %d \n", a_pstHostDetails->m_nHostArchitecture);
  DistPipeDebug(HOST_DEBUG, "(InitializeNode)Host User Name -> %s \n", a_pstHostDetails->m_pszHostLogin);
  DistPipeDebug(HOST_DEBUG, "(InitializeNode)Host Process ID -> %d\n", a_pstHostDetails->m_nHostProcessID);
  
  return 1;
}


/***********************************************************************************************************
 * Description : 
 *               To delete the host strucutre details from the link list of host details
 * Arguments   :
 *               a_pstListofHosts - Address of pointer the start of the link list of host details
 *               a_pstHostDetails - Pointer to the Host structure
 *               a_nWhichAddress - Intialize the TCP / UNIX Stream address structure
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int DeleteHostFromList(a_pstListofHosts, a_pstHostNode, a_nWhichSocket)
     PHOST_DESCRIPTOR_COLLECTION* a_pstListofHosts;
     PHOST_DESCRIPTOR_COLLECTION a_pstHostNode;
     int a_nWhichSocket;
{

  /* First node is to be deleted */
  if (NULL == a_pstHostNode->m_pstPrevious)
    {
      /* make the head point to next of this */
      *a_pstListofHosts = a_pstHostNode->m_pstNext;
      /* Check if it is not the only node present */
      if (*a_pstListofHosts != NULL)
	{
	  /* make the head nodes previous pointer to null */
	  (*a_pstListofHosts)->m_pstPrevious = NULL;
	}
    }
  else
    /* Check if last node is to be deleted */
    if (NULL == a_pstHostNode->m_pstNext)
      {
	/* make the previous node next pointer to  null */
	a_pstHostNode->m_pstPrevious->m_pstNext = NULL;
      }
    else
      /* any node that can be in the center */
      {
	/* make the previous nodes next to next of this node */
	a_pstHostNode->m_pstPrevious->m_pstNext = a_pstHostNode->m_pstNext;
	/* make the next nodes previous to previous of this node */
	a_pstHostNode->m_pstNext->m_pstPrevious = a_pstHostNode->m_pstPrevious;
      }


  DeleteHost(&a_pstHostNode->m_stHost, a_nWhichSocket);

  DISTPIPES_FREE(a_pstHostNode);

  return 1;
}


/***********************************************************************************************************
 * Description : 
 *               Delete a host detail structure intialized using IntializeHost call
 * Arguments   :
 *               a_pstHostDetails - Pointer to the Host structure
 *               a_nWhichAddress - Intialize the TCP / UNIX Stream address structure
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int DeleteHost(a_pstHostNode, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostNode;
     int a_nWhichSocket;
{


   /* memory information */
  DistPipeDebug(MEM_DEBUG, "Deleting Host [%s] information \n", a_pstHostNode->m_pszHostName);

  CloseHost(a_pstHostNode, a_nWhichSocket);

  DISTPIPES_FREE(a_pstHostNode->m_pszHostName);
  if ((TCP_HOST == a_nWhichSocket) || (TCP_AND_UNIXSTREAM_HOST == a_nWhichSocket))
    DISTPIPES_FREE(a_pstHostNode->m_pstTCPHostAddress);
  if ((UNIXSTREAM_HOST == a_nWhichSocket) || (TCP_AND_UNIXSTREAM_HOST == a_nWhichSocket))
    DISTPIPES_FREE(a_pstHostNode->m_pstUNIXStreamHostAddress);
  DISTPIPES_FREE(a_pstHostNode->m_pszHostLogin);
  DISTPIPES_FREE(a_pstHostNode->m_pszDaemonPath);


  return 1;
}

/***********************************************************************************************************
 * Description : 
 *               Add a new host detail to the link list of host details.
 * Arguments   :
 *               a_pstHostDetails - Pointer to the Host structure
 *               a_pstListofHost - Address of the starting pointer to the link list
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
PHOST_DESCRIPTOR_COLLECTION AddNewHost(a_pstListofHosts)
     PHOST_DESCRIPTOR_COLLECTION* a_pstListofHosts;
{
  
  PHOST_DESCRIPTOR_COLLECTION l_pstHostListNode;
  /*
   * Allocate a New Host Detail
   */

  DistPipeDebug(MEM_DEBUG, "Adding a New Host Information ...\n");

  DISTPIPES_CALLOC(l_pstHostListNode, HOST_DESCRIPTOR_COLLECTION, 1);

  l_pstHostListNode->m_pstPrevious = NULL;
  l_pstHostListNode->m_pstNext = NULL;

  /*
   * Add the element in the start location
   */
  if (NULL == *a_pstListofHosts)
    {
      DistPipeDebug(MEM_DEBUG, "Starting Node of the Link List\n");
      *a_pstListofHosts = l_pstHostListNode;
    }
  else
    {
      l_pstHostListNode->m_pstNext = *a_pstListofHosts;
      (*a_pstListofHosts)->m_pstPrevious = l_pstHostListNode;
      *a_pstListofHosts = l_pstHostListNode;
    }
  
  
  DistPipeDebug(MEM_DEBUG, "Previous Pointer --> [%X]\n", l_pstHostListNode->m_pstPrevious);
  DistPipeDebug(MEM_DEBUG, "Next Pointer --> [%X]\n", l_pstHostListNode->m_pstNext);  

  return l_pstHostListNode;
}


/***********************************************************************************************************
 * Description : 
 *               Get the Next host detail in the link list of nodes
 * Arguments   :
 *               a_pstListNode - Pointer to the host detail node in link list
 * Return      :
 *               Pointer to the Host detail node in the link list
 ************************************************************************************************************/
PHOST_DESCRIPTOR_COLLECTION GetNextHost(a_pstHostListNode)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostListNode;
{
  
  if (NULL == a_pstHostListNode)
    return NULL;
  else
    return a_pstHostListNode->m_pstNext;

}

/***********************************************************************************************************
 * Description : 
 *               Get the Previous host detail in the link list of nodes
 * Arguments   :
 *               a_pstListNode - Pointer to the host detail node in link list
 * Return      :
 *               Pointer to the Host detail node in the link list

 ************************************************************************************************************/
PHOST_DESCRIPTOR_COLLECTION GetPreviousHost(a_pstHostListNode)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostListNode;
{

  if (NULL == a_pstHostListNode)
    return NULL;
  else
    return a_pstHostListNode->m_pstPrevious;

}
