/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeLocalCoordinator.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for Local Coordinator's functions.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 26/September/98
 *
 ***********************************************************************************************/

/* #ifdef LINUX_ARCHITECTURE */

#include <linux/kernel.h>  
#include <linux/sys.h>

/* #endif */

#include "DistPipeSystemCoordinator.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeError.h"
#include "DistPipeNetLevel.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeXDRRoutines.h"
#include "StProbMigration.h"



typedef struct tagWriteProcessBlocked
{

  PHOST_DESCRIPTOR_COLLECTION m_pstWriteProcessToBeBlocked;
  char* m_pszPipeName;

  struct tagWriteProcessBlocked* m_pstNext;

} WRITE_PROCESSES_BLOCKED, *PWRITE_PROCESSES_BLOCKED;


HOST_DESCRIPTOR g_stLocalCoordinator;
HOST_DESCRIPTOR g_stSystemCoordinator;
PHOST_DESCRIPTOR_COLLECTION g_pstListofUserProcesses = NULL;
PWRITE_PROCESSES_BLOCKED g_pstListofWriteProcessesBlocked = NULL;

int g_nNumberofUserProcesses = 0;

char* g_pszProgramName;


/* local prototypes */
int InitializeLocalCoordinator(char* a_pszSystemCoordinator);
int AcceptMessages();
void CleanUpLocalCoordinator(int a_nSignalNumber);
PHOST_DESCRIPTOR_COLLECTION WakeUpWriteEnd(PWRITE_PROCESSES_BLOCKED* a_pstWriteProcessBlocked, char* a_pszPipeName);
int BlockWriteEnd(PWRITE_PROCESSES_BLOCKED* a_pstWriteProcessBlocked, PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, char* a_pszPipeName);
int SystemLocalHi();
int SystemLocalWritePendingForPipe();
int SystemLocalFindLoad();
int ProcessSystemCoordinatorMessage();
int UserLocalHi(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalCreatePipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalExistPipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalOpenWriteMode(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, char* a_pszPipeName);
int UserLocalOpenReadMode(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, char* a_pszPipeName);
int UserLocalOpenPipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalGetLoadFactor(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalPurgePipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int UserLocalGoodBye(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int ProcessUserProcessesMessage(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);

/************************************************************************************************************
 * Description : 
 *               Main program 
 * Arguments   :
 *               a_nNumberofArguments - Number of command line arguments passed.
 *               a_pszArguments - command line arguments.
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int main(a_nNumberofArguments, a_pszArguments)
     int a_nNumberofArguments;
     char** a_pszArguments;
{

  InitializeDistPipeDebug();

  /*  Usage(a_nNumberofArguments);   */

  InitializeLocalCoordinator(a_pszArguments[1]);

  AcceptMessages();

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Find the Load of the Machine
 * Arguments   :
 *               ------
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
unsigned long int FindLoadOfMachine()
{
  unsigned long int l_ulLoad;
  
  /* #ifdef LINUX_ARCHITECTURE */
  
  struct sysinfo l_stLoadInformation;
  
  if (sysinfo(&l_stLoadInformation) < 0)
    {
      l_ulLoad = 0;
    }
  else
    {
      l_ulLoad = (20.0 / 100) * l_stLoadInformation.loads[0] +
	(30.0 / 100) * l_stLoadInformation.loads[1] +
	(50.0 / 100) * l_stLoadInformation.loads[2];
    }

  /* #endif */

  return l_ulLoad;
}

/************************************************************************************************************
 * Description : 
 *               Intialization of the local coordinator in which in contacts the system coordiantor and 
 *               sends it details.
 * Arguments   :
 *               a_pszSystemCoordinator - Name of the machine where system coordiantor is running
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int InitializeLocalCoordinator(a_pszSystemCoordinator)
     char* a_pszSystemCoordinator;
{

  int l_nErrorCode;
  int l_nMessageID;

 /* Trap the termination signal so that when the
   * kill command is iisued to the process it does
   * a gracefull exit from the code */
#ifdef SIGTERM
  signal(SIGTERM, CleanUpLocalCoordinator);
#endif

  /* check and delete the unix stream socket file */
  unlink(LOCAL_COORDINATOR_PORT);

  /*
   * Get the Details of the Host
   */
  InitializeHost(&g_stLocalCoordinator, TCP_AND_UNIXSTREAM_HOST);
 
  
  /*
   * Create a Socket and Connect to Big Boss (System coordinators)
   */
  
  if ((l_nErrorCode = CreateTCPClientHostAndConnect(&g_stLocalCoordinator, a_pszSystemCoordinator, SYSTEM_COORDINATOR_PORT)) < 0)
    {
      return l_nErrorCode;
    }

  /*
   * Create a UNIX STREAM socket and listen on it
   */

  if ((l_nErrorCode = CreateUNIXStreamServerHostAndListen(&g_stLocalCoordinator, LOCAL_COORDINATOR_PORT)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send a Hi Message to System Coordinator */
  l_nMessageID = LOCAL_SYSTEM_HI;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 



  /* Send the Local Coordinator Details to System Coordinator */

  /* Send Host Name */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_HI, &g_stLocalCoordinator, (void* ) g_stLocalCoordinator.m_pszHostName, 
				     HOSTNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

   /* Send Architecture */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_HI, &g_stLocalCoordinator, (void* ) &g_stLocalCoordinator.m_nHostArchitecture,
					    sizeof(g_stLocalCoordinator.m_nHostArchitecture), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  /* Send Address */
  if ((l_nErrorCode = WriteMessageAndData(LOCAL_SYSTEM_HI, (char*) g_stLocalCoordinator.m_pstTCPHostAddress, 
					  sizeof(struct sockaddr_in), &g_stLocalCoordinator, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send Process Id */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_HI, &g_stLocalCoordinator, (void* ) &g_stLocalCoordinator.m_nHostProcessID,
					  sizeof(g_stLocalCoordinator.m_nHostProcessID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Send Login Name*/
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_HI, &g_stLocalCoordinator, (void*) g_stLocalCoordinator.m_pszHostLogin, 
				     LOGINNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Send Daemon path*/
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_HI, &g_stLocalCoordinator, (void*) g_stLocalCoordinator.m_pszDaemonPath, 
				     PATH_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Go to accept message stage and scan for information from system coordiantor / user programs
 * Arguments   :
 *               ---
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int AcceptMessages()
{

  int l_nErrorCode;
  PHOST_DESCRIPTOR_COLLECTION l_pstNewUserProcess;
  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentUserProcess;
  int l_nNewSocketDescriptor;

  while (1)
    {
      /*
       * check for Big Boss information
       */
      if (AnyDataToReadOnHost(&g_stLocalCoordinator, 0, 100, TCP_HOST))
	{
	  /*
	   * Yes Boss !!! iam ready to get u'r
	   * information
	   */
	  if ((l_nErrorCode = ProcessSystemCoordinatorMessage()) < 0)
	    {
	      return l_nErrorCode;
	    }
	}
      
      /* 
       * get the audieance response and process them if any
       */
      if (AnyDataToReadOnHost(&g_stLocalCoordinator, 0, 100, UNIXSTREAM_HOST))
	{
	  /*
	   * Hey U are using Distributed Pipes
	   * welcome
	   */
	  /*
	   * A New Guy as joined the poll of machines
	   * Welcome buddy!!!!
	   */
	  if ((l_nNewSocketDescriptor = AcceptOnHost(&g_stLocalCoordinator, UNIXSTREAM_HOST)) < 0)
	    {
	      return l_nNewSocketDescriptor;
	    }
	  

	  /*
	   * Add the New fellow among other fellow mates
	   */
	  l_pstNewUserProcess = AddNewHost(&g_pstListofUserProcesses);	  
	  
	  /*
	   * I will speak to the new fellow in this socket
	   */
	  l_pstNewUserProcess->m_stHost.m_nUNIXStreamSocketDescriptor = l_nNewSocketDescriptor;
	  /*
	   * Process the message that is the initialization of user process
	   */
	   if ((l_nErrorCode = ProcessUserProcessesMessage(l_pstNewUserProcess)) < 0)
	     {
	       return l_nErrorCode;
	     }
	  
	  /*
	   * One more member in the family
	   */
	  
	  g_nNumberofUserProcesses++;
	  
	}

      /*
       * cheak all the user process socket if they have 
       * anything to say
       */
      l_pstCurrentUserProcess = g_pstListofUserProcesses;
      while (l_pstCurrentUserProcess != NULL)
	{
	  if (AnyDataToReadOnHost(&(l_pstCurrentUserProcess->m_stHost), 0, 100, UNIXSTREAM_HOST))
	    {
	      if ((l_nErrorCode = ProcessUserProcessesMessage(l_pstCurrentUserProcess)) < 0)
		{
		  return l_nErrorCode;
		}
	      /* This node may be deleted also so need to restart again */
	      break;
	    }
	  l_pstCurrentUserProcess = GetNextHost(l_pstCurrentUserProcess);
	}      
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               this deletes all the memory allocated during the execution of the local coordiantor
 * Arguments   :
 *               a_nSignalNumber - Number of signal because which this signal was called
 * Return      :
 *               None
 ************************************************************************************************************/
void CleanUpLocalCoordinator(a_nSignalNumber)
     int a_nSignalNumber;
{

  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentUserProcess;
  PHOST_DESCRIPTOR_COLLECTION l_pstNextUserProcess;

  DISTPIPES_FREE(g_stSystemCoordinator.m_pszHostName);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pstTCPHostAddress);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pszHostLogin);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pszDaemonPath);


  CloseHost(&g_stLocalCoordinator, TCP_AND_UNIXSTREAM_HOST);
  DISTPIPES_FREE(g_stLocalCoordinator.m_pszHostName);
  DISTPIPES_FREE(g_stLocalCoordinator.m_pstTCPHostAddress);
  DISTPIPES_FREE(g_stLocalCoordinator.m_pszHostLogin);
  DISTPIPES_FREE(g_stLocalCoordinator.m_pszDaemonPath);


  /*
   * Delete all the user process information
   * ifnormation that are existing now
   */
  
  l_pstCurrentUserProcess = g_pstListofUserProcesses;
  while (l_pstCurrentUserProcess != NULL)
    {
      l_pstNextUserProcess = GetNextHost(l_pstCurrentUserProcess);
      DeleteHost(&l_pstCurrentUserProcess->m_stHost, UNIXSTREAM_HOST);
      DISTPIPES_FREE(l_pstCurrentUserProcess);
      l_pstCurrentUserProcess = l_pstNextUserProcess;
    }

  exit(1);

}

/************************************************************************************************************
 * Description : 
 *               Wake up the write end process which was waiting for the read end to be opened by another process
 * Arguments   :
 *               a_pstWriteProcessBlocked - pointer to the start of link list structure having writer blocked processes
 *               a_pszPipeName - Name of the pipe associated with the writer process
 * Return      :
 *               Pointer to the host strucutre of user process in the link list maintianed by local coordinator 
 ************************************************************************************************************/
PHOST_DESCRIPTOR_COLLECTION WakeUpWriteEnd(a_pstWriteProcessBlocked, a_pszPipeName)
     PWRITE_PROCESSES_BLOCKED* a_pstWriteProcessBlocked;
     char* a_pszPipeName;
{
  
  PWRITE_PROCESSES_BLOCKED l_pstCurrent;
  PWRITE_PROCESSES_BLOCKED l_pstPrevious;
  PHOST_DESCRIPTOR_COLLECTION l_pstFoundHost;



  l_pstCurrent = *a_pstWriteProcessBlocked;
  l_pstPrevious = NULL;
  while ((NULL != l_pstCurrent) && (strcmp(a_pszPipeName, l_pstCurrent->m_pszPipeName)))
    {
      l_pstPrevious = l_pstCurrent;
      l_pstCurrent = l_pstCurrent->m_pstNext;
    }
  
  if (NULL == l_pstCurrent)
    {
      /* Entry not found */
      return NULL;
    }
  else
   if ((NULL != l_pstCurrent) && (NULL == l_pstPrevious))
     {
       /* This the head of the list */
       *a_pstWriteProcessBlocked = l_pstCurrent->m_pstNext;
     }
   else
     if ((NULL != l_pstCurrent) && (NULL != l_pstPrevious))
       {
	 /* not the head of the list */
	 l_pstPrevious->m_pstNext = l_pstCurrent->m_pstNext;
       }

  l_pstFoundHost = l_pstCurrent->m_pstWriteProcessToBeBlocked;

  DISTPIPES_FREE(l_pstCurrent->m_pszPipeName);
  DISTPIPES_FREE(l_pstCurrent);


  DistPipeDebug(GENERAL_DEBUG, "Waking Up Write Process of Host [%s] for Pipe [%s] \n",
		l_pstFoundHost->m_stHost.m_pszHostName, a_pszPipeName);

  return l_pstFoundHost;
}

/************************************************************************************************************
 * Description : 
 *               Since the read end of the pipe is not opened block the writer process 
 * Arguments   :
 *               a_pstWriteProcessBlocked - pointer to the start of link list structure having writer blocked processes
 *               a_pstHostDetails - host that is to be blocked for opening pipe for write 
 *               a_pszPipeName - Name of the pipe associated with the writer process
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int BlockWriteEnd(a_pstWriteProcessBlocked, a_pstHostDetails, a_pszPipeName)
     PWRITE_PROCESSES_BLOCKED* a_pstWriteProcessBlocked;
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     char* a_pszPipeName;
{
  
  PWRITE_PROCESSES_BLOCKED l_pstNewWriteProcess;
  
  DistPipeDebug(GENERAL_DEBUG, "Blocking the User Process of Host [%s] for WRITE to Pipe Name [%s] \n",
		a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName);

  DISTPIPES_CALLOC(l_pstNewWriteProcess, WRITE_PROCESSES_BLOCKED, 1);

  DISTPIPES_CALLOC(l_pstNewWriteProcess->m_pszPipeName, char, PIPENAME_LENGTH);
  
  /* Set the Structure fields */
  strcpy(l_pstNewWriteProcess->m_pszPipeName, a_pszPipeName);
  l_pstNewWriteProcess->m_pstWriteProcessToBeBlocked = a_pstHostDetails;
  l_pstNewWriteProcess->m_pstNext = NULL;

  if (NULL == *a_pstWriteProcessBlocked)
    {
      *a_pstWriteProcessBlocked = l_pstNewWriteProcess;
    }
  else
    {
      l_pstNewWriteProcess->m_pstNext = *a_pstWriteProcessBlocked;
      *a_pstWriteProcessBlocked = l_pstNewWriteProcess;
    }
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               System Coordinator to Local Coordiantor Hi Message Routine
 * Arguments   :
 *               ----
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int SystemLocalHi()
{

  int l_nErrorCode;
  
  DISTPIPES_CALLOC(g_stSystemCoordinator.m_pszHostName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(g_stSystemCoordinator.m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
  DISTPIPES_CALLOC(g_stSystemCoordinator.m_pszHostLogin, char, LOGINNAME_LENGTH);
  DISTPIPES_CALLOC(g_stSystemCoordinator.m_pszDaemonPath, char, PATH_LENGTH);
  
  /* Read Host name */
  if ((l_nErrorCode = ReadXDRString(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) g_stSystemCoordinator.m_pszHostName, 
				    HOSTNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_HI) From %s \n", g_stSystemCoordinator.m_pszHostName);
  
  /* Read Architecture */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) &g_stSystemCoordinator.m_nHostArchitecture,
					   sizeof(g_stSystemCoordinator.m_nHostArchitecture), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_HI) [%s] Architecture [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, g_stSystemCoordinator.m_nHostArchitecture);
  
  /* Read Address */
  if ((l_nErrorCode = ReadMessageAndData(SYSTEM_LOCAL_HI, (char*) g_stSystemCoordinator.m_pstTCPHostAddress, 
					 &g_stLocalCoordinator, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Read ProcessID */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) &g_stSystemCoordinator.m_nHostProcessID,
					   sizeof(g_stSystemCoordinator.m_nHostProcessID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_HI) [%s] ProcessID [%d] \n", g_stSystemCoordinator.m_pszHostName, 
		g_stSystemCoordinator.m_nHostProcessID);
  
  /* Read Login Name */
  if ((l_nErrorCode = ReadXDRString(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) g_stSystemCoordinator.m_pszHostLogin, 
				    LOGINNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_HI) [%s] Login Name [%s] \n", g_stSystemCoordinator.m_pszHostName, 
		g_stSystemCoordinator.m_pszHostLogin);
  
  /* Read Daemon Path */
  if ((l_nErrorCode = ReadXDRString(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) g_stSystemCoordinator.m_pszDaemonPath, 
					 PATH_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_HI) [%s] Daemon Path [%s] \n", g_stSystemCoordinator.m_pszHostName, 
		g_stSystemCoordinator.m_pszDaemonPath); 

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               System Coordinator to Local Coordiantor Write Pending  Message Routine
 * Arguments   :
 *               ---------
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int SystemLocalWritePendingForPipe()
{

  int l_nErrorCode;
  int l_nWritePendingStatus;

  char* l_pszPipeName;


  PHOST_DESCRIPTOR_COLLECTION l_pstWriteHostDetails;

  /* Get the status if any of the Process is blocked for this pipe to
   * be opened in the read mode from the system coordinator */
  /* Read the Write Pending Status from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_WRITE_PENDING_PIPE, &g_stLocalCoordinator, (void*) &l_nWritePendingStatus, 
					   sizeof(l_nWritePendingStatus), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_WRITE_PENDING_PIPE) Write Pending Status [%d]\n", 
		l_nWritePendingStatus);
  

  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);

  /* Read the Pipe Name for which the write is pending */
  if ((l_nErrorCode = ReadXDRString(SYSTEM_LOCAL_HI, &g_stLocalCoordinator, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_WRITE_PENDING_PIPE) Write Pending PipeName [%s] \n", 
		l_pszPipeName);
  
  if (OPEN_PIPE_WRITE_PENDING == l_nWritePendingStatus)
    {
      /* Write is pending for this pipe name therefore find 
       * the process that is waiting for the pipe to be opened
       * in write mode and send a message */
      l_pstWriteHostDetails = WakeUpWriteEnd(&g_pstListofWriteProcessesBlocked, l_pszPipeName);
      
      if (NULL == l_pstWriteHostDetails)
	{
	  DISTPIPES_FREE(l_pszPipeName);
	  return -1;
	}

      /* Here Normal process for exchange for open write pipe
       * information protocol must be written */
      return UserLocalOpenWriteMode(l_pstWriteHostDetails, l_pszPipeName);
    }
  else
    {
      /* An Wrong message is obtained must add appropriate
       * error message */
      DISTPIPES_FREE(l_pszPipeName);
    }
  
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               System Coordinator to Local Coordiantor Write Pending  Message Routine
 * Arguments   :
 *               ---------
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int SystemLocalFindLoad()
{

  int l_nErrorCode;
  unsigned long int l_ulLoad;

  l_ulLoad = FindLoadOfMachine();

  /* Send the Load to the system coordinator */
  if ((l_nErrorCode = WriteXDRUnSignedLongInteger(LOCAL_SYSTEM_FIND_LOAD, &g_stLocalCoordinator, 
						  (void* ) &l_ulLoad,
						  sizeof(&l_ulLoad), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Processing of the system coordinator messages
 * Arguments   :
 *               ---
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ProcessSystemCoordinatorMessage()
{

  int l_nErrorCode;
  int l_nMessageID;

  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_ANY_MESSAGE, &g_stLocalCoordinator, (void* ) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  switch (l_nMessageID)
    {

    case SYSTEM_LOCAL_HI :      
      DistPipeDebug(SCTOLC_DEBUG, "Recieved [SYSTEM_LOCAL_HI] \n");
      return SystemLocalHi();
      break;

    case SYSTEM_LOCAL_CREATE_PIPE :
      break;

    case SYSTEM_LOCAL_OPEN_PIPE :
      break;

    case SYSTEM_LOCAL_WRITE_PENDING_PIPE :
      DistPipeDebug(SCTOLC_DEBUG, "Recieved [SYSTEM_LOCAL_WRITE_PENDING_PIPE] \n");
      return SystemLocalWritePendingForPipe();
      break;


    case SYSTEM_LOCAL_FIND_LOAD :
      DistPipeDebug(SCTOLC_DEBUG, "Recieved [SYSTEM_LOCAL_FIND_LOAD] \n");
      return SystemLocalFindLoad();
      break;


    case SYSTEM_LOCAL_GOODBYE :
      break;
    }
  
  return 1;

}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Hi Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 *               a_pstReceivedMessage - Message that is Received from the system coordiantor 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalHi(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;

  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszHostName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszHostLogin, char, LOGINNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszDaemonPath, char, PATH_LENGTH);
  
  /* Read Host name */
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_HI, &a_pstHostDetails->m_stHost, 
				    (void*) a_pstHostDetails->m_stHost.m_pszHostName, 
				    HOSTNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_HI) From %s \n", a_pstHostDetails->m_stHost.m_pszHostName);
  
  /* Read Architecture */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_HI, &a_pstHostDetails->m_stHost, 
					 (void*) &a_pstHostDetails->m_stHost.m_nHostArchitecture, 
					 sizeof(a_pstHostDetails->m_stHost.m_nHostArchitecture), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_HI) [%s] Architecture [%d] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_nHostArchitecture);
  
  /* Read ProcessID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_HI, &a_pstHostDetails->m_stHost, 
					 (void*) &a_pstHostDetails->m_stHost.m_nHostProcessID, 
					 sizeof(a_pstHostDetails->m_stHost.m_nHostProcessID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_HI) [%s] ProcessID [%d] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_nHostProcessID);
  
  /* Read Login Name */
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_HI, &a_pstHostDetails->m_stHost, 
				    (void*) a_pstHostDetails->m_stHost.m_pszHostLogin, 
				    LOGINNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_HI) [%s] Login Name [%s] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_pszHostLogin);
  
  /* Read Daemon Path */
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_HI,  &a_pstHostDetails->m_stHost, 
				    (void*) a_pstHostDetails->m_stHost.m_pszDaemonPath, 
				    PATH_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_HI) [%s] Daemon Path [%s] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_pszDaemonPath);       
  
  
  return 1;
  
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Create pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalCreatePipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nPipeCreated;
  int l_nMessageID;
  
  char* l_pszPipeName;
      
  /* Read the Pipe Name */
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_ANY_MESSAGE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_CREATE_PIPE) From [%s] PipeName [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);

  /* Send the Message to System coordinator create pipe  */
  l_nMessageID = LOCAL_SYSTEM_CREATE_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, (void*) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Send the Pipe Name to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_CREATE_PIPE, &g_stLocalCoordinator, (void*) l_pszPipeName, 
					  PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }

  
  /* Read the status if the pipe is a valid pipe name */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_CREATE_PIPE, &g_stLocalCoordinator, (void*) &l_nPipeCreated, 
					   sizeof(l_nPipeCreated), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_CREATE_PIPE) PipeName [%s] Pipe Created [%d]\n", 
		l_pszPipeName, l_nPipeCreated);

  /* Send the Status to the user program */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_CREATE_PIPE, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nPipeCreated, 
					    sizeof(l_nPipeCreated), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DISTPIPES_FREE(l_pszPipeName);


  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Pipe Exist Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalExistPipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nPipeExist;
  int l_nMessageID;
  char* l_pszPipeName;

  /* Read the Pipe Name */
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_EXIST_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				 PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_EXIST_PIPE) From [%s] PipeName [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);


  /* Send the Message to System coordinator if the pipe exist */
  l_nMessageID = LOCAL_SYSTEM_EXIST_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, (void*) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Send the Pipe Name to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_EXIST_PIPE, &g_stLocalCoordinator, (void*) l_pszPipeName, 
				     PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Read from the system coodinator the pipe name exist */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_EXIST_PIPE, &g_stLocalCoordinator, (void*) &l_nPipeExist, 
					   sizeof(l_nPipeExist), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }

  DistPipeDebug(USERTOLC_DEBUG, "(SYSTEM_LOCAL_EXIST_PIPE) From [%s] Exist PipeName [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nPipeExist);

  /* Write on to the user process the status returned by the system coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_EXIST_PIPE,  &a_pstHostDetails->m_stHost, 
					    (void*) &l_nPipeExist, 
					    sizeof(l_nPipeExist),UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DISTPIPES_FREE(l_pszPipeName);

  return 1;

}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Open Write Mode Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 *               a_pszPipeName - Name of the pipe to be opened in write mode 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalOpenWriteMode(a_pstHostDetails, a_pszPipeName)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     char* a_pszPipeName;
{

  int l_nErrorCode;
  int l_nUserProcessID;
  int l_nPortNumber;
  int l_nPipeOpen;
  int l_nWriteStatus;
  int l_nSocketDescriptor;
  int l_nConnectStatus;

  char* l_pszMachineName;


  /* Read from the system coodinator the write status */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &g_stLocalCoordinator, 
					   (void*) &l_nWriteStatus, 
					   sizeof(l_nWriteStatus), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_OPEN_PIPE) From System Coordinator Pipe [%s] Write Status [%d]\n", 
		a_pszPipeName, l_nWriteStatus);
  
  if (OPEN_PIPE_WRITE_PENDING == l_nWriteStatus)
    {
      /* Write mode is not yet ready with the data from sc */
      
      /* Don't Send any message to this process b'c the status is not
       * obtained from the read end of the pipe */
      BlockWriteEnd(&g_pstListofWriteProcessesBlocked, a_pstHostDetails, a_pszPipeName);
    }
  else
    if (OPEN_PIPE_WRITE_NOT_PENDING != l_nWriteStatus)
      {
	/* And status that is not defined there must add 
	 * appropiate error message */
      }
    else
      {
	/* Open pipe for write is ok */
	
	DISTPIPES_CALLOC(l_pszMachineName, char, HOSTNAME_LENGTH);
	
	/* Read from the system coodinator the write status */
	if ((l_nErrorCode = ReadXDRString(SYSTEM_LOCAL_OPEN_PIPE, &g_stLocalCoordinator, (void*) l_pszMachineName, 
					  HOSTNAME_LENGTH, TCP_HOST)) < 0)
	  {
	    DISTPIPES_FREE(l_pszMachineName);
	    DISTPIPES_FREE(a_pszPipeName);
	    return l_nErrorCode;
	  }
	DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_OPEN_PIPE) From System Coordinator Pipe [%s] Read m/c name [%s]\n", 
		      a_pszPipeName, l_pszMachineName);
	
	/* Read from the system coodinator the read port number */
	if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nPortNumber, 
						 sizeof(l_nPortNumber), TCP_HOST)) < 0)
	  {
	    DISTPIPES_FREE(l_pszMachineName);
	    DISTPIPES_FREE(a_pszPipeName);
	    return l_nErrorCode;
	  }
	DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_OPEN_PIPE) From System Coordinator Pipe [%s] Port number [%d]\n", 
		      a_pszPipeName, l_nPortNumber);
	
	
	/* Write the details from sc to lc to user program */
	/* Write on to the user process read machine name returned by the system coordinator */
	if ((l_nErrorCode = WriteXDRString(LOCAL_USER_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
					   (void*) l_pszMachineName, 
					   HOSTNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
	  {
	    DISTPIPES_FREE(l_pszMachineName);
	    DISTPIPES_FREE(a_pszPipeName);
	    return l_nErrorCode;
	  }
	
	if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						  (void*) &l_nPortNumber, 
						  sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0)
	  {
	    DISTPIPES_FREE(l_pszMachineName);
	    DISTPIPES_FREE(a_pszPipeName);
	    return l_nErrorCode;
	  }
	DISTPIPES_FREE(l_pszMachineName);
	
	/* Get the Details if Connection Succeded or not succeded from user program */
	if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						 (void*) &l_nConnectStatus, 
						 sizeof(l_nConnectStatus), UNIXSTREAM_HOST)) < 0)
	  {
	    DISTPIPES_FREE(a_pszPipeName);
	    return l_nErrorCode;
	  }	  
	DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Connect Status [%d]\n", 
		      a_pstHostDetails->m_stHost.m_pszHostName, l_nConnectStatus);
	
	
	if (CONNECT_READ_WRITE_PIPE_FAILED == l_nConnectStatus)
	  {
	    /* connect has failed */
	    /* inform the system coordinator that the connection has failed */
	    l_nConnectStatus = CONNECT_READ_WRITE_PIPE_FAILED;
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nConnectStatus, 
						    sizeof(l_nConnectStatus), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	  }
	else
	  { 
	    /* Connect has succeded */
	    /* inform the system coordinator that the connection has succeded */
	    l_nConnectStatus = CONNECT_READ_WRITE_PIPE_SUCCEDED;
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nConnectStatus, 
						    sizeof(l_nConnectStatus), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	    
	    /* Read Process ID */
	    if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						     (void*) &l_nUserProcessID, 
						     sizeof(l_nUserProcessID), UNIXSTREAM_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }	  
	    DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Process Id [%d]\n", 
			  a_pstHostDetails->m_stHost.m_pszHostName, l_nUserProcessID);
	    
	    /* Read the WRITE Port Number */
	    if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						     (void*) &l_nPortNumber, 
						     sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }	  
	    DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Port Number [%d]\n", 
			  a_pstHostDetails->m_stHost.m_pszHostName, l_nPortNumber);
	    
	    
	    /* Read the WRITE Socket Descriptor */
	    if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						     (void*) &l_nSocketDescriptor, 
						     sizeof(l_nSocketDescriptor), UNIXSTREAM_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }	  
	    DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Socket Descriptor [%d]\n", 
			  a_pstHostDetails->m_stHost.m_pszHostName, l_nSocketDescriptor);
	    
	    
	    /* Send the User Process ID to the System Coodrinator */
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, 
						      (void*) &l_nUserProcessID, 
						      sizeof(l_nUserProcessID), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	    
	    /* Send the User Process port number to the System Coodrinator */
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nPortNumber, 
						    sizeof(l_nPortNumber), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	    
	    
	    /* Send the User Process Write socket Descriptor to the System Coodrinator */
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nSocketDescriptor, 
						    sizeof(l_nSocketDescriptor), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	    
	    /* Wait for the status of updation */
	    /* Read from the system coodinator the pipe name exist */
	    if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nPipeOpen, 
						   sizeof(l_nPipeOpen), TCP_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	    DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_OPEN_PIPE) From System Coordinator Pipe [%s] Open Status [%d]\n", 
			  a_pszPipeName, l_nPipeOpen);
	    
	    /* Write on to the user process the status returned by the system coordinator */
	    if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
						      (void*) &l_nPipeOpen, 
						      sizeof(l_nPipeOpen), UNIXSTREAM_HOST)) < 0)
	      {
		DISTPIPES_FREE(a_pszPipeName);
		return l_nErrorCode;
	      }
	  }
      }

  DISTPIPES_FREE(a_pszPipeName);	    
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Open Read Mode Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 *               a_pszPipeName - Name of the pipe to be opened in write mode 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalOpenReadMode(a_pstHostDetails, a_pszPipeName)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     char* a_pszPipeName;
{

  int l_nErrorCode;
  int l_nUserProcessID;
  int l_nPortNumber;
  int l_nPipeOpen;
  int l_nSocketDescriptor;

  /* Read Process ID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nUserProcessID, 
					   sizeof(l_nUserProcessID), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }	  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Process Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nUserProcessID);
  
  /* Read the READ Port Number */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost,  
					   (void*) &l_nPortNumber, 
					   sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }	  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Port Number [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nPortNumber);
  
  
  /* Read the READ Socket Descriptor */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE,  &a_pstHostDetails->m_stHost, 
					   (void*) &l_nSocketDescriptor, 
					   sizeof(l_nSocketDescriptor), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }	  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] Socket Descriptor [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nSocketDescriptor);
  


   /* Send the User Process ID to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nUserProcessID, 
					    sizeof(l_nUserProcessID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Send the User Process port number to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nPortNumber, 
					    sizeof(l_nPortNumber), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  
  
  /* Send the User Process Write socket Descriptor to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nSocketDescriptor, 
					    sizeof(l_nSocketDescriptor), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Wait for the status of updation */
  /* Read from the system coodinator the pipe name exist */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nPipeOpen, 
					   sizeof(l_nPipeOpen), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_OPEN_PIPE) From System Coordinator Pipe [%s] Open Status [%d]\n", 
		a_pszPipeName, l_nPipeOpen);
  
  

  /* Write on to the user process the status returned by the system coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nPipeOpen, 
					    sizeof(l_nPipeOpen), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Open Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalOpenPipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nAccessMode;
  int l_nMessageID;

  char* l_pszPipeName;

  /* Read the Pipe Name */
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				 PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] PipeName [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  
  /* Read the Access mode in which the pipe is opened */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nAccessMode, 
					   sizeof(l_nAccessMode), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_OPEN_PIPE) From [%s] PipeName [%s] Access [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName, l_nAccessMode);

  
  /* Send the Message to System coordinator eopn pipe  */
  l_nMessageID = LOCAL_SYSTEM_OPEN_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, (void*) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  
  /* Send the Pipe Name to the System Coodrinator */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) l_pszPipeName, 
				     PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Send the Access Mode to the system coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &g_stLocalCoordinator, (void*) &l_nAccessMode, 
					  sizeof(l_nAccessMode), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  
  if (READ_ACCESS_MODE == l_nAccessMode)
    {
      /* Read Mode */
      return UserLocalOpenReadMode(a_pstHostDetails, l_pszPipeName);
    }
  else
    if (WRITE_ACCESS_MODE != l_nAccessMode)
      {
	/* Invalid Access Mode obtained from user program */
	
      }
    else
      {
	/* Write Mode */
	return UserLocalOpenWriteMode(a_pstHostDetails, l_pszPipeName);
      }  
  return 1;

}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Get Load Factor Message processing
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalGetLoadFactor(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nNumberofMachines;
  int l_nNumberofAvailableMachines;
  
  /* Obtain the Number of Machine required by the user program */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_GET_LOAD_FACTOR, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberofMachines, 
					   sizeof(l_nNumberofMachines), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(USER_LOCAL_GET_LOAD_FACTOR) Number of Machines Required [%d]\n", 
		l_nNumberofMachines);

  /* Send Number of Machines required to system coordinator */
  l_nMessageID = LOCAL_SYSTEM_GET_LOAD_FACTOR;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void*) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_GET_LOAD_FACTOR, &g_stLocalCoordinator, 
					    (void*) &l_nNumberofMachines, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* System Coordinator Finds the Number of Machines that are available now */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_GET_LOAD_FACTOR, &g_stLocalCoordinator, 
					   (void*) &l_nNumberofAvailableMachines, 
					   sizeof(l_nNumberofAvailableMachines), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_GET_LOAD_FACTOR) Number of Available Machines [%d]\n", 
		l_nNumberofAvailableMachines);

  /* Send the information to the user program */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_GET_LOAD_FACTOR, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nNumberofAvailableMachines              , 
					    sizeof(l_nNumberofAvailableMachines), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Purge Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalPurgePipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nPurgeReturn;
  int l_nMessageID;

  char* l_pszPipeName;
      
  /* Read the Pipe Name */
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_PURGE_PIPE, &a_pstHostDetails->m_stHost, 
				    (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_PURGE_PIPE) From [%s] PipeName [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  

  /* Send the Message to System coordinator create pipe  */
  l_nMessageID = LOCAL_SYSTEM_PURGE_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, (void*) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Send the Pipe Name to the System Coordinator */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_PURGE_PIPE, &g_stLocalCoordinator, (void*) l_pszPipeName, 
				     PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  /* Read status of deletion from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_PURGE_PIPE, &g_stLocalCoordinator, (void*) &l_nPurgeReturn, 
					   sizeof(l_nPurgeReturn), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_PURGE_PIPE) From System Coordinator Pipe [%s] Purge Status [%d]\n", 
		l_pszPipeName, l_nPurgeReturn);
  
  /* Write the status to the user process */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_PURGE_PIPE, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nPurgeReturn, 
					    sizeof(l_nPurgeReturn), UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DISTPIPES_FREE(l_pszPipeName);

  return 1;

}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor GoodBye Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalGoodBye(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nDummyMessage;

  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_GOODBYE, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nDummyMessage, 
					   sizeof(l_nDummyMessage), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  DeleteHostFromList(&g_pstListofUserProcesses, a_pstHostDetails, UNIXSTREAM_HOST);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Steady State problem initialization
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalSteadyStateInitialize(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nWorkID;

  /* Send the Intialization Message to System coordinator */
  l_nMessageID = LOCAL_SYSTEM_STEADY_STATE_INITIALIZE;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  /* Read the Work Identification */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_INITIALIZE, &g_stLocalCoordinator, 
					   (void*) &l_nWorkID,
					   sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_INITIALIZE) [%s] Work Identifier [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nWorkID);


  /* Send the Work ID to the user Process */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_INITIALIZE, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nWorkID, 
					    sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Steady State problem Send the Matrix information to system coordinator.
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalSteadyStateMatrixInformation(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nWorkID;
  int l_nStatus;
  int l_nNumberOfRows;
  int l_nNumberOfColumns;
  
  /* Send the message to system coordinator */
  l_nMessageID = LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 


  /* Read the Work ID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION) [%s] Work Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

  /* Read Number of Rows */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberOfRows, 
					   sizeof(l_nNumberOfRows), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION) [%s] No. of Rows [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nNumberOfRows);

  /* Read the Number od Columns */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberOfColumns, 
					   sizeof(l_nNumberOfColumns), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION) [%s] No. of Columns[%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nNumberOfColumns);



  /* Send the Work ID to System coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &g_stLocalCoordinator, 
					    (void* ) &l_nWorkID, 
					    sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  /* Send the Number of Rows System Coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &g_stLocalCoordinator, 
					    (void* ) &l_nNumberOfRows, 
					    sizeof(l_nNumberOfRows), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  /* Send the Number of Columns System Coordinator*/
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &g_stLocalCoordinator, 
					    (void* ) &l_nNumberOfColumns, 
					    sizeof(l_nNumberOfColumns), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Read the Updation status from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &g_stLocalCoordinator, 
					   (void*) &l_nStatus,
					   sizeof(l_nStatus), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_MATRIX_INFORMATIONRMATION) [%s] Status [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nStatus);

  /* Send the status to the user program */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nStatus, 
					    sizeof(l_nStatus), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Steady State problem Find Number of Splits
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalSteadyStateNumberOfSplits(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nWorkID;
  int l_nNumberOfSplits;
  unsigned long int l_ulLoad;

  /* Send the message to system coordinator */
  l_nMessageID = LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 


  /* Read the Work ID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS) [%s] Work Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

  /* Send the Work ID to System coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS, &g_stLocalCoordinator, 
					    (void* ) &l_nWorkID, 
					    sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 
  

  /* Send the Load of this machine also to system adminstrator */
  l_ulLoad = FindLoadOfMachine();

  /* Send the Load to the system coordinator */
  if ((l_nErrorCode = WriteXDRUnSignedLongInteger(LOCAL_SYSTEM_FIND_LOAD, &g_stLocalCoordinator, 
						  (void* ) &l_ulLoad,
						  sizeof(&l_ulLoad), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Read the number of splits from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS, &g_stLocalCoordinator, 
					   (void*) &l_nNumberOfSplits,
					   sizeof(l_nNumberOfSplits), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS) [%s] No. of Splits [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nNumberOfSplits);

  /* Send the status to the user program */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_NUMBER_OF_SPLITS, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nNumberOfSplits, 
					    sizeof(l_nNumberOfSplits), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Steady State problem Find Split information
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalSteadyStateSplitInformation(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nWorkID;
  int l_nStartRow;
  int l_nSplitID;
  int l_nNumberOfRows;


  /* Send the message to system coordinator */
  l_nMessageID = LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 


  /* Read the Work ID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION) [%s] Work Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

  /* Read the Split Indentification */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nSplitID, 
					   sizeof(l_nSplitID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION) [%s] Work Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nSplitID);

  /* Send the Work ID to System coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION, &g_stLocalCoordinator, 
					    (void* ) &l_nWorkID, 
					    sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 

  /* Send Split Information */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION, &g_stLocalCoordinator, 
					    (void* ) &l_nSplitID, 
					    sizeof(l_nSplitID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Read the Start Row from System coordinator*/
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &g_stLocalCoordinator, 
					   (void*) &l_nStartRow,
					   sizeof(l_nStartRow), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION) [%s] Start Row [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nStartRow);

  /* Read the Start Row from System coordinator*/
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &g_stLocalCoordinator, 
					   (void*) &l_nNumberOfRows,
					   sizeof(l_nNumberOfRows), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION) [%s] No. of Rows [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nNumberOfRows);

  /* Send Start Row to user process */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nStartRow, 
					    sizeof(l_nStartRow), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send Number of Rows to user process */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nNumberOfRows, 
					    sizeof(l_nNumberOfRows), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}


/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Steady State problem Send Work for routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int UserLocalSteadyStateSendWork(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nWorkID;
  int l_nSplitID;
  int l_nSubTaskID;

  char* l_pszMigrateFile;
  char* l_pszResultPipe;

  /* Send the message to system coordinator */
  l_nMessageID = LOCAL_SYSTEM_STEADY_STATE_SEND_WORK;
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &g_stLocalCoordinator, 
					    (void* ) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    } 


  /* Read the Work ID */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SEND_WORK) [%s] Work Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

  /* Read the Split Indentification */
  if ((l_nErrorCode = ReadXDRSignedInteger(USER_LOCAL_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nSplitID, 
					   sizeof(l_nSplitID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SEND_WORK) [%s] Split Id [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nSplitID);


  DISTPIPES_CALLOC(l_pszMigrateFile, char, PATH_LENGTH);

  /* Read the Migrate File Name*/
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) l_pszMigrateFile, 
					   PATH_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SEND_WORK) [%s] Migrate File Name [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszMigrateFile);


  DISTPIPES_CALLOC(l_pszResultPipe, char, PIPENAME_LENGTH);

  /* Read the Result Pipe Name */
  if ((l_nErrorCode = ReadXDRString(USER_LOCAL_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) l_pszResultPipe, 
					   PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      DISTPIPES_FREE(l_pszResultPipe);
      return l_nErrorCode;
    }
  
  DistPipeDebug(USERTOLC_DEBUG, "(USER_LOCAL_STEADY_STATE_SEND_WORK) [%s] Result Pipe Name [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszResultPipe);

  /* Send the Work ID to System coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &g_stLocalCoordinator, 
					    (void* ) &l_nWorkID, 
					    sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      DISTPIPES_FREE(l_pszResultPipe);
      return l_nErrorCode;
    } 

  /* Send Split Information */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &g_stLocalCoordinator, 
					    (void* ) &l_nSplitID, 
					    sizeof(l_nSplitID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      DISTPIPES_FREE(l_pszResultPipe);
      return l_nErrorCode;
    }


  /* Send the Result Pipe Name */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &g_stLocalCoordinator, 
				     (void* ) l_pszResultPipe, 
				     PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      DISTPIPES_FREE(l_pszResultPipe);
      return l_nErrorCode;
    }

  /* Migrate the file to the system coordinator and that will start the new task */
  if (SendDetailsForMigration(&g_stLocalCoordinator, l_pszMigrateFile) < 0)
    {
      DISTPIPES_FREE(l_pszMigrateFile);
      DISTPIPES_FREE(l_pszResultPipe);
      return -1;
    }

  /* Read the Sub Task Id from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &g_stLocalCoordinator, 
					   (void*) &l_nSubTaskID,
					   sizeof(l_nSubTaskID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(SCTOLC_DEBUG, "(SYSTEM_LOCAL_STEADY_STATE_SEND_WORK) [%s] Sub Task ID [%d] \n", 
		g_stSystemCoordinator.m_pszHostName, l_nSubTaskID);


  /* Send the Sub task id to the user process */
  if ((l_nErrorCode = WriteXDRSignedInteger(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nSubTaskID, 
					    sizeof(l_nSubTaskID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  DISTPIPES_FREE(l_pszMigrateFile);
  DISTPIPES_FREE(l_pszResultPipe);

  return 1;
}
/************************************************************************************************************
 * Description : 
 *               User process to Local Coordiantor Processing of messages routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the local coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ProcessUserProcessesMessage(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;
  
  if ((l_nErrorCode = ReadXDRSignedInteger(SYSTEM_LOCAL_ANY_MESSAGE, &a_pstHostDetails->m_stHost,
					   (void* ) &l_nMessageID, 
					   sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  switch (l_nMessageID)
    {

    case USER_LOCAL_HI :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_HI] Message\n");
      return UserLocalHi(a_pstHostDetails);
      break;


    case USER_LOCAL_GET_LOAD_FACTOR :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_GET_LOAD_FACTOR] Message\n");
      return UserLocalGetLoadFactor(a_pstHostDetails);
      break;

    case USER_LOCAL_CREATE_PIPE :      
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_CREATE_PIPE] Message\n");
      return UserLocalCreatePipe(a_pstHostDetails);
      break;

    case USER_LOCAL_EXIST_PIPE :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_EXIST_PIPE] Message\n"); 
      return UserLocalExistPipe(a_pstHostDetails);
      break;

    case USER_LOCAL_OPEN_PIPE :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_OPEN_PIPE] Message\n");
      return UserLocalOpenPipe(a_pstHostDetails);
      break;

    case USER_LOCAL_PURGE_PIPE :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_PURGE_PIPE] Message\n");
      return UserLocalPurgePipe(a_pstHostDetails);
      break;

    case USER_LOCAL_GOODBYE :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_GOODBYE] Message\n");      
      return UserLocalGoodBye(a_pstHostDetails);
      break;

    case USER_LOCAL_STEADY_STATE_INITIALIZE :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_STEADY_STATE_INIIALIZE] Message\n");      
      return UserLocalSteadyStateInitialize(a_pstHostDetails);
      break;

    case USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION] Message\n");      
      return UserLocalSteadyStateMatrixInformation(a_pstHostDetails);
      break;

    case USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_STEADY_STATE_NUMBER_OF_SPILTS] Message\n");      
      return UserLocalSteadyStateNumberOfSplits(a_pstHostDetails);
      break;
      
    case USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION] Message\n");      
      return UserLocalSteadyStateSplitInformation(a_pstHostDetails);
      break;

    case USER_LOCAL_STEADY_STATE_SEND_WORK :
      DistPipeDebug(USERTOLC_DEBUG, "Received [USER_LOCAL_STEADY_STATE_SEND_WORK] Message\n");      
      return UserLocalSteadyStateSendWork(a_pstHostDetails);
      break;

    }

  return 1;

}



