/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/





#ifndef __DISTPIPEMESG_H__

#define __DISTPIPEMESG_H__


#include <rpc/types.h>
#include <rpc/xdr.h>


#define PIPENAME_LENGTH 256
#define HOSTNAME_LENGTH 256
#define LOGINNAME_LENGTH 256
#define PATH_LENGTH 256

#define TCP_HOST 1
#define UNIXSTREAM_HOST 2
#define TCP_AND_UNIXSTREAM_HOST 3


/*
 * User Process -> Local Coordintor Messages
 */

#define USER_LOCAL_CREATE_PIPE 1
#define USER_LOCAL_EXIST_PIPE  2
#define USER_LOCAL_OPEN_PIPE   3
#define USER_LOCAL_GOODBYE     4
#define USER_LOCAL_PURGE_PIPE  5
#define USER_LOCAL_HI          6
#define USER_LOCAL_GET_LOAD_FACTOR 7
#define USER_LOCAL_ANY_MESSAGE 500

/*
 * Local Coordintor -> System Coordinator Messages
 */

#define LOCAL_SYSTEM_CREATE_PIPE 21
#define LOCAL_SYSTEM_EXIST_PIPE  22
#define LOCAL_SYSTEM_OPEN_PIPE   23
#define LOCAL_SYSTEM_GOODBYE     24
#define LOCAL_SYSTEM_PURGE_PIPE  25
#define LOCAL_SYSTEM_HI          26
#define LOCAL_SYSTEM_GET_LOAD_FACTOR 27
#define LOCAL_SYSTEM_FIND_LOAD 28
#define LOCAL_SYSTEM_ANY_MESSAGE 500

/*
 * System Coordinator -> Local Coordintor Messages
 */

#define SYSTEM_LOCAL_CREATE_PIPE 41
#define SYSTEM_LOCAL_EXIST_PIPE   42
#define SYSTEM_LOCAL_OPEN_PIPE   43
#define SYSTEM_LOCAL_GOODBYE     44
#define SYSTEM_LOCAL_PURGE_PIPE  45
#define SYSTEM_LOCAL_HI          46
#define SYSTEM_LOCAL_WRITE_PENDING_PIPE 47
#define SYSTEM_LOCAL_GET_LOAD_FACTOR 48
#define SYSTEM_LOCAL_FIND_LOAD 49
#define SYSTEM_LOCAL_ANY_MESSAGE 500

/*
 * Local Coordintor -> User Process Messages
 */

#define LOCAL_USER_CREATE_PIPE 61
#define LOCAL_USER_EXIST_PIPE  62
#define LOCAL_USER_OPEN_PIPE   63
#define LOCAL_USER_PURGE_PIPE  64
#define LOCAL_USER_GOODBYE     65
#define LOCAL_USER_HI          66
#define LOCAL_USER_GET_LOAD_FACTOR 67




/* Steady State Problem Messages */

/* User to local coordinator messages 801 - 820*/
#define USER_LOCAL_STEADY_STATE_INITIALIZE 801
#define USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION 802
#define USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS 803
#define USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION 804
#define USER_LOCAL_STEADY_STATE_SEND_WORK 805

/* Local coordinator to user messages 821 - 840*/
#define LOCAL_USER_STEADY_STATE_INITIALIZE 821
#define LOCAL_USER_STEADY_STATE_MATRIX_INFORMATION 822
#define LOCAL_USER_STEADY_STATE_NUMBER_OF_SPLITS 823
#define LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION 824
#define LOCAL_USER_STEADY_STATE_SEND_WORK 825

/* local coordinator to system coordinator messages 841 - 860 */
#define LOCAL_SYSTEM_STEADY_STATE_INITIALIZE 841
#define LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION 842
#define LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS 843
#define LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION 844
#define LOCAL_SYSTEM_STEADY_STATE_SEND_WORK 845

/* system coordinator to local coordinator messages 861 - 880*/
#define SYSTEM_LOCAL_STEADY_STATE_INITIALIZE 861
#define SYSTEM_LOCAL_STEADY_STATE_MATRIX_INFORMATION 862
#define SYSTEM_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS 863
#define SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION 864
#define SYSTEM_LOCAL_STEADY_STATE_SEND_WORK 865

typedef struct 
{

  int m_nMessageType;

  int m_nFollowingMessageLength;

  int m_nXDRDataLength;
  char* m_pXDRData;
  XDR m_XDRVariable;


} MESSAGE, *PMESSAGE;



#endif
