/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeMigration.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for that are required for migrating the source 
 * c code to another machine and compling the code there.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 15/October/98
 *
 ***********************************************************************************************/




/************************************************************************************************************
 * Description : 
 *               Creates a new file with the file name passed.
 * Arguments   :
 *               a_pszFileName - Name of the file to be created
 * Return      :
 *               ErrorCode / Created File Descriptor
 ************************************************************************************************************/
int CreateFileForCompilation(char* a_pszFileName)
{
  
  int l_nCreartedFileDescriptor;

  if ((l_nCreatedFileDescriptor = open(a_pszFileName, O_CREAT) < 0))
    {
      DistPipeSystemError("open");
      return CANNOT_CREATE_FILE;
    }

  return l_nCreatedFileDescriptor;
}

/************************************************************************************************************
 * Description : 
 *               Writes to the file.
 * Arguments   :
 *               a_nFileDescriptor - File descriptor to be written to.
 *               a_pszWriteBuffer - Buffer to be written to file.
 *               a_nWriteBufferSize - Amount of bytes to be written.
 * Return      :
 *               ErrorCode / Number of bytes written
 ************************************************************************************************************/
int WriteToFile(int a_nFileDescriptor, char* a_pszWriteBuffer, int a_nWriteBufferSize)
{
  
  int l_nBytesWritten;

  if ((l_nBytesWritten = write(a_nFileDescriptor, a_pszWriteBuffer, a_nWriteBufferSize)) < 0)
    {
      DistPipeSystemError("write");
      return cANNOT_WRITE_TO_FILE;
    }

  return l_nBytesWritten;
}

/************************************************************************************************************
 * Description : 
 *               Compiles the C source code depending on the architecutre it is compiled on.
 * Arguments   :
 *               a_pszFileName - C Source code to be compiled.
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CompileMigratedCFile(char* a_pszFileName)
{

  char *tmpstr;
  char *tstr;

  strcpy (tstr, a_pszFileName);
  tmpstr = strrchr (tstr, '.');
  *tmpstr = '\0';

  tmpstr = "gcc -o ";
  strcat (tmpstr, tstr);
  strcat (tmpstr, " ");
  strcat (tmpstr, a_pszFileName);

  system (tmpstr);
  fprintf(stderr, "Entered CompileMigratedCFile in DistPipeMigration.c - received %s\n\n", a_pszFileName);

  printf("!!!!!!!!!!!!!!!!!!!!!!\n\n\n\n\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  return 1;
 
}


/************************************************************************************************************
 * Description : 
 *               Deletes a C source code and the executable file created.
 * Arguments   :
 *               a_pszFileName - file to be deleted
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DeleteMigratedFiles(char* a_pszFileName)
{


  DISTPIPES_CALLOC(l_pszExecutableFileName, char, (strlen(a_pszFileName) + strlen(EXECUTABLE_FILENAME_EXTENSION)));

  unlink(a_pszFileName);
  strcpy(l_pszExecutableFileName, a_pszFileName);
  strcat(l_pszExecutableFileName, EXECUTABLE_FILENAME_EXTENSION);

  unlink(l_pszExecutableFileName);

  DISTPIPES_FREE(l_pszExecutableFileName);

  return 1;

}

/************************************************************************************************************
 * Description : 
 *               This migrates the code to the specified machine in the array depending on the load factor
 * Arguments   :
 *               a_nNumberofMachineNames - Number of machines in the set.
 *               a_pszMachineNames - set of name of the machine that the code may be migrated.
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int Migrate(int a_nNumberofMachineNames, char** a_pszMachineNames)
{
  return 1;
}
