/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeNetLevel.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the socket related function calls which are used bu
 * others programs to manipulates on sockets.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 24/September/98
 *
 ***********************************************************************************************/


/*
 *must find the header files for each paltform and then 
 * depending on that current files have to be included
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <netdb.h>
#include <signal.h>
#include <sys/un.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "DistPipeNetLevel.h"
#include "DistPipeError.h"
#include "DistPipeDebug.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeXDRRoutines.h"

char* g_pszProgramName;


/************************************************************************************************************
 * Description : 
 *               Creates a TCP Socket
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateTCPHost(a_pstHostDetails)
     PHOST_DESCRIPTOR a_pstHostDetails;
{
  /* 
   * create the socket for this host
   */
  if ((a_pstHostDetails->m_nTCPSocketDescriptor = socket(AF_INET, SOCK_STREAM, 0)) <= 0)
    {
      DistPipeLogSystemError("create");
      return SOCKET_CREATION_ERROR;
    }


  DistPipeDebug(SOCKET_DEBUG, "(CreateSocket)[HostName->m_%s] TCP Socket Descriptor->%d\n", a_pstHostDetails->m_pszHostName, 
	 a_pstHostDetails->m_nTCPSocketDescriptor);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Binds a TCP Socket
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int BindTCPHost(a_pstHostDetails)
     PHOST_DESCRIPTOR a_pstHostDetails;
{

  /*
   * Bind the Socket the port specified
   */
  if (bind(a_pstHostDetails->m_nTCPSocketDescriptor, (struct sockaddr*) a_pstHostDetails->m_pstTCPHostAddress, 
	   sizeof(struct sockaddr)) < 0)
    {
      DistPipeLogSystemError("bind");
      return SOCKET_BIND_ERROR;
    }

  DistPipeDebug(SOCKET_DEBUG, "(BindSocket)[HostName->%s] TCP Socket Port Number->%d\n", 
		a_pstHostDetails->m_pszHostName, ntohs(a_pstHostDetails->m_pstTCPHostAddress->sin_port));

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Creates a UNIX Stream Socket
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateUNIXStreamHost(a_pstHostDetails)
     PHOST_DESCRIPTOR a_pstHostDetails;
{
  /* 
   * create the socket for this host
   */
  if ((a_pstHostDetails->m_nUNIXStreamSocketDescriptor = socket(AF_UNIX, SOCK_STREAM, 0)) <= 0)
    {
      DistPipeLogSystemError("socket");
      return SOCKET_CREATION_ERROR;
    }

  DistPipeDebug(SOCKET_DEBUG, "(CreateSocket)[HostName->%s] UNIX Socket Descriptor->%d\n", a_pstHostDetails->m_pszHostName, 
	 a_pstHostDetails->m_nUNIXStreamSocketDescriptor);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Binds a UNIX Stream Socket
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int BindUNIXStreamHost(a_pstHostDetails)
     PHOST_DESCRIPTOR a_pstHostDetails;
{

  /*
   * Bind the Socket the port specified
   */
  if (bind(a_pstHostDetails->m_nUNIXStreamSocketDescriptor, 
	   (struct sockaddr*) a_pstHostDetails->m_pstUNIXStreamHostAddress, 
	   sizeof(struct sockaddr)) < 0)
    {
      DistPipeLogSystemError("bind");
      return SOCKET_BIND_ERROR;
    }

  DistPipeDebug(SOCKET_DEBUG, "(BindSocket)[HostName->%s] UNIX Socket Connection file ->%s\n", 
		a_pstHostDetails->m_pszHostName, a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_path);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Listens on a Socket
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 *               a_nWhichSocket - TCP/ UNIX Domain sockets
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ListenOnHost(a_pstHostDetails, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nWhichSocket;
{

  int l_nSocketDescriptor;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      l_nSocketDescriptor = a_pstHostDetails->m_nUNIXStreamSocketDescriptor;

  
  listen(l_nSocketDescriptor, 5);


  DistPipeDebug(SOCKET_DEBUG, "(ListenOnSocket)-> Listening for Host [%s] Socket [%d] \n",
		a_pstHostDetails->m_pszHostName, l_nSocketDescriptor);
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Accepts a new connection on the sockets
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 *               a_nWhichSocket - TCP/ UNIX Domain sockets
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int AcceptOnHost(a_pstHostDetails, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nWhichSocket;
{

  int l_nNewSocketDescriptor;

  int l_nSocketDescriptor;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      l_nSocketDescriptor = a_pstHostDetails->m_nUNIXStreamSocketDescriptor;

  /* Accept Connection */
  l_nNewSocketDescriptor = accept(l_nSocketDescriptor, (struct sockaddr*)0, (int*)0);


  DistPipeDebug(SOCKET_DEBUG,"(AcceptOnSocket)[HostName->%s] TCP New Socket Descriptor->%d\n", 
		a_pstHostDetails->m_pszHostName, l_nNewSocketDescriptor);

  
  if (l_nNewSocketDescriptor < 0)
    {
      DistPipeLogSystemError("accept");
      return SOCKET_ACCEPT_ERROR;
    }

  return l_nNewSocketDescriptor;
}


/************************************************************************************************************
 * Description : 
 *               Reads on a Socket Descriptor value
 * Arguments   :
 *               a_nSocketDescriptor - Socket Descriptor
 *               a_pszReadBuffer - Buffer to read into
 *               a_nBufferSize - Size of the buffer to be read
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ReadFromSocketDescriptor(a_nSocketDescriptor, a_pszReadBuffer, a_nBufferSize)
     int a_nSocketDescriptor;
     char* a_pszReadBuffer;
     int a_nBufferSize;
{
  int l_nBytesRead;
  int l_nRemainingSize;
  char* l_pszRemainingBuffer;

  DISTPIPES_ZERO(a_pszReadBuffer, a_nBufferSize);


  l_pszRemainingBuffer = a_pszReadBuffer;
  l_nRemainingSize = a_nBufferSize;

  do
    {
      if ((l_nBytesRead = read(a_nSocketDescriptor,
			       l_pszRemainingBuffer, l_nRemainingSize)) < 0)
	{
	  DistPipeLogSystemError("read");
	  return SOCKET_READ_ERROR;
	}
      if (0 == l_nBytesRead)
	{
	  return l_nBytesRead;
	}
      l_nRemainingSize = l_nRemainingSize - l_nBytesRead;
      l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesRead;

    } while (l_nRemainingSize > 0);
  

  return a_nBufferSize;


}

/************************************************************************************************************
 * Description : 
 *               Reads on a Socket Descriptor value
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details structure
 *               a_pszReadBuffer - Buffer to read into
 *               a_nBufferSize - Size of the buffer to be read
 *               a_nWhichSocket - TCP/ UNIX Domain sockets
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ReadOnHost(a_pstHostDetails, a_pszReadBuffer, a_nBufferSize, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     char* a_pszReadBuffer; 
     int a_nBufferSize; 
     int a_nWhichSocket;
{
  
  int l_nSocketDescriptor;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      l_nSocketDescriptor = a_pstHostDetails->m_nUNIXStreamSocketDescriptor;

  return ReadFromSocketDescriptor(l_nSocketDescriptor, a_pszReadBuffer, a_nBufferSize);
  
}

/************************************************************************************************************
 * Description : 
 *               Writes on a Socket Descriptor value
 * Arguments   :
 *               a_nSocketDescriptor - Socket Descriptor
 *               a_pszWriteBuffer - Buffer to write into
 *               a_nBufferSize - Size of the buffer to be written
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int WriteToSocketDescriptor(a_nSocketDescriptor, a_pszWriteBuffer, a_nBufferSize)
     int a_nSocketDescriptor;
     char* a_pszWriteBuffer;
     int a_nBufferSize;
{
  int l_nBytesWritten;
  int l_nRemainingSize;
  char* l_pszRemainingBuffer;

  l_pszRemainingBuffer = a_pszWriteBuffer;
  l_nRemainingSize = a_nBufferSize;

  do
    {
      if ((l_nBytesWritten = write(a_nSocketDescriptor, l_pszRemainingBuffer, l_nRemainingSize)) <= 0)
	{
	  DistPipeLogSystemError("write");
	  return SOCKET_WRITE_ERROR;
	}
      
      
      l_nRemainingSize = l_nRemainingSize - l_nBytesWritten;
      l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesWritten;

    } while (l_nRemainingSize > 0);
 	 
  return a_nBufferSize;
}

/************************************************************************************************************
 * Description : 
 *               Writes on a Socket Descriptor value
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_pszWriteBuffer - Buffer to write into
 *               a_nBufferSize - Size of the buffer to be written
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int WriteOnHost(a_pstHostDetails, a_pszWriteBuffer, a_nBufferSize, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     char* a_pszWriteBuffer; 
     int a_nBufferSize; 
     int a_nWhichSocket;
{

  
  int l_nSocketDescriptor;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      l_nSocketDescriptor = a_pstHostDetails->m_nUNIXStreamSocketDescriptor;

  return WriteToSocketDescriptor(l_nSocketDescriptor, a_pszWriteBuffer, a_nBufferSize);
}

/************************************************************************************************************
 * Description : 
 *               Creates a TCP Server Socket and listens on it
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nPortNumber - Port Number to listen on
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateTCPServerHostAndListen(a_pstHostDetails, a_nPortNumber)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nPortNumber;
{


  int l_nCreateSocket;
  int l_nBindCall;

  if (NULL == a_pstHostDetails->m_pstTCPHostAddress)
    {
      /*
       * Error Message 
       */
      return -1;
    }

  /* Create the Socket */
  if ((l_nCreateSocket = CreateTCPHost(a_pstHostDetails)) <= 0)
    {
      return l_nCreateSocket;
    }

  a_pstHostDetails->m_pstTCPHostAddress->sin_port = htons(a_nPortNumber);
  a_pstHostDetails->m_pstTCPHostAddress->sin_addr.s_addr = htonl(INADDR_ANY);
  if ((l_nBindCall = BindTCPHost(a_pstHostDetails)) < 0)
    {
      return l_nBindCall;
    }
  
  /* Listen on the Socket */
  ListenOnHost(a_pstHostDetails, TCP_HOST);

  return l_nCreateSocket;
}

/************************************************************************************************************
 * Description : 
 *               Creates a TCP Clients Socket and connects to server machine
 * Arguments   :
 *               a_pstClientsDetails - Pointer to the host details
 *               a_pszServerName - Name of the server to connect to
 *               a_nServerPortNumber - Port Number for the server to connect
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateTCPClientHostAndConnect(a_pstClientDetails, a_pszServerName, a_nServerPortNumber)
     PHOST_DESCRIPTOR a_pstClientDetails;
     char* a_pszServerName; 
     int a_nServerPortNumber;
{

  int nCreatedSocket;
  struct sockaddr_in l_stServerAddress;
  struct hostent* l_pstServerHostEntry;


  if (NULL == a_pstClientDetails->m_pstTCPHostAddress)
    {
      /*
       * Error Message 
       */
      return -1;
    }

  /* Create the Socket */
  if ((nCreatedSocket = CreateTCPHost(a_pstClientDetails)) <= 0)
    {
      return nCreatedSocket;
    }

  DISTPIPES_ZERO(&l_stServerAddress, sizeof(struct sockaddr_in));

  l_pstServerHostEntry = gethostbyname(a_pszServerName);

  if (NULL == l_pstServerHostEntry)
    {
      DistPipeLogSystemError("gethostbyname");
      return HOST_GETHOSTBYNAME_ERROR;
    }

  DISTPIPES_COPY(l_pstServerHostEntry->h_addr, &l_stServerAddress.sin_addr, l_pstServerHostEntry->h_length);

  l_stServerAddress.sin_family = AF_INET;
  l_stServerAddress.sin_port = htons(a_nServerPortNumber);

  DistPipeDebug(SOCKET_DEBUG, "(CreateTCPClientSocketAndConnect)->Host [%s], Port [%d]\n",
		a_pszServerName, a_nServerPortNumber);

  /* Connect to the server node */
  if (connect(a_pstClientDetails->m_nTCPSocketDescriptor, (struct sockaddr *)&l_stServerAddress, sizeof(struct sockaddr)) < 0)
    {
      DistPipeLogSystemError("conenct");
      return SOCKET_CONNECT_ERROR;
    }

  DistPipeDebug(SOCKET_DEBUG, "(CreateTCPClientSocketAndConnect)->Connected to Host [%s], Port [%d]\n",
		a_pszServerName, a_nServerPortNumber);
  
  return a_pstClientDetails->m_nTCPSocketDescriptor;
}

/************************************************************************************************************
 * Description : 
 *               Creates a UNIX domain stream Server Socket and listens on it
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_pszUNIXStreamPath - path of UNIX domain stream socket file
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateUNIXStreamServerHostAndListen(a_pstHostDetails, a_pszUNIXStreamPath)
     PHOST_DESCRIPTOR a_pstHostDetails;
     char* a_pszUNIXStreamPath;
{
  int l_nCreateSocket;
  int l_nBindCall;

  if (NULL == a_pstHostDetails->m_pstUNIXStreamHostAddress)
    {
      /*
       * Error Message 
       */
      return -1;
    }

  strcpy(a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_path, a_pszUNIXStreamPath);

  DistPipeDebug(SOCKET_DEBUG, "server path --> %s\n", a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_path);

  /* Create the Socket */
  if ((l_nCreateSocket = CreateUNIXStreamHost(a_pstHostDetails)) <= 0)
    {
      return l_nCreateSocket;
    }


  if ((l_nBindCall = BindUNIXStreamHost(a_pstHostDetails)) < 0)
    {
      return l_nBindCall;
    }

  /* Listen on the Socket */
  ListenOnHost(a_pstHostDetails, UNIXSTREAM_HOST);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Creates a TCP Clients Socket and connects to server machine
 * Arguments   :
 *               a_pstClientDetails - Pointer to the gost details
 *               a_pszServerPath - path of UNIX domain stream server socket file
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateUNIXStreamClientHostAndConnect(a_pstClientDetails, a_pszServerPath)
     PHOST_DESCRIPTOR a_pstClientDetails;
     char* a_pszServerPath;
{

  int nCreatedSocket;
  struct sockaddr_un l_stServerAddress;


  if (NULL == a_pstClientDetails->m_pstUNIXStreamHostAddress)
    {
      /*
       * Error Message 
       */
      return -1;
    }

  /* Create the Socket */
  if ((nCreatedSocket = CreateUNIXStreamHost(a_pstClientDetails)) <= 0)
    {
      return nCreatedSocket;
    }

  DISTPIPES_ZERO(&l_stServerAddress, sizeof(struct sockaddr_in));

  l_stServerAddress.sun_family = AF_UNIX;
  strcpy(l_stServerAddress.sun_path, a_pszServerPath);

  DistPipeDebug(SOCKET_DEBUG, "connecting to sever in port [%s]\n", a_pszServerPath);

  
  /* Connect to the server node */
  if (connect(a_pstClientDetails->m_nUNIXStreamSocketDescriptor, 
	      (struct sockaddr *)&l_stServerAddress, 
	      (sizeof(l_stServerAddress.sun_family) + strlen(l_stServerAddress.sun_path))) < 0)
    {
      DistPipeLogSystemError("connect");
      return SOCKET_CONNECT_ERROR;
    }
  
  return a_pstClientDetails->m_nUNIXStreamSocketDescriptor;
}

/************************************************************************************************************
 * Description : 
 *               close the Socket Descriptor 
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CloseHost(a_pstHostDetails, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nWhichSocket;
{

  if (TCP_HOST == a_nWhichSocket)
    {
      close(a_pstHostDetails->m_nTCPSocketDescriptor);
      DistPipeDebug(SOCKET_DEBUG,"(CloseSocket)[HostName->%s] Closing ... TCP Socket Descriptor->%d \n", 
		    a_pstHostDetails->m_pszHostName, a_pstHostDetails->m_nTCPSocketDescriptor);
    }
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      {
	close(a_pstHostDetails->m_nUNIXStreamSocketDescriptor);
	DistPipeDebug(SOCKET_DEBUG,"(CloseSocket)[HostName->%s] Closing ... TCP Socket Descriptor->%d \n", 
		      a_pstHostDetails->m_pszHostName, a_pstHostDetails->m_nUNIXStreamSocketDescriptor);
      }
    else
      if (TCP_AND_UNIXSTREAM_HOST == a_nWhichSocket)
      {
	close(a_pstHostDetails->m_nTCPSocketDescriptor);
	close(a_pstHostDetails->m_nUNIXStreamSocketDescriptor);	
	DistPipeDebug(SOCKET_DEBUG,"(CloseSocket)[HostName->%s] Closing ... TCP Socket Descriptor->%d \n", 
		      a_pstHostDetails->m_pszHostName, a_pstHostDetails->m_nUNIXStreamSocketDescriptor);
	DistPipeDebug(SOCKET_DEBUG,"(CloseSocket)[HostName->%s] Closing ... TCP Socket Descriptor->%d \n", 
		    a_pstHostDetails->m_pszHostName, a_pstHostDetails->m_nTCPSocketDescriptor);
      }
  
  DISTPIPES_ZERO(a_pstHostDetails, sizeof(HOST_DESCRIPTOR));

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               checks if any data is available on the sockets
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nlSeconds - Seconds to wait for
 *               a_nlMicroSeconds - Micro Seconds to wait for
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int AnyDataToReadOnHost(a_pstHostDetails, a_nlSeconds, a_nlMicroSeconds, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     long a_nlSeconds; 
     long a_nlMicroSeconds; 
     int a_nWhichSocket;
{
  

  fd_set l_stReadDescriptors;
  struct timeval l_stTimeOutPeriod;
  int l_nIsReady;
  int l_nNumberofDescriptors;
  int l_nSocketDescriptor;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      l_nSocketDescriptor = a_pstHostDetails->m_nUNIXStreamSocketDescriptor;



  l_nNumberofDescriptors  = getdtablesize() + 1;

  FD_ZERO(&l_stReadDescriptors);

  l_stTimeOutPeriod.tv_sec = a_nlSeconds;
  l_stTimeOutPeriod.tv_usec = a_nlMicroSeconds;

  FD_SET(l_nSocketDescriptor, &l_stReadDescriptors);
  
  /*
   * Using the select call which polls and returns immediatly
   * if no data is avialable since Seconds is 0
   */
  l_nIsReady = select(l_nNumberofDescriptors, &l_stReadDescriptors, 
		      (fd_set* )NULL, (fd_set* )NULL, &l_stTimeOutPeriod);

  FD_CLR(l_nSocketDescriptor, &l_stReadDescriptors);
  
  return l_nIsReady;

}

/************************************************************************************************************
 * Description : 
 *               Write the message and the data to the socket
 * Arguments   :
 *               a_nMessageID - Message ID to be sent
 *               a_pszMessage - Message to tbe sent
 *               a_nMessageSize - size of the message to be sent
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int WriteMessageAndData(a_nMessageID, a_pszMessage, a_nMessageSize, a_pstHostDetails, a_nWhichHost)
     int a_nMessageID;
     char* a_pszMessage; 
     int a_nMessageSize; 
     PHOST_DESCRIPTOR a_pstHostDetails; 
     int a_nWhichHost;
{

  int l_nErrorCode;
  MESSAGE l_stSendMessage;

  l_stSendMessage.m_nMessageType = htons(a_nMessageID);
  l_stSendMessage.m_nFollowingMessageLength = htons(a_nMessageSize);
  if ((l_nErrorCode = WriteOnHost(a_pstHostDetails, (char*) &l_stSendMessage, 
				    sizeof(MESSAGE), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    } 
  if ((l_nErrorCode = WriteOnHost(a_pstHostDetails, (char*) a_pszMessage, 
				    ntohs(l_stSendMessage.m_nFollowingMessageLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Read the message and the data from the socket
 * Arguments   :
 *               a_nMessageID - Message ID to be read
 *               a_pszMessage - Message to be read
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ReadMessageAndData(a_nMessageID, a_pszMessage, a_pstHostDetails, a_nWhichHost)
     int a_nMessageID; 
     char* a_pszMessage; 
     PHOST_DESCRIPTOR a_pstHostDetails; 
     int a_nWhichHost;
{


  int l_nErrorCode;
  MESSAGE l_stReceivedMessage;

  if ((l_nErrorCode = ReadOnHost(a_pstHostDetails, (char*) &l_stReceivedMessage, 
				   sizeof(MESSAGE), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }
  if ((l_nErrorCode = ReadOnHost(a_pstHostDetails, (char*) a_pszMessage, 
				   ntohs(l_stReceivedMessage.m_nFollowingMessageLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}



/************************************************************************************************************
 * Description : 
 *               Get Socket Details 
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int GetHostPortAddress(a_pstHostDetails, a_nWhichSocket)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nWhichSocket;
{

  int l_nSocketDescriptor;
  int l_nSizeofAddress;
  struct sockaddr_in l_stGetSockNameAddress;

  if (TCP_HOST == a_nWhichSocket)
    l_nSocketDescriptor = a_pstHostDetails->m_nTCPSocketDescriptor;
  else 
    if (UNIXSTREAM_HOST == a_nWhichSocket)
      {
	/* Not yet implemented */
	return -1;
      }

  l_nSizeofAddress = sizeof(struct sockaddr);
  if (getsockname(l_nSocketDescriptor, (struct sockaddr*) &l_stGetSockNameAddress, &l_nSizeofAddress) < 0)
    {
      DistPipeLogSystemError("getsockname");
      return SOCKET_GETSOCKNAME_ERROR;
    }
  DistPipeDebug(SOCKET_DEBUG, "Port number returned by gethostname -> [%d]\n", ntohs(l_stGetSockNameAddress.sin_port));
  a_pstHostDetails->m_pstTCPHostAddress->sin_port = l_stGetSockNameAddress.sin_port;

  return l_stGetSockNameAddress.sin_port;
}

/************************************************************************************************************
 * Description : 
 *               Close socket descriptor
 * Arguments   :
 *               a_nSocketDescriptor - socket descriptor to be closed
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CloseSocketDescriptor(a_nSocketDescriptor)
     int a_nSocketDescriptor;
{

  return close(a_nSocketDescriptor);

}
