/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/

#include <stdio.h>
#include <endian.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include <strings.h>

#define MAX_FLOAT_ARRAY_SIZE 16

/*
*  (float)1.0 =
*  IEEE single ll   00 00 80 3f
*  IEEE double ll   00 00 00 00 00 00 f0 3f
*  IEEE single hh   3f 80 00 00
*  IEEE double hh   3f f0 00 00 00 00 00 00
*  CRAY       (hh)  40 01 80 00 00 00 00 00
*  Convex s   (hh)  40 80 00 00
*  Convex d   (hh)  40 10 00 00 00 00 00 00
*  IBM370 s
*  IBM370 d
*  VAX xxxx?
*/

typedef struct {

  int m_nLength;
  unsigned char *m_pBytes;
  char* m_pszName;

} FLOAT_SIGNATURE;

/* signatures for float = 1.0 in hh byte order, all formats */

static unsigned char g_pFloatSignature1[] = { 0x3f, 0x80, 0x00, 0x00 };
static unsigned char g_pFloatSignature2[] = { 0x3f, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static unsigned char g_pFloatSignature4[] = { 0x40, 0x01, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00 };
static unsigned char g_pFloatSignature5[] = { 0x40, 0x80, 0x00, 0x00 };
static unsigned char g_pFloatSignature6[] = { 0x40, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

FLOAT_SIGNATURE g_pstFloatSignatureDetails[MAX_FLOAT_ARRAY_SIZE] = {
	{ 0, 0, "UNKNOWN" },
	{ 4, g_pFloatSignature1, "IEEE_single" },
	{ 8, g_pFloatSignature2, "IEEE_double" },
	{ 0, 0, "UNDEFINED" },
	{ 8, g_pFloatSignature4, "Cray" },
	{ 4, g_pFloatSignature5, "Convex_single" },
	{ 8, g_pFloatSignature6, "Convex_double" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
	{ 0, 0, "UNDEFINED" },
};

int g_nDataSignature = -1;


/************************************************************************************************************
 * Description : 
 *               Finds the if the machine is little endian / big endian format
 * Arguments   :
 *               a_pDataType - type of the dats for which the format is to be indentified
 *               a_nDataSize - size of the data type
 *               a_nBitPosition - Position of the bits to be set in the format
 * Return      :
 *               The corresponding Bits are set to indicate the byte order and size
 ************************************************************************************************************/

int IntegerSignature(a_nBitPosition, a_pDataType, a_nDataSize)
     int a_nBitPosition;
     char* a_pDataType; 
     int a_nDataSize;
{
  
  int l_nByteOrder;
  int l_nFirstBitSet;

  if (0 == a_pDataType[0])
    /* Low Order Byte in Lower End */
    l_nByteOrder = 0;
  else
    if ((a_nDataSize - 1) == a_pDataType[0])
      /* High Order Byte in Higher End */
      l_nByteOrder = 3;
    else
      if ((a_nDataSize / 2) == a_pDataType[0])
	/* High Order Byte in Lower End */
	l_nByteOrder = 2;
      else
	if ((a_nDataSize / 2 - 1) == a_pDataType[0])
	  /* Low Order Byte in Higher End */
	  l_nByteOrder = 1;
	else
	  {
	    /* Cannot find the byte order */
	    /*
	     * DistPipeAbort();
	     */
	  }
  
  l_nFirstBitSet = ffs(a_nDataSize) - 1;

  /*
  DistPipeDebug(GENERAL_DEBUG, ".%d%d.%d%d%d\n", 
		(l_nByteOrder & 2) ? 1 : 0, 
		(l_nByteOrder & 1) ? 1 : 0,
		(l_nFirstBitSet & 4) ? 1 : 0, 
		(l_nFirstBitSet & 2) ? 1 : 0, 
		(l_nFirstBitSet & 1) ? 1 : 0);
  */

  printf(".%d%d.%d%d%d\n", 
	 (l_nByteOrder & 2) ? 1 : 0, 
	 (l_nByteOrder & 1) ? 1 : 0,
	 (l_nFirstBitSet & 4) ? 1 : 0, 
	 (l_nFirstBitSet & 2) ? 1 : 0, 
	 (l_nFirstBitSet & 1) ? 1 : 0);

  return ((l_nByteOrder << 3) | l_nFirstBitSet) << a_nBitPosition;
  
}



/************************************************************************************************************
 * Description : 
 *               Finds the if the machine is little endian / big endian format for floats
 * Arguments   :
 *               a_pDataType - type of the dats for which the format is to be indentified
 *               a_nDataSize - size of the data type
 *               a_nBitPosition - Position of the bits to be set in the format
 * Return      :
 *               The corresponding Bits are set to indicate the byte order and size
 ************************************************************************************************************/

int FloatSignature(a_nBitPosition, a_pDataType, a_nDataSize)
     int a_nBitPosition;
     unsigned char* a_pDataType;
     int a_nDataSize;
{
  
  int l_nFloatArrayIndex;
  int l_nBytesIndex;


  /* Scan the array and find which of the size matches with the size passed */
  for(l_nFloatArrayIndex = 0; l_nFloatArrayIndex < MAX_FLOAT_ARRAY_SIZE; l_nFloatArrayIndex++)
    {
      if (g_pstFloatSignatureDetails[l_nFloatArrayIndex].m_nLength == a_nDataSize)
	{
	  /* Find out the signature of the float value depending on the stored */
	  for(l_nBytesIndex = 0; l_nBytesIndex < a_nDataSize; l_nBytesIndex++)
	    if (a_pDataType[l_nBytesIndex] != g_pstFloatSignatureDetails[l_nFloatArrayIndex].m_pBytes[l_nBytesIndex])
	
	  if (l_nBytesIndex == a_nDataSize)
	    return ((3 << 4) | l_nFloatArrayIndex) << a_nBitPosition;

	  for(l_nBytesIndex = 0; l_nBytesIndex < a_nDataSize; l_nBytesIndex++)
	    if (a_pDataType[a_nDataSize - 1 - l_nBytesIndex] != g_pstFloatSignatureDetails[l_nFloatArrayIndex].m_pBytes[l_nBytesIndex])
		  break;
	  
	  if (l_nBytesIndex == a_nDataSize)
	    return l_nFloatArrayIndex << a_nBitPosition;
	}
    }

  /* error cannot obtain the byte order for this float/double */
  /*
   * DistPipeAbort();
   */
  return 0;
  
}

/************************************************************************************************************
 * Description : 
 *               finds out the signature of byte, short, int, long, float, double.
 * Arguments   :
 *              ----- 
 * Return      :
 *              Bits:
 *              0:2  short int size
 *              3:4  short int byte order
 *              5:7  int size
 *              8:9  int byte order
 *              10:12  long int size
 *              13:14  long int byte order
 *              15:18  float format
 *              19:20  float byte order
 *              21:24  double float format
 *              25:26  double float byte order
 *              27:31  0
 *
 *	        Sizes are 2 ** (0..7) bytes
 *	        Byte order specifies overall order and half-folding:
 *		   00  Little endian (e.g. 0 1 2 3)
 *		   01  Little, but half-swapped (1 0 3 2 or 3 2 1 0 7 6 5 4)
 *		   10  Big, but half swapped (2 3 0 1 or 4 5 6 7 0 1 2 3)
 *		   11  Big endian (3 2 1 0)
 ************************************************************************************************************/
int FindSignatures()
{
  short l_nShortInt;
  int l_nInt;
  long l_nLongInt;
  float l_fFloat;
  double l_fDouble;
  int l_nSizeIndex;
  int l_nSignatureFormat;
  
  if (g_nDataSignature != -1)
    return g_nDataSignature;
  
  l_nSignatureFormat = 0;
  
  l_nShortInt = 0;
  for (l_nSizeIndex = 0; l_nSizeIndex < sizeof(l_nShortInt); l_nSizeIndex++)
    l_nShortInt += (short)l_nSizeIndex << (l_nSizeIndex * 8);
  l_nSignatureFormat |= IntegerSignature(0, (char *) &l_nShortInt, (int) sizeof(l_nShortInt));
  
  printf("Size of short int [%d] and Endian [%d]\n", (l_nSignatureFormat & 0x00000003) * 2, (l_nSignatureFormat & 0x00000018) >> 3);


  l_nInt = 0;
  for (l_nSizeIndex = 0; l_nSizeIndex < sizeof(l_nInt); l_nSizeIndex++)
    l_nInt += (int)l_nSizeIndex << (l_nSizeIndex * 8);
  l_nSignatureFormat |= IntegerSignature(5, (char *) &l_nInt, (int) sizeof(l_nInt));
  printf("Size of int [%d] and Endian [%d]\n", ((l_nSignatureFormat & 0x000000E0) >> 5) * 2, (l_nSignatureFormat & 0x00000300) >> 8);

  
  l_nLongInt = 0;
  for (l_nSizeIndex = 0; l_nSizeIndex < sizeof(l_nLongInt); l_nSizeIndex++)
    l_nLongInt += (long)l_nSizeIndex << (l_nSizeIndex * 8);
  l_nSignatureFormat |= IntegerSignature(10, (char *) &l_nLongInt, (int) sizeof(l_nLongInt));
   printf("Size of long [%d] and Endian [%d]\n", ((l_nSignatureFormat & 0x00001C00) >> 10) * 2, (l_nSignatureFormat & 0x00006000) >> 13);

  l_fFloat = 1.0;
  l_nSignatureFormat |= FloatSignature(15, (unsigned char *) &l_fFloat, (int) sizeof(l_fFloat));
  printf("format of float [%d] and Endian [%d]\n", ((l_nSignatureFormat & 0x00078000) >> 15), (l_nSignatureFormat & 0x00180000) >> 19);
 

  l_fDouble = 1.0;
  l_nSignatureFormat |= FloatSignature(21, (unsigned char *) &l_fDouble, (int) sizeof(l_fDouble));
  printf("format of double [%d] and Endian [%d]\n", ((l_nSignatureFormat & 0x01E00000) >> 21), (l_nSignatureFormat & 0x06000000) >> 25);

  g_nDataSignature = l_nSignatureFormat;
  return l_nSignatureFormat;
}


int main()
{

  int l_nFormat;

  l_nFormat = FindSignatures();

  printf("%0X, %d %d %d %d \n", l_nFormat, sizeof(int), sizeof(long), sizeof(float), sizeof(double));
  return 1;

}
