/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeRoutines.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the distributed pipe related function call that the user program
 * will invoke. 
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 3/October/98
 *
 ***********************************************************************************************/
#include "DistPipeSystemCoordinator.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeError.h"
#include "DistPipeNetLevel.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeTable.h"
#include "DistPipeRoutines.h"
#include "DistPipeXDRRoutines.h"


PDIST_PIPE_TABLE g_pstListofDistributedPipes = NULL;
HOST_DESCRIPTOR g_stUserProcess;
HOST_DESCRIPTOR g_stDistributedPipes;
int g_nIntializedUserProcess;
char* g_pszProgramName;


/************************************************************************************************************
 * Description : 
 *               Intialization of the user program
 * Arguments   :
 *               ---
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int InitializeUserProcess()
{

  int l_nErrorCode;
  int l_nMessageID;

  /* The user process has already been intialized its data */

  if (1 == g_nIntializedUserProcess)
    {
      return 1;
    }


  InitializeDistPipeDebug();

  g_nIntializedUserProcess = 1;

  /* 
   * connect to the local boss for distributed pipes
   */
  InitializeHost(&g_stUserProcess, UNIXSTREAM_HOST);


   /*
   * Create a Socket and Connect to Local Boss (Local coordinators)
   */
  
  if ((l_nErrorCode = CreateUNIXStreamClientHostAndConnect(&g_stUserProcess, LOCAL_COORDINATOR_PORT)) < 0)
    {
      return l_nErrorCode;
    }
 
  /*
   * Send a Hi Message to Local Coordinator
   */
  /*
   * Send the User Process  Details to Local Coordinator
   */
  l_nMessageID = USER_LOCAL_HI;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Send Host Name */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_HI, &g_stUserProcess, 
				     (void*) g_stUserProcess.m_pszHostName, 
				     HOSTNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Architecture */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_HI, &g_stUserProcess, 
					    (void*) &g_stUserProcess.m_nHostArchitecture, 
					    sizeof(g_stUserProcess.m_nHostArchitecture), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
 
  /* Send Process Id */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_HI, &g_stUserProcess, 
					    (void*) &g_stUserProcess.m_nHostProcessID, 
					    sizeof(g_stUserProcess.m_nHostProcessID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Login Name*/
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_HI, &g_stUserProcess, 
				     (void*) g_stUserProcess.m_pszHostLogin, 
				     LOGINNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Daemon path*/
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_HI, &g_stUserProcess, 
				     (void*) g_stUserProcess.m_pszDaemonPath, 
				     PATH_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
 
}

/************************************************************************************************************
 * Description : 
 *               Creation of the Distributed pipe 
 * Arguments   :
 *               a_pszPipeName - Pipe to tbe created
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CreateDistPipe(a_pszPipeName)
     char* a_pszPipeName;
{

  int l_nErrorCode;
  int l_nPipeCreated;
  int l_nMessageID;


  GetLoadFactor(1);

  l_nErrorCode = DoesPipeExist(a_pszPipeName);
  if ((l_nErrorCode < 0) || (PIPENAME_DOES_EXIST == l_nErrorCode ))
    {
      /* Pipe Name Already Exist therefore cannot create it again */
      return USER_ERROR_MESSAGE_PIPENAME_ALREADY_EXIST;
    }

  /* Create a new pipe descriptor details */

  /* insert the distributed pipe detail into the list of details */
  AppendTable(&g_pstListofDistributedPipes, g_stUserProcess.m_pszHostName, a_pszPipeName);

  l_nMessageID = USER_LOCAL_CREATE_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send Create Pipe Name */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_CREATE_PIPE,  &g_stUserProcess, 
				     (char*) a_pszPipeName, 
				     PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* check if the pipe could be created */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_CREATE_PIPE, &g_stUserProcess, 
					   (void*) &l_nPipeCreated, 
					   sizeof(l_nPipeCreated), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_CREATE_PIPE) [%s] Pipe Create Status [%d] \n", 
		g_stUserProcess.m_pszHostName, l_nPipeCreated);


  if (CREATE_PIPENAME_ALREADY_EXIST == l_nPipeCreated)
    return USER_ERROR_MESSAGE_PIPENAME_ALREADY_EXIST;

   return 1;
}

/************************************************************************************************************
 * Description : 
 *               Check for existence of the Distributed pipe 
 * Arguments   :
 *               a_pszPipeName - Pipe to tbe checked
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DoesPipeExist(a_pszPipeName)
     char* a_pszPipeName;
{

  int l_nErrorCode;
  int l_nPipeExist;
  int l_nMessageID;

  l_nMessageID = USER_LOCAL_EXIST_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send Create Pipe Name */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_EXIST_PIPE, &g_stUserProcess, 
				     (void*) a_pszPipeName, 
				     PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Read from the local coordinator if the pipe name exist */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_EXIST_PIPE, &g_stUserProcess, 
					   (void*) &l_nPipeExist, 
					   sizeof(l_nPipeExist), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "[LOCAL_USER_EXIST_PIPE] for Pipe [%s] Exist Code [%d]\n",
		a_pszPipeName, l_nPipeExist);

  return l_nPipeExist;
}

/************************************************************************************************************
 * Description : 
 *               Open a Distributed pipe 
 * Arguments   :
 *               a_pszPipeName - Pipe to tbe created
 *               a_nAccessMode - Read / Write Mode
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int OpenDistPipe(a_pszPipeName, a_nAccessMode)
     char* a_pszPipeName;
     int a_nAccessMode;
{

    int l_nErrorCode;
    int l_nPipeDescriptor;
    int l_nPipeOpen;
    int l_nPortNumber;
    int l_nConnectStatus;
    int l_nMessageID;

    char* l_pszMachineName;
    HOST_DESCRIPTOR l_stDistributedPipeHost;


    l_nErrorCode = DoesPipeExist(a_pszPipeName);
    if ((l_nErrorCode < 0) || (PIPENAME_DOESNOT_EXIST == l_nErrorCode))
    {
        /* Pipe Name does not exist cannot open then must create it first */
        return USER_ERROR_MESSAGE_PIPENAME_DOESNOT_EXIST;
    }

    l_nMessageID = USER_LOCAL_OPEN_PIPE;
    if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess, (void*) &l_nMessageID, sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
        return l_nErrorCode;
    }

    /* Send Create Pipe Name */
    if ((l_nErrorCode = WriteXDRString(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) a_pszPipeName, PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
        return l_nErrorCode;
    }
  
    /* Send the Access Mode the user program is opening it */
    if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &a_nAccessMode, sizeof(a_nAccessMode), UNIXSTREAM_HOST)) < 0)
    {
        return l_nErrorCode;
    }
  

    InitializeHost(&l_stDistributedPipeHost, TCP_HOST);

    /* Create a New TCP csocket coonection depending on Read / Write to the Pipe */
    if (READ_ACCESS_MODE == a_nAccessMode)
    {

        DistPipeDebug(GENERAL_DEBUG, "Opening Pipe [%s] in READ mode \n", a_pszPipeName);

        /* Create a New socket that will listen to the port */
        if ((l_nErrorCode = CreateTCPServerHostAndListen(&l_stDistributedPipeHost, 0)) < 0)
        {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nErrorCode;
	    }

        /* get the address in which the address is bind to */
        if ((l_nPortNumber = GetHostPortAddress(&l_stDistributedPipeHost, TCP_HOST)) < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nPortNumber;
	    }
      
        l_nPortNumber = ntohs(l_nPortNumber);
        DistPipeDebug(SOCKET_DEBUG, "Distributed Pipe [%s] Address : [%d] Read Pipe Descriptor : [%d] \n", a_pszPipeName, l_nPortNumber, l_stDistributedPipeHost.m_nTCPSocketDescriptor);
        
        /* Update the entry about the distributed pipes */
        UpdateEntry(g_pstListofDistributedPipes, a_pszPipeName, READ_ACCESS_MODE, g_stUserProcess.m_pszHostName, l_nPortNumber, g_stUserProcess.m_nHostProcessID, l_stDistributedPipeHost.m_nTCPSocketDescriptor); 

        /* Send the details to the local coordinator */
        /* Send Process ID */
        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_stDistributedPipeHost.m_nHostProcessID, sizeof(l_stDistributedPipeHost.m_nHostProcessID), UNIXSTREAM_HOST)) < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nErrorCode;
	    }

        /* Send Port Number */
        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_nPortNumber, sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0)
		{
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nErrorCode;
	    }

        /* Send the Socket Descriptor number for this pipe */
        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_stDistributedPipeHost.m_nTCPSocketDescriptor, sizeof(l_stDistributedPipeHost.m_nTCPSocketDescriptor), UNIXSTREAM_HOST)) < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nErrorCode;
	    }

        /* read the open read status from the local coordinator */
        if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &g_stUserProcess, (void*) &l_nPipeOpen, sizeof(l_nPipeOpen), UNIXSTREAM_HOST)) < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nErrorCode;
	    }

        DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_OPEN_PIPE] Pipe Name [%s] Read Status [%d]\n", a_pszPipeName, l_nPipeOpen);
      
        if (l_nPipeOpen < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nPipeOpen;
	    }

        /* This blocks and accepts the connection that is come from write end */
        if ((l_nPipeDescriptor = AcceptOnHost(&l_stDistributedPipeHost, TCP_HOST)) < 0)
	    {
	        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	        return l_nPipeDescriptor;
	    }
      
        /* delete the entry for distributed pipe host */
        DeleteHost(&l_stDistributedPipeHost, TCP_HOST);      
      
    } else  
        if (WRITE_ACCESS_MODE == a_nAccessMode)
        {
            DistPipeDebug(GENERAL_DEBUG, "Opening Pipe [%s] in WRITE mode\n", a_pszPipeName);

	        /* Write end of the pipe */
	        /* enquiry about the machine name and the port in which the read end is waiting for me to write to the DISTRIBUTED PIPE */

	        DISTPIPES_CALLOC(l_pszMachineName, char, HOSTNAME_LENGTH);

	        DistPipeDebug(GENERAL_DEBUG, "Blocked At Read Machine Name from local coordinator \n");


	        /* Read the Read end of the machine name */
	        if ((l_nErrorCode = ReadXDRString(LOCAL_USER_OPEN_PIPE, &g_stUserProcess, (void*) l_pszMachineName, HOSTNAME_LENGTH, UNIXSTREAM_HOST)) < 0)
	        {
                printf("CUPPED!!\n");
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            DISTPIPES_FREE(l_pszMachineName);
	            return l_nErrorCode;
	        }

	        DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_OPEN_PIPE] Pipe Name [%s] Read Machine [%s]\n", a_pszPipeName, l_pszMachineName);

	        /* Read the port number of Read end */
	        if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &g_stUserProcess, (void*) &l_nPortNumber, sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0) 
            {
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            DISTPIPES_FREE(l_pszMachineName);
	            return l_nErrorCode;
	        }
	
            DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_OPEN_PIPE] Pipe Name [%s] Read Port [%d]\n", a_pszPipeName, l_nPortNumber);

	        /* Create a TCP Client socket and connect to the server end */
	        if ((l_nErrorCode = CreateTCPClientHostAndConnect(&l_stDistributedPipeHost, l_pszMachineName, l_nPortNumber)) < 0)
	        {
	            l_nConnectStatus = CONNECT_READ_WRITE_PIPE_FAILED;
	            /* Inform the local coordinator that write end connection to read end not succeded */
	            if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_nConnectStatus, sizeof(l_nConnectStatus), UNIXSTREAM_HOST)) < 0)
	            {
	                DISTPIPES_FREE(l_pszMachineName);
	                DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	                return l_nErrorCode;
	            } 

	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            DISTPIPES_FREE(l_pszMachineName);
	            return l_nErrorCode;
	        }

	        l_nPipeDescriptor = l_stDistributedPipeHost.m_nTCPSocketDescriptor;

	        /* get the address in which the address is bind to */
	        if ((l_nPortNumber = GetHostPortAddress(&l_stDistributedPipeHost, TCP_HOST)) < 0)
	        {
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nPortNumber;
	        }

            l_nPortNumber = ntohs(l_nPortNumber);
	        DistPipeDebug(SOCKET_DEBUG, "Distributed Pipe [%s] Address : [%d] Write Pipe Descriptor : [%d] \n", a_pszPipeName, l_nPortNumber, l_stDistributedPipeHost.m_nTCPSocketDescriptor);

	        /* Update the entry for this distributed pipe */
	        UpdateEntry(g_pstListofDistributedPipes, a_pszPipeName, WRITE_ACCESS_MODE, g_stUserProcess.m_pszHostName, l_nPortNumber, g_stUserProcess.m_nHostProcessID, l_stDistributedPipeHost.m_nTCPSocketDescriptor); 

	        l_nConnectStatus = CONNECT_READ_WRITE_PIPE_SUCCEDED;
	        /* Inform the local coordinator that write end connection to read end not succeded */
	        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_nConnectStatus, sizeof(l_nConnectStatus), UNIXSTREAM_HOST)) < 0)
	        {
	            DISTPIPES_FREE(l_pszMachineName);
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nErrorCode;
	        } 

	        /* Must Send the Write details to the local coordinator here */
	        /* Send the details to the local coordinator */
	        /* Send Process ID */
	        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_stDistributedPipeHost.m_nHostProcessID, sizeof(l_stDistributedPipeHost.m_nHostProcessID), UNIXSTREAM_HOST)) < 0)
	        {
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nErrorCode;
	        }
	
	        /* Send Port Number */
	        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_nPortNumber, sizeof(l_nPortNumber), UNIXSTREAM_HOST)) < 0)
	        {
	            DISTPIPES_FREE(l_pszMachineName);
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nErrorCode;
	        }
	
	
	        /* Send the Socket Descriptor number for this pipe */
	        if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_OPEN_PIPE, &g_stUserProcess, (void*) &l_stDistributedPipeHost.m_nTCPSocketDescriptor, sizeof(l_stDistributedPipeHost.m_nTCPSocketDescriptor), UNIXSTREAM_HOST)) < 0)
	        {
	            DISTPIPES_FREE(l_pszMachineName);
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nErrorCode;
	        }

	
	        /* read the open read status from the local coordinator */
	        if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_OPEN_PIPE, &g_stUserProcess, (char*) &l_nPipeOpen, sizeof(l_nPipeOpen), UNIXSTREAM_HOST)) < 0)
	        {
	            DISTPIPES_FREE(l_pszMachineName);
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nErrorCode;
	        }
	        
            DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_OPEN_PIPE] Pipe Name [%s] Write Status [%d]\n", a_pszPipeName, l_nPipeOpen);
	
	        if (l_nPipeOpen < 0)
	        {
	            DISTPIPES_FREE(l_pszMachineName);
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            return l_nPipeOpen;
	        }
	
	        /* Should not close the pipe descriptor here */
	        DISTPIPES_FREE(l_stDistributedPipeHost.m_pszHostName);
	        DISTPIPES_FREE(l_stDistributedPipeHost.m_pstTCPHostAddress);
	        DISTPIPES_FREE(l_stDistributedPipeHost.m_pstUNIXStreamHostAddress);
	        DISTPIPES_FREE(l_stDistributedPipeHost.m_pszHostLogin);
	        DISTPIPES_FREE(l_stDistributedPipeHost.m_pszDaemonPath);

	        DISTPIPES_FREE(l_pszMachineName);
        } else
            {
	            DeleteHost(&l_stDistributedPipeHost, TCP_HOST);
	            /* Invalid Access Mode is specified */
	            return INVALID_ACCESS_MODE;
            }
 
        return l_nPipeDescriptor;
}

/************************************************************************************************************
 * Description : 
 *               Get the Number of Machines Available
 * Arguments   :
 *               a_nNumberofRequiredMachines- required machines total
 * Return      :
 *               Erro
 ************************************************************************************************************/
int GetLoadFactor(int a_nNumberofRequiredMachines)
{
  
  int l_nErrorCode;
  int l_nMessageID;
  int l_nAvailableMachines;

  /* Send the message to the local coordiantor */
  l_nMessageID = USER_LOCAL_GET_LOAD_FACTOR;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send the Number of Machnies required */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_GET_LOAD_FACTOR, &g_stUserProcess, 
					    (void*) &a_nNumberofRequiredMachines, 
					    sizeof(a_nNumberofRequiredMachines), 
					    UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_GET_LOAD_FACTOR, &g_stUserProcess, 
					   (char*) &l_nAvailableMachines, 
					   sizeof(l_nAvailableMachines), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_GET_LOAD_FACTOR] Available Machines [%d]\n",
		l_nAvailableMachines);

  return l_nAvailableMachines;

}
/************************************************************************************************************
 * Description : 
 *               Read from Distributed pipe 
 * Arguments   :
 *               a_nPipeDescriptor - Pipe Descriptor
 *               a_pszBuffer - Buffer to read into
 *               a_nBufferSize - Size of buffer to be read
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ReadDistPipe(a_nPipeDescriptor, a_pszBuffer, a_nBufferSize)
     int a_nPipeDescriptor;
     char* a_pszBuffer; 
     int a_nBufferSize;
{
  return ReadFromSocketDescriptor(a_nPipeDescriptor, a_pszBuffer, a_nBufferSize);
}

/************************************************************************************************************
 * Description : 
 *               Write to Distributed pipe 
 * Arguments   :
 *               a_nPipeDescriptor - Pipe Descriptor
 *               a_pszBuffer - Buffer to write
 *               a_nBufferSize - Size of buffer to be written
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int WriteDistPipe(a_nPipeDescriptor, a_pszBuffer, a_nBufferSize)
     int a_nPipeDescriptor;
     char* a_pszBuffer;
     int a_nBufferSize;
{
  return WriteToSocketDescriptor(a_nPipeDescriptor, a_pszBuffer, a_nBufferSize);
}

/************************************************************************************************************
 * Description : 
 *               Close Distributed pipe 
 * Arguments   :
 *               a_nPipeDescriptor - Pipe Descriptor
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CloseDistPipe(a_nPipeDescriptor)
     int a_nPipeDescriptor;
{
  return CloseSocketDescriptor(a_nPipeDescriptor);
}

/************************************************************************************************************
 * Description : 
 *               Purge Distributed pipe 
 * Arguments   :
 *               a_pszPipeName - Name of the distributed pipe
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DeleteDistPipe(a_pszPipeName)
     char* a_pszPipeName;
{

  int l_nErrorCode;
  int l_nPurgeReturn;
  int l_nMessageID;

  DeleteEntry(&g_pstListofDistributedPipes, a_pszPipeName);


  l_nMessageID = USER_LOCAL_PURGE_PIPE;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Local Coordinator that the pipe has been deleted */
  /* Send Pipe Name */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_PURGE_PIPE, &g_stUserProcess, 
					    (void*) a_pszPipeName, 
					    PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Read the status of deletion */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_PURGE_PIPE, &g_stUserProcess, 
					   (void*) &l_nPurgeReturn, 
					   sizeof(l_nPurgeReturn), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
	
  DistPipeDebug(USERTOLC_DEBUG, "[LOCAL_USER_PURGE_PIPE] Pipe Name [%s] deletion Status [%d]\n",
		a_pszPipeName, l_nPurgeReturn);

  return l_nPurgeReturn;

}

/************************************************************************************************************
 * Description : 
 *               clean the user process details
 * Arguments   :
 *               ---
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CleanUpUserProcess()
{

  int l_nErrorCode;
  int l_nMessageID;

  l_nMessageID = USER_LOCAL_GOODBYE;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

   /* Send USER_LOCAL_GOODBYE_MESSAGE */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_GOODBYE, &g_stUserProcess, 
					    (void*) &l_nErrorCode, 
					    sizeof(l_nErrorCode), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  DeleteHost(&g_stUserProcess, UNIXSTREAM_HOST);
  DeleteTable(g_pstListofDistributedPipes);
  
 
  return 1;
}





