/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeSystemCoordinator.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for system coordinator's functions.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 28/September/98
 *
 ***********************************************************************************************/


#include "DistPipeError.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeSystemCoordinator.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeNetLevel.h"
#include "DistPipeTable.h"
#include "DistPipeXDRRoutines.h"

#include "StProbMigration.h"
#include "Work.h"





int g_nNumberofLocalCoordinators = 0;
PWORK_INFORMATION_COLLECTION g_pstWorkCollection = NULL;
PHOST_DESCRIPTOR_COLLECTION g_pstListOfLocalCoordinators = NULL;
HOST_DESCRIPTOR g_stSystemCoordinator;
PDIST_PIPE_TABLE g_pstDistributedPipes = NULL;

char* g_pszProgramName;
int g_nWorkIdentifier= 0;

/* Prototypes that are not exported out */
int InitializeSystemCoordinator();
int AcceptMessages();
void CleanUpSystemCoordinator(int a_nSignalNumber);
int FindHostLoads(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, PWORK_INFORMATION_COLLECTION a_pstWorkInformation);
int LocalSystemHi(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemExistPipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemCreatePipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemOpenWriteMode(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, char* a_pszPipeName);
int LocalSystemOpenReadMode(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails, char* a_pszPipeName);
int LocalSystemOpenPipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemGetLoadFactor(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemPurgePipe(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);
int LocalSystemGoodBye(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetail);
int LocalSystemSteadyStateinitialize(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetail);
int LocalSystemSteadyStateMatrixInformation(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetail);
int LocalSystemSteadyStateNumberOfSplits(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetail);
int LocalSystemSteadyStateSplitInformation(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetail);
int ProcessLocalCoordinatorMessage(PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails);



/************************************************************************************************************
 * Description : 
 *               Main routine
 * Arguments   :
 *               ---
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int main(void)
{

  InitializeDistPipeDebug();
  
  InitializeSystemCoordinator();

  AcceptMessages();
  
  return 1;
}




/************************************************************************************************************
 * Description : 
 *               Intialization of the system coordinator, creates a TCP sockets and listens on it
 * Arguments   :
 *               ---
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int InitializeSystemCoordinator()
{


  int l_nErrorCode;

  
  /* Trap the termination signal so that when the
   * kill command is iisued to the process it does
   * a gracefull exit from the code */
#ifdef SIGTERM
  signal(SIGTERM, CleanUpSystemCoordinator);
#endif

  /* Get the Details of the Host */
  InitializeHost(&g_stSystemCoordinator, TCP_HOST);

  /* Create a Socket and Just Listen on it */
  if ((l_nErrorCode = CreateTCPServerHostAndListen(&g_stSystemCoordinator, SYSTEM_COORDINATOR_PORT)) < 0)
    {
      return l_nErrorCode;
    }
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Checks if the local coordinators are trying to contact
 * Arguments   :
 *               ---
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int AcceptMessages()
{  


  int l_nNewSocketDescriptor;
  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentLocalCoordinator;
  PHOST_DESCRIPTOR_COLLECTION l_pstNewLocalCoordinator;
  int l_nErrorCode;

  while (1)
    {

      /*
       * check if any local coordinators are trying to
       * make a new connection to the system coordinator
       * poll for 100 micro seconds and will not block
       * since the seconds is specified as 0
       */
      
      if (AnyDataToReadOnHost(&g_stSystemCoordinator, 0, 100, TCP_HOST))
	{

	  
	  /*
	   * A New Guy as joined the poll of machines
	   * Welcome buddy!!!!
	   */
	  if ((l_nNewSocketDescriptor = AcceptOnHost(&g_stSystemCoordinator, TCP_HOST)) < 0)
	    {
	      return l_nNewSocketDescriptor;
	    }
	  

	  /*
	   * Add the New fellow among other fellow mates
	   */
	  l_pstNewLocalCoordinator = AddNewHost(&g_pstListOfLocalCoordinators);
	  
	  

	  /*
	   * I will speak to the new fellow in this socket
	   */
	  l_pstNewLocalCoordinator->m_stHost.m_nTCPSocketDescriptor = l_nNewSocketDescriptor;
	  /*
	   * Find Out the New Guy
	   */
	   if ((l_nErrorCode = ProcessLocalCoordinatorMessage(l_pstNewLocalCoordinator)) < 0)
	     {
	       return l_nErrorCode;
	     }
	  
	  /*
	   * One more member in the family
	   */
	  
	  g_nNumberofLocalCoordinators++;
	  
	}
      
      
      /*
       * Find out any of the local coordinator have anything
       * to say 
       */
      l_pstCurrentLocalCoordinator = g_pstListOfLocalCoordinators;
      while (l_pstCurrentLocalCoordinator != NULL)
	{
	  /*
	   * Find if as any to say
	   */
	  if (AnyDataToReadOnHost(&(l_pstCurrentLocalCoordinator->m_stHost), 0, 100, TCP_HOST))
	    {
	      /*
	       * Yeah Local Coordinator has something to say
	       * Iam All ears!!!!!
	       */
	      if ((l_nErrorCode = ProcessLocalCoordinatorMessage(l_pstCurrentLocalCoordinator)) < 0)
		{
		  return l_nErrorCode;
		}
	    } 
	  
	  l_pstCurrentLocalCoordinator = GetNextHost(l_pstCurrentLocalCoordinator);
	}
      

    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               this deletes all the memory allocated during the execution of the sytstem coordiantor
 * Arguments   :
 *               a_nSignalNumber - Number of signal because which this rotuine was called
 * Return      :
 *               None
 ************************************************************************************************************/
void CleanUpSystemCoordinator(a_nSignalNumber)
     int a_nSignalNumber;
{

  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentLocalCoordinator;
  PHOST_DESCRIPTOR_COLLECTION l_pstNextLocalCoordinator;
  /*
   * Delete all the local coordinator 
   * ifnormation that are existing now
   */
  
  l_pstCurrentLocalCoordinator = g_pstListOfLocalCoordinators;
  while (l_pstCurrentLocalCoordinator != NULL)
    {

      l_pstNextLocalCoordinator = GetNextHost(l_pstCurrentLocalCoordinator);
      DeleteHost(&l_pstCurrentLocalCoordinator->m_stHost, TCP_HOST);
      DISTPIPES_FREE(l_pstCurrentLocalCoordinator);
      l_pstCurrentLocalCoordinator = l_pstNextLocalCoordinator;
    }


  CloseHost(&g_stSystemCoordinator, TCP_HOST);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pszHostName);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pstTCPHostAddress);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pszHostLogin);
  DISTPIPES_FREE(g_stSystemCoordinator.m_pszDaemonPath);

  DeleteAllSplitInformation(&g_pstWorkCollection);

  exit(1);

}

/************************************************************************************************************
 * Description : 
 *              Split the work into sun tasks and update the work information
 * Arguments   :
 *               a_pstHostDetails - Pointer to host details
 *               a_pstWorkInformation - pointer to work information
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int SplitWork(a_pstWorkInformation)
      PWORK_INFORMATION_COLLECTION a_pstWorkInformation;
{
  
  int l_nNumberOfSplits = 0;
  int l_nStart;
  int l_nIncrement;
  int l_nSplitIndex;
  unsigned long int l_ulTotalLoad;

  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentLocalCoordinator;


  l_pstCurrentLocalCoordinator = g_pstListOfLocalCoordinators;
  while (l_pstCurrentLocalCoordinator != NULL)
    {
    
      /* Must get a good idea to split the load on the machines
       * depending on the load of individual machines */

      l_ulTotalLoad = l_ulTotalLoad + l_pstCurrentLocalCoordinator->m_stHost.m_ulLoad;
      l_nNumberOfSplits++;

      l_pstCurrentLocalCoordinator = GetNextHost(l_pstCurrentLocalCoordinator);
    }


  /* Allocate Meomory for split information */
  SetNumberOfSplitInformation(a_pstWorkInformation, l_nNumberOfSplits);
  
  /* Store the split information */
  l_pstCurrentLocalCoordinator = g_pstListOfLocalCoordinators;
  l_nStart = 0;
  l_nIncrement = a_pstWorkInformation->m_pstWorkDetail->m_nMaximumColumns / l_nNumberOfSplits;
  for(l_nSplitIndex = 0; l_nSplitIndex < l_nNumberOfSplits; l_nSplitIndex++)
    {
      /* For the Time being equally divided the work among
	 the machines */
      SetSplitInformation(a_pstWorkInformation, l_nSplitIndex, &l_pstCurrentLocalCoordinator->m_stHost,
			  l_nStart, l_nIncrement);

      l_nStart = l_nStart + l_nIncrement;      
      l_pstCurrentLocalCoordinator = GetNextHost(l_pstCurrentLocalCoordinator);
    }

  return l_nNumberOfSplits;

}


/************************************************************************************************************
 * Description : 
 *               Find the loads on each of the machines
 * Arguments   :
 *               a_pstHostDetails - Pointer to host details
 *               a_pstWorkInformation - pointer to work information
 * Return      :
 *               Error Code / 1
 ************************************************************************************************************/
int FindHostLoads(a_pstHostDetails, a_pstWorkInformation)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     PWORK_INFORMATION_COLLECTION a_pstWorkInformation;
{

  int l_nErrorCode;
  int l_nMessageID;


  PHOST_DESCRIPTOR_COLLECTION l_pstCurrentLocalCoordinator;

  /* Send the Message to each of the host and find their loads */
  l_pstCurrentLocalCoordinator = g_pstListOfLocalCoordinators;
  while (l_pstCurrentLocalCoordinator != NULL)
    {
      
      if (l_pstCurrentLocalCoordinator != a_pstHostDetails)
	{
	  /* Send the Message to the Local coordinator to find the load */
	  l_nMessageID = SYSTEM_LOCAL_FIND_LOAD;
	  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_ANY_MESSAGE, &l_pstCurrentLocalCoordinator->m_stHost, 
						    (void*) &l_nMessageID, 
						    sizeof(l_nMessageID), TCP_HOST)) < 0)
	    {
	      return l_nErrorCode;
	    }

	  /* Get the Load returned by the coordinator */
	  if ((l_nErrorCode = ReadXDRUnSignedLongInteger(LOCAL_SYSTEM_FIND_LOAD, &l_pstCurrentLocalCoordinator->m_stHost, 
						   (void*) &l_pstCurrentLocalCoordinator->m_stHost.m_ulLoad, 
						   sizeof(l_pstCurrentLocalCoordinator->m_stHost.m_ulLoad), 
						   TCP_HOST)) < 0)
	    {
	      return l_nErrorCode;
	    }
	  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_FIND_LOAD) [%s] Load [%d] \n", 
			l_pstCurrentLocalCoordinator->m_stHost.m_pszHostName, 
			l_pstCurrentLocalCoordinator->m_stHost.m_ulLoad);
	}


      l_pstCurrentLocalCoordinator = GetNextHost(l_pstCurrentLocalCoordinator);
    }



  return 1;

}



/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Hi Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pstReceivedMessage - Message that is Received from the system coordiantor 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemHi(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nMessageID;

  /*
   * Hi Message from local coordinator
   */
  
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszHostName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszHostLogin, char, LOGINNAME_LENGTH);
  DISTPIPES_CALLOC(a_pstHostDetails->m_stHost.m_pszDaemonPath, char, PATH_LENGTH);
  
  /* Read Host name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_HI, &a_pstHostDetails->m_stHost, (void* ) a_pstHostDetails->m_stHost.m_pszHostName, 
				    HOSTNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_HI) From %s \n", a_pstHostDetails->m_stHost.m_pszHostName);
  
  /* Read Architecture */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_HI, &a_pstHostDetails->m_stHost, 
					   (void*) &a_pstHostDetails->m_stHost.m_nHostArchitecture, 
					   sizeof(a_pstHostDetails->m_stHost.m_nHostArchitecture), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_HI) [%s] Architecture [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, a_pstHostDetails->m_stHost.m_nHostArchitecture);
  
  /* Read Address */
  if ((l_nErrorCode = ReadMessageAndData(LOCAL_SYSTEM_HI, (char*) a_pstHostDetails->m_stHost.m_pstTCPHostAddress, 
					 &a_pstHostDetails->m_stHost, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Read ProcessID */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_HI, &a_pstHostDetails->m_stHost, (char*) &a_pstHostDetails->m_stHost.m_nHostProcessID, 
					 sizeof(a_pstHostDetails->m_stHost.m_nHostProcessID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_HI) [%s] ProcessID [%d] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_nHostProcessID);
  
  /* Read Login Name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_HI, &a_pstHostDetails->m_stHost, (void*) a_pstHostDetails->m_stHost.m_pszHostLogin, 
				    LOGINNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_HI) [%s] Login Name [%s] \n", a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_pszHostLogin);
  
  /* Read Daemon Path */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_HI, &a_pstHostDetails->m_stHost, (void*) a_pstHostDetails->m_stHost.m_pszDaemonPath, 
				    PATH_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_HI) [%s] Daemon Path [%s] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, a_pstHostDetails->m_stHost.m_pszDaemonPath);       
  
  /* Send a Hi Message to Local Coordinator */
  l_nMessageID = SYSTEM_LOCAL_HI;
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_ANY_MESSAGE, &a_pstHostDetails->m_stHost, (void*) &l_nMessageID, 
					    sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  

  /* Send the Local Coordinator Details of System Coordinator */
  
  /* Send Host Name */
  if ((l_nErrorCode = WriteXDRString(SYSTEM_LOCAL_HI, &a_pstHostDetails->m_stHost, (void*) g_stSystemCoordinator.m_pszHostName, 
				     HOSTNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Architecture */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_HI, &a_pstHostDetails->m_stHost, (void*) &g_stSystemCoordinator.m_nHostArchitecture, 
					  sizeof(g_stSystemCoordinator.m_nHostArchitecture), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Address */
  if ((l_nErrorCode = WriteMessageAndData(SYSTEM_LOCAL_HI, (char*) g_stSystemCoordinator.m_pstTCPHostAddress,
					  sizeof(struct sockaddr_in), &a_pstHostDetails->m_stHost, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Process Id */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_HI, &a_pstHostDetails->m_stHost, 
					    (void* ) &g_stSystemCoordinator.m_nHostProcessID,
					    sizeof(g_stSystemCoordinator.m_nHostProcessID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Login Name*/
  if ((l_nErrorCode = WriteXDRString(SYSTEM_LOCAL_HI, &a_pstHostDetails->m_stHost, (void*) g_stSystemCoordinator.m_pszHostLogin,
				     LOGINNAME_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send Daemon path*/
  if ((l_nErrorCode = WriteXDRString(SYSTEM_LOCAL_HI, &a_pstHostDetails->m_stHost, (char*) g_stSystemCoordinator.m_pszDaemonPath,
				     PATH_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Exist Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pstReceivedMessage - Message that is Received from the system coordiantor 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemExistPipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{
  
  int l_nErrorCode;
  int l_nPipeExist;
  
  char* l_pszPipeName;

  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  /* Read Pipe name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_EXIST_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }  
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_EXIST_PIPE) From [%s] PipeName [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  
  /* Must Find if the Pipe Exist from the table and return the Error code */
  if (NULL == IfPipeExists(g_pstDistributedPipes, l_pszPipeName))
    l_nPipeExist = PIPENAME_DOESNOT_EXIST;
  else
    l_nPipeExist = PIPENAME_DOES_EXIST;
  
  /* Write the Pipe Descriptor the Local coordiantor */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_EXIST_PIPE, &a_pstHostDetails->m_stHost, (void*)&l_nPipeExist,
					    sizeof(l_nPipeExist), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }

  
  DISTPIPES_FREE(l_pszPipeName);
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Create Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pstReceivedMessage - Message that is Received from the system coordiantor 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemCreatePipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nPipeCreate;
  
  char* l_pszPipeName;
  
  
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  /* Read Pipe name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_CREATE_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_CREATE_PIPE) From [%s] PipeName [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  
  
  /* Must create an entry in the pipe table entry */
  l_nPipeCreate = AppendTable(&g_pstDistributedPipes, a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  
  /* Write the error code status to the local coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_CREATE_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPipeCreate,
					  sizeof(l_nPipeCreate), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  DISTPIPES_FREE(l_pszPipeName);

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordinator To System Coordinator Open Pipe Write mode
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pszPipeName - Name of the Pipe which need to be opened in write mode
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemOpenWriteMode(a_pstHostDetails, a_pszPipeName)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     char* a_pszPipeName;
{

    int l_nErrorCode;
    int l_nSocketDescriptor;
    int l_nPortNumber;
    int l_nUserProcessID;
    int l_nWriteStatus;
    int l_nConnectStatus;
    int l_nPipeOpen;

    PDIST_PIPE_TABLE l_pstDistPipe;

    /* Check if the Read entry is a valid entry */
    if ((l_pstDistPipe = IfPipeExists(g_pstDistributedPipes, a_pszPipeName)) == NULL) {
        /* pipe is not found */		
    } else {
        if (1 == l_pstDistPipe->m_stDistPipeEntry.m_nReadEntryValid) {
	        /* read has happened */
	        l_nWriteStatus = OPEN_PIPE_WRITE_NOT_PENDING;
	        /* Send to the local coordinator that read has happened */
	        if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nWriteStatus, sizeof(l_nWriteStatus), TCP_HOST)) < 0) {
	            DISTPIPES_FREE(a_pszPipeName);
	            return l_nErrorCode;
	        }
	        /* Send the Machine Name and the Port number it is listening to local coordinator */
	        /* send the read machine name */
	        if ((l_nErrorCode = WriteXDRString(SYSTEM_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pstDistPipe->m_stDistPipeEntry.m_pszReadMachineName, HOSTNAME_LENGTH, TCP_HOST)) < 0)
	        {
	            DISTPIPES_FREE(a_pszPipeName);
	            return l_nErrorCode;
	        }
	        /* send the read port number */
	        if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_pstDistPipe->m_stDistPipeEntry.m_nReadPortNumber, sizeof(l_pstDistPipe->m_stDistPipeEntry.m_nReadPortNumber), TCP_HOST)) < 0) {
	            DISTPIPES_FREE(a_pszPipeName);
	            return l_nErrorCode;
	        }
	  
	        /* Read the details of the connected pipe so that it can update the details */
	        /* Read the status of the conenction of the distributed pipe from the system coordinator */
	        if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nConnectStatus, sizeof(l_nConnectStatus), TCP_HOST)) < 0) {
	            DISTPIPES_FREE(a_pszPipeName);
	            return l_nErrorCode;
	        }
	        DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Connection Status [%d]\n", a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nConnectStatus);
	  
	        if (CONNECT_READ_WRITE_PIPE_FAILED == l_nConnectStatus) {
	        /* connect has failed */
	      
	        } else {
	            /* Read the Process Id of the user process */
	            if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nUserProcessID, sizeof(l_nUserProcessID), TCP_HOST)) < 0) {
		            DISTPIPES_FREE(a_pszPipeName);
		            return l_nErrorCode;
		        }
        	    
                DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Process Id [%d]\n", a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nUserProcessID);
	      
	            /* Read the Port Number of the user process */
	            if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPortNumber, sizeof(l_nPortNumber), TCP_HOST)) < 0) {
		            DISTPIPES_FREE(a_pszPipeName);
		            return l_nErrorCode;
		        }
	                
                DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Port Number [%d]\n", a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nPortNumber);
	      
	            /* Read the Socket Descriptor of the user process */
	            if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nSocketDescriptor, sizeof(l_nSocketDescriptor), TCP_HOST)) < 0) {
		            DISTPIPES_FREE(a_pszPipeName);
		            return l_nErrorCode;
		        }
	            DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Socket Descriptor [%d]\n", a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nSocketDescriptor);
	      
	            /* Must update the entry in the user level pipes */
	            l_nPipeOpen = UpdateEntry(g_pstDistributedPipes, a_pszPipeName, WRITE_ACCESS_MODE, a_pstHostDetails->m_stHost.m_pszHostName, l_nPortNumber, l_nUserProcessID, l_nSocketDescriptor);
	     
	            /* send the status of the opening */
	            /* Write the error code status to the local coordinator */
	            if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPipeOpen, sizeof(l_nPipeOpen), TCP_HOST)) < 0) {
		            DISTPIPES_FREE(a_pszPipeName);
		            return l_nErrorCode;
		        }	  
	        }
        } else {
            /* Read has not yet happened */
	        /* set write pending */
	        l_pstDistPipe->m_stDistPipeEntry.m_nWritePending = 1;
	        l_pstDistPipe->m_stDistPipeEntry.m_pstLCWritePending = a_pstHostDetails;
	        l_nWriteStatus = OPEN_PIPE_WRITE_PENDING;
	        /* Send to the local coordinator that read has happened */
	        if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_CREATE_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nWriteStatus, sizeof(l_nWriteStatus), TCP_HOST)) < 0) {
	            DISTPIPES_FREE(a_pszPipeName);
	            return l_nErrorCode;
	        }
	    }
    }

    DISTPIPES_FREE(a_pszPipeName);
    return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordinator To System Coordinator Open Pipe Read mode
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pszPipeName - Name of the Pipe which need to be opened in write mode
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemOpenReadMode(a_pstHostDetails, a_pszPipeName)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     char* a_pszPipeName;
{

  int l_nWritePendingStatus;
  int l_nErrorCode;
  int l_nSocketDescriptor;
  int l_nPortNumber;
  int l_nUserProcessID;
  int l_nPipeOpen;
  int l_nMessageID;


  PDIST_PIPE_TABLE l_pstDistPipe;

  /* Read the Process Id of the user process */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nUserProcessID, 
					   sizeof(l_nUserProcessID), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Process Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nUserProcessID);
  
  /* Read the Port Number of the user process */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPortNumber, 
					   sizeof(l_nPortNumber), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Port Number [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nPortNumber);
  
  
  /* Read the Socket Descriptor of the user process */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nSocketDescriptor, 
					   sizeof(l_nSocketDescriptor), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Socket Descriptor [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, a_pszPipeName, l_nSocketDescriptor);
  
  
  /* Must update the entry in the user level pipes */
  l_nPipeOpen = UpdateEntry(g_pstDistributedPipes, a_pszPipeName, READ_ACCESS_MODE, a_pstHostDetails->m_stHost.m_pszHostName,
			    l_nPortNumber, l_nUserProcessID, l_nSocketDescriptor);
  
  /* send the status of the opening */
  /* Write the error code status to the local coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPipeOpen,
					    sizeof(l_nPipeOpen), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(a_pszPipeName);
      return l_nErrorCode;
    }
  
  /* check if any process waiting for this pipe to opened in 
   * read mode has to be awakened */
  if ((l_pstDistPipe = IfPipeExists(g_pstDistributedPipes, a_pszPipeName)) == NULL)
    {
      /* Pipe Entry is not Found */
    }
  else
    {
      /* Pipe Entry is Found */
      /* Check if the write for this pipe is pending */
      if (1 == l_pstDistPipe->m_stDistPipeEntry.m_nWritePending)
	{
	  /* Write is pending */
	  
	  /* Inform the local coordinator that the user process with
	   * pipe opened for writing can be unblocked */
	  
	  l_nWritePendingStatus = OPEN_PIPE_WRITE_PENDING;
	  /* Send a Pending Message */
	  l_nMessageID = SYSTEM_LOCAL_WRITE_PENDING_PIPE;
	  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_ANY_MESSAGE, &l_pstDistPipe->m_stDistPipeEntry.m_pstLCWritePending->m_stHost, (void*) &l_nMessageID,
						    sizeof(l_nMessageID), TCP_HOST)) < 0)
	    {
	      DISTPIPES_FREE(a_pszPipeName);
	      return l_nErrorCode;
	    }
	  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_WRITE_PENDING_PIPE, &l_pstDistPipe->m_stDistPipeEntry.m_pstLCWritePending->m_stHost, (void*) &l_nWritePendingStatus,
						    sizeof(l_nWritePendingStatus), TCP_HOST)) < 0)
	    {
	      DISTPIPES_FREE(a_pszPipeName);
	      return l_nErrorCode;
	    }
	  
	  /* Send the pipe name for which the write is pending */
	  /* Send the status to the local coordiantor */
	  if ((l_nErrorCode = WriteXDRString(SYSTEM_LOCAL_WRITE_PENDING_PIPE,  
					     &l_pstDistPipe->m_stDistPipeEntry.m_pstLCWritePending->m_stHost, (void*) a_pszPipeName,
					     PIPENAME_LENGTH, TCP_HOST)) < 0)
	    {
	      DISTPIPES_FREE(a_pszPipeName);
	      return l_nErrorCode;
	    }

	  return LocalSystemOpenWriteMode(l_pstDistPipe->m_stDistPipeEntry.m_pstLCWritePending, a_pszPipeName);
	}
    }

  DISTPIPES_FREE(a_pszPipeName);
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Open Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 *               a_pstReceivedMessage - Message that is Received from the system coordiantor 
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemOpenPipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nAccessMode;

  char* l_pszPipeName;

  
  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  /* Read Pipe name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_EXIST_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }  
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);

  /* Access Mode */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_OPEN_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nAccessMode, 
					   sizeof(l_nAccessMode), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_OPEN_PIPE) From [%s] PipeName [%s] Acess Mode [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName, l_nAccessMode);
  
  if (READ_ACCESS_MODE == l_nAccessMode)
    {
      return LocalSystemOpenReadMode(a_pstHostDetails, l_pszPipeName);
    }
  else
    if (WRITE_ACCESS_MODE == l_nAccessMode)
      {
	return LocalSystemOpenWriteMode(a_pstHostDetails, l_pszPipeName);	
      }
    else
      {
	/* Invalid Access Mode Read */
	DISTPIPES_FREE(l_pszPipeName);	
      }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Get Load Factor
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemGetLoadFactor(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{
  
  int l_nErrorCode;
  int l_nNumberofAvailableMachines;
  int l_nNumberofMachines;

  /* Read the Number of Machines that are required */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_GET_LOAD_FACTOR, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberofMachines, 
					   sizeof(l_nNumberofMachines), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_GET_LOAD_FACTOR) From [%s] Number of Machines Required [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nNumberofMachines);

  /* Send Number of machines available */
  l_nNumberofAvailableMachines = g_nNumberofLocalCoordinators;
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_GET_LOAD_FACTOR, &a_pstHostDetails->m_stHost,
					    (void*) &l_nNumberofAvailableMachines,
					    sizeof(l_nNumberofAvailableMachines), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Purge Pipe Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemPurgePipe(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nPurgeReturn;

  char* l_pszPipeName;

  DISTPIPES_CALLOC(l_pszPipeName, char, PIPENAME_LENGTH);
  
  /* Read Pipe name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_EXIST_PIPE, &a_pstHostDetails->m_stHost, (void*) l_pszPipeName, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_PURGE_PIPE) From [%s] PipeName [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszPipeName);
  
  l_nPurgeReturn = DeleteEntry(&g_pstDistributedPipes, l_pszPipeName);
  
  /* Send the status to the local coordiantor */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_PURGE_PIPE, &a_pstHostDetails->m_stHost, (void*) &l_nPurgeReturn,
					  sizeof(l_nPurgeReturn), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszPipeName);
      return l_nErrorCode;
    }
  
  
  DISTPIPES_FREE(l_pszPipeName);
      
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator GoodBye Message Routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemGoodBye(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{
  /* One Local Coordinator Bidding Farewell */
  g_nNumberofLocalCoordinators--;
  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Steady State Problem Initialization
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemSteadyStateInitialize(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;

  /* Increment the Work Identification */
  /* Since it is integer the wrap around time for this is very high */
  g_nWorkIdentifier++;


  /* Intialize the Data Structure to store the work information*/
  AddSplitInformation(&g_pstWorkCollection, g_nWorkIdentifier);


  /* Send the Work Indetifier to local coordinator */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_INITIALIZE, &a_pstHostDetails->m_stHost, 
					    (void*) &g_nWorkIdentifier,
					    sizeof(g_nWorkIdentifier), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Steady State Problem Matrix Information collection routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemSteadyStateMatrixInformation(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nWorkID;
  int l_nNumberOfRows;
  int l_nNumberOfColumns;
  int l_nStatus;

  PWORK_INFORMATION_COLLECTION l_pstWorkInformation;


  /* Read WorkId from system coordinator */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION) From [%s] Work Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

  /* Read Number of Rows */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberOfRows, 
					   sizeof(l_nNumberOfRows), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION) From [%s] Number of Rows [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nNumberOfRows);

   /* Read Number of Columns */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nNumberOfColumns, 
					   sizeof(l_nNumberOfColumns), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION) From [%s] Number of Columns [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nNumberOfColumns);


  l_pstWorkInformation = FindWork(g_pstWorkCollection, l_nWorkID);

  l_nStatus = UpdateRowColumnForWork(l_pstWorkInformation, l_nNumberOfRows, l_nNumberOfColumns);

  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nStatus,
					    sizeof(l_nStatus), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Steady State Problem Number of Split routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemSteadyStateNumberOfSplits(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nWorkID;
  int l_nNumberOfSplits;

  PWORK_INFORMATION_COLLECTION l_pstWorkInformation;

  /* Read WorkId */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS) From [%s] Work Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);
  


  /* Read the Load of this machine */
  if ((l_nErrorCode = ReadXDRUnSignedLongInteger(LOCAL_SYSTEM_FIND_LOAD, &a_pstHostDetails->m_stHost, 
						 (void*) &a_pstHostDetails->m_stHost.m_ulLoad, 
						 sizeof(a_pstHostDetails->m_stHost.m_ulLoad), 
						 TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_FIND_LOAD) [%s] Load [%d] \n", 
		a_pstHostDetails->m_stHost.m_pszHostName, 
		a_pstHostDetails->m_stHost.m_ulLoad);
  
  l_pstWorkInformation = FindWork(g_pstWorkCollection, l_nWorkID);

  /* Routine to find and set the Number of Splits */
  /* Send a Message to all the local coordinators and find the load in each of them */
  FindHostLoads(a_pstHostDetails, l_pstWorkInformation);

  /* Set the Split Informations */
  l_nNumberOfSplits = SplitWork(l_pstWorkInformation);

  /* Send Number of splits */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nNumberOfSplits,
					    sizeof(l_nNumberOfSplits), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Steady State Problem Number of Split routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemSteadyStateSplitInformation(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nWorkID;
  int l_nSplitID;
  int l_nStartRow;
  int l_nNumberOfRows;

  PWORK_INFORMATION_COLLECTION l_pstWorkInformation;

   /* Read WorkId */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION) From [%s] Work Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

   /* Read SpiltId from  */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nSplitID, 
					   sizeof(l_nSplitID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION) From [%s] Split Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nSplitID);



  l_pstWorkInformation = FindWork(g_pstWorkCollection, l_nWorkID);

  /* Get the Start Row and the number of Row for this work and split */
  FindStartAndNumberOfRowsForSplit(l_pstWorkInformation, l_nSplitID, &l_nStartRow, &l_nNumberOfRows);
  
  /* Send Start Row */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nStartRow,
					    sizeof(l_nStartRow), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send Number of Row */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nNumberOfRows,
					    sizeof(l_nNumberOfRows), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;

}


/************************************************************************************************************
 * Description : 
 *               Local Coordiantor to System Coordinator Steady State Problem Number of Split routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int LocalSystemSteadyStateSendWork(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{

  int l_nErrorCode;
  int l_nWorkID;
  int l_nSplitID;
  int l_nSubTaskID;
  int l_nMigrated;

  char* l_pszResultPipe;

  PWORK_INFORMATION_COLLECTION l_pstWorkInformation;


   /* Read WorkId */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] Work Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nWorkID);

   /* Read SpiltId from  */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					   (void*) &l_nSplitID, 
					   sizeof(l_nSplitID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] Split Id [%d]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_nSplitID);


  DISTPIPES_CALLOC(l_pszResultPipe, char, PIPENAME_LENGTH);

  /* Read the Result Pipe Name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
				    (void*) l_pszResultPipe, 
				    PIPENAME_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszResultPipe);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] Result Pipe Name [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszResultPipe);

  DISTPIPES_FREE(l_pszResultPipe);

  l_pstWorkInformation = FindWork(g_pstWorkCollection, l_nWorkID);

  /* Do the migration of the file to the other end */
  l_nMigrated = DoMigration(a_pstHostDetails, l_pstWorkInformation, l_nSplitID);
 
  if (l_nMigrated < 0)
    l_nSubTaskID = -1;
  else
    l_nSubTaskID = l_nSplitID + 1;

  /* Send the Sub Task Id */
  if ((l_nErrorCode = WriteXDRSignedInteger(SYSTEM_LOCAL_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
					    (void*) &l_nSubTaskID,
					    sizeof(l_nSubTaskID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Local Coordiantor Message handler routine
 * Arguments   :
 *               a_pstHostDetails - Pointer to the host details in the link list maintained by the system coordinator
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int ProcessLocalCoordinatorMessage(a_pstHostDetails)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
{


  int l_nMessageID;
  int l_nErrorCode;
  
  /*
   * Find what the local coordinator wants to say
   */
  
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_SYSTEM_ANY_MESSAGE, &a_pstHostDetails->m_stHost, 
					   (void *) &l_nMessageID, 
					   sizeof(l_nMessageID), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  switch (l_nMessageID)
    {

    case LOCAL_SYSTEM_HI :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_HI] Message From [%s]\n", a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemHi(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_GET_LOAD_FACTOR :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_GET_LOAD_FACTOR] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemGetLoadFactor(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_CREATE_PIPE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_CREATE_PIPE] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemCreatePipe(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_EXIST_PIPE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_EXIST_PIPE] Message From [%s]\n", a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemExistPipe(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_OPEN_PIPE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_OPEN_PIPE] Message From [%s]\n", a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemOpenPipe(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_PURGE_PIPE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_PURGE_PIPE] Message From [%s]\n", a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemPurgePipe(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_GOODBYE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_GOODBYE] Message From [%s]\n", a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemGoodBye(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_STEADY_STATE_INITIALIZE :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_STEADY_STATE_INITIALIZE] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemSteadyStateInitialize(a_pstHostDetails);
      break;
      
    case LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_STEADY_STATE_MATRIX_INFORMATION] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemSteadyStateMatrixInformation(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_STEADY_STATE_NUMBER_OF_SPLITS] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemSteadyStateNumberOfSplits(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_STEADY_STATE_SPLIT_INFORMATION] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemSteadyStateSplitInformation(a_pstHostDetails);
      break;

    case LOCAL_SYSTEM_STEADY_STATE_SEND_WORK :
      DistPipeDebug(LCTOSC_DEBUG, "Received [LOCAL_SYSTEM_STEADY_STATE_SEND_WORK] Message From [%s]\n", 
		    a_pstHostDetails->m_stHost.m_pszHostName);
      return LocalSystemSteadyStateSendWork(a_pstHostDetails);
      break;

    }  

  return 1;

}
