/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/

/*******************************************************************
*
* File Name	: DistPipeTable.c
* Purpose	: This file contains the implementation of the
*		  data structure for distributed pipe table
*		  maintained at the System Coordinator and all its
*		  interfaces.
* 
* Developers	: Binu K.J. and Karthikeyan R.
* Known Bugs	: 
* Date		: 3rd October 1998
* Modifications : 
* Modification Details	:
*
*********************************************************************/


#include <stdio.h>
#include "DistPipeSystemCoordinator.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeError.h"
#include "DistPipeNetLevel.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeTable.h"


/*****************************************************************
* Function : PrintTable
* Purpose  : To print partial contents of the PipeTable
* Args	   : PDIST_PIPE_TABLE
* Return Value : None 
* Known Bugs :
*****************************************************************/


int PrintTable(PDIST_PIPE_TABLE a_pstDistPipeTable)
{
  int l_nSerialNo;

  l_nSerialNo = 0;
  while (a_pstDistPipeTable != NULL)
    {
      l_nSerialNo++;
      printf("\n %d M/c Name: %s; PipeName: %s \n", l_nSerialNo, 
	     a_pstDistPipeTable->m_stDistPipeEntry.m_pszCreateMachineName, a_pstDistPipeTable->m_stDistPipeEntry.m_pszPipeName);
      a_pstDistPipeTable = a_pstDistPipeTable->m_pstNext;
    }

  return 1;
}


/*****************************************************************
* Function : PrintFullTable
* Purpose  : To print all the contents of the PipeTable
* Args	   : PDIST_PIPE_TABLE
* Return Value : None 
* Known Bugs :
*****************************************************************/


int PrintFullTable(a_pstDistPipeTable)
     PDIST_PIPE_TABLE a_pstDistPipeTable;
{
  int l_nSerialNo=0;
  
  while (a_pstDistPipeTable != NULL)
    {
      l_nSerialNo++;
      printf("\n %d %s %s %s %d %d %s %d %d\n", l_nSerialNo, \
	     a_pstDistPipeTable->m_stDistPipeEntry.m_pszCreateMachineName, a_pstDistPipeTable->m_stDistPipeEntry.m_pszPipeName, \
	     a_pstDistPipeTable->m_stDistPipeEntry.m_pszReadMachineName, a_pstDistPipeTable->m_stDistPipeEntry.m_nReadProcessId, \
	     a_pstDistPipeTable->m_stDistPipeEntry.m_nReadPortNumber, a_pstDistPipeTable->m_stDistPipeEntry.m_pszWriteMachineName, \
	     a_pstDistPipeTable->m_stDistPipeEntry.m_nWritePortNumber, a_pstDistPipeTable->m_stDistPipeEntry.m_nWriteProcessId);
      
      a_pstDistPipeTable = a_pstDistPipeTable->m_pstNext;
    }

  return 1;
}



/*****************************************************************
* Function Name : AppendTable
* Purpose 	: Append an entry to the table.
* Args 		: 
* Return Value	: 
* Purpose 	: To Append the table entry
******************************************************************/


int AppendTable(a_pstDistPipeTable, a_pszCreateMachineName, a_pszPipeName)
     PDIST_PIPE_TABLE* a_pstDistPipeTable;
     char* a_pszCreateMachineName;
     char* a_pszPipeName;
{

  PDIST_PIPE_TABLE l_pstCurrent;

  DistPipeDebug(GENERAL_DEBUG, "Create Machine Name [%s] & Pipe Name [%s]\n", a_pszCreateMachineName, a_pszPipeName);

  if (NULL != IfPipeExists(*a_pstDistPipeTable, a_pszPipeName))
    return CREATE_PIPENAME_ALREADY_EXIST;
 

  DISTPIPES_CALLOC(l_pstCurrent, DIST_PIPE_TABLE, 1);
 
  DISTPIPES_CALLOC(l_pstCurrent->m_stDistPipeEntry.m_pszCreateMachineName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(l_pstCurrent->m_stDistPipeEntry.m_pszPipeName, char, PIPENAME_LENGTH);
  
  DISTPIPES_CALLOC(l_pstCurrent->m_stDistPipeEntry.m_pszReadMachineName, char, HOSTNAME_LENGTH);
  DISTPIPES_CALLOC(l_pstCurrent->m_stDistPipeEntry.m_pszWriteMachineName, char, HOSTNAME_LENGTH);

  strcpy(l_pstCurrent->m_stDistPipeEntry.m_pszCreateMachineName, a_pszCreateMachineName);
  strcpy(l_pstCurrent->m_stDistPipeEntry.m_pszPipeName, a_pszPipeName);


  l_pstCurrent->m_pstNext = NULL;
  if (NULL == *a_pstDistPipeTable)
    {
      *a_pstDistPipeTable = l_pstCurrent;
    }
  else
    {
     l_pstCurrent->m_pstNext = *a_pstDistPipeTable;
     *a_pstDistPipeTable = l_pstCurrent;
    }

  l_pstCurrent->m_stDistPipeEntry.m_nReadEntryValid = 0;
  l_pstCurrent->m_stDistPipeEntry.m_nWritePending = 0;

  return CREATE_PIPENAME_DONE;
}


/*****************************************************************
* Function Name : IfPipeExists
* Purpose	: To see if a named Pipe Name already has an entry
*			in the table.
* Return Value 	: 1 if it already exist in the table, 0 othewise.
*****************************************************************/


PDIST_PIPE_TABLE IfPipeExists(a_pstDistPipeTable, a_pszPipeName)
     PDIST_PIPE_TABLE a_pstDistPipeTable;
     char* a_pszPipeName;
{

  PDIST_PIPE_TABLE l_pstCurrent;

  l_pstCurrent = a_pstDistPipeTable;
  while ((l_pstCurrent != NULL) && (strcmp(a_pszPipeName, l_pstCurrent->m_stDistPipeEntry.m_pszPipeName) != 0))
    {
      l_pstCurrent = l_pstCurrent->m_pstNext;
    }

  if (NULL != l_pstCurrent)
    {
      DistPipeDebug(GENERAL_DEBUG, "GetDistPipeDetails Found ... Pipe Name [%s]\n", 
		    l_pstCurrent->m_stDistPipeEntry.m_pszPipeName);
      return l_pstCurrent;
    }
  else
    return NULL;
}



/******************************************************************
* Function Name : DeleteEntry
* Purpose	: To delete an entry by PipeName.
* Args		:
* Return Value 	: Returns 1 on succesful deletion. Returns 0 if
*			it hits the bottom without finding the
*			entry.
*******************************************************************/

int DeleteEntry(a_pstDistPipeTable, a_pszPipeName)
     PDIST_PIPE_TABLE* a_pstDistPipeTable;
     char* a_pszPipeName;
{

  PDIST_PIPE_TABLE l_pstCurrent;
  PDIST_PIPE_TABLE l_pstPrevious;


  l_pstCurrent = *a_pstDistPipeTable;
  l_pstPrevious = NULL;

  while ((l_pstCurrent != NULL) && (strcmp(a_pszPipeName, l_pstCurrent->m_stDistPipeEntry.m_pszPipeName) != 0))
    {
      l_pstPrevious = l_pstCurrent;
      l_pstCurrent = l_pstCurrent->m_pstNext;
    }

  if (NULL == l_pstCurrent)
    {
      /* Entry not found */
      return PIPENAME_DOESNOT_EXIST;
    }
  else
   if ((NULL != l_pstCurrent) && (NULL == l_pstPrevious))
     {
       /* This the head of the list */
       *a_pstDistPipeTable = l_pstCurrent->m_pstNext;
     }
   else
     if ((NULL != l_pstCurrent) && (NULL != l_pstPrevious))
       {
	 /* not the head of the list */
	 l_pstPrevious->m_pstNext = l_pstCurrent->m_pstNext;
       }



  DistPipeDebug(GENERAL_DEBUG, "Deleting Pipe Name [%s]\n", a_pszPipeName);

  DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszCreateMachineName);
  DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszReadMachineName);
  DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszWriteMachineName);
  DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszPipeName);
  DISTPIPES_FREE(l_pstCurrent);

  return 1;

}


/*****************************************************************
*Function Name	: UpdateEntry
*Purpose	: To fill in the rest of data when a distributed
*			pipe is opened.
*Args		: Mode is 1 for read and 0 for write
*Return		: returns 1 on success, 0 otherwise.
******************************************************************/

int UpdateEntry(a_pstDistPipeTable, a_pszPipeName, a_nAccessMode, a_pszMachineName, a_nPortNumber, a_nProcessId, a_nSocketDescriptor)
     PDIST_PIPE_TABLE a_pstDistPipeTable;
     char* a_pszPipeName; 
     int a_nAccessMode; 
     char* a_pszMachineName; 
     int a_nPortNumber; 
     int a_nProcessId; 
     int a_nSocketDescriptor;
{
  
  PDIST_PIPE_TABLE l_pstFoundEntry;

  if ((l_pstFoundEntry = IfPipeExists(a_pstDistPipeTable, a_pszPipeName)) == NULL)
    {
      /* Entry not found */
      return PIPENAME_DOESNOT_EXIST;
    }
  
  if (READ_ACCESS_MODE == a_nAccessMode)
    {
      strcpy(l_pstFoundEntry->m_stDistPipeEntry.m_pszReadMachineName, a_pszMachineName);
      l_pstFoundEntry->m_stDistPipeEntry.m_nReadPortNumber = a_nPortNumber;
      l_pstFoundEntry->m_stDistPipeEntry.m_nReadProcessId = a_nProcessId;
      l_pstFoundEntry->m_stDistPipeEntry.m_nReadEntryValid = 1;
      l_pstFoundEntry->m_stDistPipeEntry.m_nReadSocketDescriptor = a_nSocketDescriptor;
      DistPipeDebug(GENERAL_DEBUG, "Read Machine Name [%s] : Pipe Name [%s] : Port Number [%d] Socket Desriptor [%d] & ProcessId [%d]\n", 
		    a_pszMachineName, a_pszPipeName, a_nPortNumber, a_nSocketDescriptor, a_nProcessId);
    }
  else
    if (WRITE_ACCESS_MODE == a_nAccessMode)
    {
      strcpy(l_pstFoundEntry->m_stDistPipeEntry.m_pszWriteMachineName, a_pszMachineName);
      l_pstFoundEntry->m_stDistPipeEntry.m_nWritePortNumber = a_nPortNumber;
      l_pstFoundEntry->m_stDistPipeEntry.m_nWriteProcessId = a_nProcessId;
      l_pstFoundEntry->m_stDistPipeEntry.m_nWriteSocketDescriptor = a_nSocketDescriptor;
      DistPipeDebug(GENERAL_DEBUG, "Write Machine Name [%s] : Pipe Name [%s] : Port Number [%d] Socket Desriptor [%d] & ProcessId [%d]\n", 
		    a_pszMachineName, a_pszPipeName, a_nPortNumber, a_nSocketDescriptor, a_nProcessId);
    }


  return 1;

}


/******************************************************************
* Function Name : DeleteTable
* Purpose	: To delete Table
* Args		:
* Return Value 	: Returns 1 on succesful deletion. Returns 0 if
*			it hits the bottom without finding the
*			entry.
*******************************************************************/
int DeleteTable(a_pstDistPipeTable)
     PDIST_PIPE_TABLE a_pstDistPipeTable;
{

  PDIST_PIPE_TABLE l_pstCurrent;
  PDIST_PIPE_TABLE l_pstNext;

  l_pstCurrent = a_pstDistPipeTable;
  l_pstNext = NULL;

  while (l_pstCurrent != NULL)
    {

      l_pstNext = l_pstCurrent->m_pstNext;

      DistPipeDebug(GENERAL_DEBUG, "Deleting ... Pipe Name [%s]\n", 
		    l_pstCurrent->m_stDistPipeEntry.m_pszPipeName);

      DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszCreateMachineName);
      DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszReadMachineName);
      DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszWriteMachineName);
      DISTPIPES_FREE(l_pstCurrent->m_stDistPipeEntry.m_pszPipeName);
      DISTPIPES_FREE(l_pstCurrent);

      l_pstCurrent = l_pstNext;
      
    }

  return 1;
}


