/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/


#include <rpc/types.h>
#include <rpc/xdr.h>


#include "DistPipeNetLevel.h"
#include "DistPipeError.h"
#include "DistPipeDebug.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeXDRRoutines.h"

char* g_pszProgramName;



/*********************************************************************************************************************
 * Write Fucntions / Encoders
 *********************************************************************************************************************/

/************************************************************************************************************
 * Description : 
 *              Sends a XDR Format Message
 * Arguments   :
 *              a_pstHostDetails - The host detail pointer to which message is to be sent
 *              a__pstMessage - Message to be Sent
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRMessage(a_pstHostDetails, a_pstMessage, a_nWhichHost)
     PHOST_DESCRIPTOR a_pstHostDetails;
     PMESSAGE a_pstMessage;
     int a_nWhichHost;
{
  
  int l_nErrorCode;

   /* Send the Message Type */
  if ((l_nErrorCode = WriteOnHost(a_pstHostDetails, (char*) &a_pstMessage->m_nMessageType, 
				  sizeof(a_pstMessage->m_nMessageType), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send the Message Length */
  if ((l_nErrorCode = WriteOnHost(a_pstHostDetails, (char*) &a_pstMessage->m_nXDRDataLength, 
				  sizeof(a_pstMessage->m_nXDRDataLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send the Message */
  if ((l_nErrorCode = WriteOnHost(a_pstHostDetails, (char*) a_pstMessage->m_pXDRData, 
				  ntohs(a_pstMessage->m_nXDRDataLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;

}


/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRSignedChar(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_char(&l_stWriteMessage.m_XDRVariable, (signed char *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding Signed character error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }

  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);
 

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRUnSignedChar(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);
  if (TCP_HOST == a_nWhichHost)
    {
    
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_char(&l_stWriteMessage.m_XDRVariable, (unsigned char *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding unsigned character Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRSignedShortInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_short(&l_stWriteMessage.m_XDRVariable, (signed short *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding signed short integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRUnSignedShortInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_short(&l_stWriteMessage.m_XDRVariable, (unsigned short int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding unsigned short integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRSignedInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_int(&l_stWriteMessage.m_XDRVariable, (signed int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding signed integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRUnSignedInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_int(&l_stWriteMessage.m_XDRVariable, (unsigned int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding unsigned integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRSignedLongInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_long(&l_stWriteMessage.m_XDRVariable, (signed long int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding signed long integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRUnSignedLongInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_long(&l_stWriteMessage.m_XDRVariable, (unsigned long int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding unsigned long integer Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRString(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;
  
  char* l_pszPointerToTheString;

  int l_nErrorCode;
  unsigned int l_nBytesToBeEncoded;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, LENGTH_FIELD_FOR_XDR_STRING + a_nSize);
  
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData,  LENGTH_FIELD_FOR_XDR_STRING + a_nSize, XDR_ENCODE);
      
      l_pszPointerToTheString = (char*) a_pDataElement;
      l_nBytesToBeEncoded = a_nSize;
      /* Convert to the XDR Data representation */
      /* convert each of them to XDR bytes and then send the data */
      if (FALSE == xdr_string(&l_stWriteMessage.m_XDRVariable, (char**) &l_pszPointerToTheString, a_nSize + LENGTH_FIELD_FOR_XDR_STRING))
	{
	  DistPipeLogError("XDR Encoding String Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize +  LENGTH_FIELD_FOR_XDR_STRING);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRFloat(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_float(&l_stWriteMessage.m_XDRVariable, (float *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding float Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}


/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int WriteXDRDouble(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stWriteMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stWriteMessage.m_pXDRData, char, a_nSize);

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stWriteMessage.m_XDRVariable, l_stWriteMessage.m_pXDRData, a_nSize, XDR_ENCODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_double(&l_stWriteMessage.m_XDRVariable, (double *)a_pDataElement))
	{
	  DistPipeLogError("XDR Encoding double Error\n");
	  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(a_pDataElement, l_stWriteMessage.m_pXDRData, a_nSize);
      }
  /* Set the length of the message */
  l_stWriteMessage.m_nXDRDataLength = htons(a_nSize);
					 
  /* Set the Message Id */
  l_stWriteMessage.m_nMessageType = htons(a_nMessageID);

  /* Send the Data to the other end */
  l_nErrorCode = WriteXDRMessage(a_pstHostDetails, &l_stWriteMessage, a_nWhichHost);

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stWriteMessage.m_pXDRData);

  return l_nErrorCode;
}



/*********************************************************************************************************************
 * Read Functions / Decodders
 *********************************************************************************************************************/

/************************************************************************************************************
 * Description : 
 *              Sends a XDR Format Message
 * Arguments   :
 *              a_pstHostDetails - The hsot detail pointer to which message is to be sent
 *              a__pstMessage - Message to be Sent
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRMessage(a_pstHostDetails, a_pstMessage, a_nWhichHost)
     PHOST_DESCRIPTOR a_pstHostDetails;
     PMESSAGE a_pstMessage;
     int a_nWhichHost;
{

  int l_nErrorCode;
  
   /* Send the Message Type */
  if ((l_nErrorCode = ReadOnHost(a_pstHostDetails, (char*) &a_pstMessage->m_nMessageType, 
				  sizeof(a_pstMessage->m_nMessageType), a_nWhichHost)) < 0)
    {
        printf("CUPPED!\n\n");
      return l_nErrorCode;
    }
  
  /* Send the Message Length */
  if ((l_nErrorCode = ReadOnHost(a_pstHostDetails, (char*) &a_pstMessage->m_nXDRDataLength, 
				  sizeof(a_pstMessage->m_nXDRDataLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }

  
  /* Send the Message */
  if ((l_nErrorCode = ReadOnHost(a_pstHostDetails, (char*) a_pstMessage->m_pXDRData, 
				  ntohs(a_pstMessage->m_nXDRDataLength), a_nWhichHost)) < 0)
    {
      return l_nErrorCode;
    }

  return 1;

}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRSignedChar(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_char(&l_stReadMessage.m_XDRVariable, (signed char *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding signed character Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRUnSignedChar(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);
  
   /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_char(&l_stReadMessage.m_XDRVariable, (unsigned char *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding unsigned character Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRSignedShortInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);
 
  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_short(&l_stReadMessage.m_XDRVariable, (signed short *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding signed short integer Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRUnSignedShortInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

   /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_short(&l_stReadMessage.m_XDRVariable, (unsigned short int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding unsigned short integer Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRSignedInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_int(&l_stReadMessage.m_XDRVariable, (signed int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding signed integer Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRUnSignedInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }
  
  if (TCP_HOST == a_nWhichHost)
    {
  
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_int(&l_stReadMessage.m_XDRVariable, (signed int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding unsigned integer Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRSignedLongInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }
  
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_long(&l_stReadMessage.m_XDRVariable, (signed long int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding signed long Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRUnSignedLongInteger(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }
  
  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_u_long(&l_stReadMessage.m_XDRVariable, (unsigned long int *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding unsigned long Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRString(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  char* l_pszPointerToTheString;

  int l_nErrorCode;
  int l_nBytesDecoded;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char,  LENGTH_FIELD_FOR_XDR_STRING + a_nSize);
  
  /* Read the XDR Message */
  /*
   * Freezing here
   */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }


  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData,  LENGTH_FIELD_FOR_XDR_STRING + a_nSize, XDR_DECODE);
      
      l_pszPointerToTheString = (char*) a_pDataElement;
      l_nBytesDecoded =  a_nSize;
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_string(&l_stReadMessage.m_XDRVariable, (char**) &l_pszPointerToTheString, a_nSize))
	{
	  DistPipeLogError("XDR Decoding String Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
      }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }

  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}

/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRFloat(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);

  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }
  

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_float(&l_stReadMessage.m_XDRVariable, (float *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding float Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}


/************************************************************************************************************
 * Description : 
 *              sends a short value in XDR format
 * Arguments   :
 *              a_nMessageID - Message Identifier to be written
 *              a_pDataElement - pointer to the data element to be sent
 *              a_nSize - size of each data item
 *              a_nWhichHost - Type of the Host TCP/UNIX Host
 * Return      :
 *              TRUE/FALSE
 ************************************************************************************************************/
int ReadXDRDouble(a_nMessageID, a_pstHostDetails, a_pDataElement, a_nSize, a_nWhichHost)
     int a_nMessageID;
     PHOST_DESCRIPTOR a_pstHostDetails; 
     void* a_pDataElement;
     int a_nSize;
     int a_nWhichHost;
{


  MESSAGE l_stReadMessage;

  int l_nErrorCode;

  DISTPIPES_CALLOC(l_stReadMessage.m_pXDRData, char, a_nSize);
  
  /* Read the XDR Message */
  if ((l_nErrorCode = ReadXDRMessage(a_pstHostDetails, &l_stReadMessage, a_nWhichHost)) < 0)
    {
      /* Delete the Memory allocated for storing the XDR data format */
      DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
      return l_nErrorCode;
    }

  if (TCP_HOST == a_nWhichHost)
    {
  
      /* Create the memory for the XDR format data representation */
      xdrmem_create(&l_stReadMessage.m_XDRVariable, l_stReadMessage.m_pXDRData, a_nSize, XDR_DECODE);
      
      /* Convert to the XDR Data representation */
      if (FALSE == xdr_double(&l_stReadMessage.m_XDRVariable, (double *)a_pDataElement))
	{
	  DistPipeLogError("XDR Decoding double Error\n");
	  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);
	  return XDR_ERROR;
	}
    }
  else
    if (UNIXSTREAM_HOST == a_nWhichHost)
      {
	DISTPIPES_COPY(l_stReadMessage.m_pXDRData, a_pDataElement, a_nSize);
      }
  /* Delete the Memory allocated for storing the XDR data format */
  DISTPIPES_FREE(l_stReadMessage.m_pXDRData);

  return l_nErrorCode;
}












