/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
 /***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * DistPipeMigration.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for that are required for migrating the source 
 * c code to another machine and compling the code there.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 15/October/98
 *
 ***********************************************************************************************/
#include <sys/stat.h>

#include "DistPipeError.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeSystemCoordinator.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeNetLevel.h"
#include "DistPipeTable.h"
#include "DistPipeXDRRoutines.h"

#include "Work.h"

#include "StProbMigration.h"


/************************************************************************************************************
 * Description : 
 *               Creates a new file with the file name passed.
 * Arguments   :
 *               a_pszFileName - Name of the file to be created
 * Return      :
 *               ErrorCode / Created File Descriptor
 ************************************************************************************************************/
int CreateFileForCompilation(a_pszFileName)
     char* a_pszFileName;
{
  
  int l_nCreatedFileDescriptor;

  fprintf(stderr, "Entered CreateFileForCompilation in StProbMigration.c\n\n");

  if ((l_nCreatedFileDescriptor = open(a_pszFileName, O_CREAT | O_WRONLY) < 0))
    {
      DistPipeLogSystemError("open");
      return CANNOT_CREATE_FILE;
    }

  close(l_nCreatedFileDescriptor);
  return l_nCreatedFileDescriptor;
}

/************************************************************************************************************
 * Description : 
 *               Writes to the file.
 * Arguments   :
 *               a_pszFileName - File Name to be written to.
 *               a_pszWriteBuffer - Buffer to be written to file.
 *               a_nWriteBufferSize - Amount of bytes to be written.
 * Return      :
 *               ErrorCode / Number of bytes written
 ************************************************************************************************************/
int WriteToFile(a_pszFileName, a_pszWriteBuffer, a_nWriteBufferSize)
     char* a_pszFileName;
     char* a_pszWriteBuffer;
     int a_nWriteBufferSize;
{
  
  int l_nBytesWritten;
  int l_nFileDescriptor;

  fprintf(stderr, "Entered WriteToFile in StProbMigration.c\n\n");
  
  if ((l_nFileDescriptor = open(a_pszFileName, O_CREAT | O_WRONLY, S_IRWXU)) <= 0)
    {
      DistPipeLogSystemError("open");
      return CANNOT_OPEN_FILE;
    }

  DistPipeDebug(GENERAL_DEBUG, "Opened File [%s] Descriptor [%d] \n", a_pszFileName, l_nFileDescriptor);

  if ((l_nBytesWritten = write(l_nFileDescriptor, a_pszWriteBuffer, a_nWriteBufferSize)) < 0)
    {
      DistPipeLogSystemError("write");
      return CANNOT_WRITE_TO_FILE;
    }
  close(l_nFileDescriptor);
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Read from the file.
 * Arguments   :
 *               a_pszFileName - File Name to be read from.
 *               a_pszWriteBuffer - Buffer to be read to.
 *               a_nWriteBufferSize - Amount of bytes to be read.
 * Return      :
 *               ErrorCode / Number of bytes written
 ************************************************************************************************************/
int ReadFromFile(a_pszFileName, a_pszReadBuffer, a_nReadBufferSize)
     char* a_pszFileName;
     char* a_pszReadBuffer;
     int a_nReadBufferSize;
{
  
  int l_nBytesRead;
  int l_nFileDescriptor;

  fprintf(stderr, "Entered ReadFromFile in StProbMigration.c\n\n");

  if ((l_nFileDescriptor = open(a_pszFileName, O_RDONLY)) <= 0)
    {
      DistPipeLogSystemError("open");
      return CANNOT_OPEN_FILE;
    }
  
  DistPipeDebug(GENERAL_DEBUG, "Opened File [%s] Descriptor [%d] \n", a_pszFileName, l_nFileDescriptor);
  
  if ((l_nBytesRead = read(l_nFileDescriptor, a_pszReadBuffer, a_nReadBufferSize)) < 0)
    {
      DistPipeLogSystemError("read");
      return CANNOT_READ_FROM_FILE;
    }
/*   DistPipeDebug(GENERAL_DEBUG, "file [%s]\n", a_pszReadBuffer); */

  close(l_nFileDescriptor);
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Compiles the C source code depending on the architecutre it is compiled on.
 * Arguments   :
 *               a_pszFileName - C Source code to be compiled.
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int CompileMigratedCFile(a_pszFileName)
     char* a_pszFileName;
{
  char *tmpstr;
  char *tstr;

  strcpy (tstr, a_pszFileName);
  tmpstr = strrchr (tstr, '.');
  *tmpstr = '\0';

  tmpstr = "gcc -o ";
  strcat (tmpstr, tstr);
  strcat (tmpstr, " ");
  strcat (tmpstr, a_pszFileName);

  system (tmpstr);
  fprintf(stderr, "Entered CompileMigratedCFile in StProbMigration.c - received %s\n\n", a_pszFileName);
  
  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Deletes a C source code and the executable file created.
 * Arguments   :
 *               a_pszFileName - file to be deleted
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DeleteMigratedFiles(a_pszFileName)
     char* a_pszFileName;
{


  char* l_pszExecutableFileName;

  fprintf(stderr, "Entered DeleteMigratedFiles in StProbMigration.c - got %s\n\n", a_pszFileName);

  DISTPIPES_CALLOC(l_pszExecutableFileName, char, PATH_LENGTH);

  unlink(a_pszFileName);
  strcpy(l_pszExecutableFileName, a_pszFileName);
  strcat(l_pszExecutableFileName, EXECUTABLE_FILENAME_EXTENSION);

  unlink(l_pszExecutableFileName);

  DISTPIPES_FREE(l_pszExecutableFileName);

  return 1;

}

/************************************************************************************************************
 * Description : 
 *               Sends the details for migrating a file to another machine
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details
 *               a_pszFileName - file to be migrated
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int SendDetailsForMigration(a_pstHostDetails, a_pszFileName)
     PHOST_DESCRIPTOR a_pstHostDetails;
     char* a_pszFileName;
{


  int l_nErrorCode;
  
  unsigned long int l_ulSizeOfFile;
  
  char* l_pszFileContent;

  struct stat l_stFileInformation;

  fprintf(stderr, "Entered SendDetailsForMigration in StProbMigration.c\n\n");

  /* Send the File Name to the System coordinator */
  if ((l_nErrorCode = WriteXDRString(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, a_pstHostDetails, 
				     (void* ) a_pszFileName, 
				     PATH_LENGTH, TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  
  if (stat(a_pszFileName, &l_stFileInformation) < 0)
    {
      DistPipeLogSystemError("stat");
      l_ulSizeOfFile = 0;
      if ((l_nErrorCode = WriteXDRUnSignedLongInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, a_pstHostDetails, 
						      (void* ) &l_ulSizeOfFile, 
						      sizeof(l_ulSizeOfFile), TCP_HOST)) < 0)
	{
	  return l_nErrorCode;
	}
      return -1;
    }
  

  l_ulSizeOfFile = l_stFileInformation.st_size;

  /* Send the Size of the file to System coordinator */
  if ((l_nErrorCode = WriteXDRUnSignedLongInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, a_pstHostDetails, 
						  (void* ) &l_ulSizeOfFile, 
						  sizeof(l_ulSizeOfFile), TCP_HOST)) < 0)
    {
      return l_nErrorCode;
    }

 

  DISTPIPES_CALLOC(l_pszFileContent, char, l_ulSizeOfFile);

  if (ReadFromFile(a_pszFileName, l_pszFileContent, l_ulSizeOfFile) < 0)
    {

      DISTPIPES_FREE(l_pszFileContent);
      return -1;
    }

  /* Send the Content of the File to the System Coordinator */
  if ((l_nErrorCode = WriteMessageAndData(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, 
					  (char*) l_pszFileContent, l_ulSizeOfFile,
					  a_pstHostDetails, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszFileContent);
      return l_nErrorCode;
    }


  DISTPIPES_FREE(l_pszFileContent);
  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Migrates the file to another machine
 * Arguments   :
 *               a_pstHostDetails - pointer to the host details
 *               a_pstWorkDetails -  pointer to the work details
 *               a_pszFileName - file to be migrated
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DoMigration(a_pstHostDetails, a_pstWorkDetails, a_nSplitID)
     PHOST_DESCRIPTOR_COLLECTION a_pstHostDetails;
     PWORK_INFORMATION_COLLECTION a_pstWorkDetails;
     int a_nSplitID;
{

  int l_nErrorCode;
  

  char* l_pszFileName;
  char* l_pszFileContent;

  unsigned long int l_ulSizeOfFile;

  fprintf(stderr, "Entered DoMigration in StProbMigration.c:%s\n\n", a_pstHostDetails->m_stHost.m_pszHostName );

  DISTPIPES_CALLOC(l_pszFileName, char, PATH_LENGTH);

  /* Read the File Name */
  if ((l_nErrorCode = ReadXDRString(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
				    (void*) l_pszFileName, 
				    PATH_LENGTH, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszFileName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] Migrate File Name [%s]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_pszFileName);
  

  strcat(l_pszFileName, ".Migrated.c");

  UpdateMigrateFile(a_pstWorkDetails, l_pszFileName);

  /* Read the File Size */
  if ((l_nErrorCode = ReadXDRUnSignedLongInteger(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, &a_pstHostDetails->m_stHost, 
						 (void*) &l_ulSizeOfFile, 
						 sizeof(l_ulSizeOfFile), TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszFileName);
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] Size of file [%ld]\n", 
		a_pstHostDetails->m_stHost.m_pszHostName, l_ulSizeOfFile);


  if (0 == l_ulSizeOfFile)
    return -1;

  /* Read the File Content */
  DISTPIPES_CALLOC(l_pszFileContent, char, l_ulSizeOfFile);

   /* Read the file from the local coordinator */
  if ((l_nErrorCode = ReadMessageAndData(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK, 
					 (char*) l_pszFileContent, 
					 &a_pstHostDetails->m_stHost, TCP_HOST)) < 0)
    {
      DISTPIPES_FREE(l_pszFileName);
      DISTPIPES_FREE(l_pszFileContent);
      return l_nErrorCode;
    }

  /*  DistPipeDebug(LCTOSC_DEBUG, "(LOCAL_SYSTEM_STEADY_STATE_SEND_WORK) From [%s] file [%s]\n",  */
  /* 		a_pstHostDetails->m_stHost.m_pszHostName, l_pszFileContent); */

  if (WriteToFile(l_pszFileName, l_pszFileContent, l_ulSizeOfFile) < 0)
    {
      DISTPIPES_FREE(l_pszFileName);
      DISTPIPES_FREE(l_pszFileContent);
      return -1;
    }

  DISTPIPES_FREE(l_pszFileContent); 
  DISTPIPES_FREE(l_pszFileName); 
  
  return 1;
}





