/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
#include <stdio.h>


#include "DistPipeSystemCoordinator.h"
#include "DistPipeLocalCoordinator.h"
#include "DistPipeError.h"
#include "DistPipeNetLevel.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeTable.h"
#include "DistPipeRoutines.h"
#include "DistPipeXDRRoutines.h"

#include "Work.h"
#include "SteadyState.h"


extern HOST_DESCRIPTOR g_stUserProcess;



/************************************************************************************************************
 * Description : 
 *               Initialize the steady state probelm and obtain the work id
 * Arguments   :
 *               ----
 * Return      :
 *               ErrorCode / WorkID
 ************************************************************************************************************/
int IntializeSteadyStateProblem()
{


  int l_nWorkID;
  int l_nErrorCode;
  int l_nMessageID;


  InitializeUserProcess();

  /* Send the Message of intialization */
  l_nMessageID = USER_LOCAL_STEADY_STATE_INITIALIZE;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Obtain the Work id for this steady state problem */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_INITIALIZE, &g_stUserProcess, 
					   (void*) &l_nWorkID, 
					   sizeof(l_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_INITIALIZE) Work Identifier [%d]\n", 
		l_nWorkID);


 return l_nWorkID;

}

/************************************************************************************************************
 * Description : 
 *               Send the Details of the Matrix Number of Rows and Columns 
 * Arguments   :
 *               a_nWorkID - Work identification 
 *               a_nRows - Number of Rows of the matrix
 *               a_nColumns - Number of Columns of the matrix
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int SendMatrixInformation(a_nWorkID, a_nRows, a_nColumns)
     int a_nWorkID;
     int a_nRows;
     int a_nColumns;
{


  int l_nErrorCode;
  int l_nStatus;
  int l_nMessageID;
  
  /* Send the message for sending matrix information */
  l_nMessageID = USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Work Identification */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &g_stUserProcess,
					    (void*) &a_nWorkID,
					    sizeof(a_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Send the number of rows */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &g_stUserProcess,
					    (void*) &a_nRows,
					    sizeof(a_nRows), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Number of columns */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_MATRIX_INFORMATION, &g_stUserProcess,
					    (void*) &a_nColumns,
					    sizeof(a_nColumns), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Recieve the success/failure status */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_MATRIX_INFORMATION, &g_stUserProcess, 
					   (void*) &l_nStatus, 
					   sizeof(l_nStatus), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_MATRIX_INFORMATION) Send Matrix Infomation Status [%d]\n", 
		l_nStatus);

  return l_nStatus;
}


/************************************************************************************************************
 * Description : 
 *               Obtain number of splits
 * Arguments   :
 *               a_nWorkID - Work identification 
 *               
 * Return      :
 *               ErrorCode / Number of Splits
 ************************************************************************************************************/
int ObtainNumberOfSplits(a_nWorkID)
     int a_nWorkID;
{

  int l_nErrorCode;
  int l_nNumberOfMachines;
  int l_nMessageID;
  
  /* Send the message for sending matrix information */
  l_nMessageID = USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Work Identification */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_NUMBER_OF_SPLITS, &g_stUserProcess,
					    (void*) &a_nWorkID,
					    sizeof(a_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }


  /* Obtain number of machines */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_NUMBER_OF_SPLITS, &g_stUserProcess, 
					   (void*) &l_nNumberOfMachines, 
					   sizeof(l_nNumberOfMachines), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_NUMBER_OF_SPLITS) Number of Machine [%d]\n", 
		l_nNumberOfMachines);
  
 
  fprintf( stderr, "%s %d", "ObtainNoOfSplits returning ", l_nNumberOfMachines );
  return l_nNumberOfMachines;
}




/************************************************************************************************************
 * Description : 
 *               Obtain Split Information 
 * Arguments   :
 *               a_nWorkID - Work identification 
 *               a_nSplitID - Split indenitifier for this work id
 *               a_pnStartRow - Address of the starting row information
 *               a_pnNumberOfRows - Address of Number of Rows for this split
 *               
 * Return      :
 *               ErrorCode / Number of Splits
 ************************************************************************************************************/
int ObtainSplitInformation(a_nWorkID, a_nSplitID, a_pnStartRow, a_pnNumberOfRows)
     int a_nWorkID;
     int a_nSplitID;
     int* a_pnStartRow;
     int* a_pnNumberOfRows;
{

  int l_nErrorCode;
  int l_nMessageID;

  /* Send the message for sending matrix information */
  l_nMessageID = USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Work Identification */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &g_stUserProcess,
					    (void*) &a_nWorkID,
					    sizeof(a_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  
  /* Send the split information */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_SPLIT_INFORMATION, &g_stUserProcess,
					    (void*) &a_nSplitID,
					    sizeof(a_nSplitID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Obtain the Starting Row */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION, &g_stUserProcess, 
					   (void*) a_pnStartRow, 
					   sizeof(int), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION) Start Rows [%d]\n", 
		*a_pnStartRow);


   /* Obtain Number of Rows */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION, &g_stUserProcess, 
					   (void*) a_pnNumberOfRows, 
					   sizeof(int), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_SPLIT_INFORMATION) Number of Rows [%d]\n", 
		*a_pnNumberOfRows);

  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Send the work for execution
 * Arguments   :
 *               a_nWorkID - Work identification 
 *               a_nSplitID - Split indenitifier for this work id
 *               
 *               
 * Return      :
 *               ErrorCode / Number of Splits
 ************************************************************************************************************/
int SendWorkAndExecute(a_nWorkID, a_nSplitID, l_pszMigrateFile, l_pszData, l_nDataSize, l_pszResultPipe)
     int a_nWorkID;
     int a_nSplitID;
     char* l_pszMigrateFile;
     char* l_pszData;
     int l_nDataSize;
     char* l_pszResultPipe;
{

  int l_nErrorCode;
  int l_nMessageID;
  int l_nSubTaskID;

  /* Send the Send Work Message to local coordinator */
  l_nMessageID = USER_LOCAL_STEADY_STATE_SEND_WORK;
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_ANY_MESSAGE, &g_stUserProcess,
					    (void*) &l_nMessageID,
					    sizeof(l_nMessageID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Work Identification */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_SEND_WORK, &g_stUserProcess,
					    (void*) &a_nWorkID,
					    sizeof(a_nWorkID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  
  /* Send the split information */
  if ((l_nErrorCode = WriteXDRSignedInteger(USER_LOCAL_STEADY_STATE_SEND_WORK, &g_stUserProcess,
					    (void*) &a_nSplitID,
					    sizeof(a_nSplitID), UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }

  /* Send the Migrate File Name to Local coordinator */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_STEADY_STATE_SEND_WORK, &g_stUserProcess,
				     (void*) l_pszMigrateFile,
				     PATH_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Send the Result Pipe Name to Local coordinator */
  if ((l_nErrorCode = WriteXDRString(USER_LOCAL_STEADY_STATE_SEND_WORK, &g_stUserProcess,
				     (void*) l_pszResultPipe,
				     PIPENAME_LENGTH, UNIXSTREAM_HOST)) < 0)
    {
      return l_nErrorCode;
    }
  
  /* Get the Task Id for the Send Work */
  if ((l_nErrorCode = ReadXDRSignedInteger(LOCAL_USER_STEADY_STATE_SEND_WORK, &g_stUserProcess,  
 					   (void*) &l_nSubTaskID,  
 					   sizeof(l_nSubTaskID), UNIXSTREAM_HOST)) < 0) 
    { 
      return l_nErrorCode; 
    } 
  DistPipeDebug(LCTOUSER_DEBUG, "(LOCAL_USER_STEADY_STATE_SEND_WORK) Sub Task Id [%d]\n",  
 		l_nSubTaskID); 

  return 1;
}
