/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
#include "DistPipeError.h"
#include "DistPipeHost.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"


#include "Work.h"

/************************************************************************************************************
 * Description : 
 *               Add a new entry to the table
 * Arguments   :
 *               a_pstWorkCollection - Address of the starting pointer to the link list of work information
 *               a_nWorkID - Work identifier
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int AddSplitInformation(a_pstWorkCollection, a_nWorkID)
     PWORK_INFORMATION_COLLECTION* a_pstWorkCollection;
     int a_nWorkID;
{
  PWORK_INFORMATION_COLLECTION l_pstNewWork;

  DISTPIPES_CALLOC(l_pstNewWork, WORK_INFORMATION_COLLECTION, 1);

  DISTPIPES_CALLOC(l_pstNewWork->m_pstWorkDetail, WORK_INFORMATION, 1);
  l_pstNewWork->m_pstNext = NULL;

  l_pstNewWork->m_pstWorkDetail->m_nWorkID = a_nWorkID;
  l_pstNewWork->m_pstWorkDetail->m_pstSplitInformation = NULL;
  
  /* Not Start Node */
  if (NULL != *a_pstWorkCollection)
    {
      l_pstNewWork->m_pstNext = *a_pstWorkCollection;
    }
  *a_pstWorkCollection = l_pstNewWork;
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Delete a entry to the table
 * Arguments   :
 *               a_pstWorkCollection - Address of the starting pointer to the link list of work information
 *               a_nWorkID - Work identifier
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DeleteSplitInformation(a_pstWorkCollection, a_nWorkID)
     PWORK_INFORMATION_COLLECTION* a_pstWorkCollection;
     int a_nWorkID;
{

  PWORK_INFORMATION_COLLECTION l_pstCurrent;
  PWORK_INFORMATION_COLLECTION l_pstPrevious;

  l_pstCurrent = *a_pstWorkCollection;
  l_pstPrevious = NULL;
  while((l_pstCurrent != NULL) && (l_pstCurrent->m_pstWorkDetail->m_nWorkID != a_nWorkID))
    {
      l_pstPrevious = l_pstCurrent;
      l_pstCurrent = l_pstCurrent->m_pstNext;
    }


  if (NULL != l_pstCurrent)
    {
      /* First Node */
      if (NULL == l_pstPrevious)
	{
	  *a_pstWorkCollection = l_pstCurrent->m_pstNext;
	}
      else /* Not First Node */
	{
	  l_pstPrevious->m_pstNext = l_pstCurrent->m_pstNext;
	}

      DISTPIPES_FREE((l_pstCurrent->m_pstWorkDetail->m_pstSplitInformation));
      DISTPIPES_FREE(l_pstCurrent->m_pstWorkDetail);
      DISTPIPES_FREE(l_pstCurrent);

    }
  return 1;
}


/************************************************************************************************************
 * Description : 
 *               Delete All the entries
 * Arguments   :
 *               a_pstWorkCollection - Address of the starting pointer to the link list of work information
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/
int DeleteAllSplitInformation(a_pstWorkCollection)
     PWORK_INFORMATION_COLLECTION* a_pstWorkCollection;
{
  PWORK_INFORMATION_COLLECTION l_pstCurrent;
  PWORK_INFORMATION_COLLECTION l_pstNext;

  l_pstCurrent = *a_pstWorkCollection;
  l_pstNext = NULL;
  while(l_pstCurrent != NULL)
    {
      l_pstNext = l_pstCurrent->m_pstNext;

      DISTPIPES_FREE(l_pstCurrent->m_pstWorkDetail->m_pstSplitInformation);
      DISTPIPES_FREE(l_pstCurrent->m_pstWorkDetail);
      DISTPIPES_FREE(l_pstCurrent);
      
      l_pstCurrent = l_pstNext;
    }

  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Find the Node with Work Id
 * Arguments   :
 *               a_pstWorkCollection - Address of the starting pointer to the link list of work information
 *               a_nWorkID - work identification number
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
PWORK_INFORMATION_COLLECTION FindWork(a_pstWorkCollection, a_nWorkID)
     PWORK_INFORMATION_COLLECTION a_pstWorkCollection;
     int a_nWorkID;
{
  
  PWORK_INFORMATION_COLLECTION l_pstCurrent;

  l_pstCurrent = a_pstWorkCollection;
  while((l_pstCurrent != NULL) && (l_pstCurrent->m_pstWorkDetail->m_nWorkID != a_nWorkID))
    {
      l_pstCurrent = l_pstCurrent->m_pstNext;
    }
  
  return l_pstCurrent;
}

/************************************************************************************************************
 * Description : 
 *               Set Number of splits that the work is divided into
 * Arguments   :
 *               a_pstWorkCollection - Address of the Work detail.
 *               a_nNumberOfSplits  - Number of splits of this work
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
int SetNumberOfSplitInformation(a_pstWork, a_nNumberOfSplits)
     PWORK_INFORMATION_COLLECTION a_pstWork;
     int a_nNumberOfSplits;
{

  DISTPIPES_CALLOC(a_pstWork->m_pstWorkDetail->m_pstSplitInformation, SPLIT_INFORMATION, a_nNumberOfSplits);
  a_pstWork->m_pstWorkDetail->m_nNumberOfSplits = a_nNumberOfSplits;
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Store the split information
 * Arguments   :
 *               a_pstWorkCollection - Address of the Work detail.
 *               a_nNumberOfSplits  - Number of splits of this work
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
int SetSplitInformation(a_pstWork, a_nSplitID, a_pstHostDetail, a_nStart, a_nTotal)
     PWORK_INFORMATION_COLLECTION a_pstWork;
     int a_nSplitID;
     PHOST_DESCRIPTOR a_pstHostDetail;
     int a_nStart;
     int a_nTotal;
{

  DistPipeDebug(GENERAL_DEBUG, "Split ID [%d], Start [%d], Total [%d]\n",
		a_nSplitID, a_nStart, a_nTotal);

  if (a_nSplitID < a_pstWork->m_pstWorkDetail->m_nNumberOfSplits)
    {
      a_pstWork->m_pstWorkDetail->m_pstSplitInformation[a_nSplitID].m_pstSplitHostDetail = a_pstHostDetail; 
      a_pstWork->m_pstWorkDetail->m_pstSplitInformation[a_nSplitID].m_nStarting = a_nStart;
      a_pstWork->m_pstWorkDetail->m_pstSplitInformation[a_nSplitID].m_nTotal = a_nTotal;
    }
  return 1;
  
}

/************************************************************************************************************
 * Description : 
 *               Store Maximum Rows and Columns
 * Arguments   :
 *               a_pstWorkCollection - Address of the Work detail.
 *               a_nRows - Number of Rows
 *               a_nColumns - Number of Columns
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
int UpdateRowColumnForWork(a_pstWork, a_nRows, a_nColumns)
     PWORK_INFORMATION_COLLECTION a_pstWork;
     int a_nRows;
     int a_nColumns;
{

  a_pstWork->m_pstWorkDetail->m_nMaximumRows = a_nRows;
  a_pstWork->m_pstWorkDetail->m_nMaximumColumns = a_nColumns;
  return 1;
  
}

/************************************************************************************************************
 * Description : 
 *               Store the Migrate file name
 * Arguments   :
 *               a_pstWorkCollection - Address of the Work detail.
 *               a_pszFileName - Migrate file name 
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
int UpdateMigrateFile(a_pstWork, a_pszFileName)
     PWORK_INFORMATION_COLLECTION a_pstWork;
     char* a_pszFileName;
{

  strcpy(a_pstWork->m_pstWorkDetail->m_szMigrateFileName, a_pszFileName);  
  return 1;
  
}

/************************************************************************************************************
 * Description : 
 *               Find the Number of rows and columns for a split uner a work
 * Arguments   :
 *               a_pstWorkCollection - Address of the Work detail.
 *               a_nSplitID - split identification number
 *               a_pnStart - Address of the start row value
 *               a_pnTotal - Address of the number of rows value
 * Return      :
 *               NULL / Node Found
 ************************************************************************************************************/
int FindStartAndNumberOfRowsForSplit(a_pstWork, a_nSplitID, a_pnStart, a_pnTotal)
     PWORK_INFORMATION_COLLECTION a_pstWork;
     int a_nSplitID;
     int* a_pnStart;
     int* a_pnTotal;
{
  
  if (a_nSplitID < a_pstWork->m_pstWorkDetail->m_nNumberOfSplits)
    {
      *a_pnStart = a_pstWork->m_pstWorkDetail->m_pstSplitInformation[a_nSplitID].m_nStarting;
      *a_pnTotal = a_pstWork->m_pstWorkDetail->m_pstSplitInformation[a_nSplitID].m_nTotal;
    }
  return 1;

}


