/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/*
 * Main program for the steady state pipeline
 * Refer juSteadyState.h for details
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>

#include "DistPipeError.h"
#include "DistPipeDefines.h"
#include "DistPipeMesg.h"
#include "DistPipeDebug.h"
#include "DistPipeRoutines.h"

#include "juSteadyState.h"
#include "juSteadyStateAux.h"

struct timeval currentTime, startTime;

int isourcepDPD, idestpDPD; // the source and destination dist. pipe descriptors
int isourcenDPD, idestnDPD; // the source and dest pipe desc for the next node
int ipipePos;
int inumNodes;
int idone; // flag indicating steady state for the local node
char ssourcepPipe[60], sdestpPipe[60]; // Dist. Pipes we read from and write to the previous node respectively
char sdestnPipe[60], ssourcenPipe[60]; // dist pipes for the next node
char slocalNodeName[30];

long sec, usec, syncsec, syncusec;

double   flocalMatrix[INUMGRIDS][2]; // the state of the rod as seen by the local node
int ibeginRange, iendRange; // the we compute the locations flocalMatrix[i]
double   fcolumnToSend[INUMGRIDS];


void *CreateSourcePipes( void *arg ){
  char szPrevNodeName[20];
  char szNextNodeName[20];
  int iPipePos = *(int *)arg;
  int err;
  
  // create a source pipe from the previous node
  strcpy (szPrevNodeName, spositionToName(iPipePos - 1));
  strcpy (ssourcepPipe, szPrevNodeName);
  strcat (ssourcepPipe, slocalNodeName);

  // create a source pipe from the next node
  strcpy( szNextNodeName, spositionToName(iPipePos +1));
  strcpy( ssourcenPipe, szNextNodeName );
  strcat ( ssourcenPipe, slocalNodeName );

  if( (err = CreateDistPipe(ssourcepPipe)) != 1 ) // sc should take care of two nodes creating same pipe
    fprintf( stderr, "[CreateSourcepPipe] CreateDistPipe returned %d for %s\n", err, ssourcepPipe );
  gettimeofday(&currentTime, NULL );
  fprintf( stderr, "[%3d.%06d] Created the source pipe : %s\n", currentTime.tv_sec, currentTime.tv_usec,ssourcepPipe);
  
if( (isourcepDPD = OpenDistPipe(ssourcepPipe, READ_ACCESS_MODE) ) < 0 )
    fprintf( stderr, "[CreateSourcepPipe] Error %d opening the read pipe on %s as %s\n", isourcepDPD, slocalNodeName, ssourcepPipe );

  // create and open the pipe from the next node
  if( (err = CreateDistPipe(ssourcenPipe)) != 1 ) // sc should take care of two nodes creating same pipe
    fprintf( stderr, "[CreateSourcenPipe] CreateDistPipe returned %d\n for %s", err, ssourcenPipe );
  gettimeofday(&currentTime, NULL );
  fprintf( stderr, "[%3d.%06d] Created the source pipe : %s\n", currentTime.tv_sec, currentTime.tv_usec, ssourcenPipe);

  if( (isourcenDPD = OpenDistPipe(ssourcenPipe, READ_ACCESS_MODE) ) < 0 )
    fprintf( stderr, "[CreateSourcenPipe] Error %d opening the read pipe on %s as %s\n", isourcenDPD, slocalNodeName, ssourcenPipe );  
  return NULL;
}

void *CreateDestPipes( void *arg ){
  char szNextNodeName[20];
  char szPrevNodeName[20];
  int err, iPipePos = *(int *)arg;
  
  strcpy (szNextNodeName, spositionToName(iPipePos + 1));
  strcpy (sdestnPipe, slocalNodeName);
  strcat (sdestnPipe, szNextNodeName);

  strcpy (szPrevNodeName, spositionToName(iPipePos - 1));
  strcpy (sdestpPipe, slocalNodeName);
  strcat (sdestpPipe, szPrevNodeName);


  if( (err = CreateDistPipe(sdestnPipe) ) != 1 ) // sc should take care of two nodes creating same pipe
    fprintf( stderr, "[CreateDestnPipe] CreateDistPipe returned %d for %s\n", err, sdestnPipe );
  gettimeofday( &currentTime, NULL );
  fprintf(stderr, "[%3d.%06d] Created the dest pipe : %s\n", currentTime.tv_sec, currentTime.tv_usec, sdestnPipe);
  

  if( (idestnDPD = OpenDistPipe(sdestnPipe, WRITE_ACCESS_MODE) ) < 0 )
    fprintf( stderr, "[CreateDestnPipe] Error %d opening the write pipe on %s as %s\n", idestnDPD, slocalNodeName, sdestnPipe );
  
  if( (err = CreateDistPipe(sdestpPipe) ) != 1 ) // sc should take care of two nodes creating same pipe
    fprintf( stderr, "[CreateDestpPipe] CreateDistPipe returned %d for %s\n", err, sdestpPipe );
  gettimeofday(&currentTime, NULL );
  fprintf(stderr, "[%3d.%06d] Created the dest pipe : %s\n", currentTime.tv_sec, currentTime.tv_usec, sdestpPipe);


  if( (idestpDPD = OpenDistPipe(sdestpPipe, WRITE_ACCESS_MODE) ) < 0 )
    fprintf( stderr, "[CreateDestPipe] Error %d opening the write pipe on %s as %s\n", idestpDPD, slocalNodeName, sdestpPipe );
  

  return NULL;
}

/*
 * doComputation is the main steady state equilibrium solver
 */
void doComputation() {
    int i, j;
    int prev,next;
    
    idone = 1;
    
    for (i = ibeginRange; i <= iendRange; i++)
    {
	    prev = i - 1;
    
    	if( prev < 0 )
    	    prev = INUMGRIDS - 1 ;
    
    	next = i + 1;
    
    	if( next >= INUMGRIDS )
    	    next = 0;
   
    	// remember that the first and the last grid
    	// are in the heat bath, and their vals. don't change
    	if( (!i)||(i == INUMGRIDS-1))
    	    flocalMatrix[i][1] = flocalMatrix[i][0];
    	else 
    	    flocalMatrix [i][1] =  steadyState(flocalMatrix[i][0], flocalMatrix[prev][0],flocalMatrix[next][0]); // correct only if both FT1 and FT2 are the same temperature!! - in this case, 500

        if (flocalMatrix[i][1] != flocalMatrix[i][0])
            idone = 0;

    }

    // copy the computed values into the zero'th column for the next iteration
    for ( i = ibeginRange; i <= iendRange; i++ )
        flocalMatrix[i][0] = flocalMatrix[i][1];

} // doComputation

#define JU_MSGSIZE 64
// all messages exchanged are fixed size data of 64 bytes
int sendMatrix() {
    int i;
    char *buffer;
    int buffersz = 0;
    int bufferlocn = 0;

    // send this to the next node through the destination pipe
    if( idestnDPD < 1 )
      return -1;
    
    buffersz = JU_MSGSIZE;
    buffer = (char *)malloc( buffersz );

    // copy the boundary value to send to the previous node
    memcpy( buffer, &flocalMatrix[ibeginRange][1], sizeof(double) );
    // send it
    if ( WriteDistPipe( idestpDPD, &buffer[bufferlocn], JU_MSGSIZE ) < 0)
      return -1;

    // copy the boundary value to send to the next node
    memcpy( buffer, &flocalMatrix[iendRange][1], sizeof(double) );
    // send it
    if( WriteDistPipe( idestnDPD, &buffer[bufferlocn], JU_MSGSIZE ) < 0)
      return -1;

    free(buffer);
    return 1;
}


int recvMatrix() {
    int i, locn;
    char buffer[JU_MSGSIZE];
    // first, read the matrix from the previous node
    
    if( isourcepDPD < 1 || isourcenDPD <1 )
      return -1;

    // read data from the previous pipe
    if (ReadDistPipe( isourcepDPD, buffer, JU_MSGSIZE ) < 0)
        return -1;
    
    // augment it into our matrix
    locn = ibeginRange - 1;
    if( locn < 0 )
        return 1;

    flocalMatrix[locn][0] = *((double *)buffer);

    // read data from the next pipe
    if (ReadDistPipe( isourcenDPD, buffer, JU_MSGSIZE ) < 0)
        return -1;
    
    // augment it into our matrix
    locn = iendRange + 1;
    if( locn >= INUMGRIDS )
        return 1;

    flocalMatrix[locn][0] = *((double *)buffer);
    
    return 1;
}

/*
 * initializePipes creates and opens the distributed pipes through which we read/write
 * the steady state data
 */
void initializePipes(void) {
    int  i;
   
    ipipePos = inameToPosition(slocalNodeName); // get the local node's position in the pipeline

    ibeginRange = ipipePos * (INUMGRIDS/NUMNODES);
    iendRange = (ipipePos +1 ) * (INUMGRIDS/NUMNODES) - 1;

    fprintf( stderr, "my range is %d to %d\n", ibeginRange, iendRange );

    if( (INUMGRIDS % NUMNODES) && (ipipePos == (NUMNODES -1)))
      iendRange = INUMGRIDS - 1;

    fprintf( stderr, "[initializePipes] pipe no. %d\n",ipipePos );
    if( ipipePos % 2 ){
      CreateSourcePipes( &ipipePos );
      CreateDestPipes( &ipipePos );
    } else {
      CreateDestPipes( &ipipePos );
      CreateSourcePipes( &ipipePos );
    }
    
    gettimeofday( &currentTime, NULL );
    memcpy (&startTime, &currentTime, sizeof(struct timeval));
    
    fprintf( stderr, "[%3d.%06d] Begin computations\n", currentTime.tv_sec, currentTime.tv_usec );
    fprintf( stderr, "CONFIRMATION: [%3d.%06d] Begin computations\n", startTime.tv_sec, startTime.tv_usec );
    flocalMatrix[0][0] = FT2;              // Initialize the rod's state at the first node:
    flocalMatrix[INUMGRIDS - 1][0] = FT1;  // All the internal grids of the rod are
    for (i = 1; i < INUMGRIDS - 2; i++)            // set to 0.0 while the ends are in the
      flocalMatrix[i][0] = 0.0;           // heat baths (FT1, FT2).
    
    // start doing the computation
    doComputation();
      
    // write our results to sdestPipe
    sendMatrix();

    //    }
    
            
} // initializePipes
    
int main() {
    int i,inumIterations, first;
    char *finddot;

    first = 1;
    idone = 0;
    syncsec = 0;
    syncusec = 0;

    fprintf(stderr, "NUMNODES = %d\n\n", NUMNODES);
    fprintf(stderr, "NUMGRIDS = %d\n\n", INUMGRIDS);
    gethostname (slocalNodeName, 30); // find out which machine we're on
    
    InitializeUserProcess();

    initializePipes(); // initialize all necessary pipes on the local machine

    for (inumIterations = 0; inumIterations < 999 ; inumIterations++) {
        // read the rod state sent by the previous node and the next node in the pipeline through ssourcePipe's
        if( !(inumIterations % 50 )){
	        gettimeofday( &currentTime, NULL );
        	fprintf( stderr, "[%3d.%06d] Begin iteration %d\n", currentTime.tv_sec, currentTime.tv_usec, inumIterations );
        }

        gettimeofday(&currentTime, NULL );
        sec = currentTime.tv_sec;
        usec = currentTime.tv_usec;
        
        recvMatrix();

        gettimeofday(&currentTime, NULL );
        syncsec += currentTime.tv_sec - sec;
        usec = currentTime.tv_usec - usec;

        if (usec < 0) {
            syncsec--;
            usec += 1000000;
            syncusec += usec;
        }

        // start doing the computation
        doComputation();
     
        // write our results to the destination pipes
        sendMatrix();
    
    }
    
    gettimeofday( &currentTime, NULL );
    fprintf( stderr, "[%3d.%06d] Computation complete in %d iterations\n", currentTime.tv_sec, currentTime.tv_usec, inumIterations+1 );    // see what the results are

    sec = currentTime.tv_sec - startTime.tv_sec;
    usec = currentTime.tv_usec - startTime.tv_usec;
    if (usec < 0) {
        sec--;
        usec += 1000000;
    }

    fprintf(stderr, "\n\nEXECUTION TIME : \n\n%d seconds, %d useconds\n\n", sec, usec);
    fprintf(stderr, "\n\nSYNCH. TIME : \n\n%d seconds, %d useconds \n\n", syncsec, syncusec);

    fprintf(stderr, "The final state of the rod between points %d and %d: |", ibeginRange, iendRange);
    for (i = ibeginRange; i <= iendRange; i++)
        fprintf(stderr, " %f |", flocalMatrix[i][0]);

    // get rid of the pipes
    CloseDistPipe (isourcenDPD);
    DeleteDistPipe (ssourcenPipe);
    CloseDistPipe (idestpDPD);
    DeleteDistPipe (sdestpPipe);
    CloseDistPipe (isourcepDPD);
    DeleteDistPipe (ssourcepPipe);
    CloseDistPipe (idestnDPD);
    DeleteDistPipe (sdestnPipe);

    CleanUpUserProcess();
    
    return 0;
} // main
