/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
/*
 * HEADER FILE FOR THE STEADY STATE PIPELINE
 *
 * Following type convention for naming variables : <datatype initial><var>
 *
 * DESCRIPTION :
 
 
     ---- 
  { | T2 |
N {  ----
O { | ?? |
D {  ----
E { |  . |
  { |  . |
O { |  . |
N {  ----   _   _
E { | ?? |   \ /  
     ----     \   boundary value exchange  
N { | ?? |  _/ -- 
O {  ----
D { |  . |
E { |  . |
  { |  . |
T {  ----
W { | ?? |
O {  ----   _   _
  { | ?? |   \ /
     ----     \   boundary value exchange                ================>>>>>>  proceeds in time until steady state
    |  . |  _/ --
    |  . |
    |  . |
    |  . |
    |  . |  _   _
    |  . |   \ /
     ----     \   boundary value exchange
N { | ?? |  _/ --
O {  ----
D { | ?? |
E {  ----
  { | ?? |
N {  ----
U { | ?? |
M {  ----
N { | ?? |
O {  ----
D { | ?? |
E {  ----
S { | T1 |
     ----

*
*/

#ifndef JU_STEADYSTATE_H
#define JU_STEADYSTATE_H

#define INUMGRIDS 4800 // INUMGRIDS is the number of grids on the rod
#define FT1 500.0 // FT1 is the constant temperature of one end's heat bath - IF ALTERING, MAKE IT EQUAL TO FT2
#define FT2 500.0 // FT2 is the constant temperature of the other end's heat bath - IF ALTERING, MAKE IT EQUAL TO FT1
#define ITIMESTEP 5

void initializePipes(void);
void doComputation(void);
char *scolumnToString(const double *);
char *spositionToName(int); 
double steadyState(double, double, double);
int inameToPosition(const char *);
int inameToPosition(const char *);
int sendMatrix(void);
int recvMatrix(void);

#endif

