/***************************************************************
Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.		*
								*
This program is free software; you can redistribute it and/or	*
modify it under the terms of the GNU General Public License	*
as published by the Free Software Foundation; either version 2	*
of the License, or (at your option) any later version.		*
								*
This program is distributed in the hope that it will be useful,	*
but WITHOUT ANY WARRANTY; without even the implied warranty of	*
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	*
GNU General Public License for more details.			*
								*
You should have received a copy of the GNU General Public License*	
along with this program; if not, write to Prof. D. Janakiram, 	*
DOS Lab, Department of CS & E, Indian Institute of Technology Madras,* 
Chennai - 600036, India. Alternatively, you can send a mail to 	*
d.janakiram@cs.iitm.ernet.in.					*	
*****************************************************************/
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <math.h>

#include "juSteadyState.h"
#define u2_ST_AUX
#include "juSteadyStateAux.h"

/* 
 * sColumnToString converts the array of doubleing point numbers in the pipeline 
 * to a string
 * to write into the pipe as input to the next node
 * Encoding format is as follows :-
 * strings are padded with zeros
 */
char *scolumnToString( const double *numlist ){
  int i;
  int temp;
  char *bufferptr, *newbufferptr;
  
  bufferptr = (char *)malloc( 256 );
  bufferptr[0] = NULL;

  for( i = 0; i < INUMGRIDS; i++ ){
    sprintf( bufferptr, "%s %f", bufferptr, numlist[i] );
    if( (temp = strlen( bufferptr )) > 220 ){
      newbufferptr = (char *)malloc( (temp / 256 + 2) * 256 );  
      strcpy( newbufferptr, bufferptr );
      free( bufferptr );
      bufferptr = newbufferptr;
      newbufferptr = NULL;
    }
  }

  return bufferptr;

} //scolumnToString

/*
 * inameToPosition returns the position of the node in the manually laid out pipeline
 *
 * returns -1 if the machine is not part of the pipeline
 */
int inameToPosition(const char *nodeName) {
  int i;
  
  for( i = 0; i < NUMNODES; ++i ){
    if( !strcmp( szNodeList[i], nodeName ) )
      return i;
  }

  fprintf( stderr, "FATAL: not supposed to reach here!" );
  return -1; // machine not part of pipeline!

}

/*
 * spositionToName is the mirror function of inameToPosition. Given an integer, it returns
 * the node in the pipeline which corresponds to that integer. If that many nodes do not
 * exist in the pipeline, it returns NULL
 */
char *spositionToName(int pos) {
  if( pos < 0 )
    return( szNodeList[NUMNODES-1] );
  else if ( pos >= NUMNODES )
    return( szNodeList[0] );
  else
    return szNodeList[pos];
} // spositionToName

/*
 * steadyState is the central steadyState computation : have a dummy for now - just takes the average
 *
double steadyState (double farg1, double farg2, double farg3) {
    double i;
  
    for( i = 0.0; i < 1.0; i += 0.05 )
        i += 0.005;
  
    return fabs(fabs(farg2) + (fabs(farg3)  + fabs(farg1) - fabs(farg2) * 2.0)/1.93576);
}*/

double steadyState (double farg1, double farg2, double farg3) {
  //  return (farg1+farg2+farg3)/3;
  double i;
  for( i = 0.0; i < 1.0; i += 0.004 ){
    i += 0.0005;
    i /= 1.0;
  }
  return fabs(1.0 * fabs(farg2) + 1.0 * (fabs(farg3) + 1.0 * fabs(farg1) - 1.0 * fabs(farg2) * 2.0)/3.1756);
}

