/* main() allocates memory for all arrays used */

/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */


#include<stdio.h>
#include "primitives.c"

#define MAX_LEN 20

typedef struct 
  {
	float a[MAX_LEN + 1];
	float b[MAX_LEN + 1];
	int n;
  } arg_struct_t ;

typedef struct 
  {
	int row;
	int col;
  } map_struct_t;

map_struct_t *global_map;

void get_col(float **array, int col, int n, float *ret_col)
  {
	int i;

	for(i = 1; i <= n; i++)
	  {
		ret_col[i] = array[i][col];
	  }
  }

void get_row(float **array, int row, int n, float *ret_row)
  {
	int i;

	for(i = 1; i <= n; i++)
	  {
		ret_row[i] = array[row][i];
	  }
  }

void generate_map(int n)
  {
	int i,j;
	int index;

	index = 1;
	for(i = 1; i <= n; i++)
		for(j = 1; j <= n; j++)
		  {
			global_map[index].row = i;
			global_map[index++].col = j;
		  }
  }

char *matmul(char *argument)
  {
	arg_struct_t args;
	int i;
	float *p_product;

	p_product = (float *)malloc(sizeof(float));
	(*p_product) = 0;
	args = *(arg_struct_t *)(argument);
	for(i = 1;i <= args.n;i++)
	  {
		(*p_product) += (args.a[i] * args.b[i]);
	  }
	
	return((char *)(p_product));
  }

void main()
  {
	int n;
	int i,j,k;
	int global_num, global_low, global_high;
	int num_needed;
	int *tag;
		
	float **a, **b;
	float *dummy_row;
	float product;
	float *result;

	arg_struct_t args;
	LFmessage lf;
	struct datapack dp;
	char *argument;
	char *tempcp;

	/* Map generation */

	printf("\nEnter the dimension of the matrix :");
	scanf("%d", &n);
	global_map =(map_struct_t *)malloc((n * n + 1) * sizeof(map_struct_t));
	generate_map(n);

	/* Allocation of memory space for the matrices */

	a = (float **)malloc((n + 1) * sizeof(float *));
	b = (float **)malloc((n + 1) * sizeof(float *));
	for(i = 1; i <= n; i++)
	  {
		a[i] = (float *)malloc((n + 1) * sizeof(float));
		b[i] = (float *)malloc((n + 1) * sizeof(float));
	  }
	
	dummy_row = (float *)malloc((n + 1) * sizeof(float));
	
	args.n = n;

	/* Initialization of values for the matrices */

	for(i = 1; i <= n; i++)
		for(j = 1; j <= n; j++)
		  {
			a[i][j] = i+j;
			b[i][j] = 1.0/(i+j);
		  }
	/* -----------------------------------------*/

	global_num = 1;
	num_needed = n * n;
	argument = (char *)malloc(sizeof(arg_struct_t));
	tag = (int *)malloc((n * n + 1) * sizeof(int));
	result = (float *)malloc((n * n + 1) * sizeof(float)); 
	initialize_ARC();
		
	while(global_num <= n*n)
	  {
		lf = get_load_factor(num_needed);
		printf("Number obtained : %d\n",lf.number);
		num_needed -= lf.number;
		global_low = global_num;
		global_high = global_num + lf.number - 1;
		for(i = global_low; i <= global_high; i++)
		  {
			get_row(a, global_map[i].row, n, args.a);
			get_col(b, global_map[i].col, n, args.b);
			for(k = 0; k < sizeof(arg_struct_t); k++)
			  {
			    argument[k] = *((char *)(& args) + k);
			  }
			dp = open_datapack();
			new_insert_datapack(&dp,argument,sizeof(arg_struct_t));
			close_datapack(&dp);
			printf("Sending work to %d",lf.LFinfo[i-global_low].index);
			fflush(stdout);
			tag[i] = send_work_outside("sparc.1.c",&dp,sizeof(float),600,3,lf.LFinfo[i-global_low].index);
		  }
		global_num = global_high + 1;
		
	  }


	for(i = 1; i <= n * n; i++)
	  {
	    tempcp = obtain_result(tag[i]);
	    if(tempcp == NULL)
	      {
		printf("Result %d not obtained\n",i);
		continue;
	      }
	    else
	      result[i] = *(float *)tempcp;
	  }

	for(i = 1; i <= n * n; i++)
	  {
	    printf("i:%f ", result[i]);
	  }

   	/* Clean up */ 

	free(a);
	free(b);
	free(dummy_row);
	printf("ALL OVER...\n");
  }

