/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */

/*
extern "C" {
#include "mytime.c"
}
*/
#define MAX_DIM 50  /* maximum dimension of the matrix */
#include <math.h>

typedef struct {
  int startRow;
  int endRow;
  int dim;
  int a[MAX_DIM][MAX_DIM];      /* generic argument to each task */
  int b[MAX_DIM][MAX_DIM];
} argumentType;

argumentType args;              /* initialised matrix */
argumentType *taskArgs;         /* arguments to tasks */
argumentType *result;          

/* ********************************************************************** 
   Function      : partition() 
   Purpose       : Calculates the fragments (granularity) for each task based 
   on load 
   Arguments     : Total granularity and the load factors
   Return Values : array of task granularities

   ********************************************************************** */

int *partition(int total, float *loads)
{
  int noOfMachines;
  int *fragments;
  int i;
  int totalFrag;
  float totalLoad;
  
  totalLoad = 0.0;
  noOfMachines = (int)loads[0];
  for(i = 1; i <= noOfMachines; i++)
    totalLoad += loads[i];

  fragments = (int *)malloc((noOfMachines + 1) * sizeof(int));
  fragments[0] = noOfMachines;
  totalFrag = 0;
  for(i = 1; i < noOfMachines; i++){
    fragments[i] = (int)((loads[i]/totalLoad) * total);
    totalFrag += fragments[i];
  }
  fragments[noOfMachines] = total - totalFrag;

  return(fragments);
}

/* **********************************************************************
   Function      : matmul
   Purpose       : Multipiles 2 matrices
   Arguments     : generic task arguments
   Return Values : generic task argument containing the result matrix

   ********************************************************************** */

char* matmul(char* data)
{
  argumentType* argument;
  argumentType* result;
  int sum;
  int i,j,k;

  argument = (argumentType *)data;
  result = (argumentType *)malloc(sizeof(argumentType));
  
  result->startRow = argument->startRow;
  result->endRow = argument->endRow;
  result->dim = argument->dim;

  for(i=argument->startRow; i<=argument->endRow; i++){
    for(j=0; j<argument->dim; j++){
      sum = 0;
      for(k=0; k<argument->dim; k++){
	sum+= ((argument)->a[i][k]) * ((argument)->b[k][j]);
      }
      (result)->a[i][j] = sum;
    }
  }
  return((char *)result);
}

int main()
{
  int i,j,k;
  int numIter;
  int numOfMachines;
  int granOver,granCurrent,dim,startRow;
  int *tag;
  int *fragments;
  float *loads;
  LFmessage lf;
  

  printf("Enter Number Of Iterations :");
  scanf("%d", &numIter);
  printf("Enter dimension : ");
  scanf("%d", &(args.dim));
  dim = args.dim;
  printf("Enter the maximum number of machines available to you : ");
  scanf("%d", &numOfMachines);

      

        /* Initialization of the matrices 'a' and 'b' */

  for(i=0;i<args.dim;i++){
    for(j=0;j<args.dim;j++){
      args.a[i][j] = 0;
      args.b[i][j] = rand() % 10;  /* to ensure a number between 0 and 9 */
    }
  }
  
  for(i=0;i<args.dim;i++)
    args.a[i][i] = 1;

  printf("\nMatrix a :\n");
  for(i = 0; i < args.dim; i++){
    for(j = 0; j < args.dim; j++)
      printf("%d ",args.a[i][j]);
    printf("\n");
  }
  printf("\nMatrix b :\n");
  for(i = 0; i < args.dim; i++){
    for(j = 0; j < args.dim; j++)
      printf("%d ",args.b[i][j]);
    printf("\n");
  }

  /* **************************** */

  

  initialize_ARC();

      /* Now starts the iterative process */

  granOver = granCurrent = 0;
  startRow = 0;
  while(granOver < args.dim){
    lf = get_load_factor(numOfMachines);
    tag = (int *)malloc(sizeof(int) * lf.number);
    loads = (float *)malloc(sizeof(float) * (lf.number + 1));
    loads[0] = lf.number;
    taskArgs = (argumentType *)malloc(lf.number * sizeof(argumentType));
    for(j = 1; j <= lf.number; j++){
      loads[j] = 1.0/((lf.LFinfo[j - 1]).loadFactor);
    }
    
    if((dim - granOver) < dim/numIter){
      granCurrent = dim - granOver;
      granOver = dim;
    }                                     /* Decide on current granularity */
    else{
      granCurrent = dim/numIter;
      granOver += granCurrent;
    }
    
    fragments = partition(granCurrent,loads);
    
    for(j = 1; j <= lf.number; j++){
      printf("machine = %d, load = %f, grain = %d\n",(lf.LFinfo[j-1]).index,(lf.LFinfo[j-1]).loadFactor,fragments[j]);
    }
  
       /* The following sets up the task arguments for the ensuing calls */

    taskArgs[0].startRow = startRow;
    for(j = 0; j < lf.number - 1; j++){
      taskArgs[j].endRow = taskArgs[j].startRow + fragments[j + 1] - 1;
      taskArgs[j + 1].startRow = taskArgs[j].endRow + 1;
    }
    taskArgs[lf.number - 1].endRow = taskArgs[lf.number - 1].startRow + fragments[lf.number] - 1;
    startRow = taskArgs[lf.number - 1].endRow + 1;
    
    for(j = 0; j < lf.number; j++){
      taskArgs[j].dim = dim;
      for(i = 0; i < dim; i++)
	for(k = 0; k < dim; k++){
	  taskArgs[j].a[i][k] = args.a[i][k];
	  taskArgs[j].b[i][k] = args.b[i][k];
	}
    }

          /* submit the tasks */
    
    for(j = 0; j < lf.number; j++){
ARC_function_call(matmul, 100, 3, &taskArgs[j], sizeof(argumentType), lf.LFinfo[j].index, tag[j],  sizeof(argumentType));
    }

          /* obtain the results */
    
    free(loads);
    free(taskArgs);
    printf("\nThe Result Matrix is :\n");
    for(j = 0; j < lf.number; j++){
      result = (argumentType *)obtain_result(tag[j]);
      if(result != NULL){
	for(i = result->startRow; i <= result->endRow; i++){
	  for(k = 0; k < dim; k++)
	    printf("%d ", (result)->a[i][k]);
	  printf("\n");
	}
      }
      else{
	printf("Result for tag %d not available",tag[j]);
      }
    }
    free(tag);
        
  
  }  /* end of iteration */   
  

  
  /*
  for(i=0; i<args.dim; i++){
    for(j=0; j<args.dim; j++){
      printf("%d ",args.a[i][j]);
    }
    printf("\n");
  }

  printf("\n");

  for(i=0; i<args.dim; i++){
    for(j=0; j<args.dim; j++){
      printf("%d ",args.b[i][j]);
    }
    printf("\n");
  }
  
   printf("\n");

  for(i=0; i<result.dim; i++){
    for(j=0; j<result.dim; j++){
      printf("%d ",result.a[i][j]);
    }
    printf("\n");
  }
  */

  printf("over\n");

  close_ARC();
}


