/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */

/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * GCError.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the error routines used for displaying the error message.
 * this is used for logging the error details.
 *
 * Developer
 * ~~~~~~~~~~
 * N V Palankeswara Rao
 *
 * Date 
 * ~~~~
 * 06/March/2003
 *
 ***********************************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include "GCError.h"


int g_nHowToLog = FILE_ERROR_LOG;
int g_nLogFileDescriptor;

static char g_szGCText[2048];		


int GCLogError(a_pszErrorMessage)
	char *a_pszErrorMessage;
{
  	static char *l_pszLoggingTooMuch = "\n*** logging truncated\n";
  	static int l_nHowMuchAlreadyLogged = 0;	
  	int l_nErrorMessageLength = strlen(a_pszErrorMessage);
  	char l_szBuffer[40];
  	int l_nBufferLength;
  
  	time_t l_stNow;
  	struct tm *l_pstLocalTime;
  
  	time(&l_stNow);
  	l_pstLocalTime = localtime(&l_stNow);
  
  
  	if((g_nHowToLog & STANDARD_ERROR_LOG) == STANDARD_ERROR_LOG)
    	{
      		fprintf(stderr, "[PID %d] %02d/%02d %02d:%02d:%02d ", getpid(),
	      	l_pstLocalTime->tm_mon + 1,
	      	l_pstLocalTime->tm_mday,
	      	l_pstLocalTime->tm_hour,
	      	l_pstLocalTime->tm_min,
	      	l_pstLocalTime->tm_sec);
    	}
  	else if(l_nHowMuchAlreadyLogged < GC_LOGMAX) 
      	{
	
		sprintf(l_szBuffer, "[PID %d] %02d/%02d %02d:%02d:%02d ", getpid(),
		l_pstLocalTime->tm_mon + 1,
		l_pstLocalTime->tm_mday,
		l_pstLocalTime->tm_hour,
		l_pstLocalTime->tm_min,
		l_pstLocalTime->tm_sec);
	
		l_nBufferLength = strlen(l_szBuffer);
	
		(void)write(g_nLogFileDescriptor, l_szBuffer, l_nBufferLength);
		(void)write(g_nLogFileDescriptor, a_pszErrorMessage, l_nErrorMessageLength);
	
		if((l_nHowMuchAlreadyLogged += l_nErrorMessageLength + l_nBufferLength) >= GC_LOGMAX)
	  		(void)write(g_nLogFileDescriptor, l_pszLoggingTooMuch, strlen(l_pszLoggingTooMuch));
     
      	}

  	return 1;
}
  

int GCLogClusterError(a_pszErrorMessage)
     	char *a_pszErrorMessage;
{
  
  	(void)sprintf(g_szGCText, "%s: %s\n", a_pszErrorMessage,
		((errno >= 0 && errno < sys_nerr)
		 ? sys_errlist[errno] : "Unknown Error"));

  	return GCLogError(g_szGCText);
}


int GCOpenLog()
{
  	char l_szBuffer[160];
  
  	if(!(g_nHowToLog & STANDARD_ERROR_LOG))
    	{
      		(void)sprintf(l_szBuffer, GC_LOGFILE, getpid());
      
      		if((g_nLogFileDescriptor = open(l_szBuffer, O_WRONLY|O_TRUNC|O_CREAT|O_APPEND, 0600)) != -1)
			g_nHowToLog |= STANDARD_ERROR_LOG;
    }

  return 1;
}
