/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * GCHost.c
 *
 * Purpose 
 * ~~~~~~~
 * This contains a function to initialize a host descriptor 
 *
 * Developer 
 * ~~~~~~~~~~
 * N V Palankeswara Rao
 *
 * Date 
 * ~~~~
 * 04/March/2003
 *
 ***********************************************************************************************/

#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <sys/un.h>
#include <pwd.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "GCHost.h"
#include "GCDefines.h"
#include "GCError.h"
#include "GCNetLevel.h"
#include "GCMesg.h"

char *g_pszProgramName;


/************************************************************************************************************
 * Description: 
 * 	This Function is used to intialize the host details which may be a TCP or UNIX stream host
 * 
 * Arguments:
 * 	a_pstHostDetails - Pointer to the Host structure
 *      a_nWhichAddress - Intialize the TCP / UNIX Stream address structure
 * 
 * Return:
 * 	Error Code / 1
 *       
 * Note:- This function also initializes some un-used fields of the structure in the program. 
 * But they may be useful in future for providing extra functionality such as security. 
 * 
 ************************************************************************************************************/

int InitializeHost(a_pstHostDetails, a_nWhichAddress)
	PHOST_DESCRIPTOR a_pstHostDetails;
     	int a_nWhichAddress;
{
  	struct hostent *l_pstHostEntry;
  	char *hostname;
  	struct in_addr addr;
  	uid_t l_uidtUserID;
  	struct passwd *l_pstPasswordEntry;
  
	/* 
	 * Allocating memory for all pointers to structures, 
	 * which is very much required, otherwise memory errors may come.
	 */
	
//	a_pstHostDetails = (PHOST_DESCRIPTOR)malloc(sizeof(HOST_DESCRIPTOR));
  
	l_pstHostEntry = (struct hostent *)malloc(sizeof(struct hostent));
	
	l_pstPasswordEntry = (struct passwd *)malloc(sizeof(struct passwd));
	
		
	GC_ZERO(a_pstHostDetails, sizeof(HOST_DESCRIPTOR));

  	/* Allcoate memory for Host Name, Login Name, Path of the executable, socket address */
  
  	GC_CALLOC(a_pstHostDetails->m_pszHostName, char, HOSTNAME_LENGTH);
  	GC_CALLOC(a_pstHostDetails->m_pszHostLogin, char, LOGINNAME_LENGTH);
  	GC_CALLOC(a_pstHostDetails->m_pszDaemonPath, char, PATH_LENGTH);

  	if(a_nWhichAddress == TCP_AND_UNIXSTREAM_HOST)
    	{
      		GC_CALLOC(a_pstHostDetails->m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
      		GC_CALLOC(a_pstHostDetails->m_pstUNIXStreamHostAddress, struct sockaddr_un, sizeof(struct sockaddr_un));
    	}
  
	if(a_nWhichAddress == UNIXSTREAM_HOST)
    	{
      		GC_CALLOC(a_pstHostDetails->m_pstUNIXStreamHostAddress, struct sockaddr_un, sizeof(struct sockaddr_un));
    	}

  	if(a_nWhichAddress == TCP_HOST)
    	{
      		GC_CALLOC(a_pstHostDetails->m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
    	}
  
  	hostname = (char *)malloc(HOSTNAME_LENGTH);
  	gethostname(hostname, HOSTNAME_LENGTH);
  	a_pstHostDetails->m_nHostProcessID = getpid();
  	
	printf("\nInitialized Host Name: %s\n", hostname);
	//write(1,hostname,sizeof(hostname));

  	l_uidtUserID = getuid();
  	l_pstPasswordEntry = getpwuid(l_uidtUserID);

  	if(l_pstPasswordEntry->pw_name == NULL)
    	{
      		strcpy(a_pstHostDetails->m_pszHostLogin, "      ");
    	}
  	else
    	{
      		strcpy(a_pstHostDetails->m_pszHostLogin, l_pstPasswordEntry->pw_name);
    	}


  	getcwd(a_pstHostDetails->m_pszDaemonPath, PATH_LENGTH);

 	
	/* Set the members in the strtucture variable */
  
	l_pstHostEntry = gethostbyname(hostname);

  	if(l_pstHostEntry == NULL)
    	{
      		//GCLogClusterError("gethostbyname");
		return HOST_GETHOSTBYNAME_ERROR;
    	}
 
	
	/* Must understand this better */
  
	if((a_nWhichAddress == TCP_AND_UNIXSTREAM_HOST) || (a_nWhichAddress == TCP_HOST))
    	{
      		a_pstHostDetails->m_pstTCPHostAddress->sin_family = AF_INET;
    	}
  
	if((a_nWhichAddress == TCP_AND_UNIXSTREAM_HOST)  || (a_nWhichAddress == UNIXSTREAM_HOST))
    	{
      		a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_family = AF_UNIX;
    	}

  	memcpy(&addr, l_pstHostEntry->h_addr_list[0], sizeof(struct in_addr));
  	strcpy(a_pstHostDetails->m_pszHostName, inet_ntoa(addr));
  	
	write(1,"IP Addres :", 20);
	write(1, a_pstHostDetails->m_pszHostName, sizeof(a_pstHostDetails->m_pszHostName)+20);
  	write(1, "\n", 2);
	
	free(hostname);
  	
	return 1;
}
