/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */
#ifndef __GCMESG_H__

#define __GCMESG_H__


#include <rpc/types.h>
#include <rpc/xdr.h>


#define PIPENAME_LENGTH 256
#define HOSTNAME_LENGTH 256
#define LOGINNAME_LENGTH 256
#define PATH_LENGTH 256

/*Friend machine*/
#define SELECT_LENGTH 256

#define TCP_HOST 1
#define UNIXSTREAM_HOST 2
#define TCP_AND_UNIXSTREAM_HOST 3

/* Common message types for the first time sent to the cluster by the acceptor or the donor */
#define AD_CC_DATA	1000
#define AD_CC_CONTROL	1001

#define CC_AC_DATA	1000
#define CC_AC_CONTROL	1001

#define AC_CC_DATA	1000
#define AC_CC_CONTROL	1001

#define CC_DR_DATA	1000
#define CC_DR_CONTROL	1001

#define DR_CC_DATA	1000
#define DR_CC_CONTROL	1001

#define CC_CC_DATA	1000
#define CC_CC_CONTROL	1001

#define AC_DR_DATA	1000
#define AC_DR_CONTROL   1001

#define DR_AC_DATA      1000
#define DR_AC_CONTROL   1001

#define UP_AC_DATA	1000
#define UP_AC_CONTROL	1001

#define UP_CC_DATA	1000
#define UP_CC_CONTROL	1001


/* Acceptor -> Cluster Coordinator Messages */

#define AC_CC_GOODBYE		11
#define AC_CC_ADDREQ		12
#define AC_CC_GETHPF		13
#define AC_CC_TASK		14


/* Acceptor -> Donor Messages */

#define AC_DR_ADDREQ	21
#define AC_DR_GOODBYE	22


/* Donor -> Acceptor Messages */

#define DR_AC_ADDACK	31
#define DR_AC_GOODBYE	32


/* Cluster Coordinator -> Acceptor Messages */

#define CC_AC_SENDHPF	41
#define CC_AD_GOODBYE	42
#define CC_AC_ADDACK	43
#define CC_AC_DONORS    44
#define CC_AD_ADDREJ	45
#define CC_ACG_SENDHPF	46
#define CC_AC_TASKRESULT	47
#define CC_AC_CLUSTERS	48


/* Cluster Coordinator -> Donor Messages */

#define CC_DR_ADDACK	51
#define CC_DR_GETHPF   52
#define CC_DR_TASK      53


/* Cluster Coordinator -> Cluster Coordinator Messages */

#define CC_CC_ADDACK 	61
#define CC_CC_ADDREQ 	62
#define CC_CC_TASK 	63
#define CC_CC_ADDREJ    64
#define CC_CC_GETHPF 65
#define CC_CC_SENDHPF	66
#define CC_CC_DONORS	67
#define CC_CC_TASKRESULT	68


/* Donor-> Cluster Coordinator Messages */

#define DR_CC_ADDREQ 	71
#define DR_CC_TASK 	72
#define DR_CC_GOODBYE 	73
#define DR_CC_SENDHPF 	74
#define DR_CC_STATUSCHANGE  75


/* User Process -> Acceptor */
#define UP_AC_ADDREQ	81
#define UP_AC_PROBINITIALIZE	82


/* User Process -> Cluster Coordinator */
#define UP_CC_ADDCLIENT 	91
#define UP_CC_REMOVECLIENT	92
#define UP_CC_ADDFRIEND		93
#define UP_CC_REMOVEFRIEND	94


typedef struct
{
	int m_nMessageType;
  	int m_nFollowingMessageLength;
  	int m_nXDRDataLength;
  	char *m_pXDRData;
  	XDR m_XDRVariable;

} MESSAGE, *PMESSAGE;



#endif
