/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */
 

/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * acceptor.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for Acceptor's functions.
 *
 * Developers 
 * ~~~~~~~~~~
 * N V Palankeswara Rao, Hiren Ladva and Nilesh Baldwa
 *
 * Date 
 * ~~~~
 * 25/March/2003
 *
 ***********************************************************************************************/

#include <linux/kernel.h>  
#include <linux/sys.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "GCError.h"
#include "GCHost.h"
#include "GCMesg.h"
#include "GCXDRRoutines.h"
#include "Acceptor.h"
#include "GCDefines.h"
#include "GCNetLevel.h"
#include <unistd.h>

HOST_DESCRIPTOR g_stAcceptor;

int accsock;  // the socket descriptor which connects with the cluster
int usersock; // the socket descriptor which connects with the user process

int g_nNumberofDesc = 0;
int g_lSocketDescriptors[50];
int CurrentDescIndex = 0;
int NoOfDonors = 0, NoOfDonors1 = 0;
int NoOfClusters = 0;
int TaskID = 1;
double g_result;

struct timeval time1;
long start_s, start_m, end_s, end_m;

DRLIST  DRLStart, *pDRLStart = NULL, *pDRLEnd = NULL;
TASK TStart, *pTStart = NULL, *pTEnd = NULL;

int InitializeAcceptor(char *a_pszClusterCoordinator);
int AcceptMessages(void );
void CleanUpAcceptor(int a_nSignalNumber);
int CloseHost();

typedef void Sigfunc(int);


/************************************************************************************************************
 * Description : 
 *               Main program 
 * Arguments   :
 *               a_nNumberofArguments - Number of command line arguments passed.
 *               a_pszArguments - command line arguments.
 * Return      :
 *               Error Code or 1
 ************************************************************************************************************/

int main(a_nNumberofArguments, a_pszArguments)
     	int a_nNumberofArguments;
 	char **a_pszArguments;
{

/**********
	#ifdef SIGTERM
  	signal(SIGTERM, CleanUpAcceptor);
	#endif

************************************/

	if(a_nNumberofArguments != 2)
	{
		
		//write(1, "\nAcceptor: Enter CC id as a command line argument..\n", 60);
		
		printf("\nAcceptor: Enter CC id as a command line argument..\n");
 
		
		return(1);
	}
	
  	if(InitializeAcceptor(a_pszArguments[1]) < 0)
  	{
		//write(1, "\nAcceptor: Unable to Initialise the Acceptor...", 50);
		printf("\nAcceptor: Unable to Initialise the Acceptor...\n");
		exit(1);
  	}

   	if(AcceptMessages() < 0)
	{
		//write(1, "\nAcceptor: Not able to accept the messages....", 45);
		printf("\nAcceptor: Not able to accept the messages....\n");
  		exit(1);
	}

	return 1;
}


/**********************************************************************************************
 * Description : 
 *               Intialization of the acceptor in which it contacts 
 *               the cluster coordiantor and sends its details.
 *               
 * Arguments   :
 *               a_pszClusterCoordinator - Name of the machine where 
 *               		the cluster coordiantor is running
 *               		
 * Return      :
 *               Error Code or 1
 ***********************************************************************************************/

int InitializeAcceptor(a_pszClusterCoordinator)
     	char *a_pszClusterCoordinator;
{
	
	int l_nErrorCode;
	int l_nMessageID;

	/* Trap the termination signal so that when the
	 kill command is issued to the process it does
	 a gracefull exit from the code
	*/
/**********************************	
	#ifdef SIGTERM
  	signal(SIGTERM, CleanUpAcceptor);
	#endif
	*********************************************/
  	
	pDRLStart = NULL;
  	
	if(!InitializeHost(&g_stAcceptor, TCP_AND_UNIXSTREAM_HOST))
	{
		//write(1, "\nAcceptor: Unable to Initialise the host", 45);
		printf("\nAcceptor: Unable to Initialise the host\n");
		return -1;
	}
 	 	 	
 	if((usersock = CreateUNIXStreamServerHostAndListen(&g_stAcceptor, AC_UP_PORT )) < 0)
		return usersock;
 	 	
 	/* Create a Socket and Connect to the Cluster
 	 	 g_stAcceptor is host structure of global...*/
   	if((accsock = CreateTCPClientHostAndConnect(&g_stAcceptor, 
					a_pszClusterCoordinator, CC_AC_PORT)) < 0)
    	{
      		return accsock;
    	}
  	
/****	write(1, "\nAcceptor: Connected to the CC", 35); ********/
  	
  	/* Send a *Add Request* Message to Cluster Coordinator */

  	l_nMessageID = AC_CC_ADDREQ;
  	
  	if((l_nErrorCode = WriteXDRSignedInteger(accsock, AC_CC_CONTROL, (void *)&l_nMessageID,
						 sizeof(l_nMessageID), TCP_HOST)) < 0)
      		return l_nErrorCode;

  	/* Send the Acceptor Details to Cluster Coordinator */

  	/* Send Host Name */
  	
  	if((l_nErrorCode = WriteXDRString(accsock, AC_CC_DATA, (void *)g_stAcceptor.m_pszHostName,
						HOSTNAME_LENGTH, TCP_HOST)) < 0)
	      	return l_nErrorCode;

  	/* Send Address */
	if((l_nErrorCode = WriteMessageAndData(accsock, AC_CC_DATA, 
			   	(char *)g_stAcceptor.m_pstTCPHostAddress, 
			    	sizeof(struct sockaddr_in))) < 0)
	      	return l_nErrorCode;

	/* Send Process Id */
	if((l_nErrorCode = WriteXDRSignedInteger(accsock, AC_CC_DATA, 
				(void *)&g_stAcceptor.m_nHostProcessID, 
				sizeof(g_stAcceptor.m_nHostProcessID), TCP_HOST)) < 0)
	      	return l_nErrorCode;

	/* Send Login Name*/
	if((l_nErrorCode = WriteXDRString(accsock, AC_CC_DATA, (void *)g_stAcceptor.m_pszHostLogin, 
						LOGINNAME_LENGTH, TCP_HOST)) < 0)
		return l_nErrorCode;

  	/* Send Daemon path*/
  	if((l_nErrorCode = WriteXDRString(accsock, AC_CC_DATA, (void *)g_stAcceptor.m_pszDaemonPath, 
						PATH_LENGTH, TCP_HOST)) < 0)
	      	return l_nErrorCode;
	
	/* Receive the Acknowledgment*/
	if((l_nErrorCode = ReadXDRSignedInteger(accsock, CC_AC_CONTROL, (void *)&l_nMessageID,
						sizeof(l_nMessageID), TCP_HOST)) < 0)
	      	return l_nErrorCode;	
      		
	if(l_nMessageID == CC_AC_ADDACK )
	     	//write(1, "\nAcceptor: Ack received from the CC", 45);
	     	printf("\nAcceptor: Ack received from the CC\n");
	else
	      	return -1;

	return 1;
}


/********************************************************************************************
 * Description : The acceptor listens for the messages from the cluster coordinator, the
 *		 donor and the user process.
 * Arguments   : none
 * Return      : Error Code / 1
 *********************************************************************************************/

int AcceptMessages()
{
	int nsock;
	int j;
	int l_nErrorCode;
	fd_set rfd;
	g_nNumberofDesc = 0;

	FD_ZERO(&rfd);
		
	/*wait for messages from anywhere*/
	while(1)
	{
		FD_SET(accsock, &rfd);
		FD_SET(usersock, &rfd);
		
		for(j = 0; j < g_nNumberofDesc; j++)
 		     	FD_SET(g_lSocketDescriptors[j], &rfd);
		
		if(AnyDataToRead(&rfd) == 1)
	        {
			for(j = (g_nNumberofDesc-1); j >= 0; j--)
			{
			  	if(FD_ISSET(g_lSocketDescriptors[j], &rfd))
		     		{
	        			nsock = g_lSocketDescriptors[j];
		        		CurrentDescIndex = j;
			        	if((l_nErrorCode = ProcessMessage(nsock)) < 0)
  			         	{
						//write(1, "\nAcceptor: Error in processing a Message", 45);
						printf("\nAcceptor: Error in processing a Message\n");
						CloseHost();
	 		         	}
		      		}
		  	}
	  		
 			/* Has a message come from the cluster coordinator */
	  		if(FD_ISSET(accsock, &rfd))
		     	{
	    	        	if((l_nErrorCode = ProcessMessage(accsock)) < 0)
  		         	{
					//write(1, "\nAcceptor: Error in processing Message", 45);
					printf("\nAcceptor: Error in processing Message\n");
					CleanUpAcceptor(SIGTERM);
 	         		}
	      		}
			
 			/* Has a message come from the user process */
			if(FD_ISSET(usersock, &rfd))
		     	{
	     			nsock = accept(usersock, (struct sockaddr *)NULL, 0);
				if((l_nErrorCode = ProcessUserMessage(nsock)) < 0)
				{
					//write(1, "\nAcceptor: Error in processing Message", 45);
					printf("\nAcceptor: Error in processing Message\n");
					close(nsock);
				}
	      		}
   		}
  	}
	
	return 1;
}


/**********************************************************************
 * Description: this function checks if any data is available
 * Argumsents : rfd - pointer to the set of socket descriptors
 * Return     : socket descriptor
 **********************************************************************/

int AnyDataToRead(rfd)
	fd_set *rfd;
{
  
	int l_nIsReady;
  	int l_nNumberofDescriptors;

  	l_nNumberofDescriptors  = getdtablesize() + 1;

  /*
    Using the select call which polls and returns immediatly
    if no data is avialable since Timeval is NULL
  */

  	if((l_nIsReady = select(l_nNumberofDescriptors, rfd, (fd_set *)NULL, (fd_set *)NULL, NULL)) > 0)
  	{
  	  	return 1;
  	}
   	
	else
   		return -1;
}


/***********************************************************************************************
 * Description: this function processes an arrived message and calls appropriate 
 * 		action to be taken depending on the sender and the message sent
 * 		
 * Arguments  : sock - socket descriptor on which the message has come
 * 				 
 * Return     : Error Code / 1

*************************************************************************************************/

int ProcessMessage(sock)
   	int sock;
{
	
	int l_nMessageID;
	int l_nErrorCode;
	
	/*get the message type */
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, CC_AC_CONTROL, (void *)&l_nMessageID,
						sizeof(l_nMessageID), TCP_HOST)) < 0)
	     	return l_nErrorCode;

	switch(l_nMessageID)
	{
		case CC_AD_GOODBYE:
			//write(1,"\nAcceptor: The CC is shutting down.", 36);
			printf("\nAcceptor: The CC is shutting down.\n");
	        	CleanUpAcceptor(SIGTERM);
			_exit(1);
			
		case CC_AC_SENDHPF:
			//write(1, "\nAcceptor: The HPF has come", 30);
			printf("\nAcceptor: The HPF has come\n");
			return ReceiveHPF(sock, l_nMessageID);
			
		case CC_ACG_SENDHPF:
			//write(1, "\nAcceptor: The HPF has come from other cluster ", 48);
			printf("\nAcceptor: The HPF has come from other cluster \n");
			return ReceiveHPF(sock, l_nMessageID);
			
		case CC_AC_DONORS:
			//write(1, "\nAcceptor: Receiving the no of donors", 40);
			printf("\nAcceptor: Receiving the no of donors\n");
			return GetDonors(sock);
			
		case CC_AC_CLUSTERS:
			//write(1, "\nAcceptor: Receiving the no of clusters", 41);
			printf("\nAcceptor: Receiving the no of clusters\n");
			return GetClusters(sock);			
			
		case DR_AC_ADDACK:
		        //write(1, "\nAcceptor: Donor Ack Received", 30);
		        printf("\nAcceptor: Donor Ack Received\n");
		        return ReceiveResults(sock);
			
		case DR_AC_GOODBYE:
		        //write(1, "\nAcceptor: donor is shutting down", 35);
		        printf("\nAcceptor: donor is shutting down\n");
		        close(sock);
		        return -1;
			
	        case CC_AD_ADDREJ:
	        	//write(1, "\nAcceptor: CC rejected this Acceptor connection", 48);
	        	printf("\nAcceptor: CC rejected this Acceptor connection\n");
	        	CleanUpAcceptor(SIGTERM);
			return 1;
			
	        case CC_AC_TASKRESULT:
	        	//write(1, "\nAcceptor: Result Received ", 30);
	        	printf("\nAcceptor: Result Received \n");
	        	return ReceiveResults(sock);
	}
	
	return 1;
}


/**********************************************************************
 * Description : this deletes all the memory allocated during the
  		 execution of the Acceptor daemon.
 * Arguments   : a_nSignalNumber - Number of signal because which this
  		 signal was called
 * Return      : None
 **********************************************************************/

void CleanUpAcceptor(a_nSignalNumber)
    	int a_nSignalNumber;
{
   
	int j;
   	int l_nMessageID = AC_CC_GOODBYE;
  	//write(1, "\nAcceptor: sending goodbye", 30);
  	printf("\nAcceptor: sending goodbye\n");
  	
   	WriteXDRSignedInteger(accsock, AC_CC_CONTROL, (void *)&l_nMessageID, 
				sizeof(l_nMessageID), TCP_HOST);

   	l_nMessageID = AC_DR_GOODBYE;
   	
	for(j = 0; j < g_nNumberofDesc; j++)
		WriteXDRSignedInteger(g_lSocketDescriptors[j], AC_CC_CONTROL, 
					(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST);

  	close(accsock);
  	close(usersock);

  	GC_FREE(g_stAcceptor.m_pszHostName);
  	GC_FREE(g_stAcceptor.m_pstTCPHostAddress);
  	GC_FREE(g_stAcceptor.m_pszHostLogin);
  	GC_FREE(g_stAcceptor.m_pszDaemonPath);

  	exit(1);
}


/***************************************************************
Description:  the signal function which will be called when the process
              is killed
Arguments:     default..of the signal function
****************************************************************/

Sigfunc *signal(int signo, Sigfunc *func)
{
	struct sigaction act, oact;
	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	
	if(sigaction(SIGTERM, &act, &oact) < 0)
		//write(1, "\nAcceptor: Not able to process the signal", 45);
		printf("\nAcceptor: Not able to process the signal\n");
	
	return(oact.sa_handler);
}


/*******************************************************************************
 * Description : this function updates the array of the open socket descriptors

*******************************************************************************/

int CloseHost()
{
  	int i;

  	/* Update the array of Descriptors opened */		
  	for(i = CurrentDescIndex; i < (g_nNumberofDesc-1); i++)
      		g_lSocketDescriptors[i] = g_lSocketDescriptors[i+1];

  	g_nNumberofDesc--;
  
	return 1;
}


/*************************************************************************
 * Description : this receives the load information from the cluster and 
 * 		 contacts the donors.
 * 		 
 * Arguments   : sock - socket descriptor of the cluster-acceptor connection.
 * 
 * Return      : 1 on success/ -ve on error
 * 
**************************************************************************/

int ReceiveHPF(sock, l_nMessageID)
	int sock;
	int l_nMessageID;
{
	
	DRLIST *t1, *t2;
	
	t2 = pDRLStart;
	
	t1 = malloc(sizeof(DRLIST));
	t1->DRHostName = (char *)malloc(HOSTNAME_LENGTH);
	t1->DRLoginName = (char *)malloc(LOGINNAME_LENGTH);
	t1->DRDaemonPath = (char *)malloc(PATH_LENGTH);
	t1->next = NULL;
		
	if(l_nMessageID == CC_ACG_SENDHPF ) //if the donors belong to a different LAN
	{
		/*Read the cluster ID*/
		if(ReadXDRSignedInteger(sock, CC_AC_DATA, (void *)&(t1->DRCSock),
					sizeof(t1->DRCSock), TCP_HOST) < 0)
			return -1;
		
		/*Read the Donor ID*/	
		if(ReadXDRSignedInteger(sock, CC_AC_DATA, (void *)&(t1->DRSock),
					sizeof(t1->DRSock), TCP_HOST) < 0)
			return -1;	
	}
	
	else //if the donors belong to the same LAN
	{
		t1->DRCSock = 0;
		t1->DRSock = 0;	
	}
		
	/*Read the hostnames*/
	if(ReadXDRString(sock, CC_AC_DATA, (void *)t1->DRHostName, HOSTNAME_LENGTH, TCP_HOST) < 0)
	      	return -1;
	
	/* Read Login Name */
	if(ReadXDRString(sock, CC_AC_DATA, (void *)t1->DRLoginName, LOGINNAME_LENGTH, TCP_HOST) < 0)
	      	return -1;
        			
	/* Read Daemon Path */
	if(ReadXDRString(sock, CC_AC_DATA, (void *)t1->DRDaemonPath, PATH_LENGTH, TCP_HOST) < 0)
		return -1;
		
	/*Read HPF Info*/
	if(ReadXDRDouble(sock, CC_AC_DATA, (void *)&(t1->DRHPF),
					sizeof(t1->DRHPF), TCP_HOST) < 0)
		return -1;

	if(pDRLStart == NULL)
	{
	 	pDRLStart = t1;
	 	pDRLEnd = t1;
	}
	
	else
	{
		pDRLEnd->next = t1;
		pDRLEnd = t1;
	}
	
	NoOfDonors--;

	printf("\n Remaining donors = %d, clusters = %d \n", NoOfDonors, NoOfClusters);
	
	if(NoOfDonors == 0 && NoOfClusters == 0)
	{
		SplitTheTask();
		ConnectDonors();
		
	}
	
/***	SearchList();   *****/
        
	return 1;
}


int SearchList()
{
	SPLIT *s;
	s = pTStart->pSStart;

	while(s != NULL)
	{
		printf("\nLower Bound: %lf \n", s->LB);
		printf("\nUpper Bound: %lf \n", s->UB);
		printf("\nCluster Socket= %d \n", s->DRCSock);
		printf("\nDonor Socket= %d \n", s->DRSock);
		s = (s->next);
	}
	
	return 1;
}


/******************************************************************************************
 * Description: this function splits the task into subtasks and updates the SPLIT table.
 * Return     : 1 on success and -1 on error
*******************************************************************************************/

int SplitTheTask()
{
	DRLIST *don;
	TASK *task;
	SPLIT *split;
	
	double t_HPF = 0;
	double l_LB;
	double g_size, temp;
	double size;
	int l_sid = 1;
	
	don = pDRLStart;
	
	task = pTStart;
	
	while(don != NULL)
	{
		t_HPF = t_HPF + (don->DRHPF);		
		don = don->next;
	}
	
	g_size = ((task->UB) - (task->LB) + 1) / (task->MinSplit);
	l_LB = task->LB;
	
	printf("\n Total HPF :: %lf\n", t_HPF);
	printf("\n Min Split  :: %lf\n", task->MinSplit);

        //size=g_size/NoOfDonors1;
		
	don = pDRLStart;
	while(don != NULL)
	{
		split = (SPLIT *)malloc(sizeof(SPLIT));
		split->DRHostName = (char *)malloc(HOSTNAME_LENGTH);
		
		split->LB = l_LB;
				
		if(don->next == NULL)
			split->UB = task->UB;
		else
		{
			temp = ((don->DRHPF)/t_HPF )*g_size;
			size = temp*(task->MinSplit);
			split->UB = l_LB+size;
		}
		
		printf("Donor HPF :: %lf\t", don->DRHPF);
		printf("Split  :%lf\n", size);
		l_LB= (split->UB) + (task->MinSplit);
			
		split->DRSock = don->DRSock;
		split->DRCSock = don->DRCSock;
		split->SID = l_sid++;
		strcpy(split->DRHostName, don->DRHostName);
		split->next = NULL;

		if(task->pSStart == NULL)
		{
			task->pSStart = split;
	 		task->pSEnd = split;
		}
		
		else
		{
			task->pSEnd->next = split;
			task->pSEnd = split;
		}
	
		don = don->next;
	}
	
	printf("HPF :: %lf", t_HPF);
	
	return 1;
}


/******************************************************************************************
 * Description: this function connects to the donors.
 * Return     : 1 on success and -1 on error.
*******************************************************************************************/

int ConnectDonors()
{
	int sock;
	int l_nMessageID;
	int len;
	int fd;
	struct stat buf;
	int l_nBytesRead;
	int l_nRemainingSize;
	char *l_pszRemainingBuffer;
	char *code, lb[25], ub[25];
	SPLIT *t1;
	TASK *t;
		
	t = pTStart;
	t1 = pTStart->pSStart;
	
	/*PREPARE THE DATA TO BE SENT TO THE DONORS*/
	
        /*reading the contents of the file passed by the user into the mem pointed by 'code' */
	//write(1, pTStart->filename, FILENAME_LENGTH);
	
	if((fd = open(pTStart->filename, O_RDONLY)) < 0)
	{
		//write(1, "\nAcceptor: Cant open the file", 30);
		printf("\nAcceptor: Cant open the file\n");
		exit(1);
	}

	fstat(fd, &buf);
	len = buf.st_size;
	printf("\nfile %s", pTStart->filename);
	printf("\nSize = %d\n", len);
	code = (char *)malloc(len);

	if(read(fd, code, len) != len)
		//write(1, "\nAcceptor: error in reading", 26);
		printf("\nAcceptor: error in reading\n");
	
	close(fd);	
	
	while(t1 != NULL)
        {
		if(t1->DRCSock == 0)
		{
		/*Sub-task has to send to a local donor */
			
			if((sock = CreateTCPClientHostAndConnect(&g_stAcceptor, 
							t1->DRHostName, DR_AC_PORT)) < 0)
	    		{
	      			return -1;
	    		}
    	                
			t1->DRSock = sock;
			g_lSocketDescriptors[g_nNumberofDesc++] = sock;

    	        	/* Send the Request*/
	 	 	l_nMessageID = AC_DR_ADDREQ;
  			if(WriteXDRSignedInteger(sock, AC_DR_CONTROL, (void *)&l_nMessageID,
						 sizeof(l_nMessageID), TCP_HOST) < 0)
      				return -1;
      			
      			/*send the size of the exe file*/	
      			if(WriteXDRSignedInteger(sock, AC_DR_DATA, (void *)&len,
						 sizeof(len), TCP_HOST) < 0)
      				return -1;	
      				
      			/*send the code to the donor*/	
			if(WriteXDRString(sock, AC_DR_DATA, (void *)code, len, TCP_HOST) < 0)
				return -1;
    		
			//write(1, "\nAcceptor: code sent to a Donor", 35);
			printf("\nAcceptor: code sent to a Donor\n");
			
			sprintf(lb, "%lf", t1->LB);
			sprintf(ub, "%lf", t1->UB);	
			
			//write(1, "\nAcceptor: sending the following bounds to the donor", 55);
			printf("\nAcceptor: sending the following bounds to the donor\n");
			write(1, lb, 25);
			write(1, ub, 25);
					
			/*send the lower bound of the task*/	
			if(WriteXDRString(sock, AC_DR_DATA, (void *)lb, strlen(lb), TCP_HOST) < 0)
    				return -1;
    	
    			/*send the upper bound of the task*/	
			if(WriteXDRString(sock, AC_DR_DATA, (void *)ub, strlen(ub), TCP_HOST) < 0)
    				return -1;
		}
		
		else
		{
			/*send the message to the cluster for forwarding the task to other networks*/
			l_nMessageID = AC_CC_TASK;
			if(WriteXDRSignedInteger(accsock, AC_CC_CONTROL, (void *)&l_nMessageID,
						 sizeof(l_nMessageID), TCP_HOST) < 0)
      				return -1;
      			
      			/*send the cluster ID*/	
    		        if(WriteXDRSignedInteger(accsock, AC_CC_DATA, (void *)&t1->DRCSock,
						 sizeof(t1->DRCSock), TCP_HOST) < 0)
      				return -1;
      			
      			/*send  the donor ID*/	
      			if(WriteXDRSignedInteger(accsock, AC_CC_DATA, (void *)&t1->DRSock,
						 sizeof(t1->DRSock), TCP_HOST) < 0)
      				return -1;
      			
      			/*send the size of the exe file*/	
      			if(WriteXDRSignedInteger(accsock, AC_CC_DATA, (void *)&len,
						 sizeof(len), TCP_HOST) < 0)
      				return -1;	
      				
      			/*send the code to the donor*/	
			if(WriteXDRString(accsock, AC_CC_DATA, (void *)code, len, TCP_HOST) < 0)
				return -1;
    		
			//write(1, "\nAcceptor: code sent to the cluster", 40);
			printf("\nAcceptor: code sent to the cluster\n");
			
			sprintf(lb, "%lf", t1->LB);
			sprintf(ub, "%lf", t1->UB);	
			
			/*send the lower bound of the task*/	
			if(WriteXDRString(accsock, AC_CC_DATA, (void *)lb, strlen(lb), TCP_HOST) < 0)
    				return -1;
    	
    			/*send the upper bound of the task*/	
			if(WriteXDRString(accsock, AC_CC_DATA, (void *)ub, strlen(ub), TCP_HOST) < 0)
    				return -1;
		}	
	    	
		t1 = t1->next;
	}
	
	free(code);
	
	return 1;
}


/**********************************************************************************************
 * Description: this function gets the count of the clusters and 
 * 		local donors from the local cluster.
 * Return     : 1 on success and -1 on error.
***********************************************************************************************/

int GetClusters(sock)
	int sock;
{
	
	int donors;
	
	/* Get the number of donors */
 	if(ReadXDRSignedInteger(sock, CC_AC_DATA, (void *)&donors,
				sizeof(donors), TCP_HOST) < 0)
	      return -1;

	/* Get the number of clusters */
 	if(ReadXDRSignedInteger(sock, CC_AC_DATA, (void *)&NoOfClusters,
				sizeof(NoOfClusters), TCP_HOST) < 0)
	      return -1;

	//write(1, "\nAcceptor: Received the no of donors and clusters", 55);
	printf("\nAcceptor: Received the no of donors and clusters\n");
	
	NoOfDonors = NoOfDonors + donors;
	NoOfDonors1 = NoOfDonors1 + donors;
	if(NoOfDonors == 0 && NoOfClusters == 0)
	{
		//write(1, "\nAcceptor: No donors available to perform the task", 55);
		printf("\nAcceptor: No donors available to perform the task\n");
		pTStart = NULL;
	}
	
	return 1;
}


/*******************************************************************************************
 * Description: this function gets the count of the donors from the other clusters.
 * Return     : 1 on success and -1 on error
********************************************************************************************/

int GetDonors(sock)
 	int sock;
{
	int donors;
	
	/* Get the number of donors */
 	if(ReadXDRSignedInteger(sock, CC_AC_DATA, (void *)&donors,
				sizeof(donors), TCP_HOST) < 0)
	  	return -1;

	//write(1, "\nAcceptor: Received the no of donors", 40);
	printf("\nAcceptor: Received the no of donors\n");
	
	NoOfDonors = NoOfDonors + donors;
	NoOfDonors1 = NoOfDonors1 + donors;
	NoOfClusters--;
	if(NoOfDonors == 0 && NoOfClusters == 0)
	{
		//write(1, "\nAcceptor: No donors available to perform the task", 55);
		printf("\nAcceptor: No donors available to perform the task\n");
		pTStart = NULL;
	}
	
	return 1;
}


/******************************************************************************************
  Description : This function is called to request the HPFs
  Arguments   : None
  Return      : 1 on Success/-1 on Error
*******************************************************************************************/
int GetHPF()
{

	/*Temporary Request for HPF Information*/
	int l_nMessageID = AC_CC_GETHPF;
	
	//write(1, "\nAcceptor: sending load req ", 30);
	printf("\nAcceptor: sending load req\n");
	if(WriteXDRSignedInteger(accsock, AC_CC_CONTROL, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)
	{
		//write(1, "\nAcceptor::unable to send", 30);
		printf("\nAcceptor::unable to send\n");
		return -1;
	}
	
	else	
		//write(1, "\nAcceptor::reqst sent ", 25);
		printf("\nAcceptor::reqst sent\n");
	
        return 1;
}


/********************************************************************************************
 * Description: this function processes an arrived message from user process and 
 * 		calls appropriate action to be taken depending on the message.
 * 		
 * Arguments  : sock - socket descriptor on which the message has come.
 * 
 * Return     : 1 on success and -1 on error.
*********************************************************************************************/

int ProcessUserMessage(sock)
   	int sock;
{
	
	int l_nMessageID;
	int l_nErrorCode;
		
	/*get the message type */
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, UP_AC_CONTROL, (void *)&l_nMessageID,
					sizeof(l_nMessageID), UNIXSTREAM_HOST )) < 0)
		return -1;

	switch(l_nMessageID)
	{
          	/*case UP_AC_ADDREQ:
	        	write(1,"Acceptor: A user process has requested to join",47);
	        	return ReceiveProcess(sock);*/
		
	  	case UP_AC_PROBINITIALIZE:
	        	//write(1, "\nAcceptor: A user process is sending data of the problem", 60);
	        	printf("\nAcceptor: A user process is sending data of the problem\n");
			if(pTStart != NULL)
			{
				//write(1, "\nAcceptor: A User Process is in progress...", 48);
				printf("\nAcceptor: A User Process is in progress...\n");
				return -1;
			}
	        	if(GetProbData(sock))
		        	return GetHPF();
		        else
		        	return -1;
	}
	
	return 1;
}


/***********************************************************************************************
 * Description: this function collects the data from the user process and 
 * 		updates the task table.
 * 		
 * Arguments  : sock - socket descriptor on which the message has come.
 * 
 * Return     : 1 on success and -1 on error
***********************************************************************************************/

int GetProbData(sock)
	int sock;
{
	
	TASK *t1;
	char lb[25], ub[25];
	
	t1 = (TASK *)malloc(sizeof(TASK));
	t1->filename = (char *)malloc(FILENAME_LENGTH);
	
	/*Register  the time of submission of the task*/
	gettimeofday(&time1, NULL);
	start_s = time1.tv_sec;
	start_m = time1.tv_usec;
		
	/* Read the min. split size of the Problem */
	if(ReadXDRDouble(sock, UP_AC_DATA, (void *)&(t1->MinSplit), 
				sizeof(t1->MinSplit), UNIXSTREAM_HOST ) < 0)
		return -1;
 	
	/* Read the Lower Bound of the Problem */
	if(ReadXDRString(sock, UP_AC_DATA, (void *)lb, 25, UNIXSTREAM_HOST ) < 0)
		return -1;
 	
	/* Read the Upper Bound of the Problem */
	if(ReadXDRString(sock, UP_AC_DATA, (void *)ub, 25, UNIXSTREAM_HOST ) < 0)
		return -1;

	write(1, lb, 25);
	write(1, ub, 25);	
	
	sscanf(lb, "%lf", &(t1->LB));
	sscanf(ub, "%lf", &(t1->UB));
				
	/* Read the name of the executable file that has to be processed */
	if(ReadXDRString(sock, UP_AC_DATA, (void *)(t1->filename), FILENAME_LENGTH, 
				UNIXSTREAM_HOST ) < 0)
		return -1;

	write(1, t1->filename, sizeof(t1->filename));			
	/* Read the mathematical operation that has to be performed on the splits */
	if(ReadXDRUnSignedChar(sock, UP_AC_DATA, (void *)&(t1->op),
				sizeof(t1->op), UNIXSTREAM_HOST ) < 0)
		return -1;
	
	t1->pSStart = NULL;
	t1->pSEnd = NULL;

 	pTStart = t1;
	
	return 1;
}


/*********************************************************************************************
*********************************************************************************************/

int ReceiveResults(sock)
	int sock;
{
        
	int l_nMessageID, j;
        char *res;
        double result;
	int len;
	char *temp;
	DRLIST *t;

	/*Receive the results*/
	/*receive the size of the results*/
	if(ReadXDRSignedInteger(sock, AC_DR_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
		return -1;

	res = (char *)malloc(len);
	
	if(ReadXDRString(sock, AC_DR_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;
		
	sscanf(res, "%lf", &result);	
	g_result = g_result + result;
	
	NoOfDonors1--; 
	
	if(NoOfDonors1 == 0)
	{
		//write(1, "\nAceeptor: results received\n", 32);
		printf("\nAceeptor: results received\n");
		temp = (char *)malloc(25);
		sprintf(temp, "%lf", g_result);
		write(1, temp, len);
	
		/*Reset the task entries in the task table*/	
		pTStart->pSStart = NULL;
		pTStart->pSEnd = NULL;
		free(pTStart);
		pTStart = NULL;
		t = pDRLStart;
		
		while(t != NULL)
		{
			t = t->next;
			free(pDRLStart);
			pDRLStart = t;
		}
		
		pDRLStart = NULL;
		g_result = 0;
		
		l_nMessageID = AC_DR_GOODBYE;
		
		for(j = 0; j < g_nNumberofDesc; j++)
			WriteXDRSignedInteger(g_lSocketDescriptors[j], AC_CC_CONTROL,
					(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST);
                
		g_nNumberofDesc = 0;

                /*Find the time of completion of entire task*/
                gettimeofday(&time1, NULL);
		end_s = time1.tv_sec;
		end_m = time1.tv_usec;
		
		end_s = end_s - start_s;
		
		if(end_m >= start_m)
            		end_m = end_m - start_m;
		else
		{
            		end_s = end_s - 1;
            		end_m = end_m + 1000000 - start_m;
            	}
            
		printf("\n ------------------------------------------------------\n");
		printf("Total time taken: sec:%ld  microsec:%ld \n", end_s, end_m);
		printf("\n ------------------------------------------------------\n");
	
                //free(temp);
	}
		
	//free(res);
	
	return 1;
}
