/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * GCClusterCoordinator.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for cluster coordinator's functions.
 *
 * Developers 
 * ~~~~~~~~~~
 * N V Palankeswara Rao, Nilesh Baldwa and Hiren Ladva
 *
 * Date 
 * ~~~~
 * 12/March/2003
 *
 ***********************************************************************************************/

#include <math.h>
#include <errno.h>
#include "GCDefines.h"
#include "GCClusterCoordinator.h"
#include "GCMesg.h"
#include "GCNetLevel.h"
#include "GCXDRRoutines.h"
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

int g_nNumberofFriends = 0;
int g_nNumberofAcceptors = 0;
int g_nNumberofDonors = 0;
int g_nNumberofDesc = 0;
int g_nSocketDescriptor[50];
int ClustersToBeDeleted[20];
int ClusterIndex = 0;
int CurrentDescIndex = 0;
int NoOfDonors = 0;

/* int usersock;   */

PHOST_DESCRIPTOR_COLLECTION g_pstListOfAcceptors = NULL;
HOST_DESCRIPTOR g_stClusterCoordinator;

ACTABLE  ACTStart, *pACTStart = NULL, *pACTEnd = NULL;
DRTABLE  DRTStart, *pDRTStart = NULL, *pDRTEnd = NULL;
FCCTABLE  FCCTStart, *pFCCTStart = NULL, *pFCCTEnd = NULL;
CTASK  *pTStart = NULL, *pTEnd = NULL;

/* FCTABLE  FCTStart, *pFCTStart = NULL, *pFCTEnd = NULL; */

CLUSTERS *pCStart = NULL, *pCEnd = NULL, *pFStart = NULL, *pFEnd = NULL;


typedef void Sigfunc(int);


/************************************************************************************************************
 * Description :
 *               Main Cluster Coordinator routine
 *               
 * Arguments   :
 *               ---
 *               
 * Return      :
 *               ErrorCode / 1
 *               
 ************************************************************************************************************/

int main(void)
{

  	write(1, "\nCluster(main): Initializing Cluster Coordinator.......\n", 60);
  	if(InitializeClusterCoordinator() < 0)
  	{
		write(1, "\nCluster(main): Failed to Initialise Cluster Coordinator....\n", 60);
		exit(1);
  	}

  	write(1, "\nCluster(main): Initialized \n", 35);
  	
	if(AcceptMessages() < 0)
	{
    		write(1, "\nCluster(main): Messages not accepted\n", 35);
		_exit(1);
	}

	return 1;
}


/************************************************************************************************************
 * Description :
 *               Intialization of the cluster coordinator, creates TCP sockets and listens on it
 *               
 * Arguments   :
 *               ---
 *               
 * Return      :
 *               Error Code / 1
 *               
 ************************************************************************************************************/

int InitializeClusterCoordinator()
{
	/*
	FILE *fp;
	
  	char *tok, *tok1, str[500], str1[500];
  	int i = 0, j;
	*/
	
  	/* Trap the termination signal so that when the
   	 * kill command is issued to the process it does
   	 * a gracefull exit from the code */
  	
	#ifdef SIGTERM
   	signal(SIGTERM, CleanUpClusterCoordinator);
  	#endif
	
  	/* Get the Details of the Host */
  	if(!InitializeHost(&g_stClusterCoordinator, TCP_AND_UNIXSTREAM_HOST))
  	{
		write(1, "\nCluster: Unable to initialise the host", 40);
		return -1;
  	}

  	/* Create a Cluster Socket and Listen on it */
  	if((ccsock[0] = CreateTCPServerHostAndListen(&g_stClusterCoordinator, CC_CC_PORT)) < 0)
  	{
      		write(1, "\nCluster: Unable to Create the Cluster Server ", 50);
      		return ccsock[0];
  	}

  	/* Create an Acceptor Socket and Listen on it */
  	if((ccsock[1] = CreateTCPServerHostAndListen(&g_stClusterCoordinator, CC_AC_PORT)) < 0)
  	{
      		write(1, "\nCluster: Unable to Create the Acceptor Server ", 50);
      		return ccsock[1];
  	}
  
	/* Create a Donor Socket and Just Listen on it */
  	if((ccsock[2] = CreateTCPServerHostAndListen(&g_stClusterCoordinator, CC_DR_PORT)) < 0)
  	{
      		write(1,"\nCluster: Unable to Create the Donor Server ", 50);
      		return ccsock[2];
  	}

  	if((usersock = CreateUNIXStreamServerHostAndListen(&g_stClusterCoordinator, CC_UP_PORT)) < 0)
  	{
      		write(1,"\nCluster: Unable to Create the user socket ", 50);
      		return usersock;
  	}
		 	
	
 	/* This piece of code makes a list of all the clusters present in the file friends.txt */

/* 	if((fp = fopen("friends.txt", "r")) == NULL)
 	{
 		perror("Unable to open the friends file");
 		return -1;
 	}
 
	while(fgets(str1, 500, fp) != NULL)
 		tok1 = strtok(str1, ":");
  	
 	i = atoi(tok1);
 
	for(j = 0; j < i; j++)
 	{
 		tok1 = strtok(NULL, ":");
 		if(strcmp(tok1, g_stClusterCoordinator.m_pszHostName) != 0)	
	        	if(InsertInFList(tok1) < 0)
	         		return -1;
 	}
 
	fclose(fp);
 	write(1, "\nCluster: Friends list is made", 35);
*/ 

	return 1;
}



/************************************************************************************************************
 * Description: this deletes all the memory allocated during the execution of cluster coordiantor
 * Arguments  : a_nSignalNumber - Signal Number because which this routine was called
 * Return     : None
 ************************************************************************************************************/

void CleanUpClusterCoordinator(a_nSignalNumber)
	int a_nSignalNumber;
{
  
	int l_nMessageID = CC_AD_GOODBYE;
  	int j;

  	for(j = 0; j < g_nNumberofDesc; j++)
		WriteXDRSignedInteger(g_nSocketDescriptor[j], CC_AC_CONTROL, (void *)&l_nMessageID, 
					sizeof(l_nMessageID), TCP_HOST);

  	write(1, "\nCluster: after closing...", 30);
  	close(ccsock[0]);
  	close(ccsock[1]);
  	close(ccsock[2]);

  	GC_FREE(g_stClusterCoordinator.m_pszHostName);
  	GC_FREE(g_stClusterCoordinator.m_pstTCPHostAddress);
  	GC_FREE(g_stClusterCoordinator.m_pszHostLogin);
  	GC_FREE(g_stClusterCoordinator.m_pszDaemonPath);

  	GC_FREE(pACTStart);
  	GC_FREE(pDRTStart);
  	GC_FREE(pFCCTStart);
  	
	exit(1);
}


/***********************************************************************************************************
 * Description : 
 * 		Signal routine
 * 		
 * Arguments   :
 * 		signo - Signal number
 * 		func  - Function to be called
 *
 * Return Value:
 * 		Pointer to function "Sigfunc".	
 *
 *********************************************************************************************************/ 		
Sigfunc *signal(int signo, Sigfunc *func)
{
	struct sigaction act, oact;
	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	if(sigaction(SIGTERM, &act, &oact) < 0)
		write(1,"\nNot able to process the signal in cleanup routine", 50);
	return(oact.sa_handler);
}


/************************************************************************************************************
 * Description :
 *               Accepts messages from other daemons
 *               
 * Arguments   :
 *               ---
 *               
 * Return      :
 *               Error Code / 1
 *               
 ************************************************************************************************************/

int AcceptMessages()
{
  	int nsock;
  	int j;
  	fd_set rfd;
  	HOST_DESCRIPTOR Host;

  	int l_nErrorCode;

  	g_nNumberofDesc = 0;

  	FD_ZERO(&rfd);

  	/* wait infinitely for messages from anywhere*/
  	
	while(1)
    	{
		/*
         	check if any other machines are trying to
         	communicate with the cluster coordinator
		*/
  	
  		/* ccsock[0] = CC to CC, ccsock[1] = AC to CC, ccsock[2] = DR to CC */
        	
		FD_SET(ccsock[0], &rfd);
        	FD_SET(ccsock[1], &rfd);
  		FD_SET(ccsock[2], &rfd);
        	
	/*	FD_SET(usersock, &rfd); */

		for(j = 0; j < g_nNumberofDesc; j++)
 	     		FD_SET(g_nSocketDescriptor[j], &rfd);
	
		if(AnyDataToRead(&rfd) == 1)
        	{
              	  	/*Has a message come from an old connection? */
	  		for(j = (g_nNumberofDesc-1); j >= 0; j--)
	  		{
				//write(1,"data from old connection",22);
	  			if(FD_ISSET(g_nSocketDescriptor[j], &rfd))
	     			{
	        			nsock = g_nSocketDescriptor[j];
	        			CurrentDescIndex = j;
	        			if((l_nErrorCode = ProcessClusterMessage(&Host, nsock)) < 0)
  	         			{
						write(1, "\nCluster: Error in processing Message ", 40);
						CloseHost(nsock);
 	         			}
	      			}
   	   		}
  	      	        		
   	   	   	/*A new connection from a Friend Cluster Coordinator*/
	   		if(FD_ISSET(ccsock[0], &rfd))
	   		{
	  	  		g_nSocketDescriptor[g_nNumberofDesc] = accept(ccsock[0], 
									(struct sockaddr *)NULL, 0);
	          		nsock = g_nSocketDescriptor[g_nNumberofDesc++];
		   		if((l_nErrorCode = ProcessClusterMessage(&Host, nsock)) < 0)
		     		{
					write(1,"\nCluster: Error in processing Message ", 40);
					g_nNumberofDesc--;
		     		}
		    	}
	
          		/*A new connection from an Acceptor*/
	  		if(FD_ISSET(ccsock[1], &rfd))
	    		{
	    			g_nSocketDescriptor[g_nNumberofDesc] = accept(ccsock[1], 
									(struct sockaddr *)NULL, 0);
                		nsock = g_nSocketDescriptor[g_nNumberofDesc++];
	        		if((l_nErrorCode = ProcessClusterMessage(&Host, nsock)) < 0)
	        		{
       		  			write(1,"\nCluster: Error in processing Message ", 40);
		  			g_nNumberofDesc--;
				}
			}
	
           		/*A new connection from a Donor*/
	  		if(FD_ISSET(ccsock[2], &rfd))
	    		{
	    			g_nSocketDescriptor[g_nNumberofDesc] = accept(ccsock[2], 
									(struct sockaddr *)NULL, 0);
                		nsock = g_nSocketDescriptor[g_nNumberofDesc++];
	        		
				if((l_nErrorCode = ProcessClusterMessage(&Host, nsock)) < 0)
	        		{
       		 			write(1, "\nCluster: Error in processing Message ", 40);
		  			g_nNumberofDesc--;
				}
	    		}
	
	 		/****************************************
			 
			if(FD_ISSET(usersock, &rfd))
		     	{
	     			nsock = accept(usersock, (struct sockaddr *)NULL, 0);
				
				if((l_nErrorCode = ProcessUserMessage(nsock)) < 0)
				{
					write(1, "\nCluster: Error in processing Message", 40);
					close(nsock);
				}
	      		}

			*****************************************/
			
		}
      	}

  	return 1;
}


/************************************************************************************************************
 * Description: this function checks if any data is available
 * 
 * Arguments  : rfd - pointer to the set of socket descriptors
 *		
 * Return     : socket descriptor
 * 
 ************************************************************************************************************/

int AnyDataToRead(rfd)
	fd_set *rfd;
{
  	
	int l_nIsReady;
  	int l_nNumberofDescriptors;

  	l_nNumberofDescriptors = getdtablesize() + 1;

  /*
    Using the select call which polls and returns immediatly
    if no data is avialable since Timeval is NULL
  */
	if((l_nIsReady = select(l_nNumberofDescriptors, rfd, (fd_set *)NULL, (fd_set *)NULL, NULL)) > 0)
  	  	return 1;
   	else
   		return -1;
}


/*************************************************************************************************************
 * Description: this function processes an arrived message and calls appropriate action 
 * 		to be taken depending on the sender and the message sent
 * 
 * Arguments  : l_pstHost - pointer to the Host Descriptor Structure to store 
 * 		the info of the sender
 *		sock - socket descriptor on which the message has come
 * 
 * Return     : It depends on the message type.
 *
***************************************************************************************************************/

int ProcessClusterMessage(l_pstHost, sock)
	PHOST_DESCRIPTOR l_pstHost;
	int sock;
{
  
	int l_nMessageID;
  	int l_nErrorCode;

  	/*get the message type */
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, AC_CC_CONTROL, (void *)&l_nMessageID, 
						sizeof(l_nMessageID), TCP_HOST)) < 0)
      		return l_nErrorCode;

   	switch(l_nMessageID)
   	{
        	/*Messages from the Acceptors*/
		case AC_CC_ADDREQ:
			write(1, "\nCluster: I will try to add an acceptor", 40);	
	      		return ClusterRead(l_pstHost, sock, l_nMessageID);
	
		case AC_CC_GETHPFFACTOR:
      			write(1, "\nCluster: I will find the HPF ", 35);
			return AskACHPFInfo(sock);  	
	
		case AC_CC_GOODBYE:
			write(1, "\nCluster: I am disconnecting you ", 35);
			return DeleteAcceptor(sock);
        
		case AC_CC_TASK:
        		write(1, "\nCluster: Forwarding the task...", 35);
        		return ForwardTask(sock);
		
        	/*Messages from the Donors*/
		case DR_CC_ADDREQ:
			write(1, "\nCluster: I will try to add a donor ", 40);	
			return ClusterRead(l_pstHost, sock, l_nMessageID);
	
		case DR_CC_GOODBYE:
			write(1, "\nCluster: I am disconnecting you ", 40);
			return DeleteDonor(sock);
	
		case DR_CC_SENDHPF:
			write(1, "\nCluster: I will take the HPF ", 40);
			return ReceiveHPFInfo(sock); 
			/* This function receives the HPF from the local donor 
			 * and send to the respective Acceptor who requested for it */
	
		case DR_CC_TASK:
	 		write(1, "\nCluster: Donor Sending Completed Task", 40);
	 		return ReturnResultsToCluster(sock);
	
		case DR_CC_STATUSCHANGE:
			return ChangeStatusofDonor(sock);
	        	 			
		/*Messages from other cluster coordinators*/        	 			
        	case CC_CC_SENDHPF:
			write(1, "\nCluster: Received the HPF info from Cluster ", 50);
			return SendHPF(sock); 
			/* This function receives the HPF from the other cluster 
			 * and send it to the respective Acceptor who requested for it */
	
		
/*-----------------------------------------------------------------------------------------
		case CC_CC_ADDACK:
			write(1, "\nCluster: Thanks for adding me as your friend ", 50);
	 		return ClusterRead(l_pstHost, sock, l_nMessageID);
	
		case CC_CC_ADDREQ:
			write(1, "\nCluster: Let me see if i can add this cluster...", 50);	
			return ClusterAdd(sock);
	
		case CC_CC_ADDREJ:
			write(1, "\nCluster: my req has not been accepted ", 40);
			return 1;
	
		case CC_AD_GOODBYE:
			write(1, "\nCluster: I am disconnecting you ", 40);
			if(DeleteClient(sock) < 0)
				return DeleteFriend(sock);
			else
				return 1;
--------------------------------------------------------------------------------------------*/
			
	
		case CC_CC_GETHPFFACTOR:
			write(1, "\nCluster: Request for the HPF from Cluster ", 50);
			return AskCCHPFInfo(sock);  	
        
		case CC_CC_DONORS:
        		write(1, "\nCluster: Received the no of donors from Cluster ", 50);
        		return SendDonorsCount(sock);
        
		case CC_CC_TASK:
        		return DistributeTask(sock);
        
		case CC_CC_TASKRESULT:
        		return SendResultToAcceptor(sock);
	
		default :
		  	return 1;   /* If the message is not appropriate just return */
    	}

  	return 1;
}


/*************************************************************************************************************
 * Description: this function accepts an acceptor, a donor or a friend cluster coordinator to 
 * 		insert into their respective tables.
 * 
 * Arguments  : l_pstACHost - pointer to the Host Descriptor Structure of the sender 
 * 		sock - socket descriptor on which the message has come
 *
 * Return     : Error Code / 1
 *
***************************************************************************************************************/

int ClusterRead(l_pstHost, sock, MessageID)
	PHOST_DESCRIPTOR l_pstHost;  /* to add to the list of hosts*/
     	int sock;
     	int MessageID;
{
  	
	int l_nErrorCode;
  	int l_nMessageID;

  	/* ADDREQ Message from Acceptor or Donor or ADDACK from Cluster*/

  	/*allocate memory for the host structure in the acceptor table*/
  	GC_CALLOC(l_pstHost->m_pszHostName, char, HOSTNAME_LENGTH);
  	GC_CALLOC(l_pstHost->m_pstTCPHostAddress, struct sockaddr_in, sizeof(struct sockaddr_in));
  	GC_CALLOC(l_pstHost->m_pszHostLogin, char, LOGINNAME_LENGTH);
  	GC_CALLOC(l_pstHost->m_pszDaemonPath, char, PATH_LENGTH);

  	/* Read Host name */
  	if((l_nErrorCode = ReadXDRString(sock, AC_CC_DATA, (void *)l_pstHost->m_pszHostName,
				    HOSTNAME_LENGTH, TCP_HOST)) < 0)
    	{
      		return l_nErrorCode;
    	}
  
	/* Read Address */
  	if((l_nErrorCode = ReadMessageAndData(sock, AC_CC_DATA, 
						(char *)l_pstHost->m_pstTCPHostAddress)) < 0)
    	{
      		return l_nErrorCode;
    	}

  	/* Read ProcessID */
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, AC_CC_DATA, (char *)&l_pstHost->m_nHostProcessID,
						sizeof(l_pstHost->m_nHostProcessID), TCP_HOST)) < 0)
    	{
      		return l_nErrorCode;
    	}
  
	/* Read Login Name */
  	if((l_nErrorCode = ReadXDRString(sock, AC_CC_DATA, (void *)l_pstHost->m_pszHostLogin, 
					 LOGINNAME_LENGTH, TCP_HOST)) < 0)
    	{
      		return l_nErrorCode;
    	}
  
	/* Read Daemon Path */
    	if((l_nErrorCode = ReadXDRString(sock, AC_CC_DATA, (void *)l_pstHost->m_pszDaemonPath, 
					 PATH_LENGTH, TCP_HOST)) < 0)
    	{
      		return l_nErrorCode;
    	}


/*---------------------------------------------------------------------------------------*/

	/* Insert the cluster in the friends cluster table */
	
/*	if(MessageID == CC_CC_ADDACK)
	{
		//write(1,"Cluster : Inserting Cluster",27);
		if(InsertClient(sock, l_pstHost) < 0)
		{
			write(1, "\nCluster: A Cluster can't be inserted ", 40);
			return -1;
		}
		
		write(1, "\nCluster: A Cluster is added  ", 35);
		
		return 1;
	} //end cluster req
*/
	
	/* Processing of ADDREQ from acceptors and donors of the same LAN  and sending the ADDACK*/
/*	if((MessageID == AC_CC_ADDREQ) || (MessageID == DR_CC_ADDREQ))
	{
		write(1," checking validity......",25);
		if(!IsValidHost("machines.txt",l_pstACHost->m_pszHostName))
		{
			write(1,"An ADD REQ came from a host which cant be added",47);
			l_nMessageID = CC_AD_ADDREJ;
			if(WriteXDRSignedInteger(sock,CC_AC_CONTROL,(void*) &l_nMessageID,sizeof(l_nMessageID),TCP_HOST) < 0)
				return -1;
				
			return 1 ;
		}
	}

	//fflush(stdin);
*/

/*-------------------------------------------------------------------------------------*/

	
	/*Insert the Acceptor in the Acceptor Table*/
	if(MessageID == AC_CC_ADDREQ)
	{
		if(!InsertAcceptor(sock, l_pstHost))
		{
			write(1, "\nCluster: An Acceptor REQ has been Rejected", 50);
			return -1;
		}
		   	           	   	
		/* Send an ADDACK Message to the acceptor */
		l_nMessageID = CC_AC_ADDACK;
		
		if((l_nErrorCode = WriteXDRSignedInteger(sock, CC_AC_CONTROL, 
				    (void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST)) < 0)
		{
			return l_nErrorCode;
		}
		
		write(1, "\nCluster: An Acceptor has been added ", 50);
		
/*		ClusterAddReq(NULL);  */
		
		return 1;
	}

	if(MessageID == DR_CC_ADDREQ)
	{
		/* Insert the Donor in the Donor Table*/
		if(!InsertDonor(sock, l_pstHost))
		{
			write(1, "\nCluster: A Donor REQ has been Rejected!", 50);
			return -1;
		}
 	
		/* Send an ADDACK Message to the donor */
		l_nMessageID = CC_DR_ADDACK;

		if((l_nErrorCode = WriteXDRSignedInteger(sock, CC_DR_CONTROL, 
				    (void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST)) < 0)
		{
			return l_nErrorCode;
		}
		
		write(1, "\nCluster: A donor has been added ", 35);
		
/*		ClusterAckReq(NULL);     */
		
		return 1;
	} //end if

	return 1;
}


/*************************************************************************************************************
 * Description: this function adds the acceptor info to the Acceptor Table
 * 
 * Arguments  : sd	-socket descriptor
 *		pHost	-Structure of the Host descriptor of the Acceptor
 *		
 * Return     :
 * 		Error Code / 1
 *
***************************************************************************************************************/

int InsertAcceptor(sd, pHost)
	int sd;	
	PHOST_DESCRIPTOR pHost;
{
	
	ACTABLE *t1, *t2;
	
	t2 = pACTStart;

	/*allocate memory for the structure of the Acceptor Table*/
	t1 = malloc(sizeof(ACTABLE));
	t1->ACHostName = (char *)malloc(HOSTNAME_LENGTH);
	t1->ACLoginName = (char *)malloc(LOGINNAME_LENGTH);
	t1->ACDaemonPath = (char *)malloc(PATH_LENGTH);
	
	/*Insert the values into the node*/
	strcpy(t1->ACHostName, pHost->m_pszHostName);
	strcpy(t1->ACLoginName, pHost->m_pszHostLogin);
        strcpy(t1->ACDaemonPath, pHost->m_pszDaemonPath);
	t1->ACPID = pHost->m_nHostProcessID;
	t1->ACSockDesc = sd;
	t1->next = NULL;
	
        /*if no acceptors in the table right now,*/
	if(pACTStart == NULL )
	{
		pACTStart = t1;
		pACTEnd = t1;
		g_nNumberofAcceptors++;
		write(1, "\nCluster: first acceptor inserted", 40);
		return 1;
	}
	
	/*if not the first acceptor, move to the end of the list of acceptors*/
	while(t2->next != NULL)
	{
		t2 = t2->next;
	}
	
	t2->next = t1;
	pACTEnd = t1;
	g_nNumberofAcceptors++;
	
	return 1;
}


/*************************************************************************************************************
 * Description: this function adds the donor info to the donor Table
 * 
 * Arguments  : sd	-socket descriptor
 *		pHost	-Structure of the Host descriptor of the Donor
 *		
 * Return     :
 * 		Error Code / 1
 *
***************************************************************************************************************/

int InsertDonor(sd, pHost)
	int sd;	
	PHOST_DESCRIPTOR pHost;
{
	
	DRTABLE *t1, *t2;
	
	t2 = pDRTStart;
 
	write(1,"\nCluster: inside InsertDonor", 30);
	
	t1 = malloc(sizeof(DRTABLE));
	t1->DRHostName = (char *)malloc(HOSTNAME_LENGTH);
	t1->DRLoginName = (char *)malloc(LOGINNAME_LENGTH);
	t1->DRDaemonPath = (char *)malloc(PATH_LENGTH);
	
	strcpy(t1->DRHostName, pHost->m_pszHostName);
	strcpy(t1->DRLoginName, pHost->m_pszHostLogin);
        strcpy(t1->DRDaemonPath, pHost->m_pszDaemonPath);
	t1->DRPID = pHost->m_nHostProcessID;
	t1->DRSockDesc = sd;
	t1->DRStatus = 0;
	t1->next = NULL;

	if(pDRTStart == NULL )
	{
		pDRTStart = t1;
		pDRTEnd = t1;
		g_nNumberofDonors++;
		return 1;
	}
	
	while(t2->next != NULL)
	{
		t2 = t2->next;
	}
	
	t2->next = t1;
	pDRTEnd = t1;
	g_nNumberofDonors++;
	return 1;
}


/*******************************************************************************
 * Description : This function asks for the HPF information from donors
 *		 when a request comes from the acceptor
*******************************************************************************/

int AskACHPFInfo(sock)
	int sock;
{
	
	int l_nMessageID, count = 0;
	DRTABLE *d;
	FCCTABLE *f;
	d = pDRTStart;
	f = pFCCTStart;
	
        while(d != NULL)
        {
    		if(d->DRStatus == 0)
    	        {
    	    	        /* Send the Request*/
 		 	l_nMessageID = CC_DR_GETHPF;
  			
			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_CONTROL, (void *)&l_nMessageID, 
						 sizeof(l_nMessageID), TCP_HOST) < 0)
      				return -1;
      		
	      		/* Send the Cluster Socket ID */		
    			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_CONTROL, (void *)&d->DRSockDesc, 
						 sizeof(d->DRSockDesc), TCP_HOST) < 0)
      				return -1;
      		
	      		/* Send the Acceptor Socket Desc*/
  			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_DATA, (void *)&sock, 
						 sizeof(sock), TCP_HOST) < 0)
      				return -1;
	    	
		    	NoOfDonors++;
		}
	        
		d = d->next;
	}//end while d loop

 	while(f != NULL)
 	{
 	 	/* Send the Request*/
 	 	l_nMessageID = CC_CC_GETHPFFACTOR;
  		
		if(WriteXDRSignedInteger(f->CCSockDesc, CC_DR_CONTROL, (void *)&l_nMessageID,
					 sizeof(l_nMessageID), TCP_HOST) < 0)
      			return -1;
      		
      		/* Send the Acceptor Socket ID */		
    		if(WriteXDRSignedInteger(f->CCSockDesc, CC_DR_CONTROL, (void *)&sock,
					 sizeof(sock), TCP_HOST) < 0)
      			return -1;
      		
		f = f->next;
      		count++;
      	}
 	
	l_nMessageID = CC_AC_CLUSTERS;	
	
 	/*send message type*/
 	if(WriteXDRSignedInteger(sock, CC_AC_DATA, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)
		return -1;
	
    	 /*send the number of donors*/
	 if(WriteXDRSignedInteger(sock, CC_AC_DATA, (void *)&NoOfDonors,
				  sizeof(NoOfDonors), TCP_HOST) < 0)
		return -1;
	
	 /*send the number of clusters*/
	 if(WriteXDRSignedInteger(sock, CC_AC_DATA, (void *)&count,
				  sizeof(count), TCP_HOST) < 0)
		return -1;
		
	 write(1, "\nCluster: no of donors and the no of clusters sent back to acceptor", 70);	
	 
	 NoOfDonors = 0;
     	 
	 return 1;
}


/*******************************************************************************************
 * Description:	This function deletes an acceptor from the acceptor table
 * 
 * Arguments  :	sd - socket descriptor
 * 
 * Return     :	status of operation
 * 
********************************************************************************************/

int DeleteAcceptor(sock)
	int sock;
{
	
	ACTABLE *t1, *t2;

	t1 = pACTStart;
	t2 = pACTStart;
	
	/*if no acceptors in the Acceptor Table*/
	if(pACTStart == NULL)
	{
		/*Error*/
		write(1, "\nCluster: No Acceptors in the table to delete", 50);
		return -1;
	}
	
	/*if it is the first acceptor, then delete it*/
	if((pACTStart->ACSockDesc) == sock)
	{
		pACTStart = pACTStart->next;
		write(1, "\nCluster: The Acceptor has been deleted", 40);
		g_nNumberofAcceptors--;
		CloseHost(sock);
		return 1;
	}		
	
	/*else scan the list*/
	else
	{
		/*If only a single node*/
	/*	if(pACTStart->next == NULL)
		{
			write(1,"The Only Acceptor cant be deleted",33);
			return 1;
	
		}
	
	*/
		
		t2 = t2->next;
		/*scan the table for the right acceptor*/
		while(t2 != NULL)
		{	
			/*if found delete and return*/
			if(t2->ACSockDesc == sock)
			{
				t1->next = t2->next;
				write(1, "\nCluster: The Acceptor has been deleted", 45 );
				g_nNumberofAcceptors--;
				CloseHost(sock);
				return 1;
			}
			
			/*else move to next acceptor*/
			t1 = t2;
			t2 = t2->next;
		}
	}
	
	/*if the acceptor is not in the list, then return error*/
	return -1;
}


/*******************************************************************************
 * Description : this function updates the array of the open socket descriptors
*******************************************************************************/

int CloseHost(sock)
	int sock;
{
  
	int i;
  	int index;

  	/*Search the array for the descriptor*/
  	for(i = (g_nNumberofDesc-1); i >= 0; i--)
  	{
  		if(g_nSocketDescriptor[i] == sock)
  		{
  			index=i;
  			break;
  		}
  	}

  	/* Update the array of Descriptors opened */		
  	for(i = index; i < (g_nNumberofDesc-1); i++)
      		g_nSocketDescriptor[i] = g_nSocketDescriptor[i+1];

  	g_nNumberofDesc--;
  
	return 1;
}


/*************************************************************************************************************
 Description : The Cluster Forwards the task info sent by the acceptor 
 		to the respective donors through the clusters
*************************************************************************************************************/
 
int ForwardTask(sock)
	int sock;
{
 	
	int csock, dsock;
        int l_nMessageID;
        char *code;
	
//        long l_LB,l_UB;
        
	int len;
        char lb[25], ub[25];
        CTASK *t;

 	/*Receive Info from the Acceptor*/
 	
 	/*Read the Cluster ID*/
	if(ReadXDRSignedInteger(sock, AC_CC_DATA, (void *)&csock, sizeof(csock), TCP_HOST) < 0)
		return -1;
	 	
 	/*Read the Donor ID*/
 	if(ReadXDRSignedInteger(sock, AC_CC_DATA, (void *)&dsock, sizeof(dsock), TCP_HOST) < 0)
	      	return -1;
	
	/*Read the size of the exe file*/
	if(ReadXDRSignedInteger(sock, AC_CC_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
	      	return -1;
	
	/*Read the code of the file*/
	code = (char *)malloc(len);
	if(ReadXDRString(sock, AC_CC_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
	
	/*Read the lower bound of the problem*/
	if(ReadXDRString(sock, AC_CC_DATA, (void *)lb, 25, TCP_HOST) < 0)
	      	return -1;
		
	/*Read the upper bound of the problem*/
	if(ReadXDRString(sock, AC_CC_DATA, (void *)ub, 25, TCP_HOST) < 0)
	      	return -1;
		
	/*UPDATE THE TASK TABLE WITH THE INFORMATION OF THE TASK TERMINALS*/
	t = (CTASK *)malloc(sizeof(CTASK));
	t->TACSock = sock;
	t->TFCSock = 0;
	t->TTCSock = csock;
	t->TDRSock = dsock;
	t->next = NULL;
	
	if(pTStart == NULL)
	{
		pTStart = t;
	 	pTEnd = t;
	}
	 
	else
	{
		pTEnd->next = t;
	 	pTEnd = t;
	}
	
 	/*FORWARD THE CTASK TO THE NEXT CLUSTER*/
	
 	/*send the message*/
 	l_nMessageID = CC_CC_TASK;
 	 
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&l_nMessageID, 
				 sizeof(l_nMessageID), TCP_HOST) < 0)
	 	return -1;
	
 	/*send the Donor ID to the cluster*/
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&dsock, 
				 sizeof(dsock), TCP_HOST) < 0)
	      	return -1;
	
	/*send the acceptor ID*/ 	
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&sock, 
				 sizeof(sock), TCP_HOST) < 0)
	      	return -1;

	/*send the size of the exe file*/	
      	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
      		return -1;	
      				
	/*send the code to the donor*/	
	if(WriteXDRString(csock, CC_CC_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
    		
	/*send the lower bound of the task*/	
	if(WriteXDRString(csock, CC_CC_DATA, (void *)lb, strlen(lb), TCP_HOST) < 0)
    		return -1;
    	
    	/*send the upper bound of the task*/	
	if(WriteXDRString(csock, CC_CC_DATA, (void *)ub, strlen(ub), TCP_HOST) < 0)
    		return -1;
	
 	free(code);		
	
	return 1;
}


/*******************************************************************************************
 * Description:	This function deletes a donor from the donor table
 * Arguments  : sd - socket descriptor
 * Return     : status of operation
********************************************************************************************/

int DeleteDonor(sock)
	int sock;
{
	
	DRTABLE *t1, *t2;

	t1 = pDRTStart;
	t2 = pDRTStart;
	
	/*if no donors in the Donor Table*/
	if(pDRTStart == NULL)
	{
		/*Error*/
		write(1, "\nCluster: There are no Donors which can be deleted ", 60);
		return -1;
	}
	
	/*if it is the first donor, then delete it*/
	if((pDRTStart->DRSockDesc) == sock)
	{
		write(1, "\nCluster: Deleting the 1st node in the Donor Table", 55);
		pDRTStart = pDRTStart->next;
		write(1, "\nCluster: The Donor has been deleted", 40);
		g_nNumberofDonors--;
		CloseHost(sock);
		return 1;
	}		
	
	/*else scan the list*/
	else
	{
		/*If only a single node*/
	/*	if(pDRTStart->next == NULL)
		{
			write(1,"There's only one Donor which cant be deleted",44);
			return 1;
		}
	*/	
		
		t2 = t2->next;

		/*scan the table for the right donor*/
		while(t2 != NULL)
		{
			/*if found delete and return*/
			if(t2->DRSockDesc == sock)
			{
				t1->next = t2->next;
				write(1, "\nCluster: The Donor has been deleted", 40);
				g_nNumberofDonors--;
				CloseHost(sock);
				return 1;
			}
			
			/*else move to next donor*/
			t1 = t2;
			t2 = t2->next;
		}
	}
		
	/*if the donor is not in the list, then return error*/
	return -1;
}


/***********************************************************************************************
Description: This function receives the HPF info from donor and sends it to the
	     acceptor which asked for it
	     
Arguments: sock - The socket descriptor of the connection with the donor

Return:	1 on Success / -1 on Error
**********************************************************************************************/

int ReceiveHPFInfo(sock)
	int sock;
{
	
	int accsock, csock;
	unsigned long HPF;
	int l_nMessageID;
	DRTABLE *t1;
	t1 = pDRTStart;
	
	/*Receive Cluster Socket Desc(Cluster Coordinator ID) on which it has to send the info*/
  	if(ReadXDRSignedInteger(sock, DR_CC_DATA, (void *)&csock,
				sizeof(csock), TCP_HOST) < 0)
		return -1;
	
	/*Receive Acceptor Socket Desc(Acceptor ID) to which it has to send */
  	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&accsock,
				sizeof(accsock), TCP_HOST) < 0)
	      	return -1;
	
	/*Receive the HPF from the donor*/
  	if(ReadXDRUnSignedLongInteger(sock, CC_DR_DATA, (void *)&HPF,
					sizeof(HPF), TCP_HOST) < 0)
	      	return -1;

    	if(sock == csock)    /* It means the info is to be sent on the local LAN Acceptor */
    	{
    		while(t1 != NULL)
    		{
    			if(t1->DRSockDesc == sock)
 			{
  		        	/*send the message to the local acceptor*/
				l_nMessageID = CC_AC_SENDHPF;
				if(WriteXDRSignedInteger(accsock, CC_AC_CONTROL, 
					(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST) < 0)
		    			return -1;

    				/*send hostname*/
    				if(WriteXDRString(accsock, CC_AC_DATA, (void *)t1->DRHostName,
							HOSTNAME_LENGTH, TCP_HOST) < 0)
			      		return -1;

    				/*send login name*/
    				if(WriteXDRString(accsock, CC_AC_DATA, (void *)t1->DRLoginName, 
							LOGINNAME_LENGTH, TCP_HOST) < 0)
		 	     		return -1;
          	
	          		/* Send Daemon path*/
		  		if(WriteXDRString(accsock, CC_AC_DATA,(void *)t1->DRDaemonPath, 
							PATH_LENGTH, TCP_HOST) < 0)
			      		return -1;
			
				/* Send HPF Info */
				if(WriteXDRUnSignedLongInteger(accsock, CC_AC_CONTROL, (void *)&HPF,
							sizeof(HPF), TCP_HOST) < 0)
		    			return -1;

				write(1, "\nCluster: HPF Sent to Acceptor", 35);
			
				return 1;
			}
       		
			t1 = t1->next;	
    		} //end while loop
	}//end if(sock=csock) loop
 	
       	else /* It means the info is to be sent on the cluster  */
    	{
    	 	while(t1 != NULL)
    	 	{
    			if(t1->DRSockDesc == sock)
 			{
  		        	/*send the message to the cluster */
				l_nMessageID = CC_CC_SENDHPF;
				if(WriteXDRSignedInteger(csock, CC_CC_CONTROL, (void *)&l_nMessageID,
							 sizeof(l_nMessageID), TCP_HOST) < 0)
		    			return -1;

                        	/*send the donor ID to the cluster*/
				if(WriteXDRSignedInteger(csock, CC_CC_CONTROL, (void *)&sock,
							 sizeof(sock), TCP_HOST) < 0)
		    			return -1;
    			
    				/* Send the Acceptor ID to the cluster */
    				if(WriteXDRSignedInteger(csock, CC_CC_CONTROL, (void *)&accsock,
							 sizeof(accsock), TCP_HOST) < 0)
		    			return -1;

    				/*send hostname*/
    				if(WriteXDRString(csock, CC_CC_DATA, (void *)t1->DRHostName,
							HOSTNAME_LENGTH, TCP_HOST) < 0)
			      		return -1;

    				/*send login name*/
    				if(WriteXDRString(csock, CC_CC_DATA, (void *)t1->DRLoginName, 
							LOGINNAME_LENGTH, TCP_HOST) < 0)
		 	     		return -1;
          	
	          		/* Send Daemon path*/
		  		if(WriteXDRString(csock, CC_CC_DATA, (void *)t1->DRDaemonPath, 
							PATH_LENGTH, TCP_HOST) < 0)
			      		return -1;
			
				/* Send HPF Info */
				if(WriteXDRUnSignedLongInteger(csock, CC_CC_CONTROL, (void *)&HPF,
							  sizeof(HPF), TCP_HOST) < 0)
		    			return -1;

				write(1, "\nCluster: HPF Sent to Cluster", 35);
			
				return 1;
		
       			}
       		
			t1 = t1->next;	
    	 	} //end while loop
      	}//end else loop
 		
 	return 1;
}


/**************************************************************************************************
 * Description: The Cluster receives the task results sent by donor 
 * 		and sends it to the respective cluster.
 *************************************************************************************************/
 
int ReturnResultsToCluster(sock)
	int sock;
{
	
	int acsock, csock;
	int l_nMessageID;
	int len;
	char *res;
	DRTABLE *d;
	CTASK *t1, *t2;
	int i;
	
	d = pDRTStart;
	 
	/*RECEIVE  THE RESULTS*/
	/*Receive the cluster ID*/
        if(ReadXDRSignedInteger(sock, DR_CC_DATA, (void *)&csock, sizeof(csock), TCP_HOST) < 0)
	      	return -1;
	
	/*Receive the Acceptor ID*/
	if(ReadXDRSignedInteger(sock, DR_CC_DATA, (void *)&acsock, sizeof(acsock), TCP_HOST) < 0)
	      	return -1;

	/* Receive the size of the result */
        if(ReadXDRSignedInteger(sock, DR_CC_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
	      	return -1;
	
	res = (char *)malloc(len);

	/*Receive the results*/
	if(ReadXDRString(sock, DR_CC_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;

	/*UPDATE THE CTASK TABLE TO REMOVE THE ENTRY OF THE TASK*/
	t1 = pTStart;
	t2 = pTStart;
	
	if(pTStart == NULL)
	{
		write(1, "\nCluster: Error in the TASK TABLE  ", 40);
		return -1;
	}
	
	/*if it is the node, then delete it*/
	if(((pTStart->TFCSock) == csock) && ((pTStart->TDRSock) == sock) 
			&& ((pTStart->TACSock) == acsock))
	{
		pTStart = pTStart->next;
		free(t2);
		if(pTStart == NULL)
		{
			write(1, "\nCluster: All the tasks completed.....", 40);
			
			for(i=0; i < ClusterIndex; i++)
		 	{
		 		l_nMessageID = CC_AD_GOODBYE;
		 		if(WriteXDRSignedInteger(ClustersToBeDeleted[i], CC_CC_CONTROL,
				   	(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST) < 0)
					return -1;
		 		
				CloseHost(ClustersToBeDeleted[i]);
		 	}
		 	
			ClusterIndex = 0;
		}
	}
	
	else	/*else scan the list*/
	{
		/*If only a single node*/
		if(pTStart->next == NULL)
		{
			write(1, "\nCluster: Error in the TASK TABLE  ", 35);
			return -1;
		}
		
		t2 = t2->next;
				
		/*scan the table for the right task entry*/
		while(t2 != NULL)
		{	
			/*if found delete and break*/
			if((t2->TFCSock) == csock && (t2->TDRSock == sock) 
					&& (pTStart->TACSock == acsock))
			{
				t1->next = t2->next;
				free(t2);
				break;
			}
			
			/*else move to next node*/
			t1 = t2;
			t2 = t2->next;
		}
		
	}	
		
	/*SEND THE RESULTS TO THE CLUSTERS*/
	/*send the message*/
	l_nMessageID = CC_CC_TASKRESULT;
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&l_nMessageID, 
				 sizeof(l_nMessageID), TCP_HOST) < 0)
	      	return -1;
	
	/*send the Acceptor ID*/
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&acsock, 
				 sizeof(acsock), TCP_HOST) < 0)
	      	return -1;
	
	/*send the Donor ID*/
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&sock, 
				 sizeof(sock), TCP_HOST) < 0)
	      	return -1;
	
	/* send the size of the result*/
	if(WriteXDRSignedInteger(csock, CC_CC_DATA, (void *)&len,
				 sizeof(len), TCP_HOST) < 0)
	      	return -1;
	
	/*Send the result to the cluster*/
	if(WriteXDRString(csock, CC_CC_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;

	free(res);
	
	/*UPDATE THE STATUS OF THE DONOR IN THE DONOR TABLE*/
	while(d != NULL)
	{
		if(d->DRSockDesc == sock)
		{
			d->DRStatus = 0;
			break;
		}
		
		d = d->next;
	}
	    	
	return 1;
}


/****************************************************************************************
 * Description: The Cluster changes the status of a Donor after it receives 
 * the confirmation from it that it has completed the task of a process from 
 * the local Acceptor
 *****************************************************************************************/

int ChangeStatusofDonor(sock)
	int sock;
{
        
	DRTABLE *t;

        t = pDRTStart;

        while(t != NULL)
        {
        	if(t->DRSockDesc == sock)
        	{
        		if(t->DRStatus == 0)
        			t->DRStatus = 1;
        		else
        			t->DRStatus = 0;
        		break;
        	}
        	
		t = t->next;
        }
        
	return 1;
}


/********************************************************************************************
 * Description: This function sends the HPF to the acceptor after 
 * 		receiving it from the other cluster
 *
 * Arguments: sock - The socket descriptor of the connection
 *
 * Return: 1 on Success / -1 on Error
**********************************************************************************************/

int SendHPF(sock)
	int sock;
{
	
	int accsock, dsock;
	unsigned long HPF;
	int l_nMessageID;
        char *hostname, *logname, *dpath;

        hostname = (char *)malloc(HOSTNAME_LENGTH);
        logname = (char *)malloc(LOGINNAME_LENGTH);
        dpath = (char *)malloc(PATH_LENGTH);
	
	/* Receive the Donor and HPF Data from the other Cluster */
	/*Receive the Donor ID*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&dsock,
				sizeof(dsock), TCP_HOST) < 0)
	     	return -1;
	
	/*Receive Acceptor Socket Desc(Acceptor ID) on which to send the HPF info*/
  	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&accsock, 
				sizeof(accsock), TCP_HOST) < 0)
	      	return -1;

	/*Read hostname*/
	if(ReadXDRString(sock, CC_CC_DATA, (void *)hostname, HOSTNAME_LENGTH, TCP_HOST) < 0)
	      	return -1;

	/*Read login name*/
	if(ReadXDRString(sock, CC_CC_DATA, (void *)logname, LOGINNAME_LENGTH, TCP_HOST) < 0)
 	     	return -1;
          	
       	/* Read Daemon path*/
  	if(ReadXDRString(sock, CC_CC_DATA, (void *)dpath, PATH_LENGTH, TCP_HOST) < 0)
	      	return -1;

	/*Read HPF from the Donor*/
  	if(ReadXDRUnSignedLongInteger(sock, CC_CC_DATA, (void *)&HPF, 
					sizeof(HPF), TCP_HOST) < 0)
	      	return -1;

	
	/* Send this data received to the acceptor who requested for it */	
	
        /*send the message to the local acceptor*/
	l_nMessageID = CC_ACG_SENDHPF;
	if(WriteXDRSignedInteger(accsock, CC_AC_CONTROL, (void *)&l_nMessageID, 
				 sizeof(l_nMessageID), TCP_HOST) < 0)
  		return -1;

 	/*send the cluster ID to the acceptor*/	
 	if(WriteXDRSignedInteger(accsock, CC_AC_CONTROL,(void *)&sock, 
				 sizeof(sock), TCP_HOST) < 0)
  		return -1;

  	/*send the donor ID to the acceptor*/	
	if(WriteXDRSignedInteger(accsock, CC_AC_CONTROL, (void *)&dsock,
				 sizeof(dsock), TCP_HOST) < 0)
  		return -1;

	/*send hostname*/
	if(WriteXDRString(accsock, CC_AC_DATA, (void *)hostname, 
				HOSTNAME_LENGTH, TCP_HOST) < 0)
		return -1;

	/*send login name*/
	if(WriteXDRString(accsock, CC_AC_DATA, (void *)logname, 
				LOGINNAME_LENGTH, TCP_HOST) < 0)
 		return -1;
          	
       	/* Send Daemon path*/
  	if(WriteXDRString(accsock, CC_AC_DATA, (void *)dpath, 
				PATH_LENGTH, TCP_HOST) < 0)
		return -1;

	/* Send HPF Info */
	if(WriteXDRUnSignedLongInteger(accsock, CC_AC_CONTROL, (void *)&HPF,
					sizeof(HPF), TCP_HOST) < 0)
		return -1;
			
	write(1, "\nCluster: HPF Sent to Acceptor", 40);
	free(hostname);
	free(logname);
	free(dpath);
		
	return 1;
}


/*******************************************************************************
 * Description : This function asks for the HPF information from donors
 *		 when a request comes from the other cluster
*******************************************************************************/

int AskCCHPFInfo(sock)
	int sock;
{
	
	int l_nMessageID, accsock;
	DRTABLE *d;
	d = pDRTStart;
	
        /*get the acceptor id for the cluster who send the request */
	if(ReadXDRSignedInteger(sock, CC_CC_CONTROL, (void *)&accsock,
				sizeof(accsock), TCP_HOST) < 0)
	     	return -1;

        while(d != NULL)
	{
    		if(d->DRStatus != 1)
    	        {
    	        	/* Send the Request*/
	 	 	l_nMessageID = CC_DR_GETHPF;
  			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_CONTROL, (void *)&l_nMessageID,
						 sizeof(l_nMessageID), TCP_HOST) < 0)
      				return -1;
      		
	      		/* Send the Cluster Socket ID */		
    			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_CONTROL, (void *)&sock,
						 sizeof(sock), TCP_HOST) < 0)
      				return -1;
      		
	      		/* Send the Acceptor Socket Desc*/
  			if(WriteXDRSignedInteger(d->DRSockDesc, CC_DR_DATA, (void *)&accsock,
						 sizeof(accsock), TCP_HOST) < 0)
      				return -1;
      					    	
		    	NoOfDonors++;
		}
		
		d = d->next;
	}//end while d loop

	l_nMessageID = CC_CC_DONORS;	
	
 	/*send message type*/
 	if(WriteXDRSignedInteger(sock, CC_CC_CONTROL, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)
		return -1;
	
	/* Send the Acceptor Socket Desc*/
  	if(WriteXDRSignedInteger(sock, CC_CC_DATA, (void *)&accsock,
				  sizeof(accsock), TCP_HOST) < 0)
		return -1;
      		
    	/*send the number of donors*/
	if(WriteXDRSignedInteger(sock, CC_CC_DATA, (void *)&NoOfDonors,
				 sizeof(NoOfDonors), TCP_HOST) < 0)
		return -1;
	
	write(1, "\nCluster: no. of donors sent back to cluster", 45);	
	
	NoOfDonors = 0;
     	 
	return 1;
}


/*******************************************************************************************
 * Description:	This function sends the donors count to the acceptor after 
 * 		receiving it from the other cluster
 *
 * Arguments: sock - The socket descriptor of the connection
 *
 * Return: 1 on Success / -1 on Error
********************************************************************************************/

int SendDonorsCount(sock)
	int sock;
{
	
	int accsock, donors;
	int l_nMessageID;
	
	/* Receive the Acceptor Socket Desc on which to send the donors count*/
  	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&accsock,
				sizeof(accsock), TCP_HOST) < 0)
      		return -1;
      		
    	/*Receive the number of donors*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&donors,
				sizeof(donors), TCP_HOST) < 0)
		return -1;
	
	l_nMessageID = CC_AC_DONORS;	  		
	/* Send the Message type*/
  	if(WriteXDRSignedInteger(accsock, CC_AC_DATA, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)
      		return -1;
      		
    	/*send the number of donors*/
	if(WriteXDRSignedInteger(accsock, CC_AC_DATA, (void *)&donors,
				 sizeof(donors), TCP_HOST) < 0)
		return -1;
	
	return 1;
}


/****************************************************************************************
 * Description : The Cluster receives the task info sent by other cluster and 
 * 		 sends it to the respective donor in its cluster.
 ***************************************************************************************/

int DistributeTask(sock)
	int sock;
{
 	
	int dsock, acsock;
 	int l_nMessageID;
 	char *code;
	
//        long l_LB,l_UB;
	
        int len;
        char lb[25], ub[25];
        DRTABLE *d;
        CTASK *t;

        /*RECEIVE  THE TASK INFORMATION FROM THE CLUSTER*/
 	
 	/*Receive the Donor ID*/
 	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&dsock,
				sizeof(dsock), TCP_HOST) < 0)
	      	return -1;
	
	/*Receive the Acceptor ID*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&acsock,
				sizeof(acsock), TCP_HOST) < 0)
	      	return -1;
	
	/*Read the size of the exe file*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&len,
				sizeof(len), TCP_HOST) < 0)
	      	return -1;
	
	/*Read the code of the file*/
	code = (char *)malloc(len);
	if(ReadXDRString(sock, CC_CC_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
	
	/*Read the lower bound of the problem*/
	if(ReadXDRString(sock, CC_CC_DATA, (void *)lb, 25, TCP_HOST) < 0)
	      	return -1;
		
	/*Read the upper bound of the problem*/
	if(ReadXDRString(sock, CC_CC_DATA, (void *)ub, 25, TCP_HOST) < 0)
	      	return -1;
	
	/*UPDATE THE CTASK TABLE WITH THE INFORMATION OF THE TASK TERMINALS*/
	
	t = (CTASK *)malloc(sizeof(CTASK));
	t->TACSock = acsock;
	t->TFCSock = sock;
	t->TTCSock = 0;
	t->TDRSock = dsock;
	t->next = NULL;
	
	if(pTStart == NULL)
	{
		pTStart = t;
	 	pTEnd = t;
	}
	 
	else
	{
		pTEnd->next = t;
		pTEnd = t;
	}
	
 	/*SEND THE TASK TO THE DONOR*/
	
	/*send the message to the donor about the task*/
	l_nMessageID = CC_DR_TASK;
	if(WriteXDRSignedInteger(dsock, CC_DR_DATA, (void *)&l_nMessageID, 
				 sizeof(l_nMessageID), TCP_HOST) < 0)
	      	return -1;
	
	/*send the Cluster ID*/
        if(WriteXDRSignedInteger(dsock, CC_DR_DATA, (void *)&sock, 
				 sizeof(sock), TCP_HOST) < 0)
	      	return -1;
	
	/*send the Acceptor ID*/
        if(WriteXDRSignedInteger(dsock, CC_DR_DATA, (void *)&acsock, 
				 sizeof(acsock), TCP_HOST) < 0)
	      	return -1;
	
	/*send the size of the exe file*/	
      	if(WriteXDRSignedInteger(dsock, CC_DR_DATA, (void *)&len,
				 sizeof(len), TCP_HOST) < 0)
      		return -1;	
      				
	/*send the code to the donor*/	
	if(WriteXDRString(dsock, CC_DR_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
    		
	/*send the lower bound of the task*/	
	if(WriteXDRString(dsock, CC_DR_DATA, (void *)lb, strlen(lb), TCP_HOST) <0)
    		return -1;
    	
    	/*send the upper bound of the task*/	
	if(WriteXDRString(dsock, CC_DR_DATA, (void *)ub, strlen(ub), TCP_HOST) <0)
    		return -1;
			
	d = pDRTStart;
	
	while(d != NULL)
	{
		if((d->DRSockDesc) == dsock) 
		{
			d->DRStatus = 1;
			break;
		}	
		
		d = d->next;
	}
	
	free(code);
	
	return 1;
}


/**************************************************************************************
 * Description: The Cluster receives the task results sent by cluster and 
 * 	 	sends it to the respective Acceptor
 **************************************************************************************/

int SendResultToAcceptor(sock)
	int sock;
{
        
	int acsock, dsock;
        int l_nMessageID;
	int len;
        char *res;
        CTASK *t1, *t2;
        int i;

        /*RECEIVE THE RESULTS FROM THE DONOR THRU THE CLUSTER*/

	/*Receive the Acceptor ID*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&acsock,
				sizeof(acsock), TCP_HOST) < 0)
		return -1;
	
	/*Receive the Donor ID*/
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&dsock, 
				sizeof(dsock), TCP_HOST) < 0)
		return -1;
	
	/* receive the size of the result */
	if(ReadXDRSignedInteger(sock, CC_CC_DATA, (void *)&len,
				sizeof(len), TCP_HOST) < 0)
		return -1;

	res = (char *)malloc(len);
	/*Receive the results*/
	if(ReadXDRString(sock, CC_CC_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;

	/*UPDATE THE CTASK TABLE TO REMOVE THE ENTRY OF THE TASK*/
	
	t1 = pTStart;
	t2 = pTStart;
	
	if(pTStart == NULL)
	{
		write(1, "\nCluster: Error in the TASK TABLE  ", 40);
		return -1;
	}
	
	/*if it is the node, then delete it*/
	if((pTStart->TTCSock) == sock && pTStart->TDRSock == dsock)
	{
		pTStart = pTStart->next;
		free(t1);
		if(pTStart == NULL)
		{	
			write(1, "\nCluster: All the tasks completed.....", 40);
		 	
			for(i = 0; i < ClusterIndex; i++)
		 	{
		 		l_nMessageID = CC_AD_GOODBYE;
		 		
				if(WriteXDRSignedInteger(ClustersToBeDeleted[i], CC_CC_CONTROL,
					(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST) < 0)
					return -1;
		 		CloseHost(ClustersToBeDeleted[i]);
		 	}
		 	
			ClusterIndex = 0;
		}
	}	
	
	else	/*else scan the list*/
	{
		/*If only a single node*/
		if(pTStart->next == NULL)
		{
			write(1, "\nCluster: Error in the TASK TABLE  ", 40);
			return -1;
		}
		
		t2 = t2->next;
				
		/*scan the table for the right task entry*/
		while(t2 != NULL)
		{	
			/*if found delete and break*/
			if((t2->TTCSock) == sock && t2->TDRSock == dsock)
			{
				t1->next = t2->next;
				free(t2);
				break;
			}
			
			/*else move to next node*/
			t1 = t2;
			t2 = t2->next;
		}
	}	
	
	/*SEND THE RESULTS BACK TO THE ACCEPTOR*/
	l_nMessageID = CC_AC_TASKRESULT;
	
	if(WriteXDRSignedInteger(acsock, CC_AC_DATA, (void *)&l_nMessageID, 
				 sizeof(l_nMessageID), TCP_HOST) < 0)
	     	return -1;
	
	/* send the size of the result*/	
	if(WriteXDRSignedInteger(acsock, CC_AC_DATA, (void *)&len,
				 sizeof(len), TCP_HOST) < 0)
	      	return -1;

	/*Send the result to the acceptor*/
	if(WriteXDRString(acsock, CC_AC_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;

	//free(res);
	
	return 1;		
}



/*-----------------------------------------------------------------------------------------*/

/******************************************************************************************
 * Description: this function processes an arrived message from user process and 
 * 		calls appropriate action to be taken depending on the message
 *
 * Arguments  : sock - socket descriptor on which the message has come
 * 
 * Return     : 1 on success and -1 on error
*******************************************************************************************/

/***********************************************************
 
int ProcessUserMessage(sock)
   	int sock;
{
	
	int l_nMessageID;
	int l_nErrorCode;
	char *hostname;

        hostname = (char *)malloc(HOSTNAME_LENGTH);
	
	*get the message type *
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, UP_CC_CONTROL, (void *)&l_nMessageID,
					sizeof(l_nMessageID), UNIXSTREAM_HOST )) < 0)
		return -1;

        if((hostname = GetHostName(sock)) == NULL)
        	write(1, "\nCluster: can't read the hostname ", 40);
	
	if(strcmp(hostname, g_stClusterCoordinator.m_pszHostName) == 0)
	{
		write(1, "\nCluster: Its Own Hostname Can't do this operation", 50);
		return -1;
	}
		
	switch(l_nMessageID)
	{
	  	case UP_CC_ADDCLIENT:
	        	write(1, "\nCluster: A user process is asking to insert a client", 70);
			if(SearchCCTable(hostname) > 0)
			{
				if(InsertInClientList(hostname) > 0)
					return ClusterAddReq(hostname);
				else
					return -1;
			}
			
			else
				write(1, "\nCluster: Already client to this hostname ", 45);
			
			return -1;
	  
		case UP_CC_REMOVECLIENT:
	        	write(1, "\nCluster: A user process is asking to remove a client", 60);
			return RemoveClient(hostname);
			
	  	case UP_CC_ADDFRIEND:
	        	write(1, "\nCluster: A user process is asking to add a friend ", 60);
			if(SearchFCTable(hostname) > 0)
			{
				if(InsertInFList(hostname) > 0)
					return ClusterAckReq(hostname);
				else
					return -1;
			}
			
			else
			 	write(1,"Cluster : Already friend to this client ",40);
			
			return -1;
	  
		case UP_CC_REMOVEFRIEND:
	        	write(1, "\nCluster: A user process is asking to remove a friend", 62);
			return RemoveFriend(hostname);
	}
	
	return 1;
}

***************************************************************************/

/*************************************************************************************************************
 * Description: this function checks to see if the acceptor can be added to the acceptor list
 * Arguments  : hostname	-name of the host of the sendor(acceptor)
 * Return     :

***************************************************************************************************************/

/***********************************************************************

int IsValidHost(filename,hostname)
   char *filename;
   char *hostname;
{
  FILE *fp;
  char *tok;
  char str[200];
  int i=0,j;

	/* open the file that lists the hostnames ot the possible acceptors on the LAN*	
	if( (fp = fopen(filename,"r") )==NULL )
	{
		write(1," Cluster : Unable to open file",30);
		return -1;
	}

	while(fgets(str,200,fp)!=NULL)
		tok = strtok(str,":");
  	
	i=atoi(tok);

	for(j=0;j<i;j++)
	{
		tok =strtok(NULL,":");	
		/*if hostname present in the file then close file ptr and return*
		if(strcmp(tok,hostname)== 0)
		{
			fclose(fp);
			return 1;
		}
	}
	fclose(fp);
	return 1;
}

********************************************************************************/

/*************************************************************************************************************
 * Description: This function adds the Cluster info to the Friends Clusters Table
 * Arguments  : sd	-socket descriptor
 		pHost	-Structure of the Host descriptor of the Acceptor
 * Return     :

***************************************************************************************************************/

/**************************************************************
 
int InsertFriend(sd,hostname)
	int 	  sd;	
	char *hostname;
{
	FCCTABLE *t1,*t2;
	
	t2=pFCCTStart;

	t1=malloc( sizeof(FCCTABLE) );
	t1->FCHostName=(char *)malloc(HOSTNAME_LENGTH);
	
	strcpy(t1->FCHostName,hostname);
	t1->FCSockDesc=sd;
	t1->next=NULL;

	if(pFCCTStart == NULL )
	{
		pFCCTStart=t1;
		pFCCTEnd=t1;
		g_nNumberofFriends++;
		return 1;
	}
	
	while(t2!=NULL)
	{
		if( strcmp(t2->FCHostName,hostname)==0)
		{
			write(1," Cluster : The friend has already been added ",45);
			free(t1);
			return -1;
		}
		t2=t2->next;
	}
	pFCCTEnd->next=t1;
	pFCCTEnd=t1;
	g_nNumberofFriends++;
	return 1;

} //end Insert Friend

************************************************************************/


/*******************************************************************************************
 * Description:	This function deletes a Cluster from the Friends Cluster table
 * Arguments  : 	sd	-socket descriptor
 * Return     : 	status of operation
********************************************************************************************/

/*******************************************************************************
 * 
int DeleteFriend(sock)
	int sock;
{
	FCCTABLE *t1,*t2;

	t1=pFCCTStart;
	t2=pFCCTStart;
	
	/*if no clusters in the Cluster Table*
	if(pFCCTStart == NULL)
	{
		/*Error*
		write(1,"No Clusters to be deleted",26);
		return 1;
	}
	

	/*if it is the first Cluster, then delete it*
	if(  (pFCCTStart->FCSockDesc)==sock)
	{
		InsertInFList(pFCTStart->FCHostName);
		pFCCTStart=pFCCTStart->next;
		write(1,"The Cluster has been deleted",29);
		g_nNumberofFriends--;
		CloseHost(sock);
		return 1;
	}		
	/*else scan the list*
	else
	{
		/*If only a single node*
		if(pFCCTStart->next == NULL)
		{
			write(1,"The Only Cluster cant be deleted",33);
			return 1;
		}
		t2=t2->next;
		
		/*scan the table for the right Cluster*
		while(t2!=NULL)
		{	
			/*if found delete and return*
			if(  t2->FCSockDesc==sock)
			{
				InsertInFList(t2->FCHostName);
				t1->next=t2->next;
				write(1,"The Cluster has been deleted",29);
				g_nNumberofFriends--;
				CloseHost(sock);
				return 1;
			}
			/*else move to next Cluster*
			t1=t2;
			t2=t2->next;
		}
		
	}
	
	/*if the Cluster is not in the list, then return error*
	return 1;

	
}

**************************************************************/


/*******************************************************************************
 * Description : this function searches the acceptor table

*******************************************************************************/

/***********************************************************************
 * 
int SearchAcceptor()
{
                	
	ACTABLE *t1;
	
	t1=pACTStart;

	while( t1 != NULL)
	{
		write( 1,t1->ACHostName,sizeof(t1->ACHostName) );
		t1=(t1->next);
	}
	return 1;
}

************************************************************************/


/*******************************************************************************
 * Description : This function sends CC_CC_ADDREQ to all the clusters in the list

*******************************************************************************/

/*****************************************************************
 * 
int ClusterAddReq(hname)
	char *hname;
{
   	int l_nMessageID,sock;
  	
	CLUSTERS *t;
	int flag=0;
	
	t=pCStart;
	write(1,"Sending req",11);
	
	if(hname==NULL)
		flag=1;
		
	while( t!=NULL )
	    {
	    	/* Get the hostname of the cluster one by one*
	    	if(flag==1)
		{	    	
			hname=(char *)malloc(HOSTNAME_LENGTH);
			strcpy(hname,t->Cluster);
	  	}
	    	/* Create a Socket and Connect to the Cluster *
	     		/* g_stAcceptor is host structure of global...*
	  	if ((sock =CreateTCPClientHostAndConnect(&g_stClusterCoordinator,hname,CC_CC_PORT)) > 0)
      		{
      			write(1,"request being sent to cluster",29);
      			/* Send a *Add Request* Message to Cluster Coordinator *
	 		l_nMessageID = CC_CC_ADDREQ;
		 	if (WriteXDRSignedInteger(sock,CC_CC_CONTROL,(void*) &l_nMessageID,sizeof(l_nMessageID),TCP_HOST) > 0)
    			{
    			/* Send Host Name to the other cluster to check our entry in his file*
		          if(WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszHostName ,HOSTNAME_LENGTH,TCP_HOST) > 0)
			      g_nSocketDescriptor[g_nNumberofDesc++]=sock;
			}

	        }
	        t=t->next;
	        if(flag==0)
	        	return 1;
	        else
	        	free(hname);
	
	   }
	//fclose(fp);
	return 1;
}

*****************************************************************************/


/*******************************************************************************
 * Description : This function checks whether a cluster can be sent ADDACK or not
 		 and take necessary action
   Arguments  :  sock -> the socket descriptor to which it has to send the ADDACK

   Return     :	1 on success and -1 on error
*******************************************************************************/

/**************************************************************************
 
int ClusterAdd(sock)
	int sock;
{
	char *hostname;
	int l_nMessageID;
	
	hostname=(char *)malloc(HOSTNAME_LENGTH);

	/*Read the hostname of the cluster*
	if (ReadXDRString(sock,CC_CC_DATA,  (void* ) hostname,HOSTNAME_LENGTH, TCP_HOST) < 0)
	      return -1;
	
        /*can i become his friend? *
	if( !IsValidHost("friends.txt",hostname) )
        {
        	/*If no, send a reject signal*
        	l_nMessageID = CC_CC_ADDREJ;
	 	if (WriteXDRSignedInteger(sock,CC_CC_CONTROL,(void*) &l_nMessageID,sizeof(l_nMessageID),TCP_HOST) < 0)
			return -1;
		return 1;
	}
	
	l_nMessageID = CC_CC_ADDACK;
 	if (WriteXDRSignedInteger(sock,CC_CC_CONTROL,(void*) &l_nMessageID,sizeof(l_nMessageID),TCP_HOST) < 0)
		return -1;
	
	/* Send the Cluster Details to Cluster Coordinator *
  	
  	/* Send Host Name *
  	if(WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszHostName,HOSTNAME_LENGTH,TCP_HOST) <0)
	      return -1;
	
	//write(1,"hostname sent",13);

  	/* Send Address *
  	if (WriteMessageAndData(sock,CC_CC_DATA,(char*) g_stClusterCoordinator.m_pstTCPHostAddress, sizeof(struct sockaddr_in)) < 0)
	      return -1;
	//write(1,"address sent",12);
	
	/* Send Process Id *
  	if (WriteXDRSignedInteger(sock,CC_CC_DATA,(void*) &g_stClusterCoordinator.m_nHostProcessID, sizeof(g_stClusterCoordinator.m_nHostProcessID), TCP_HOST) < 0)
	      return -1;
	//write(1,"procid sent",11);
	
	/* Send Login Name*
  	if (WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszHostLogin, LOGINNAME_LENGTH, TCP_HOST) < 0)
	      return -1;
	//write(1,"logname sent",12);
  	
  	/* Send Daemon path*
  	if (WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszDaemonPath, PATH_LENGTH, TCP_HOST) < 0)
	      return -1;
	
	write(1,"info about cluster sent on request",34);
	return 1;

}

**************************************************************************/


/*******************************************************************************
 * Description : This function sends CC_CC_ADDACK to all the clusters in the friends
 		 list when a donor of the LAN registers with it

*******************************************************************************/

/***********************************************************************
 
int ClusterAckReq(hname)
	char *hname;
{
   	int l_nMessageID,sock;
	CLUSTERS *t;
	int flag=0;
	
	t=pFStart;
	if(hname==NULL)
		flag=1;
	
	while( t!=NULL )
	    {
		if(flag==1)
		{
			hname=(char *)malloc(HOSTNAME_LENGTH);	
			strcpy(hname,t->Cluster);
		}
	    	
	    	/* Create a Socket and Connect to the Cluster *
	     		/* g_stAcceptor is host structure of global...*
	  	if ((sock = CreateTCPClientHostAndConnect(&g_stClusterCoordinator,hname,CC_CC_PORT)) > 0)
      		{
      			write(1,"ack request sent to cluster",27);
      			/* Send a *AddAck Request* Message to Cluster Coordinator *
	 		
			l_nMessageID = CC_CC_ADDACK;
		 	if (WriteXDRSignedInteger(sock,CC_CC_CONTROL,(void*) &l_nMessageID,sizeof(l_nMessageID),TCP_HOST) < 0)
				return -1;
	
			/* Send the Cluster Details to Cluster Coordinator *
  	
		  	/* Send Host Name *
		  	if(WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszHostName,HOSTNAME_LENGTH,TCP_HOST) <0)
			      return -1;
	

		  	/* Send Address *
		  	if (WriteMessageAndData(sock,CC_CC_DATA,(char*) g_stClusterCoordinator.m_pstTCPHostAddress, sizeof(struct sockaddr_in)) < 0)
			      return -1;
	
			/* Send Process Id *
		  	if (WriteXDRSignedInteger(sock,CC_CC_DATA,(void*) &g_stClusterCoordinator.m_nHostProcessID, sizeof(g_stClusterCoordinator.m_nHostProcessID), TCP_HOST) < 0)
			      return -1;
	
			/* Send Login Name*
		  	if (WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszHostLogin, LOGINNAME_LENGTH, TCP_HOST) < 0)
			      return -1;
  	
		  	/* Send Daemon path*
		  	if (WriteXDRString(sock,CC_CC_DATA,(void*) g_stClusterCoordinator.m_pszDaemonPath, PATH_LENGTH, TCP_HOST) < 0)
			      return -1;
	
			g_nSocketDescriptor[g_nNumberofDesc++]=sock;
			DeleteFromFList(hname);
			InsertFriend(sock,hname);
			write(1,"in ack info about cluster sent",29);
			
		}
		t=t->next;
		if(flag==0)
			return 1;
		else
			free(hname);
		
	    }
	
return 1;
}

*****************************************************************************************/

/************************************************************************************************************
  Description :	This function deletes an entry which has a specified hostname from the clusters list
  Arguments   :	Start Pointer of the List
  		Hostname of the host
  Return Value:	1 on Success/ -1 on Error

************************************************************************************************************/

/***************************************************************
int DeleteFromFList(hname)
	char *hname;
{
	CLUSTERS *t1,*t2;
	
	t1=pFStart;
	t2=pFStart;

	if(pFStart == NULL)
		return 1;

	/*if it is the first acceptor, then delete it*
	if(  strcmp(pFStart->Cluster,hname)==0 )
	{
		pFStart = pFStart->next;
		write(1,"Cluster : 1 Cluster deleted",27);
		return 1;
	}		
	/*else scan the list*
	else
	{
		/*If only a single node*
		if(pFStart->next == NULL)
		{
			return 1;
		}
		t2=t2->next;
		
		/*scan the table for the right acceptor*
		while(t2!=NULL)
		{	
			/*if found delete and return*
			if(  strcmp(t2->Cluster,hname)==0 )
			{
				t1->next=t2->next;
				write(1,"Cluster : other Cluster deleted",31);
				return 1;
			}
			/*else move to next acceptor*
			t1=t2;
			t2=t2->next;
		}
		
	}	
	
 return 1;	
}

********************************************************************************/

/*************************************************************************************************************
*************************************************************************************************************/

/************************************************************************
char* GetHostName(sock)
	int sock;
{
        char *hostname;

        hostname=(char *)malloc(HOSTNAME_LENGTH);

	if ( ReadXDRString(sock,UP_CC_DATA,(void* )hostname,HOSTNAME_LENGTH,UNIXSTREAM_HOST) <0)
		return NULL;
		
	return hostname;
}

*********************************************************************/

/*************************************************************************************************************
*************************************************************************************************************/

/*********************************************************************
int SearchCCTable(hostname)
	char *hostname;
{
                	
	FCCTABLE *t1;
	
	t1=pFCCTStart;

	while( t1 != NULL)
	{
		if(strcmp(t1->CCHostName,hostname)==0)
			return -1;
		else
			t1=(t1->next);
	}
	return 1;
}

***********************************************************************/

/*************************************************************************************************************
*************************************************************************************************************/

/********************************************************
int SearchFCTable(hostname)
	char *hostname;
{
                	
	FCCTABLE *t1;
	
	t1=pFCCTStart;

	while( t1 != NULL)
	{
		if(strcmp(t1->FCHostName,hostname)==0)
			return -1;
		else
			t1=(t1->next);
	}
	return 1;
}
**************************************************************/


/*************************************************************************************************************
 Description : this function removes a cluster from the list of possible/existing clusters to which this
 		cluster is a client
*************************************************************************************************************/

/**************************************************************************
int RemoveClient(hostname)
	char *hostname;
{
        int csock=-1;
        FCCTABLE *c1,*c2;
	CTASK *t;

	/*SEARCH FOR THE GIVEN CLUSTER IN THE CLIENT TABLE AND GET THE CLUSTER INFO*
	c1=pFCCTStart;
	c2=pFCCTStart;
	
	if(pFCCTStart != NULL)
	{	
		/*if it is the node, then delete it*
		if( strcmp(pFCCTStart->CCHostName,hostname)==0 )
		{
			csock=pFCCTStart->CCSockDesc;
			write(1," Removed from the Client Table ",31);
			pCCTStart = pFCCTStart->next;
			g_nNumberofClients--;
			//free(c1);
		}	
		else	/*else scan the list*
		{
			/*If not only a single node*
			if(pFCCTStart->next != NULL)
			{
				c2=c2->next;
				
				/*scan the table for the right task entry*
				while(c2!=NULL)
				{	
					/*if found delete and break*
					if(  strcmp(c2->CCHostName,hostname)==0 )
					{
						write(1," Removed from the Client Table ",31);
						csock=c2->CCSockDesc;
						c1->next=c2->next;
						g_nNumberofClients--;
						//free(c2);
						break;
					}
					/*else move to next node*
					c1=c2;
					c2=c2->next;
				}
		  	}
		}	
	}
	
	
	/*If the cluster is present in the Cluster Table,then search the Task Table*
	
	if(csock != -1 )
	{
		write(1," Searching in Task Table  ",26);
		printf("\t%d",csock);	
		t=pTStart;
		if( t!=NULL)
		{
			while(t!=NULL)
			{
				printf("* \t%d\t%d *",t->TFCSock,t->TTCSock);
				if ( t->TFCSock== csock || t->TTCSock== csock )
				{
					write(1,"Cluster : The Cluster can be deleted only after the completion of the task ",75);
					/*Add the cluster in the queue of clusters to be deleted*
					ClustersToBeDeleted[ClusterIndex++]=csock;		
				}
				t=t->next;
			}
		}
		else
		{
			write(1,"The Client is deleted ",22);
		}
		
	
	}
	
	/*Delete the cluster from the List of Clients*
	return DeleteFromClientList(hostname);
}

*********************************************************************************/


/*************************************************************************************************************
 Description : this function removes a cluster from the list of possible/existing clusters to which this
 		cluster is a client
*************************************************************************************************************/

/****************************************************************************

int RemoveFriend(hostname)
	char *hostname;
{
        int csock=-1;
        FCTABLE *f1,*f2;
	CTASK *t;

	/*SEARCH FOR THE GIVEN CLUSTER IN THE FRIEND TABLE AND GET THE CLUSTER INFO*
	f1=pFCTStart;
	f2=pFCTStart;
	
	if(pFCTStart != NULL)
	{	
		/*if it is the node, then delete it*
		if( strcmp(pFCTStart->FCHostName,hostname)==0 )
		{
			csock=pFCTStart->FCSockDesc;
			write(1," Removed from the Friend Table ",31);
			pFCTStart = pFCTStart->next;
			free(f1);
		}	
		else	/*else scan the list*
		{
			/*If not only a single node*
			if(pFCTStart->next != NULL)
			{
				f2=f2->next;
				
				/*scan the table for the right task entry*
				while(f2!=NULL)
				{	
					/*if found delete and break*
					if(  strcmp(f2->FCHostName,hostname)==0 )
					{
						write(1," Removed from the Friend Table ",31);
						csock=f2->FCSockDesc;
						f1->next=f2->next;
						free(f2);
						break;
					}
					/*else move to next node*
					f1=f2;
					f2=f2->next;
				}
		  	}
		}	
	}
	
	/*If the cluster is present in the Cluster Table,then search the Task Table*
	write(1," Searching in Task Table  ",26);
	if(csock != -1)
	{
		t=pTStart;
		if( t!=NULL)
		{
			while(t!=NULL)
			{
				if ( t->TFCSock== csock || t->TTCSock== csock )
				{
					write(1,"Cluster : The Cluster can be deleted only after the completion of the task ",75);
					/*Add the cluster in the queue of clusters to be deleted*
					ClustersToBeDeleted[ClusterIndex++]=csock;		
					
				}
				t=t->next;
			}
		}
	
	}
	
	/*Delete the cluster from the List of Friends*
	return DeleteFromFList(hostname);
}

*****************************************************************************/

/************************************************************************************************************
  Description: This function inserts a machine name in a CLUSTERS List
  Arguments  :	hostname - Friend machine name to be inserted in the list.
  return value:	1 on Success / -1 on Error
************************************************************************************************************/

/***************************************************************************
int InsertInFList(hostname)
	char *hostname;
{
	CLUSTERS *t1, *t2;
	
	t2 = pFStart;

	t1 = (CLUSTERS *)malloc(sizeof(CLUSTERS));
	t1->Cluster = (char *)malloc(HOSTNAME_LENGTH);

	strcpy(t1->Cluster, hostname);
	t1->next = NULL;
		
	if(pFStart == NULL)		
	{
		pFStart = t1;
		pFEnd = t1;
		return 1;
	}
	
	while(t2 != NULL)
	{
		if(strcmp(t2->Cluster, hostname) == 0)
		{
			write(1, "\nCluster: The host is already there in friends list", 52);
			free(t1);
			return -1;
		}
		
		t2 = t2->next;
	}       	
	
	pFEnd->next = t1;
	pFEnd = t1;	

	return 1;
}

***************************************************************************/

/*************************************************************************************************************
 * Description: This function adds the Cluster info to the Client Clusters Table
 * Arguments  : sd	-socket descriptor
 		pHost	-Structure of the Host descriptor of the Acceptor
 * Return     :

***************************************************************************************************************/

/******************************************************************
int InsertClient(sd,pHost)
	int 	  sd;	
	PHOST_DESCRIPTOR pHost;
{
	CCTABLE *t1,*t2;
	
	t2=pCCTStart;

	t1=malloc( sizeof(CCTABLE) );
	t1->CCHostName=(char *)malloc(HOSTNAME_LENGTH);
	t1->CCLoginName=(char *)malloc(LOGINNAME_LENGTH);
	t1->CCDaemonPath=(char *)malloc(PATH_LENGTH);
	
	strcpy(t1->CCHostName,pHost->m_pszHostName);
	strcpy(t1->CCLoginName,pHost->m_pszHostLogin);
        strcpy(t1->CCDaemonPath,pHost->m_pszDaemonPath);
	t1->CCPID=pHost->m_nHostProcessID;
	t1->CCSockDesc=sd;
	t1->next=NULL;

	if(pCCTStart == NULL )
	{
		write(1,"First Client has been inserted ",31);
		pCCTStart=t1;
		pCCTEnd=t1;
		g_nNumberofClients++;
		DeleteFromClientList(t1->CCHostName);
		return 1;
	}
	
	while(t2!=NULL)
	{
		if( strcmp(t2->CCHostName,pHost->m_pszHostName)==0)
		{
			write(1," Cluster : The host has already been added ",43);
			free(t1);
			return -1;
		}
		t2=t2->next;
	}
	pCCTEnd->next=t1;
	pCCTEnd=t1;
	g_nNumberofClients++;
	DeleteFromClientList(t1->CCHostName);
	return 1;

} //end Insert Cluster

************************************************************************/


/*******************************************************************************************
 * Description:	This function deletes a Cluster from the Clients Cluster table
 * Arguments  : 	sd	-socket descriptor
 * Return     : 	status of operation
********************************************************************************************/

/*************************************************************************
int DeleteClient(sock)
	int sock;
{
	CCTABLE *t1,*t2;

	t1=pCCTStart;
	t2=pCCTStart;
	
	/*if no clusters in the Cluster Table*
	if(pCCTStart == NULL)
	{
		/*Error*
		write(1,"No Clusters to be deleted",26);
		return 1;
	}
	

	/*if it is the first Cluster, then delete it*
	if(  (pCCTStart->CCSockDesc)==sock)
	{
		InsertInClientList(pCCTStart->CCHostName);
		pCCTStart=pCCTStart->next;
		write(1,"The Cluster has been deleted",29);
		g_nNumberofClients--;
		CloseHost(sock);
		return 1;
	}		
	/*else scan the list*
	else
	{
		/*If only a single node*
		if(pCCTStart->next == NULL)
		{
			write(1,"The Only Cluster cant be deleted",33);
			return 1;
		}
		t2=t2->next;
		
		/*scan the table for the right Cluster*
		while(t2!=NULL)
		{	
			/*if found delete and return*
			if(  t2->CCSockDesc==sock)
			{
				InsertInClientList(t2->CCHostName);
				t1->next=t2->next;
				write(1,"The Cluster has been deleted",29);
				g_nNumberofClients--;
				CloseHost(sock);
				return 1;
			}
			/*else move to next Cluster*
			t1=t2;
			t2=t2->next;
		}
		
	}
	
	/*if the Cluster is not in the list, then return error*
	return 1;

	
}

***************************************************************************/

/************************************************************************************************************
  Description :	This function creates a linked list of all the clusters listed in the file specified
  Arguments:	hostname
  		Start Pointer
  return      :	1 on Success/-1 on Error
************************************************************************************************************/

/*****************************************************************
int InsertInClientList(hostname)
	char *hostname;
{
	CLUSTERS *t1,*t2;
	
	t2=pCStart;
  	/* open the file that lists the hostnames ot the possible clusters *
	t1=malloc(sizeof(CLUSTERS));
	if(t1==NULL)
	{
		write(1," Cluster:: while Initialising Clusters List, Unable to allocate memory ",70);
		return -1;
	}
	t1->Cluster=(char *)malloc(HOSTNAME_LENGTH);

	strcpy(t1->Cluster,hostname);
	t1->next=NULL;
		
	if(pCStart==NULL)		
	{
		pCStart=t1;
		pCEnd=t1;
		return 1;
	}
	while(t2!=NULL)
	{
		if( strcmp(t2->Cluster,hostname)==0)
		{
			write(1," Cluster : The host is already there in cluster list ",52);
			free(t1);
			return -1;
		}
		t2=t2->next;
	}       	
	pCEnd->next=t1;
	pCEnd=t1;
	
return 1;
}

********************************************************************************/

/************************************************************************************************************
  Description :	This function deletes an entry which has a specified hostname from the clusters list
  Arguments   :	Hostname of the host
  Return Value:	1 on Success/ -1 on Error

************************************************************************************************************/

/*******************************************************************************
int DeleteFromClientList(hname)
	char *hname;
{
	CLUSTERS *t1,*t2;
	
	t1=pCStart;
	t2=pCStart;

	if(pCStart == NULL)
		return 1;

	/*if it is the first acceptor, then delete it*
	if(  strcmp(pCStart->Cluster,hname)==0 )
	{
		pCStart = pCStart->next;
		return 1;
	}		
	/*else scan the list*
	else
	{
		/*If only a single node*
		if(pCStart->next == NULL)
		{
			return 1;
		}
		t2=t2->next;
		
		//write(1,"let me delete",14);
		/*scan the table for the right acceptor*
		while(t2!=NULL)
		{	
			/*if found delete and return*
			if(  strcmp(t2->Cluster,hname)==0 )
			{
				t1->next=t2->next;
				write(1,"Cluster : other Cluster deleted",31);
				return 1;
			}
			/*else move to next acceptor*
			t1=t2;
			t2=t2->next;
		}
		
	}	
	
 return 1;	
}

**************************************************************************************/

/*--------------------------------------------------------------------------------------------*/
