/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * GCNetLevel.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the socket related function calls which are used bu
 * others programs to manipulates on sockets.
 *
 * Developers 
 * ~~~~~~~~~~
 * Karthikeyan R & Binu K J
 *
 * Date 
 * ~~~~
 * 24/September/98
 *
 ***********************************************************************************************/


/*
 *must find the header files for each paltform and then 
 * depending on that current files have to be included
 */


#include <unistd.h>
#include <netdb.h>


#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <strings.h>
#include <netdb.h>
#include <signal.h>
#include <sys/un.h>
#include <unistd.h>
#include <arpa/inet.h>

#include "GCNetLevel.h"
#include "GCError.h"
#include "GCHost.h"
#include "GCDefines.h"
#include "GCMesg.h"
#include "GCXDRRoutines.h"

char* g_pszProgramName;


/************************************************************************************************************
 * Description : 
 *               Reads on a Socket Descriptor value
 * Arguments   : sock - Socket on Which to Listen
 *               a_pszReadBuffer - Buffer to read into
 *               a_nBufferSize - Size of the buffer to be read
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int ReadOnHost(sock, a_pszReadBuffer, a_nBufferSize)
     int sock;
     char* a_pszReadBuffer; 
     int a_nBufferSize; 
{
  int l_nBytesRead;
  int l_nRemainingSize;
  char* l_pszRemainingBuffer;

  GC_ZERO(a_pszReadBuffer, a_nBufferSize);


  l_pszRemainingBuffer = a_pszReadBuffer;
  l_nRemainingSize = a_nBufferSize;

  do
    {
      if ((l_nBytesRead = read(sock,l_pszRemainingBuffer, l_nRemainingSize)) < 0)
	{
	  GCLogClusterError("read");
	  return SOCKET_READ_ERROR;
	}
      if (0 == l_nBytesRead)
	{
	  return l_nBytesRead;
	}
      l_nRemainingSize = l_nRemainingSize - l_nBytesRead;
      l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesRead;

    } while (l_nRemainingSize > 0);
  
  return a_nBufferSize;
}

/************************************************************************************************************
 * Description : 
 *               Writes on a Socket Descriptor value
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_pszWriteBuffer - Buffer to write into
 *               a_nBufferSize - Size of the buffer to be written
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int WriteOnHost(sock, a_pszWriteBuffer, a_nBufferSize)
     int sock; 	
     char* a_pszWriteBuffer; 
     int a_nBufferSize; 
{
  int l_nBytesWritten;
  int l_nRemainingSize;
  char* l_pszRemainingBuffer;

  l_pszRemainingBuffer = a_pszWriteBuffer;
  l_nRemainingSize = a_nBufferSize;

  do
    {
      if ((l_nBytesWritten = write(sock, l_pszRemainingBuffer, l_nRemainingSize)) <= 0)
	{
	  GCLogClusterError("write");
	  return SOCKET_WRITE_ERROR;
	}
      l_nRemainingSize = l_nRemainingSize - l_nBytesWritten;
      l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesWritten;

    } while (l_nRemainingSize > 0);
 	 
  return a_nBufferSize;
}

/************************************************************************************************************
 * Description : 
 *               Creates a TCP Server Socket and listens on it
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nPortNumber - Port Number to listen on
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int CreateTCPServerHostAndListen(a_pstHostDetails, a_nPortNumber)
     PHOST_DESCRIPTOR a_pstHostDetails;
     int a_nPortNumber;
{
  int l_nCreateSocket;

  if (NULL == a_pstHostDetails->m_pstTCPHostAddress)
      return -1;

  if ((l_nCreateSocket = socket(AF_INET, SOCK_STREAM, 0)) <= 0)
    {
      GCLogClusterError("create");
      return SOCKET_CREATION_ERROR;
    }
 
  a_pstHostDetails->m_pstTCPHostAddress->sin_port = htons(a_nPortNumber);
  a_pstHostDetails->m_pstTCPHostAddress->sin_addr.s_addr = htonl(INADDR_ANY);

  if ((bind(l_nCreateSocket, (struct sockaddr*) a_pstHostDetails->m_pstTCPHostAddress,sizeof(struct sockaddr))) < 0)
    {
      GCLogClusterError("bind");
      return SOCKET_BIND_ERROR;
    }

 listen(l_nCreateSocket,5);
 return l_nCreateSocket;
}
/************************************************************************************************************
 * Description :
 *               Creates a UNIX domain stream Server Socket and listens on it
 * Arguments   :
 *               a_pstHostDetails - Pointer to the gost details
 *               a_pszUNIXStreamPath - path of UNIX domain stream socket file
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int CreateUNIXStreamServerHostAndListen(a_pstHostDetails, a_pszUNIXStreamPath)
     PHOST_DESCRIPTOR a_pstHostDetails;
     char* a_pszUNIXStreamPath;
{
  int l_nCreateSocket;

  unlink(a_pszUNIXStreamPath);
  if (NULL == a_pstHostDetails->m_pstUNIXStreamHostAddress)
      return -1;

  strcpy(a_pstHostDetails->m_pstUNIXStreamHostAddress->sun_path, a_pszUNIXStreamPath);


  /* Create the Socket */
  if ((l_nCreateSocket = socket(AF_UNIX, SOCK_STREAM, 0)) <= 0)
    {
      GCLogClusterError("socket");
      return SOCKET_CREATION_ERROR;
    }


  /*Bind the socket*/
  if (bind(l_nCreateSocket,(struct sockaddr*) a_pstHostDetails->m_pstUNIXStreamHostAddress,sizeof(struct sockaddr)) < 0)
    {
      GCLogClusterError("bind");
      return SOCKET_BIND_ERROR;
    }


  /* Listen on the Socket */
  listen(l_nCreateSocket,5);

  return l_nCreateSocket;
}

/************************************************************************************************************
 * Description : 
 *               Creates a TCP Clients Socket and connects to server machine
 * Arguments   :
 *               a_pstClientsDetails - Pointer to the host details
 *               a_pszServerName - Name of the server to connect to
 *               a_nServerPortNumber - Port Number for the server to connect
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int CreateTCPClientHostAndConnect(a_pstClientDetails, a_pszServerName, a_nServerPortNumber)
     PHOST_DESCRIPTOR a_pstClientDetails;
     char* a_pszServerName; 
     int a_nServerPortNumber;
{

  int nCreatedSocket;
  struct sockaddr_in l_stServerAddress;
//  struct hostent* l_pstServerHostEntry;

  if (NULL == a_pstClientDetails->m_pstTCPHostAddress)
      return -1;

  /* Create the Socket */
  if ((nCreatedSocket = socket(AF_INET,SOCK_STREAM,0)) <= 0)
    {
      return nCreatedSocket;
    }

  GC_ZERO(&l_stServerAddress, sizeof(struct sockaddr_in));

  l_stServerAddress.sin_family = AF_INET;
  l_stServerAddress.sin_port = htons(a_nServerPortNumber);
  if(inet_pton(AF_INET,a_pszServerName,&l_stServerAddress.sin_addr) <= 0)
     {
       write(1,"error",5);
       return 0;
     }

  /* Connect to the server node */
  if (connect(nCreatedSocket, (struct sockaddr *)&l_stServerAddress, sizeof(struct sockaddr)) < 0)
    {
      GCLogClusterError("connect");
      return SOCKET_CONNECT_ERROR;
    }

  return nCreatedSocket;
}

/************************************************************************************************************
 * Description :
 *               Creates a UNIX Clients Socket and connects to server machine
 * Arguments   :
 *               a_pstClientDetails - Pointer to the gost details
 *               a_pszServerPath - path of UNIX domain stream server socket file
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int CreateUNIXStreamClientHostAndConnect( a_pszServerPath)
     char* a_pszServerPath;
{

  int l_nCreateSocket;
  struct sockaddr_un l_stServerAddress;

   /* Create the Socket */
    if ((l_nCreateSocket = socket(AF_UNIX, SOCK_STREAM, 0)) <= 0)
    {
      GCLogClusterError("socket");
      return SOCKET_CREATION_ERROR;
    }


  GC_ZERO(&l_stServerAddress, sizeof(struct sockaddr_in));

  l_stServerAddress.sun_family = AF_UNIX;
  strcpy(l_stServerAddress.sun_path, a_pszServerPath);

  /* Connect to the server node */
  if (connect(l_nCreateSocket,(struct sockaddr *)&l_stServerAddress,(sizeof(l_stServerAddress.sun_family) + strlen(l_stServerAddress.sun_path))) < 0)
    {
      GCLogClusterError("connect");
      return SOCKET_CONNECT_ERROR;
    }

  return l_nCreateSocket;
}



/************************************************************************************************************
 * Description : 
 *               Write the message and the data to the socket
 * Arguments   :
 *               a_nMessageID - Message ID to be sent
 *               a_pszMessage - Message to tbe sent
 *               a_nMessageSize - size of the message to be sent
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int WriteMessageAndData(sock,a_nMessageID, a_pszMessage, a_nMessageSize)
     int sock;	
     int a_nMessageID;
     char* a_pszMessage; 
     int a_nMessageSize; 
{

  int l_nErrorCode;
  MESSAGE l_stSendMessage;

  l_stSendMessage.m_nMessageType = htons(a_nMessageID);
  l_stSendMessage.m_nFollowingMessageLength = htons(a_nMessageSize);

  if ((l_nErrorCode = WriteOnHost(sock, (char*) &l_stSendMessage, sizeof(MESSAGE))) < 0)
      return l_nErrorCode;

  if ((l_nErrorCode = WriteOnHost(sock,(char*) a_pszMessage,ntohs(l_stSendMessage.m_nFollowingMessageLength))) < 0)
      return l_nErrorCode;
	
  return 1;
}

/************************************************************************************************************
 * Description : 
 *               Read the message and the data from the socket
 * Arguments   :
 *               a_nMessageID - Message ID to be read
 *               a_pszMessage - Message to be read
 *               a_pstHostDetails - Pointer to the gost details
 *               a_nWhichSocket - TCP/UNIX domain socket
 * Return      :
 *               ErrorCode / 1
 ************************************************************************************************************/

int ReadMessageAndData(sock,a_nMessageID, a_pszMessage)
     int sock;
     int a_nMessageID;
     char* a_pszMessage; 
{


  int l_nErrorCode;
  MESSAGE l_stReceivedMessage;

  if ((l_nErrorCode = ReadOnHost(sock, (char*) &l_stReceivedMessage,sizeof(MESSAGE))) < 0)
    {
      return l_nErrorCode;
    }
  if ((l_nErrorCode = ReadOnHost(sock, (char*) a_pszMessage,ntohs(l_stReceivedMessage.m_nFollowingMessageLength))) < 0)
    {
      return l_nErrorCode;
    }

  return 1;
}


