/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */
/***********************************************************************************************
 * Program Name 
 * ~~~~~~~~~~~~
 * donor.c
 *
 * Purpose 
 * ~~~~~~~
 * This file contains the code for Donor's functions.
 *
 * Developer 
 * ~~~~~~~~~~
 * Palankeswara Rao N V
 *
 * Date 
 * ~~~~
 * 29/March/2003
 *
 ***********************************************************************************************/

#include <linux/kernel.h>  
#include <linux/sys.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <stdio.h>
#include <sys/sysinfo.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <string.h>
#include <sys/wait.h>

#include "GCError.h"
#include "GCHost.h"
#include "GCMesg.h"
#include "GCXDRRoutines.h"
#include "Donor.h"
#include "GCNetLevel.h"
#include "GCDefines.h"

HOST_DESCRIPTOR g_stDonor;
int donsock;  /* the socket descriptor which connects with the cluster */
int g_nNumberofDesc = 0;
int l_nSocketDescriptor[50];
int CurrentDescIndex = 0;
double  HPF;

typedef void Sigfunc(int);


/*******************************************************************************************
 * Description : 
 *               Main program
 *                
 * Arguments   :
 *               a_nNumberofArguments - Number of command line arguments passed.
 *               a_pszArguments - command line arguments.
 *               
 * Return      :
 *               Error Code / 1
 ********************************************************************************************/

int main(a_nNumberofArguments, a_pszArguments)
    	int a_nNumberofArguments;
     	char **a_pszArguments;
{
	
	#ifdef SIGTERM
  	signal(SIGTERM, CleanUpDonor);
	#endif
	
	if(a_nNumberofArguments != 2)
	{
		write(1, "\nDonor: CC address has to be sent as a command line argument\n", 65);
	}
	
   	if(InitializeDonor(a_pszArguments[1]) < 0)
    	{
		write(1, "\nDonor: Unable to Initialise the Donor", 40);
		exit(1);
    	}

   	if(AcceptMessages() < 0)
		write(1, "\nDonor: Unable to accept the messages.", 40);

   	exit(1);
}


/*******************************************************************************************
 * Description : 
 *               Intialization of the Donor in which it contacts the cluster coordiantor and 
 *               sends it details.
 *               
 * Arguments   :
 *               a_pszClusterCoordinator - Name of the machine where 
 *               system coordiantor is running.
 *               
 * Return      :
 *               Error Code / 1
 *******************************************************************************************/

int InitializeDonor(a_pszClusterCoordinator)
     	char *a_pszClusterCoordinator;

{
	int l_nErrorCode;
	int l_nMessageID;

	/* Trap the termination signal so that when the
	* kill command is issued to the process it does
	* a gracefull exit from the code */

/*******************************
	#ifdef SIGTERM
  	signal(SIGTERM, CleanUpDonor);
	#endif
  	************************************************/
	
  	if(!InitializeHost(&g_stDonor, TCP_HOST))
	{
		write(1, "\nDonor: Unable to Initialise host", 36);
		return -1;
	}
 
 	/*
   	* Create a Socket and Connect to the Cluster
   	*/
     	/* g_stDonor is host structure of global...*/
	if((donsock = CreateTCPClientHostAndConnect(&g_stDonor, a_pszClusterCoordinator,
					CC_DR_PORT)) < 0)
    	{
      		return donsock;
    	}


	/* Create a Donor Socket and Listen on it for Acceptor Requests*/
	if((dsock = CreateTCPServerHostAndListen(&g_stDonor, DR_AC_PORT)) < 0)
	      	return dsock;
 	

  	/* Send an *Add Request* Message to Cluster Coordinator */
  	
  	l_nMessageID = DR_CC_ADDREQ;
  	
  	if((l_nErrorCode = WriteXDRSignedInteger(donsock, DR_CC_CONTROL,
				(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST)) < 0)
      		return l_nErrorCode;

  	/* Send the Donor Details to Cluster Coordinator */

  	/* Send Host Name */
  	if((l_nErrorCode = WriteXDRString(donsock, DR_CC_DATA, (void *)g_stDonor.m_pszHostName,
					HOSTNAME_LENGTH, TCP_HOST)) < 0)
	      	return l_nErrorCode;

  	/* Send Address */
  	if((l_nErrorCode = WriteMessageAndData(donsock, DR_CC_DATA, 
			    (char *)g_stDonor.m_pstTCPHostAddress, sizeof(struct sockaddr_in))) < 0)
	     	return l_nErrorCode;

	/* Send Process Id */
  	if((l_nErrorCode = WriteXDRSignedInteger(donsock, DR_CC_DATA,
			  	(void *)&g_stDonor.m_nHostProcessID, 
				sizeof(g_stDonor.m_nHostProcessID), TCP_HOST)) < 0)
	      	return l_nErrorCode;

	/* Send Login Name*/
  	if((l_nErrorCode = WriteXDRString(donsock, DR_CC_DATA, (void *)g_stDonor.m_pszHostLogin, 
					LOGINNAME_LENGTH, TCP_HOST)) < 0)
	      	return l_nErrorCode;

  	/* Send Daemon path*/
  	if((l_nErrorCode = WriteXDRString(donsock, DR_CC_DATA, (void *)g_stDonor.m_pszDaemonPath, 
					PATH_LENGTH, TCP_HOST)) < 0)
	      	return l_nErrorCode;
	
    	/* Receive the Acknowledgment*/
	if((l_nErrorCode = ReadXDRSignedInteger(donsock, CC_DR_CONTROL, (void *)&l_nMessageID,
					sizeof(l_nMessageID), TCP_HOST)) < 0)
	      	return l_nErrorCode;	
      		
	if(l_nMessageID == CC_DR_ADDACK )
	{
		write(1, "\nDonor: ACK Received from CC", 30); 
	     	return donsock;
	}
	
	else
	      	return -1;
}


/***************************************************************************************
 * Description : The Donor listens for messages from the cluster coordinator and
 * 		 from any local acceptor.
 * 		 
 * Arguments   : none
 * 
 * Return      : Error Code / 1
 ****************************************************************************************/

int AcceptMessages()
{
	int nsock;
	int j;
	int l_nErrorCode;
	fd_set rfd;

	FD_ZERO(&rfd);
	
	/*wait for messages from anywhere*/
	while(1)
	{
		FD_SET(donsock, &rfd);
	  	FD_SET(dsock, &rfd);

		for(j = 0; j < g_nNumberofDesc; j++)
	 	  	FD_SET(l_nSocketDescriptor[j], &rfd);
		
		if(AnyDataToRead(&rfd) == 1)
	      	{
	  		//write(1,"data from old connection",22);
			for(j = (g_nNumberofDesc-1); j >= 0; j--)
			{
				//write(1,"data from old connection",22);
			  	if(FD_ISSET(l_nSocketDescriptor[j], &rfd))
		     		{
	        			nsock = l_nSocketDescriptor[j];
		        		CurrentDescIndex = j;
			        	if((l_nErrorCode = ProcessMessage(nsock)) < 0)
  			         	{
						write(1, "\nDonor: Error in processing Message", 40);
	 		         		CloseHost();
	 		         	}
		      		}
		   	}
	  		
	  		/*Has a message come from the cluster coordinator? */
	  		if(FD_ISSET(donsock, &rfd))
	     		{
	    	        	if((l_nErrorCode = ProcessMessage(donsock)) < 0)
  		         	{
					write(1, "\nDonor: Error in processing Message", 40);
			   		CleanUpDonor(SIGTERM);
 	         		}
		      	}
		      	
		   	if(FD_ISSET(dsock, &rfd))
			{
		  		l_nSocketDescriptor[g_nNumberofDesc] = accept(dsock,
							(struct sockaddr *)NULL, 0);
			    	nsock = l_nSocketDescriptor[g_nNumberofDesc++];
				  
				if((l_nErrorCode = ProcessMessage(nsock)) < 0)
				{
					write(1, "\nDonor: Error in processing Message", 40);
					g_nNumberofDesc--;
				}
			}
		} //end AnyDataToRead
  	}
}


/**********************************************************************
 * Description: this function checks if any data is available.
 * Arguments  : rfd - pointer to the set of socket descriptors.
 * Return     : socket descriptor.
 **********************************************************************/

int AnyDataToRead(rfd)
	fd_set *rfd;
{
  	int l_nIsReady;
  	int l_nNumberofDescriptors;

  	l_nNumberofDescriptors = getdtablesize() + 1;

  /*
    Using the select call which polls and returns immediatly
    if no data is avialable since Timeval is NULL
  */
  	if((l_nIsReady = select(l_nNumberofDescriptors, rfd, (fd_set *)NULL, 
					(fd_set *)NULL, NULL)) > 0)
  	{
  	  	return 1;
  	}
   
	else
   		return -1;
}


/**********************************************************************************************
 * Description: this function processes an arrived message and calls appropriate action 
 * 		to be taken depending on the sender and the message sent.
 * 		
 * Arguments  :
 * 		sock	- socket descriptor on which the message has come.
 * Return     : 1 on success and -1 on error.
***********************************************************************************************/

int ProcessMessage(sock)
   	int sock;
{
	
	int l_nMessageID;
	int l_nErrorCode;
	
	/*get the message type */
  	if((l_nErrorCode = ReadXDRSignedInteger(sock, CC_DR_CONTROL, (void *)&l_nMessageID,
						sizeof(l_nMessageID), TCP_HOST)) < 0)
	  	return l_nErrorCode;

	switch(l_nMessageID)
	{
		case CC_DR_GETHPF:
			write(1, "\nDonor: HPF REQ has come", 30);
			return SendHPF(sock);
			
              	case CC_AD_GOODBYE:
              	        write(1, "\nDonor: CC Shutting down..\n", 30);
              	        _exit(1);
			
              	case CC_AD_ADDREJ:
              		write(1, "\nDonor: Rejected by CC\n", 30);
              		CleanUpDonor(SIGTERM);
			
              	case CC_DR_TASK:
              		write(1, "\nDonor: New Task has come ", 30);
              		GetTask(sock);
			write(1, "\nDonor: came back after task execution\n", 40);
			return 1;
			
              	case AC_DR_ADDREQ:
              		write(1, "\nDonor: Task from an acceptor\n", 35);
              	        
			if(AcceptorRead(sock) > 0)
				return SendResultToAcceptor(sock);
			else
				return -1;
			
		case AC_DR_GOODBYE:
			write(1, "\nDonor: Acceptor is shutting down", 40);
			return -1;
	}

	return 1;
}


/******************************************************************************************
 * Description : 
 *               this deletes all the memory allocated during the execution of 
 *               the donor daemon
 * Arguments   :
 *               a_nSignalNumber - Number of signal because which this signal was called
 * Return      :
 *               None
 *******************************************************************************************/

void CleanUpDonor(a_nSignalNumber)
	int a_nSignalNumber;
{
   	int j, l_nMessageID;

   	remove("/tmp/code1.c");
   	remove("/tmp/code1.out");
   	remove("/tmp/code1");

   	l_nMessageID = DR_AC_GOODBYE;
   
	//printf( "\nLoad Sent :: %ld\n",load);

   	for(j = 0; j < g_nNumberofDesc; j++)
	 	WriteXDRSignedInteger(l_nSocketDescriptor[j], DR_CC_CONTROL,
					(void *)&l_nMessageID, sizeof(l_nMessageID), TCP_HOST);

   	l_nMessageID = DR_CC_GOODBYE;
   
	WriteXDRSignedInteger(donsock, DR_CC_CONTROL, (void *)&l_nMessageID,
				sizeof(l_nMessageID), TCP_HOST);
	
   	write(1, "\nDonor: after closing\n", 30);
   
	close(donsock);
  	close(dsock);

   	GC_FREE(g_stDonor.m_pszHostName);
   	GC_FREE(g_stDonor.m_pstTCPHostAddress);
   	GC_FREE(g_stDonor.m_pszHostLogin);
   	GC_FREE(g_stDonor.m_pszDaemonPath);

  	exit(1);
}


Sigfunc *signal(int signo, Sigfunc *func)
{
   	struct sigaction act, oact;
   	act.sa_handler = func;
   	sigemptyset(&act.sa_mask);
   	act.sa_flags = 0;
   	
	if(sigaction(SIGTERM, &act, &oact) < 0)
         	write(1, "\nDonor: not able to process the signal\n", 40);
   
	return(oact.sa_handler);
}
        	

/*******************************************************************************
 * Description : this function updates the array of the open socket descriptors
*******************************************************************************/

int CloseHost()
{
  	int i;

  	/* Update the array of Descriptors opened */		
  	for(i = CurrentDescIndex; i < (g_nNumberofDesc-1); i++)
      		l_nSocketDescriptor[i] = l_nSocketDescriptor[i+1];

  	g_nNumberofDesc--;
  
	return 1;
}


int AcceptorRead(sock)
	int sock;
{
        int len;	
	char *code;
	int fd;
	int l_nBytesWritten;
	int l_nRemainingSize;
	char *l_pszRemainingBuffer;
        char *temp;
//        double l_LB,l_UB;
        char lb[25], ub[25];
        int pid;
	int l_nMessageID;

	/*Read the size of the source file from the acceptor*/
	if (ReadXDRSignedInteger(sock, AC_DR_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
	    	return -1;
        
	write(1, "\nDonor: size of the file read \n", 35);
	
	//printf("Size : %d",size);
	
	code = (char *)malloc(len);
	
	/*Receive the code of the source file from the acceptor*/
	if(ReadXDRString(sock, AC_DR_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
		
	write(1, "\nDonor: file read \n", 20);	
	
	fd = creat("/tmp/code1.c", S_IRWXU | S_IRWXG | S_IRWXO);
	l_pszRemainingBuffer = code;
	l_nRemainingSize = len;
	
	do
	{
		if((l_nBytesWritten = write(fd, l_pszRemainingBuffer, l_nRemainingSize)) <= 0)
			write(1, "\nDonor: file writing error\n", 30);
		
		l_nRemainingSize = l_nRemainingSize - l_nBytesWritten;
		l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesWritten;

	} while(l_nRemainingSize > 0);
 	
/*	if ( write(fd,code,len)!= len)
		write(1,"cant write",10);
*/
	close(fd);
		
	/*Read the lower bound of the problem*/
	if(ReadXDRString(sock, AC_DR_DATA, (void *)lb, 25, TCP_HOST) < 0)
	   	return -1;
	
	write(1, "\nDonor: Lower Bound read.. ", 30);
	
	/*Read the upper bound of the problem*/
	if(ReadXDRString(sock, AC_DR_DATA, (void *)ub, 25, TCP_HOST) < 0)
	      	return -1;
	
	write(1, "\nDonor: Upper Bound read ", 30);

	printf("\n lower bound = %s, Upper Bound = %s \n", lb, ub);
	
	/*send change status req to cluster*/
	l_nMessageID = DR_CC_STATUSCHANGE;
	if(WriteXDRSignedInteger(donsock, DR_CC_CONTROL, (void *)&l_nMessageID,
					sizeof(l_nMessageID), TCP_HOST) < 0)
		return -1;
	
	free(code);

	/*
	  sprintf(lb,"%ld",l_LB);	
	  sprintf(ub,"%ld",l_UB);	
	*/	

	temp = (char *)malloc(40);
	strcpy(temp, "g++ -o /tmp/code1");
	strcat(temp, " ");
	strcat(temp, "/tmp/code1.c");
	system(temp);

	free(temp);

	/*spawn a new process */
	if((pid = fork()) == 0)
	{
		/*In the child process*/
		execl("/tmp/code1", "/tmp/code1", &lb, &ub, "/tmp/code1.out", NULL);
		write(1, "\nDonor: Child process completed the work", 45);
		exit(0);
	}
	
	else
	{
		if(waitpid(pid, NULL, 0) == pid)
		{
			return 1;	
		}
	}
	
    	return 1;
}


/****************************************************************************************
Description : Sends the HPF info to the cluster
*****************************************************************************************/

int SendHPF(sock)
	int sock;
{
	
	int ccsock, accsock;
	int l_nMessageID = DR_CC_SENDHPF;
	struct sysinfo st;
	
	/*RECEIVE THE HPF REQUEST*/
	/*Receive Cluster Socket Desc(Cluster ID)*/
  	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&ccsock, sizeof(ccsock), TCP_HOST) < 0)
	     	return -1;
	
	/*Receive Acceptor Socket Desc(Acceptor ID)*/
  	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&accsock, sizeof(accsock), TCP_HOST) < 0)
	      	return -1;
	
	/*COMPUTE THE HPF*/	
	sysinfo(&st);
	HPF = ((1.0*st.loads[0] ) + ( 0.0*st.loads[1] ) + ( 0.0*st.loads[2] ) );
	
	if(HPF !=0){

        /* load=( (6400000/load) > 10 ? 6400000/load : 10 ); */
        HPF = 6000;
        }
	else
 		HPF=6000;
 		
	
 		
	/*SEND THE LOAD*/
	/*send the message*/
	if (WriteXDRSignedInteger(sock, DR_CC_CONTROL, (void*) &l_nMessageID, sizeof(l_nMessageID), TCP_HOST) < 0)
    		return -1;

  	/*Send Cluster Socket Desc(Cluster ID)*/
  	if (WriteXDRSignedInteger(sock,DR_CC_DATA,(void *) &ccsock,sizeof(ccsock), TCP_HOST) < 0)
	      return -1;
	
	/*send the Acceptor Socket Desc(Acceptor ID) */
  	if (WriteXDRSignedInteger(sock,DR_CC_DATA,(void *) &accsock,sizeof(accsock),TCP_HOST) < 0)
	      return -1;
	                                 	
	/*send the load*/
	if(WriteXDRDouble(sock,DR_CC_DATA,(void*) &HPF,sizeof(HPF),TCP_HOST) <0)
    		return -1;
    	else
    		write(1,"Donor::Load sent to CC ",24);	
    		
    	return 1;
}


/************************************************************************************
Description : Receives the Task Information from the Clusters
*************************************************************************************/

int GetTask(sock)
	int sock;
{
	
	int accsock, ccsock;
	char *code;
	
//        double l_LB,l_UB;

        int len, len1;
        int fd;
	int l_nBytesWritten, pid;
	int l_nRemainingSize;
	char *l_pszRemainingBuffer;
        char *temp;
        char lb[25], ub[25];
	char *res;
	int l_nMessageID;
	struct stat buf;

	/*Receive the Cluster ID*/
	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&ccsock, sizeof(ccsock), TCP_HOST) < 0)
	   	return -1;
	
	/*Receive the Acceptor ID*/
	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&accsock, sizeof(accsock), TCP_HOST) < 0)
	      	return -1;
	
	/*RECEIVE THE TASK INFO SENT BY THE ACCEPTOR THRU THE CLUSTER*/
	
	/*Read the size of the source file*/
	if(ReadXDRSignedInteger(sock, CC_DR_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
	      	return -1;
	
	/*Read the code of the file*/
	code = (char *)malloc(len);
	if(ReadXDRString(sock, CC_DR_DATA, (void *)code, len, TCP_HOST) < 0)
		return -1;
		
	write(1, "\nDonor: file read ", 20);	
	
	fd = creat("/tmp/code1.c", S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	
	l_pszRemainingBuffer = code;
	l_nRemainingSize = len;
	
	do
	{
		if((l_nBytesWritten = write(fd, l_pszRemainingBuffer, l_nRemainingSize)) < 0)
			write(1, "\nDonor: File writing error.\n", 30);
		
		l_nRemainingSize = l_nRemainingSize - l_nBytesWritten;
		l_pszRemainingBuffer = l_pszRemainingBuffer + l_nBytesWritten;

	} while(l_nRemainingSize > 0);
 	
	close(fd);
	
	/*Read the lower bound of the problem*/
	if(ReadXDRString(sock, CC_DR_DATA, (void *)lb, 25, TCP_HOST) < 0)
	      	return -1;
		
	/*Read the upper bound of the problem*/
	if(ReadXDRString(sock, CC_DR_DATA, (void *)ub, 25, TCP_HOST) < 0)
	      	return -1;
		
	free(code);

	/*PERFORM THE TASK SENT BY THE ACCEPTOR THRU THE CLUSTER*/
	
/*	sprintf(lb,"%ld",l_LB);	
	sprintf(ub,"%ld",l_UB);	
*/
	temp = (char *)malloc(40);
	strcpy(temp, "g++ -o /tmp/code1");
	strcat(temp, " ");
	strcat(temp, "/tmp/code1.c");
	system(temp);
	
	free(temp);
	
	/*spawn a new process */
	if((pid = fork()) == 0)
	{
		/*In the child process*/
		
		execl("/tmp/code1", "/tmp/code1", &lb, &ub, "/tmp/code1.out", NULL);
		exit(0);
	}
        
	else
	{
		if(waitpid(pid, NULL, 0) == pid)
		{
			/*Fetch the results from the output file*/
			if((fd = open("/tmp/code1.out", O_RDONLY)) < 0)
			{
				perror("Unable to open the output file");
				return -1;
			}
			
			stat("/tmp/code1.out", &buf);
			len1 = buf.st_size;
			res = (char *)malloc(len1);
			
			if(read(fd, res, len1) != len1 )
			{	
				write(1, "\nDonor: Unable to read from the output file\n", 50);
		                     	return -1;
			}
				
			close(fd);
				
			/*SEND THE RESULTS BACK TO THE ACCEPTOR THRU THE CLUSTER*/
		
			/*send the message*/
			l_nMessageID = DR_CC_TASK;
			if(WriteXDRSignedInteger(sock, DR_CC_CONTROL, (void *)&l_nMessageID,
						 sizeof(l_nMessageID), TCP_HOST) < 0)
				return -1;
    	
			/*send the cluster ID*/		
			if(WriteXDRSignedInteger(sock, DR_CC_DATA, (void *)&ccsock,
						 sizeof(ccsock), TCP_HOST) < 0)
			    	return -1;
    	
			/*send the acceptor ID*/	
			if(WriteXDRSignedInteger(sock, DR_CC_DATA, (void *)&accsock,
						 sizeof(accsock), TCP_HOST) < 0)
			    	return -1;
    		
			/*Send the size of the result*/
			if(WriteXDRSignedInteger(sock, DR_AC_DATA, (void *)&len1, 
						 sizeof(len1), TCP_HOST) < 0)
				return -1;
	
			/*Send the result*/	
			if(WriteXDRString(sock, DR_CC_DATA, (void *)res, len1, TCP_HOST) < 0)
				return -1;
				
			write(1, "\nDonor: result sent to the CC", 30);
				
			free(res);
			return 1; 				
		}
	}

	return 1;
}


/*******************************************************************************************
******************************************************************************************/

int SendResultToAcceptor(sock)
	int sock;
{
	
	int fd;
	struct stat buf;
	int len;
	char *res;
	int l_nMessageID = DR_AC_ADDACK;

	/*Send the result*/	
	/*Fetch the results from the output file*/
	if((fd = open("/tmp/code1.out", O_RDONLY)) < 0)
	{
		perror("Unable to open the output file");
		return -1;
	}
	
	stat("/tmp/code1.out", &buf);
	len = buf.st_size;
	res = (char *)malloc(len);

        if(read(fd, res, len) != len)
        {	
             	write(1, "\nDonor: Unable to read from output file", 40);
                       	return -1;
        }

	close(fd);
	
	/*send ACK to the acceptor*/
	if(WriteXDRSignedInteger(sock, DR_AC_CONTROL, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)				
		return -1;
	else
		write(1, "\nDonor : ADDACK sent to Acceptor", 35);

        /*send the size of the result */
	if(WriteXDRSignedInteger(sock, DR_AC_DATA, (void *)&len, sizeof(len), TCP_HOST) < 0)
		return -1;
	
	//write(1,"size sent  ",11);

	/*send the result */
	if(WriteXDRString(sock, DR_AC_DATA, (void *)res, len, TCP_HOST) < 0)
		return -1;
	
	//write(1,"result sent",11);			
	
	/*send change status req to cluster*/
	l_nMessageID = DR_CC_STATUSCHANGE;
	if(WriteXDRSignedInteger(donsock, DR_CC_CONTROL, (void *)&l_nMessageID,
				 sizeof(l_nMessageID), TCP_HOST) < 0)				
		return -1;
	
	free(res);
	
	return 1;
}
