/* Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */

#include "localDefns.h"
#include "sockfunc.c"
#include "sockplus.c"
#include "initLocalDesc.c"

void LCterminate(void) 
{
  if(DEBUG_ON) printf("LC terminating ... \n");
  system("rm mutex-file < yesfile");
  exit(1);
  return;
}

/************************ getMyIndex **********************************/
int getMyIndex(void)
{
  char name[50];
  int i;

  gethostname(name,50);
  
  for(i=0;i<MAX_NUM_LCS;i++)  
    if (strcmp(LocalDesc[i].mcName,name)==0) return i;

  printf("Index of current LC not present in file MachineNames.\n");
  LCterminate();        
}

/*********************** getMyLF **************************************/
float getMyLF(void)        
{
  /*
  char hostname[50];
  struct statstime *statp;
  float ret;

  gethostname(hostname,50);
  statp = (struct statstime *) malloc(sizeof(struct statstime));

  if (rstat(hostname,statp)==0){
    ret = (float) statp->avenrun[0];
    free(statp);
    return ret;
  }
  else{
    printf("LC unable to calculate it's load factor. Exiting ..\n");
    free(statp);
    LCterminate();
  }
*/
 return(1.5);
}

/************************* ProcessMsgSC ********************************/
void ProcessMsgSC(int sockfd)
{
  Message msg,failMsg;
  char *buffer,*filename,*curfile;
  int i,j,k,p1,p2;
  int msgSize,argSize,present,same_arch,childPid;
  char *argv_var[2],sysStr[150],exeStr[150];
  int rec_temp,repetition;
  int fd,filesize,tmpmsgtype;
  struct stat *stat_buf;

  read_on_a_sock(sockfd,(char *) &msg, sizeof msg);
  tmpmsgtype = ntohl(msg.type1);

  switch (tmpmsgtype) {

  case TAKE_LOAD_FACTOR:
    if(DEBUG_ON) printf("Got the message ..TAKE_LOAD_FACTOR from SC\n");

    /* Find the user who requested */
    for(i=0;((UserDesc[i].pid!=ntohl(msg.srcPid)) && (i<MAX_NUM_USERS));i++);
    
    /* Send him the load factors */
    if (UserDesc[i].valid){
      write_on_a_sock(UserDesc[i].sockfd,(char *) &msg, sizeof msg);

      msgSize = ntohl(msg.follMsgSize);
      buffer = (char *) malloc(msgSize);
      read_on_a_sock(sockfd,buffer,msgSize);
      write_on_a_sock(UserDesc[i].sockfd,buffer,msgSize);
      free(buffer);
    }
    break;
    

  case TAKE_WORK:
    if(DEBUG_ON) printf("Got the message ..TAKE_WORK(seqno = %d) from SC\n",ntohl(msg.srcSeqNo));
         
    msgSize = ntohl(msg.follMsgSize);
    argSize = ntohl(msg.follArgSize);
    repetition = FALSE;

    /* check for repeat work */
    for(i=0;(i<MAX_NUM_RECOVERIES);i++) {
      if(recover[i].valid && recover[i].srcLocal == ntohl(msg.srcLocal) && \
         recover[i].srcPid == ntohl(msg.srcPid) && \
         recover[i].srcSeqno == ntohl(msg.srcSeqNo)) 
	{
	  repetition = TRUE;
	  break;
	}
    }

    if (repetition) 
      {
	/* just read the remaining messages which contain the RIB name & the arguments */

	if(DEBUG_ON) printf("Repeat Work. No new process spawned..\n");
	buffer = (char *) malloc(msgSize);
	read_on_a_sock(sockfd,buffer,msgSize);
	free(buffer);

	buffer = (char *) malloc(argSize);
	read_on_a_sock(sockfd,buffer,argSize);
	free(buffer);
	break;
      }
                     
    if(DEBUG_ON) printf("New work arrived\n");        /* New work; Go on */
    for(i=0;((recover[i].valid) && (i<MAX_NUM_RECOVERIES));i++) ;   
    if(i >= MAX_NUM_RECOVERIES) printf("No more space for work. Try after sometime\n");
    recover[i].valid = TRUE;
    recover[i].srcLocal = ntohl(msg.srcLocal);
    recover[i].srcPid = ntohl(msg.srcPid);
    recover[i].destLocal = ntohl(msg.destLocal);
    recover[i].srcSeqno = ntohl(msg.srcSeqNo);
    rec_temp = i;
    recover[i].argList.resultSize = ntohl(msg.resultSize);
    recover[i].argList.arg = (char *) malloc(argSize);
    recover[i].argList.size = argSize;
    
    /* Get the RIB name from the SC */
    filename = (char *) malloc(msgSize+1);
    read_on_a_sock(sockfd,filename,msgSize);
    *(filename + msgSize) = '\0'; 

    /* Get the arguments into arglist[i] */
    read_on_a_sock(sockfd,recover[i].argList.arg,argSize);

    sprintf(exeStr,"%s.%d.exe",filename,LocalDesc[getMyIndex()].GROUPid);
    strcpy(recover[i].ribAttached,exeStr);
    
    /* Check if the file is already present in this LC */

    if((fd = open(exeStr,O_RDONLY)) == -1) 
      present = FALSE;
    else {
      present = TRUE;
      close(fd);
    }
    
    if(present == FALSE) {

      /* Check if the source code is present in this LC */
      /* This case arises for NFS file servers */
      /* Otherwise if exe file is absent, source file will also be absent */

      if((fd = open(filename,O_RDONLY)) == -1) 
	present = FALSE;
      else {
	present = TRUE;
	close(fd);
      }

      if(present == FALSE) {
	/* Even the source file is absent. So if the msg to */
	/* transfer 'curfile' was not sent to SC send it. */
	
	present = FALSE;
	for(p1 = 0,p2 = 0;(p2 < numRibsTransfered) && (p1 < MAX_NUM_RIBS);p1++) 
	  if(Ribname[p2] != NULL) {
	    if(strcmp(Ribname[p2],exeStr) == 0) {
	      present = TRUE;
	      break;
	    }
	    p2++;
	  }
	
	if(! present) {
	  msg.type1 = htonl(GET_FILE);
	  if(LocalDesc[ntohl(msg.srcLocal)].GROUPid == LocalDesc[ntohl(msg.destLocal)].GROUPid) {
	    msg.type2 = htonl(1);               /* Executable file is required */
	    msgSize = strlen(exeStr);
	    msg.follMsgSize = htonl(msgSize);
	    curfile = exeStr;
	  } else {
	    msg.type2 = htonl(0);               /* Code is required */
	    msgSize = strlen(filename);
	    msg.follMsgSize = htonl(msgSize);
	    curfile = filename;
	  }

	  /* Send the message first and the file name next */
	  write_on_a_sock(sockfd,(char *)&msg,sizeof msg);
	  write_on_a_sock(sockfd,curfile,msgSize);
	  
	  /* Update the file name in the files transfered array */
	  for(k = 0;k < MAX_NUM_RIBS;k++) 
	    if(Ribname[k] == NULL) break;
	  
	  Ribname[k] = (char *)malloc(strlen(exeStr)+1);
	  strcpy(Ribname[k],exeStr);
	  numRibsTransfered++;
	  
	} else if(DEBUG_ON) printf("Message for GET_FILE already sent\n");
	break;

      } else {

	/* Source code for the current file is present */
	/* Compile it to get the executable code */

	switch (LocalDesc[getMyIndex()].GROUPid){
	  
	case SPARCDOMAIN:
	  sprintf(sysStr,"gcc -o %s %s -lnsl -lsocket /usr/ucblib/libucb.a -DSPARC",exeStr,filename);
	  break;
	  
	case SUNDOMAIN:
	  sprintf(sysStr,"gcc -o %s %s -lrpcsvc -DSUN",exeStr,filename);
	  break;
	    
	case INTELDOMAIN:
	  sprintf(sysStr,"gcc -o %s %s -DLINUX",exeStr,filename);
	  break;
	}
	
	/* gain exclusive access */	
	if(lockf(fd_mutex_file,F_LOCK,0) == -1) {
	  printf("error locking file mutex-file\n");
	  perror("lock: ");
	  exit(1);
	}
      
	if(DEBUG_ON) printf("Compiling the rib ...%s\n",filename,sysStr);
	system(sysStr);
      
	/* release exclusive access */
	if(lockf(fd_mutex_file,F_ULOCK,0) == -1) {
	  printf("error unlocking file mutex-file\n");
	  perror("unlock :");
	  exit(1);
	}
      }
    }
    free(filename);
    
    childPid = fork();
    if (childPid < 0){
      printf("fork() failed. Exiting ..\n");
      LCterminate();
    }
    if (childPid == 0){        /* CHILD */
      /* Here the executable file is present in the LC */

      /* Execute the rib */
      argv_var[0] = (char *)malloc(sizeof(char) * (strlen(exeStr) + 1));
      argv_var[1] = NULL;
      strcpy(argv_var[0],exeStr);
      if(DEBUG_ON) printf("Executing %s(seqno = %d)\n",exeStr,ntohl(msg.srcSeqNo));

      if(execv(exeStr,argv_var) == -1) {
	printf("Error in execv \n");
	exit(1);
      }
    }

    recover[rec_temp].destPid = childPid;    
    break;

  case GET_FILE:
    if(DEBUG_ON) printf("Got the message ..GET_FILE from SC\n");

    /* Read the file name */
    msgSize = ntohl(msg.follMsgSize);
    filename = (char *)malloc(msgSize + 1);
    read_on_a_sock(sockfd,filename,msgSize);
    *(filename + msgSize) = '\0';

    /* open the file and get its size */
    if((fd = open(filename,O_RDONLY)) == -1) {
      printf("Error! File %s does not exist. Exiting ...\n",filename);
      exit(1);
    }
    stat_buf = (struct stat *)malloc(sizeof(struct stat));
    if(fstat(fd,stat_buf)==-1)  {
      printf("Error in fstat. Exiting...\n");
      exit(1);
    }   
    filesize=stat_buf->st_size;
    free(stat_buf);

    /* read the file into buffer */
    buffer = (char *)malloc(filesize);
    if(read(fd,buffer,filesize) == -1) {
      printf("Error in reading the file %s. Exiting ...\n",filename);
      exit(1);
    }

    /* Send the message TAKE_FILE to the SC */
    msg.type1 = htonl(TAKE_FILE);
    msgSize = strlen(filename);
    msg.follMsgSize = htonl(msgSize);
    msg.follArgSize = htonl(filesize);

    write_on_a_sock(sockfd,(char *)&msg,sizeof msg);
    write_on_a_sock(sockfd,filename,msgSize);

    /* Send the actual file next */
    write_on_a_sock(sockfd,buffer,filesize);

    close(fd);
    free(buffer);
    free(filename);
    break;

  case TAKE_FILE:
    if(DEBUG_ON) printf("Got the message TAKE_FILE from SC\n");

    /* Read the file name from SC */
    msgSize = ntohl(msg.follMsgSize);
    filename = (char *) malloc(msgSize+1);
    read_on_a_sock(sockfd,filename,msgSize);
    *(filename + msgSize) = '\0';

    /* Read the actual file into buffer */
    argSize = ntohl(msg.follArgSize);
    buffer = (char *)malloc(argSize);
    if(buffer==NULL) {
      printf("Couldn't allocate memory for exe\n");
      LCterminate();
    }
    read_on_a_sock(sockfd,buffer,argSize);

    if(DEBUG_ON) printf("Saving the file %s\n",filename);
    /* create the file with name buffer and store buffer in it */
    if((fd = open(filename,O_CREAT | O_WRONLY,0755)) == -1) {
      printf("error saving rib file\n");
      perror("Open: ");
      LCterminate();
    }
    if(k = (write(fd,buffer,argSize)) == -1) {
      printf("error writing output file\n");
      perror("write: ");
      LCterminate();
    }
    close(fd);
    free(buffer);

    if(ntohl(msg.type2) == 0) 
      sprintf(exeStr,"%s.%d.exe",filename,LocalDesc[getMyIndex()].GROUPid);
    else sprintf(exeStr,"%s",filename);

    /* Remove the file name from the Ribfile array */
    for(k = 0;k < MAX_NUM_RIBS;k++) 
      if(Ribname[k] != NULL)
	if(strcmp(exeStr,Ribname[k]) == 0) {
	  free(Ribname[k]);
	  Ribname[k] = NULL;
	  numRibsTransfered--;
	  break;
	}

    if(ntohl(msg.type2) == 0) { 
      /* Compilation of the code required */
      /* Compile command differs for different archirectures */
      
      switch (LocalDesc[getMyIndex()].GROUPid){
	
      case SPARCDOMAIN:
	sprintf(sysStr,"gcc -o %s %s -lnsl -lsocket /usr/ucblib/libucb.a -DSPARC",exeStr,filename);
	break;
	
      case SUNDOMAIN:
	sprintf(sysStr,"gcc -o %s %s -lrpcsvc -DSUN",exeStr,filename);
	break;
	
      case INTELDOMAIN:
	sprintf(sysStr,"gcc -o %s %s -DLINUX",exeStr,filename);
	break;
	
      default: break;
      }
      
      /* gain exclusive access */	
      if(lockf(fd_mutex_file,F_LOCK,0) == -1) {
	printf("error locking file mutex-file\n");
	perror("lock: ");
	exit(1);
      }
      
      if(DEBUG_ON) printf("Compiling the rib ...%s\n",filename,sysStr);
      system(sysStr);
      
      /* release exclusive access */
      if(lockf(fd_mutex_file,F_ULOCK,0) == -1) {
	printf("error unlocking file mutex-file\n");
	perror("unlock :");
	exit(1);
      }
    }
    
    /* Execute the ribs associated with this file */
    for(k=0;k < MAX_NUM_RECOVERIES;k++) 
      if(recover[k].valid && strcmp(recover[k].ribAttached,exeStr) == 0) {
	childPid = fork();
	if (childPid < 0) {
	  printf("fork() failed. Exiting ..\n");
	  LCterminate();
	}
	if (childPid == 0) {  
	  argv_var[0] = (char *)malloc(sizeof(char) * (strlen(exeStr) + 1));
	  argv_var[1] = NULL;
	  strcpy(argv_var[0],exeStr);
	  if(DEBUG_ON) printf("Executing %s(seqno = %d)\n",exeStr,recover[k].srcSeqno);
	  if(execv(exeStr,argv_var) == -1) {
	    printf("Error in execv. Exiting ... \n");
	    exit(1);
	  }
	}
	/* recover[k].argList.Pid = childPid; */ 
	recover[k].destPid = childPid;             
      }
  
    break;
    
  case TAKE_RESULTS:
    msgSize = ntohl(msg.follMsgSize);

    for(i=0;((resList[i].valid) && (i<MAX_NUM_RESULTS));i++) ;

    /* save the results in resList[] */
    resList[i].valid = TRUE;
    resList[i].resultPtr = (char *) malloc(msgSize);
    read_on_a_sock(sockfd,resList[i].resultPtr,msgSize);

    resList[i].resultSize = msgSize;
    resList[i].srcLocal = ntohl(msg.srcLocal);
    resList[i].destLocal = ntohl(msg.destLocal);
    resList[i].srcPid = ntohl(msg.srcPid);
    resList[i].destPid = ntohl(msg.destPid);
    resList[i].srcSeqNo = ntohl(msg.srcSeqNo);

    break;

  case GENERATOR_FAILED:
    if(DEBUG_ON) printf("Got the message ..GENERATOR_FAILED from SC\n");

    /* Kill the generator process and mark all the executors of that process as invalid */
    for(i=0;i<MAX_NUM_RECOVERIES;i++)
      if (recover[i].valid && recover[i].srcPid == ntohl(msg.srcPid) && \
	  recover[i].srcLocal == ntohl(msg.srcLocal)) {
	  for(k=0;k<MAX_NUM_USERS;k++)
	    if(UserDesc[k].valid && UserDesc[k].pid == recover[i].destPid) {
	      failMsg.type1 = htonl(TERMINATE);
	      write_on_a_sock(UserDesc[k].sockfd,(char *)&failMsg,sizeof(failMsg));
	      UserDesc[k].valid = FALSE;
	      recover[i].valid = FALSE;
	      free(recover[i].argList.arg);
	      break;
	    }
	  break;
      }
    break;

  case LC_TERMINATING:
    printf("Local Coordinator %s terminating\n",LocalDesc[ntohl(msg.srcLocal)].mcName);

    for(i=0;i<MAX_NUM_RECOVERIES;i++)
      if(recover[i].valid && recover[i].srcLocal == ntohl(msg.srcLocal)) {
	for(k = 0;k < MAX_NUM_USERS;k++)
	  if ((UserDesc[k].valid)&&(UserDesc[k].pid == recover[i].destPid)){
	    failMsg.type1 = htonl(TERMINATE);
	    write_on_a_sock(UserDesc[k].sockfd,(char *)&failMsg,sizeof(failMsg));
	    UserDesc[k].valid = FALSE;
	    recover[i].valid = FALSE;
	    free(recover[i].argList.arg);
	    break;
	  }
	break;
      }
    break;

  case TERMINATE:
    printf("Got the message ..TERMINATE from SC\n");
    fflush(stdout);
    LCterminate();
    break;

  default:
    printf("Got unknown message, %d, from SC. Terminating ...\n",tmpmsgtype);
    fflush(stdout);
    LCterminate();
    break;
    }
}

/**************************** ProcessMsgUSER ******************************/
void ProcessMsgUSER(int index)
{
  Message msg,failMsg;
  char* buffer;
  int i,k,sent;
  int msgSize,argSize;
  int tmpmsgtype;

  read_on_a_sock(UserDesc[index].sockfd,(char *) &msg, sizeof msg);
  tmpmsgtype = ntohl(msg.type1);

  switch (tmpmsgtype) {

  case GET_LOAD_FACTOR:
    if(DEBUG_ON) printf("Got the message ..GET_LOAD_FACTOR from USER process %d\n",UserDesc[index].pid);
    write_on_a_sock(SysConnxnSock,(char *) &msg, sizeof msg);
    break;

  case TAKE_WORK:
    if(DEBUG_ON) printf("Got the message..TAKE_WORK from USER process %d\n",UserDesc[index].pid);

    msg.destLocal = msg.index;
    msg.srcLocal = htonl(getMyIndex()); 

    write_on_a_sock(SysConnxnSock,(char *) &msg, sizeof msg);

    /* Get the RIB name from the user and send it to SC */
    msgSize = ntohl(msg.follMsgSize);
    buffer = (char *) malloc(msgSize);
    read_on_a_sock(UserDesc[index].sockfd,buffer,msgSize);
    write_on_a_sock(SysConnxnSock,buffer,msgSize);
    free(buffer);

    /* Get the arguments from the user and send them to SC */
    argSize = ntohl(msg.follArgSize);
    buffer = (char *) malloc(argSize);
    read_on_a_sock(UserDesc[index].sockfd,buffer,argSize);
    write_on_a_sock(SysConnxnSock,buffer,argSize);
    free(buffer);
    break;

  case CHECK_RESULTS:
    if(DEBUG_ON) printf("Got the message ..CHECK_RESULTS from USER process %d\n",UserDesc[index].pid);
                    
    /* Check whether results are available or not and send the message to the user */
    for(i=0;i<MAX_NUM_RESULTS;i++)
      if(resList[i].valid && (resList[i].srcPid == ntohl(msg.srcPid)) && \
	 (resList[i].srcSeqNo == ntohl(msg.srcSeqNo)))
	msg.type1 = htonl(RESULTS_AVAILABLE);
      else msg.type1 = htonl(RESULTS_NOT_AVAILABLE);

    write_on_a_sock(UserDesc[index].sockfd,(char *) &msg, sizeof msg);
    break;

  case WANT_RESULTS:
    if(DEBUG_ON) printf("Got the message ..WANT_RESULTS(SeqNo = %d) from USER process %d\n",ntohl(msg.srcSeqNo),UserDesc[index].pid);

    sent=FALSE;
    for(i=0;i<MAX_NUM_RESULTS;i++)
      if(resList[i].valid && resList[i].srcPid == ntohl(msg.srcPid) && \
	 resList[i].srcSeqNo == ntohl(msg.srcSeqNo)) {
	msg.type1 = htonl(RESULTS_AVAILABLE);
	msg.follMsgSize = htonl(resList[i].resultSize);
	write_on_a_sock(UserDesc[index].sockfd,(char *) &msg, sizeof msg);
	write_on_a_sock(UserDesc[index].sockfd,resList[i].resultPtr,resList[i].resultSize);
	resList[i].valid = FALSE;
	free(resList[i].resultPtr);
	sent=TRUE;
	break;
      }
    if(!sent){
      msg.type1 = htonl(RESULTS_NOT_AVAILABLE);
      write_on_a_sock(UserDesc[index].sockfd,(char *) &msg, sizeof msg);
    }
    break;

  case TAKE_RESULTS:
    if(DEBUG_ON) printf("Got the message ..TAKE_RESULTS from USER process %d\n",UserDesc[index].pid);
    
    /*Forward to SC */
    for(i=0;i < MAX_NUM_RECOVERIES && recover[i].destPid != ntohl(msg.srcPid);i++) ;
    if (recover[i].valid) {
      if(DEBUG_ON) printf("SENDING ACTUAL RESULTS(seqno = %d) TO SC NOW...\n",i,recover[i].srcSeqno);
      msg.srcLocal = htonl(recover[i].srcLocal);
      msg.srcPid = htonl(recover[i].srcPid);
      msg.destPid = htonl(recover[i].destPid);
      msg.destLocal = htonl(recover[i].destLocal);
      msg.srcSeqNo = htonl(recover[i].srcSeqno);
      
      write_on_a_sock(SysConnxnSock,(char *) &msg, sizeof msg);
      msgSize = ntohl(msg.follMsgSize);
      buffer = (char *) malloc(msgSize);
      read_on_a_sock(UserDesc[index].sockfd,buffer,msgSize);
      write_on_a_sock(SysConnxnSock,buffer,msgSize);
      free(buffer);
      recover[i].valid = FALSE; /* removing that node from recover list */
      free(recover[i].argList.arg);
    }

    break;

  case USER_TERMINATING:
    if(DEBUG_ON) printf("Got the message ..USER_TERMINATING from USER process %d\n",UserDesc[index].pid);

    /* mark user process as invalid */
    for(i=0;i<MAX_NUM_USERS;i++)
      if(UserDesc[i].valid && UserDesc[i].pid==ntohl(msg.srcPid)){
	UserDesc[i].valid = FALSE;
	close(UserDesc[i].sockfd);
	break;
      }
    break;

  case GIVE_ARGS:
    if(DEBUG_ON) printf("Got the message ..GIVE_ARGS from USER process %d\n",UserDesc[index].pid);

    for(i=0;i<MAX_NUM_RECOVERIES;i++) {
      if ((recover[i].valid)&&(recover[i].destPid == ntohl(msg.srcPid))){
	msg.type1 = htonl(TAKE_ARGS);
	msg.follMsgSize = htonl(recover[i].argList.size);
	msg.resultSize = htonl(recover[i].argList.resultSize);    /* added by maams */
	write_on_a_sock(UserDesc[index].sockfd,(char *) &msg, sizeof msg);
	write_on_a_sock(UserDesc[index].sockfd,recover[i].argList.arg,recover[i].argList.size);
      }
    }
    break;

  case TERMINATE:
    if(DEBUG_ON) printf("Got the message ..TERMINATE from USER process %d\n",UserDesc[index].pid);
    LCterminate();
    break;

  default:
    if(DEBUG_ON) printf("Got UNKNOWN message from USER process %d\n",UserDesc[index].pid);
                 
    /* mark user process as invalid */
    UserDesc[index].valid = FALSE;
    for(i=0;((i<MAX_NUM_RECOVERIES) && (recover[i].destPid != UserDesc[index].pid));i++);
    
    if (i < MAX_NUM_RECOVERIES && recover[i].valid)   
      /* an executor failed. hence invalidating that work */
      {
	recover[i].valid = FALSE; 		
	printf("An executor failed abnormally(without giving the results)\n");
      }

    /* if it is a generator, inform the SC */
    if (UserDesc[index].genORexe==0) {
      msg.type1 = htonl(GENERATOR_FAILED);
      write_on_a_sock(SysConnxnSock,(char *) &msg, sizeof msg);
    }
    break;
 
  }

}
/******************************** main ***************************************/
void main(int argc,char *argv[])
{
  char connxnSockName[50];
  char rmKRISHNA[100];
  int connxnSock,tmpSock;
  int ready;			/* if a socket is ready for reading */
  int STATE;                    /* State of the Local Coordinator */
  int myIndex,myIndexhs;
  float myLoadFactor;
  Message msg;                  /* may be unnecessary */
  struct sockaddr_in connxnSockAddr; 
  struct sockaddr_in SysConnxnSockAddr;
  int i,k,ALL_OVER=0;
  XDR p;

  if (argc != 2){
    printf("SC machine name required as the (only) argument. Exiting ..\n");
    exit(1);
  }

  signal(SIGCLD,SIG_IGN);   /* To remove all the zombie child processes */

  STATE = LC_INIT;
    
  do {
    switch (STATE) {

    case LC_INIT: 
      system("touch mutex-file");
      if((fd_mutex_file = open("mutex-file", O_RDWR)) == -1)
          {
	    printf("error opening file mutex-file\n");
	    exit(1);
          }
      
      initLocalDescArray();
      myIndex = getMyIndex();
      sprintf(rmKRISHNA,"rm %sKRISHNA < yesfile",LocalDesc[myIndex].mcName);
      system(rmKRISHNA);

      /* Mark all users as inactive */
      for (i=0; i<MAX_NUM_USERS; i++) UserDesc[i].valid=FALSE;
      for (i=0; i<MAX_NUM_RECOVERIES; i++) recover[i].valid=FALSE;
      for (i=0; i<MAX_NUM_RESULTS; i++) resList[i].valid=FALSE;
      for (i=0; i<MAX_NUM_RIBS; i++) Ribname[i] = NULL;
 

      /* Establish connection with the SC */

     SysConnxnSock = get_client_sock_and_connect(argv[1],SYSPORTNO); 
     SysConnxnSockAddr = get_sock_addr(SysConnxnSock);

     /* Send the index of the current LC in the LocalDesc array to the SC */
	  
     myIndexhs = htonl(myIndex);
     write_on_a_sock(SysConnxnSock,(char *)&myIndexhs,sizeof(int));

     /* Calculate and send the LoadFactor to the SC */
     myLoadFactor = getMyLF();
     msg.type1 = htonl(TAKE_LOAD_FACTOR);
     xdrmem_create(&p,msg.loadFactor,10,XDR_ENCODE);
     xdr_float(&p,&myLoadFactor);
     write_on_a_sock(SysConnxnSock,(char *)&msg,sizeof(msg));

     /* Set up the common connxn to the users */
     gethostname (connxnSockName, 50);
     strcat (connxnSockName,"KRISHNA");
     connxnSock = get_u_server_sock_and_listen (connxnSockName); 
     connxnSockAddr = get_sock_addr(connxnSock);
 
     STATE = LC_CHK_MSG;
     break;

    case LC_CHK_MSG:
      
      /* Check for messages from the system coordinator */
      ready = read_ready_test (SysConnxnSock,0,100);
      if (ready) ProcessMsgSC(SysConnxnSock);

      /* Check in the common(user) channel for messages */
      ready = read_ready_test (connxnSock,0,100);
      if (ready > 0) {
	tmpSock = accept_on_a_sock(connxnSock);
	read_on_a_sock(tmpSock,(char *)&msg,sizeof msg);
	if (ntohl(msg.type1) == USER_INIT){
	  if(DEBUG_ON) printf("Got the msg... USER_INIT from %d\n",ntohl(msg.srcPid));
	  /* to find out the first available free slot in UserDesc array */
	  for(i=0;((UserDesc[i].valid) && (i<MAX_NUM_USERS));i++);
	  
	  if (i>=MAX_NUM_USERS){
	    printf("Limit Reached. No more User Programs ...\n");
	    close(tmpSock);
	    break;
	  }
	  UserDesc[i].genORexe = ntohl(msg.type2);
	  UserDesc[i].valid = TRUE;
	  UserDesc[i].sockfd = tmpSock;
	  UserDesc[i].pid = ntohl(msg.srcPid);
	}
	else {
	  printf("User trying to connect in an incorrect way\n");
	  close(tmpSock); 
	}
      }
      
      /* Check in the active exclusive channels for messages*/
      for (i=0;i<MAX_NUM_USERS;i++)
	if (UserDesc[i].valid) {
	  ready = read_ready_test (UserDesc[i].sockfd,0,100);
	  k++;
	  if (ready) ProcessMsgUSER(i);
	}
      
      
      STATE = LC_CHK_MSG;
      break;

    default: 
      printf("Local Coordinator in an unknown state. Exiting ...\n");
      LCterminate();
      break;
    }
  }
  while (!ALL_OVER);
}
