/* **********************************************************************
   File Name : localDefns.h
   Purpose   : Declarations And Definitions used by lc.c

   Files Included :

   1. CommonDefns.h  - Declarations and Definitions common to all components 
   in the system.

   2. syssockfunc.c and syssockplus.c - socket libraried for underlying 
   communication

   3. Various System 'include' files 
   ********************************************************************** 

Copyright (C) DOS Lab, Dept. of CS & E, IIT Madras.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in.
   ********************************************************************** */


#include <sys/file.h>
#include <fcntl.h>
#include <unistd.h>
#include "CommonDefns.h"

#include <stdio.h>
#include <errno.h>
#ifdef SPARC
#include <string.h>
#else
#include <strings.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <netinet/in.h>


#define MAX_NUM_USERS 1000 
#define MAX_NUM_RESULTS 1000
#define MAX_NUM_RECOVERIES 1000
#define MAX_NUM_RIBS 1000

char *Ribname[MAX_NUM_RIBS];
int numRibsTransfered = 0;

/* **********************************************************************
   Structure Name : UserDescriptor
   Purpose        : To store information about the user
   
   ********************************************************************** */
   

typedef struct {
  
  int genORexe; /* task generator = 0 and task executor = 1 */
  int pid;
  int funcNum;
  int valid;
  int sockfd;
  
} UserDescriptor;

UserDescriptor UserDesc[MAX_NUM_USERS];
LocalDescriptor LocalDesc[MAX_NUM_LCS];
ResultNode resList[MAX_NUM_RESULTS];  
int SysConnxnSock;  /* TCP Socket Descriptor for connection with the SC */

/* **********************************************************************
   Structure Name : ArgumentNode
   Purpose        : To store arguments for various tasks

   ********************************************************************** */

typedef struct {
  /*  int valid; */
  int size;
  int resultSize;    /* added later */
  /*  int Pid;*/
  char *arg;
} argumentNode;

/* **********************************************************************
   Structure Name : RecoveryNode
   Purpose        : To store recovery information about various tasks 

   ********************************************************************** */

typedef struct {
  int valid;
  int srcLocal;
  int srcPid;
  int srcSeqno;
  int destLocal;
  int destPid;
  char ribAttached[30];
  argumentNode argList;  
} RecoveryNode;

/* argumentNode argList[MAX_NUM_RIBS]; */
RecoveryNode recover[MAX_NUM_RECOVERIES];
LFmessage LFmsg;
int fd_mutex_file;   /* File Descriptor for 'mutex' file */

extern void initLocalDescArray(void);






