/* This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */


/* Later, modify get_line() to read till ";" , not "\n" as is now */

#include<sys/types.h>
#include<sys/stat.h>
#include<sys/file.h>
#include<stdio.h>
#ifdef SPARC
#include<stdlib.h>
#endif
#include<fcntl.h>
#include<unistd.h>
#include "CommonDefns.h"

#define MAX_NO_RIBS 100
#define MAX_BUF_SIZE 1000
#define MAX_NAME_LEN 255
#define MAX_NO_OF_ARGS 100
#define SEMICOLON_DETECTED -2     /* should defintely be some negative value */
#define TRUE 1
#define FALSE 0

#include "rib.count.c"
#include "form_rib.c"

char* get_line(int fd, int *p_offset, int *p_nbytes)
{
  char *str,c;

  str=(char *)malloc(MAX_BUF_SIZE * sizeof(char));
  lseek(fd, *p_offset, SEEK_SET);
  (*p_nbytes)=0;
  do
    {
      read(fd, &c, 1);
      str[*p_nbytes]=c;
      (*p_offset)++;
      (*p_nbytes)++;
    } while(c!='\n');
  str[*p_nbytes]='\0';
  return(str);

}

void copy(int fd1,int fd2)
{
  struct stat * stat_buf;
  int filesize,no_bytes,offset;
  char* linebuf;
  
  stat_buf=(struct stat *)malloc(sizeof(struct stat));
  if(fstat(fd1,stat_buf)==-1)
    {
      printf("error in fstat\n");
      exit(1);
    }  
  
  filesize=stat_buf->st_size;
  offset = 0;
  
  while(offset<filesize)
    {
      linebuf = get_line(fd1,&offset,&no_bytes);
      if(write(fd2, linebuf, no_bytes)==-1)
	{
	  printf("error writing outputfile\n");
	  perror("write :");
	  exit(1);
	}
      free(linebuf);
      
    }
 
}

int search_before(char *buf, int offset, char token)
{
  int current_offset=offset;
	
  while((buf[current_offset]!=token) && (buf[current_offset]!=';'))     
    current_offset++;

  if ((buf[current_offset]==';') && (token!=';')) 
    return(SEMICOLON_DETECTED);
  else 
    return(current_offset-1);
}

int search_after(char *buf, int offset, char token)
{
  int current_offset=offset;
	
  while((buf[current_offset]!=token) && (buf[current_offset]!=';'))      
    current_offset++;
	
  if ((buf[current_offset]==';') && (token!=';'))
    return(SEMICOLON_DETECTED);
  else 
    return(current_offset+1);
}

void main(int argc, char *argv[])
{
  int fd;                     /* file descriptor for file to be parsed */
  int fd_out;                 /* file descriptor for file to be written */
  int fd_rib[MAX_NO_RIBS];    /* file descriptor for the rib to be written */
  int fd_header;              /* file descripter for the header file of user program */
  int fd_temp;
  FILE *fp_global_rib_number; /* file pointer to file containing global rib number */
  int fd_global_rib_number;   /* file descriptor for the mutex file for current global rib number */
  int offset=0;               /* current offset in user file */
  char * headername;          /* name of the header file */ 
  int current_offset;         /* offset in one run thro' a line */
  int file_size;              /* size of current file */
  int nbytes;                 /* no of bytes read in a get_line() call */ 
  int last_string_reached; 
  int no_of_args;             /* number of arguments, apart from func, timeout, and retries */
  int i,p;                    /* general counter variables */

  int rib_number;             /* current global number read from the file global.number */
  int initial_rib_number;     /* initial rib number, before the current prog starts generating ribs */
	

  struct stat *stat_buf;      /* for the fstat call */
	
  struct {
    char *name;
  } arg_list[MAX_NO_OF_ARGS]; /* the argument list, after func, timeout, and retries */
	
  char *line_buf;              /* line buffer, used to read in every line of the program */

  /* The following are internal variables of the parser, used to capture the strings parsed */
  int _func_begin, _func_end;
  int _timeout_begin, _timeout_end;
  int _retries_begin, _retries_end;
  int _prev_end, _current_begin, _current_end;

  char _func_name[MAX_NAME_LEN];
  char _timeout_name[MAX_NAME_LEN];
  char _retries_name[MAX_NAME_LEN];
  char _current_name[MAX_NAME_LEN];
	
  char code[MAX_BUF_SIZE];         /* holds the generated code, at each step in the iteration */
  int numribs = 0;
  char rib_code[MAX_BUF_SIZE];       /* holds the generated rib code */
  char rib_name[MAX_NAME_LEN];       /* holds the rib name */
  char acribname[MAX_NO_RIBS][MAX_NAME_LEN];
  char ribname[MAX_NO_RIBS][MAX_NAME_LEN];  /* holds the names of all the ribs to be compiled */
  char domain_name[MAX_NAME_LEN];           /* holds the domain name */ 
  char rib_number_string[MAX_NAME_LEN];     /* curent rib number (global) */
  char rib_file_name[MAX_NAME_LEN];         /* rib addendum file name */
  char output_file_name[MAX_NAME_LEN];      /* output file name */
  char include_string[]="extern \"C\" \n { \n#include \"primitives.c\"\n }\n";
                                      /* to be inserted into the output file  */
  rib_node *ptr;
  int curpos = 0;
  char templine_buf[400],exestr[MAX_NAME_LEN];
  int header_filesize,no_bytes;


  if(argc!=4)
    {
      printf("exec format is parse <filename> <domain> <exefile name>\n");
      exit(1);
    }

  if((fd=open(argv[1],O_RDONLY))==-1)
    {
      printf("error opening file\n");
      exit(1);
    }
  initial_rib_number = get_initial_rib_number();

  strcpy(output_file_name,argv[1]);
  strcat(output_file_name,"-output.C");

  sprintf(exestr,"rm %s <yesfile",output_file_name);
  system(exestr);

  if((fd_out=open(output_file_name,O_CREAT|O_WRONLY,0666))==-1)
    {
      printf("error opening output file\n");
      exit(1);
    }
  lseek(fd,0,SEEK_SET);
  if(write(fd_out,include_string,strlen(include_string)) == -1)
    {
      printf("error writing output file\n");
      perror("write :");
      exit(1);
    }

  stat_buf=(struct stat *)malloc(sizeof(struct stat));
  if(fstat(fd,stat_buf)==-1)
    {
      printf("error in fstat\n");
      exit(1);
    }
   
  file_size=stat_buf->st_size;

  strcpy(domain_name,argv[2]);
	
  if((fd_global_rib_number = open("global.number.mutex", O_RDWR)) == -1)
    {
      printf("error opening file global.number.mutex\n");
      exit(1);
    }

  if((fp_global_rib_number = fopen("global.number", "r+")) == NULL)
    {
      printf("error opening file global.number\n");
      exit(1);
    }

  initialize_rib_name_table();         /* initialize the rib name table */
	
  while(offset<file_size)
    {
      line_buf=get_line(fd, &offset, &nbytes);
      curpos = 0;
      while(line_buf[curpos] == ' ' || line_buf[curpos] == '\t') curpos++;
      if (!((line_buf[curpos]  =='A') && \
	    (line_buf[curpos+1]=='R') && \
	    (line_buf[curpos+2]=='C')))
	{
	  if(write(fd_out, line_buf, nbytes)==-1)
	    {
	      printf("error writing output file\n");
	      perror("write :");
	      exit(1);
	    }
	  free(line_buf);
	}
      else 
	{
	  line_buf += curpos;
	  current_offset = 0;
	  nbytes -= curpos;
	  _func_begin=search_after(line_buf, current_offset, '(');
	  if(_func_begin==SEMICOLON_DETECTED)
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }
		    
	  _func_end=search_before(line_buf, _func_begin, ',');
	  if(_func_end==SEMICOLON_DETECTED)
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }
	  memcpy(_func_name, line_buf + _func_begin, (_func_end - _func_begin + 1));
	  _func_name[(_func_end - _func_begin + 1)]='\0';
		    
	  _timeout_begin=search_after(line_buf, _func_end, ',');
	  if(_timeout_begin==SEMICOLON_DETECTED) 
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }
		    
	  _timeout_end=search_before(line_buf, _timeout_begin, ',');
	  if(_timeout_end==SEMICOLON_DETECTED) 
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }
	  memcpy(_timeout_name, line_buf + _timeout_begin, (_timeout_end - _timeout_begin + 1));
	  _timeout_name[(_timeout_end - _timeout_begin + 1)]='\0';

	  _retries_begin=search_after(line_buf, _timeout_end, ',');
	  if(_retries_begin==SEMICOLON_DETECTED) 
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }

	  _retries_end=search_before(line_buf, _retries_begin, ',');
	  if(_retries_end==SEMICOLON_DETECTED) 
	    {
	      _retries_end=search_before(line_buf, _retries_begin, ';');
	      _retries_end--;
	      last_string_reached=TRUE;
	    }
	  memcpy(_retries_name, line_buf + _retries_begin, (_retries_end - _retries_begin + 1));
	  _retries_name[(_retries_end - _retries_begin + 1)]='\0';

	  no_of_args=0; 
	  _prev_end=_retries_end;
	  last_string_reached=FALSE;
						
	  while(last_string_reached!=TRUE)
	    {
	      _current_begin=search_after(line_buf, _prev_end, ',');
	      if (_current_begin==SEMICOLON_DETECTED)
		{
		  last_string_reached=TRUE;
		  break;
		}
					
	      _current_end=search_before(line_buf, _current_begin, ',');
	      if (_current_end==SEMICOLON_DETECTED)
		{
		  _current_end=search_before(line_buf, _current_begin, ';');
		  _current_end--;          /* to get the index of last ')' */
		  last_string_reached=TRUE;
		}
	      memcpy(_current_name, line_buf + _current_begin, (_current_end - _current_begin + 1));
	      _current_name[(_current_end - _current_begin + 1)]='\0';
	      arg_list[no_of_args].name=(char *)malloc(sizeof(char)*(_current_end - _current_begin + 2));
	      strcpy(arg_list[no_of_args].name,_current_name);
	      no_of_args++;
	      _prev_end=_current_end; 

	    }
	  
	  /* check for inadequate or extra arguments in the ARC_function_call */
	  if(no_of_args != 5)
	    {
	      printf("Format for arc function call is:\n ");
	      printf("ARC_function_call(func_name, timeout, retries, argument struct, struct size, machine index, tag, result size)\n");
	      exit(1);
	    }

	  /*****************************
	  NOTE : arg_list[0],arg_list[1],arg_list[2],arg_list[3],arg_list[4]
			are :			   
	       	       	packed argument structure
       			total sizeof the above structure
  	       		machine index,
	      	       	tag(sequence_no)
	              	result size
      	       		respectively

	  ****************************/

	  /* The following gets the string domain.rib into rib_name */
	  if((ptr=search_name(_func_name))!=NULL)
	    {
	      sprintf(rib_number_string,"%d",ptr->rib_number);
	      strcpy(rib_name, domain_name);
	      strcat(rib_name,".");
	      strcat(rib_name,rib_number_string);
	      strcat(rib_name,".C");
	    }
	  else
	    {
	      lseek(fd_global_rib_number,0,SEEK_SET);
	      if(lockf(fd_global_rib_number,F_LOCK,0) == -1)
		{
		  printf("error locking file global.number.mutex\n");
		  perror("lock :");
		  exit(1);
		}
				
	      fseek(fp_global_rib_number,0,SEEK_SET);
	      fscanf(fp_global_rib_number,"%d",&rib_number);
	      sprintf(rib_number_string,"%d",rib_number);

	      insert_name(_func_name,rib_number); /* puts the function name entry into the rib name table */
	      rib_number++;
	      strcpy(acribname[numribs],_func_name);

	      strcpy(rib_name, domain_name);
	      strcat(rib_name,".");
	      strcat(rib_name,rib_number_string);
	      strcat(rib_name,".C");
	      sprintf(rib_number_string,"%d\n",rib_number);
	      fseek(fp_global_rib_number,0,SEEK_SET);
	      fprintf(fp_global_rib_number,"%d",rib_number);
	      lseek(fd_global_rib_number,0,SEEK_SET);
	      if(lockf(fd_global_rib_number,F_ULOCK,0) == -1)
		{
		  printf("error unlocking file global.number.mutex\n");
		  perror("unlock :");
		  exit(1);
		}

	      strcpy(rib_file_name,rib_name);
	      strcpy(ribname[numribs],rib_name);
	      strcat(rib_file_name,"_addendum"); 
	      
	      if((fd_rib[numribs] = open(rib_file_name,O_CREAT|O_WRONLY,0666)) == -1)
		{
		  printf("error opening rib addendum file\n");
		  exit(1);
		}

	      sprintf(rib_code,"int main()\n{\n",_func_name);
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tchar *p,*result,*current_args;\n\tMessage msg_take_work,msg_give_args,msg_take_result;\n\tint arg_size,result_size,i;\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tinitialize_RIB();\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tmsg_give_args.type1 = htonl(GIVE_ARGS);\n\tmsg_give_args.srcPid = htonl(getpid());\n\twrite_on_a_sock(parent_sock, (char *)&msg_give_args, sizeof(msg_give_args));\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tread_on_a_sock(parent_sock, (char *)&msg_take_work, sizeof(msg_take_work));\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tif(ntohl(msg_take_work.type1) != TAKE_ARGS) {\n\t\tprintf(\"Possibly TERMINATE message from local coordinator\\n\");\n\t\texit(1);\n\t}\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
	      strcpy(rib_code,"\targ_size = ntohl(msg_take_work.follMsgSize);\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}

	      strcpy(rib_code,"\tcurrent_args = (char *) malloc(arg_size);\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tread_on_a_sock(parent_sock, current_args, arg_size);\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tresult_size = ntohl(msg_take_work.resultSize);\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
					
	      sprintf(rib_code,"\tp = %s(current_args);\n", _func_name);
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tresult = (char *)malloc(result_size*sizeof(char));\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tfor(i = 0; i < result_size; i++)\n\t\tresult[i] = p[i];\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\tmsg_take_result.type1 = htonl(TAKE_RESULTS);\n\tmsg_take_result.srcPid = htonl(getpid());\n\tmsg_take_result.follMsgSize = htonl(result_size);\n\twrite_on_a_sock(parent_sock, (char *)&msg_take_result, sizeof(msg_take_result));\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\twrite_on_a_sock(parent_sock, result, result_size);\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}	

	      strcpy(rib_code,"\tclose_RIB();\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
				
	      strcpy(rib_code,"\treturn 0;\n}\n");
	      if(write(fd_rib[numribs], rib_code, strlen(rib_code))==-1)
		{
		  printf("error writing output file\n");
		  perror("write :");
		  exit(1);
		}
	      close(fd_rib[numribs]);
	      numribs ++;
				
	    }
			
	  /* insert code in output file */
	  strcpy(code, "\tdp = open_datapack();\n");
	  if(write(fd_out, code, strlen(code))==-1)
	    {
	      printf("error writing output file\n");
	      perror("write :");
	      exit(1);
	    }

	  sprintf(code,"\tnew_insert_datapack(&dp,%s,%s);\n", arg_list[0].name, arg_list[1].name);
	  if(write(fd_out, code, strlen(code))==-1)
	    {
	      printf("error writing output file\n");
	      perror("write :");
	      exit(1);
	    }

	  strcpy(code, "\tclose_datapack(&dp);\n");
	  if(write(fd_out, code, strlen(code))==-1)
	    {
	      printf("error writing output file\n");
	      perror("write :");
	      exit(1);
	    }

	  sprintf(code,"\t%s = send_work_outside(\"%s\", &dp, %s, %s, %s,%s);\n", arg_list[3].name, rib_name, arg_list[4].name ,_timeout_name, _retries_name, arg_list[2].name);
	  if(write(fd_out, code, strlen(code))==-1)
	    {
	      printf("error writing output file\n");
	      perror("write :");
	      exit(1);
	    }

	  line_buf -= curpos;
	  free(line_buf);
	}
    }
  close(fd);
  close(fd_out);
  close(fd_global_rib_number);
  fclose(fp_global_rib_number);

  /*  Second pass to add the source code of the ribs which 
      is present in the header file to the beginning of the rib */
  
  for(i=0;i<numribs;i++)
    {
      printf("Ribname = %s\n",acribname[i]);
      headername  = (char *)malloc(sizeof(char)* (strlen(acribname[i])+3));
      strcpy(headername,acribname[i]);
      strcat(headername,".c");
      if((fd_header = open(headername,O_RDONLY,0666)) == -1)
	{
	  printf("error opening header file: %s\n",headername);
	  exit(1);
	}   
      strcpy(rib_file_name,ribname[i]);
      strcat(rib_file_name,"_addendum");

      if((fd_rib[i] = open(rib_file_name,O_RDONLY,0666)) == -1)
	{
	  printf("error opening rib file: %s\n",rib_file_name);
	  exit(1);
	}   
      if((fd_temp = open(ribname[i],O_CREAT|O_WRONLY,0666)) == -1)
	{
	  printf("error opening rib file: %s\n",ribname[i]);
	  exit(1);
	}   
      lseek(fd_header,0,SEEK_SET);
      lseek(fd_rib[i],0,SEEK_SET);
      lseek(fd_temp,0,SEEK_SET);

      if(write(fd_temp,"#include\"primitives.c\"\n",23)==-1)
	{
	  printf("error writing rib file\n");
	  perror("write :");
	  exit(1);
	}
      copy(fd_header,fd_temp);
      close(fd_header);
      copy(fd_rib[i],fd_temp);
      close(fd_rib[i]);
      sprintf(exestr,"rm %s < yesfile",rib_file_name);
      system(exestr); 
      close(fd_temp);
      free(headername);
    }

  if((strcmp(argv[2],"linux") == 0)||(strcmp(argv[2],"LINUX") == 0))
    sprintf(exestr,"g++ -g -lm -o %s %s -DLINUX",argv[3],output_file_name);
  else if((strcmp(argv[2],"sparc") == 0) || (strcmp(argv[2],"SPARC") == 0))
    sprintf(exestr,"g++ -g -lm -o %s %s -lnsl -lsocket /usr/ucblib/libucb.a -DSPARC",argv[3],output_file_name);
  else if((strcmp(argv[2],"sun") == 0) || (strcmp(argv[2],"SUN") == 0))
    sprintf(exestr,"g++ -g -lm -o %s %s -lrpcsvc -DSUN",argv[3],output_file_name);
  else {
    printf("error! Unknown Domain Name %s\nExiting ...\n",argv[2]);
    exit(1);
  }
  printf("Compiling %s-output.C ...",argv[1]);
  fflush(stdout);
  system(exestr);
  printf("done\n");
  for(i=0;i<numribs;i++)
    {
      printf("Compiling RIB %s (%s)...",ribname[i],acribname[i]);
      fflush(stdout);
      if((strcmp(argv[2],"linux") == 0)||(strcmp(argv[2],"LINUX") == 0))
	sprintf(exestr,"g++ -lm -o %s.%d.exe %s -DLINUX",ribname[i],INTELDOMAIN,ribname[i]);
      else if((strcmp(argv[2],"sparc") == 0) || (strcmp(argv[2],"SPARC") == 0))
	sprintf(exestr,"g++ -lm -o %s.%d.exe %s -lnsl -lsocket /usr/ucblib/libucb.a -DSPARC",ribname[i],SPARCDOMAIN,ribname[i]);
      else if((strcmp(argv[2],"sun") == 0) || (strcmp(argv[2],"SUN") == 0))
	sprintf(exestr,"g++ -lm -o %s.%d.exe %s -lrpcsvc -DSUN",ribname[i],SUNDOMAIN,ribname[i]);
      system(exestr); 
      printf("done\n");
    }
}
