/* This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */

#include "pack.c"
#include "CommonDefns.h"
#include "sockfunc.c"
#include "userdefs.h"
#include <sys/types.h>
#include <sys/time.h>

/*
   The folowing are global variables that will be available to every user 
   program due to linking the current file 
*/

int parent_sock;
int CurrentTime;                    /* This is global virtual time */
int out_going_sequence_no;
work_queue_t work_queue[MAX_STORABLE_WORK];
int My_pid;
struct datapack dp;
int initialised = FALSE;

#ifdef SPARC
extern "C" {
  int gettimeofday(struct timeval *tp);
}
#endif

int get_time(void)
{
  struct timeval tp;

#ifdef SUN3
  struct timezone tzp;
#endif

#ifdef LINUX
  struct timezone tzp;
#endif

  int time_value;
	
#ifdef SPARC
  if(gettimeofday(&tp) == -1)
    {
      printf("\nerror in time function\n");
      exit(1);
    }
#else
#ifdef LINUX
  if(gettimeofday(&tp,&tzp) == -1)
    {
      printf("\nerror in time function\n");
      exit(1);
    }
#else
  if(gettimeofday(&tp,&tzp) == -1)
    {
      printf("\nerror in time function\n");
      exit(1);
    }
#endif
#endif
	
  time_value = tp.tv_sec;
  return(time_value);
}

void initialize_ARC(void)
{
  char common_socket_name[MAX_COMMON_SOCKET_LEN + 1];
  Message msg_connect;
   
  if(initialised == TRUE)
      printf("\nARC already initialised\n");
  else {
      gethostname (common_socket_name, 50);
      strcat (common_socket_name,"KRISHNA");  
      parent_sock = get_u_client_sock_and_connect (common_socket_name);

      My_pid = getpid();
      msg_connect.type1 = htonl(USER_INIT);
      msg_connect.type2 = htonl(0);
      msg_connect.srcPid = htonl(My_pid);
      write_on_a_sock (parent_sock, (char *)&msg_connect, sizeof (msg_connect));
      initialised = TRUE;
    }
  
  return;
}

void initialize_RIB(void)
{
  char common_socket_name[MAX_COMMON_SOCKET_LEN + 1];
  Message msg_connect;
  
  gethostname (common_socket_name, 50);
  strcat (common_socket_name,"KRISHNA");
  parent_sock = get_u_client_sock_and_connect (common_socket_name);
  
  My_pid = getpid();
  msg_connect.type1 = htonl(USER_INIT);
  msg_connect.type2 = htonl(1);
  msg_connect.srcPid = htonl(My_pid);
  write_on_a_sock (parent_sock, (char *)&msg_connect, sizeof (msg_connect));

  return;
}

void close_ARC(void)
{
  Message msg_close;
  
  if(initialised == FALSE) printf("\nARC already closed\n");
  else {
      msg_close.type1 = htonl(USER_TERMINATING);
      msg_close.srcPid = htonl(My_pid);
      
      write_on_a_sock (parent_sock, (char *)&msg_close, sizeof (msg_close));
      close (parent_sock);
    }  
} 

void close_RIB(void) 
{
  Message msg_close;

  msg_close.type1 = htonl(USER_TERMINATING);
  msg_close.srcPid = htonl(My_pid);

  write_on_a_sock(parent_sock,(char *)&msg_close,sizeof(msg_close));
  close(parent_sock);
}

LFmessage get_load_factor(int number)
{
  Message lf_req, lf_rep;
  LFmessage lf;
  int i;

  if(initialised == FALSE) {
    printf("\nARC not initialised...Quitting\n");
    exit(1);
  }
  
  lf_req.type1 = htonl(GET_LOAD_FACTOR);
  lf_req.srcPid = htonl(My_pid);
  lf_req.numLFreq = htonl(number);
  write_on_a_sock (parent_sock, (char *)&lf_req, sizeof (lf_req));

  read_on_a_sock(parent_sock, (char *)&lf_rep, sizeof(lf_rep));
  if (ntohl(lf_rep.type1) != TAKE_LOAD_FACTOR) {
    printf("Possibly TERMINATE message from local coordinator\n");
    exit(1);
  } else {
    read_on_a_sock(parent_sock, (char *)&lf, sizeof(lf));
    lf.number = ntohl(lf.number);
    for(i = 0;i < lf.number;i++) {
      lf.LFinfo[i].index = ntohl(lf.LFinfo[i].index);
    }
    return(lf);
  }
}

void init_work_storage (void)
{
  int i;

  for (i=0; i<MAX_STORABLE_WORK; i++) 
    work_queue[i].dpptr = NULL;

}

void store_work (char *func_name,
		 struct datapack *dpptr,
		 int result_size,
		 int timeout,
		 int max_retries,
		 int sequence_no,
		 int machine_id)
{
  if (sequence_no > MAX_STORABLE_WORK)
    out_going_sequence_no = 0;
    
  sequence_no = out_going_sequence_no;
  work_queue[sequence_no ].sequence_no = sequence_no;
  work_queue[sequence_no ].dpptr = dpptr;
  work_queue[sequence_no ].result_size = result_size;
  work_queue[sequence_no ].timeout = timeout;

  CurrentTime = get_time();

  work_queue[sequence_no ].start_time = CurrentTime;
  work_queue[sequence_no ].max_retries = max_retries;
  work_queue[sequence_no ].machine_id = machine_id;
  strcpy(work_queue[sequence_no].func_name, func_name);
  work_queue[sequence_no ].status = WORK_IN_PROGRESS;
  work_queue[sequence_no ].retries_made = 0;
}

void resend_work_outside (int sequence_no)
{
  Message msg_take_work;

  msg_take_work.type1 = htonl(TAKE_WORK);
  msg_take_work.follMsgSize = htonl(strlen(work_queue[sequence_no].func_name) + 1);
  msg_take_work.follArgSize = htonl((work_queue[sequence_no].dpptr)->size);
  msg_take_work.resultSize = htonl(work_queue[sequence_no].result_size);
  msg_take_work.srcSeqNo = htonl(work_queue[sequence_no].sequence_no);
  msg_take_work.srcPid = htonl(My_pid);
  msg_take_work.index = htonl(work_queue[sequence_no].machine_id);

  write_on_a_sock (parent_sock, (char *)&msg_take_work, sizeof(msg_take_work));
  write_on_a_sock (parent_sock, work_queue[sequence_no].func_name,strlen(work_queue[sequence_no].func_name) + 1);
  write_on_a_sock (parent_sock, (work_queue[sequence_no].dpptr)->data,(work_queue[sequence_no].dpptr)->size);

}

int send_work_outside(char *func_name,
		      struct datapack *dpptr,
		      int result_size,		  
		      int timeout,
		      int max_retries,
		      int machine_id)
{
  Message msg_take_work;

  if(initialised == FALSE)
    {
      printf("\nARC not initialised...Quitting\n");
      exit(1);
    }

  msg_take_work.type1 = htonl(TAKE_WORK);
  msg_take_work.follMsgSize = htonl(strlen(func_name) + 1);
  msg_take_work.follArgSize = htonl(dpptr->size);
  msg_take_work.resultSize = htonl(result_size);
  msg_take_work.srcSeqNo = htonl(out_going_sequence_no);
  msg_take_work.srcPid = htonl(My_pid);
  msg_take_work.index = htonl(machine_id);

  store_work(func_name, dpptr, result_size, timeout, max_retries, out_going_sequence_no, machine_id);
  out_going_sequence_no++;

  write_on_a_sock (parent_sock, (char *)&msg_take_work, sizeof(msg_take_work));
  write_on_a_sock (parent_sock,func_name,strlen(func_name) + 1);
  write_on_a_sock (parent_sock,dpptr->data,dpptr->size);

  return(out_going_sequence_no-1);
}

int peek_result(int sequence_no)
{
  Message msg_check_result;
  Message msg_check_result_reply;
    
  if(initialised == FALSE)
    {
      printf("\nARC not initialised...Quitting\n");
      exit(1);
    }

  if(work_queue[sequence_no].status == WORK_FAILED || work_queue[sequence_no].status == WORK_DONE) 
    return(RESULTS_AVAILABLE);
  else {
    msg_check_result.type1 = htonl(CHECK_RESULTS);
    msg_check_result.srcPid = htonl(My_pid);
    msg_check_result.srcSeqNo = htonl(work_queue[sequence_no].sequence_no);
    write_on_a_sock (parent_sock, (char *)&msg_check_result, sizeof(msg_check_result));
    read_on_a_sock(parent_sock, (char *)&msg_check_result_reply,
		   sizeof(msg_check_result_reply));
    if((ntohl(msg_check_result_reply.type1)!=RESULTS_AVAILABLE) &&
       (ntohl(msg_check_result_reply.type1)!=RESULTS_NOT_AVAILABLE)) {
      printf("Possibly TERMINATE message from local coordinator\n");
      exit(1);
    }
    return(ntohl(msg_check_result_reply.type1));
  } 
}

char *obtain_result(int sequence_no) 
{                                         
  Message msg_want_result;
  Message msg_want_result_reply;
  int got_results,msgSize;
  char *result;

  if(initialised == FALSE) {
    printf("\nARC not initialised...Quitting\n");
    exit(1);
  }

  got_results = FALSE;
  do {
    msg_want_result.type1 = htonl(WANT_RESULTS);
    msg_want_result.srcPid = htonl(My_pid);
    msg_want_result.srcSeqNo = htonl(sequence_no);
    write_on_a_sock (parent_sock, (char *)&msg_want_result, sizeof(msg_want_result));
    read_on_a_sock(parent_sock, (char *)&msg_want_result_reply, sizeof(msg_want_result_reply));
    if(ntohl(msg_want_result_reply.type1) == RESULTS_AVAILABLE) {
      got_results = TRUE;
      msgSize = ntohl(msg_want_result_reply.follMsgSize);
      result = (char *)malloc(sizeof(char) * msgSize);  
      read_on_a_sock(parent_sock, result, msgSize);
      break;
    } else {
      if(ntohl(msg_want_result_reply.type1) == RESULTS_NOT_AVAILABLE) {
	CurrentTime = get_time();
	
	if((work_queue[sequence_no].retries_made < work_queue[sequence_no].max_retries) && \
	   ((CurrentTime - work_queue[sequence_no].start_time) < work_queue[sequence_no].timeout)) {
	  sleep(work_queue[sequence_no].timeout/WAIT_WITHIN_TIMEOUT);
	  continue;
	}
	if(work_queue[sequence_no].retries_made < work_queue[sequence_no].max_retries) {
	  work_queue[sequence_no].retries_made++;
	  work_queue[sequence_no].start_time = CurrentTime;
	  resend_work_outside(sequence_no);
	  sleep(work_queue[sequence_no].timeout/WAIT_WITHIN_TIMEOUT);
	  continue;
	} else {
	  got_results = FALSE;
	  break;
	}
      } else {
	printf("Possibly TERMINATE message from local coordinator\n");
	exit(1);
      }
    }
  } while(TRUE);
  
  if(got_results == TRUE) {
    work_queue[sequence_no].status = WORK_DONE;
    return(result);
  } else {
    work_queue[sequence_no].status = WORK_FAILED;
    return(NULL);
  }
}
    
char *obtain_result_abs(int sequence_no) 
/* sleeps for 60 sec. before re-asking for result */
{              
  Message msg_want_result;
  Message msg_want_result_reply;
  int got_results;
  char *result;
  int msgSize;

  if(initialised == FALSE)
    {
      printf("\nARC not initialised...Quitting\n");
      exit(1);
    }

  got_results = FALSE;
  do {
    msg_want_result.type1 = htonl(WANT_RESULTS);
    msg_want_result.srcPid = htonl(My_pid);
    msg_want_result.srcSeqNo = htonl(sequence_no);
    write_on_a_sock (parent_sock, (char *)&msg_want_result, sizeof(msg_want_result));
    read_on_a_sock(parent_sock, (char *)&msg_want_result_reply, sizeof(msg_want_result_reply));
    if(ntohl(msg_want_result_reply.type1) == RESULTS_AVAILABLE) {
      got_results = TRUE;
      msgSize = ntohl(msg_want_result_reply.follMsgSize);  
      result = (char *)malloc(sizeof(char) * msgSize);
      read_on_a_sock(parent_sock, result, msgSize);
      break;
    }
    else {
      if(ntohl(msg_want_result_reply.type1) == RESULTS_NOT_AVAILABLE) {
	CurrentTime = get_time();
	
	if((work_queue[sequence_no].retries_made < work_queue[sequence_no].max_retries) && \
	   ((CurrentTime - work_queue[sequence_no].start_time) < work_queue[sequence_no].timeout)) {
	  sleep(ABSOLUTE_WAIT_TIME);
	  continue;
	}
	if(work_queue[sequence_no].retries_made < work_queue[sequence_no].max_retries) {
	  work_queue[sequence_no].retries_made++;
	  work_queue[sequence_no].start_time = CurrentTime;
	  resend_work_outside(sequence_no);
	  sleep(ABSOLUTE_WAIT_TIME);
	  continue;
	} else {
	  got_results = FALSE;
	  break;
	}
      } else {
	printf("Possibly TERMINATE message from local coordinator\n");
	exit(1);
      }
    }
  } while(TRUE);
    
  if (got_results == TRUE) {
    work_queue[sequence_no].status = WORK_DONE;
    return(result);
  } else {
    work_queue[sequence_no].status = WORK_FAILED;
    return(NULL);
  }  
}




