/* This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */


#define PRIME 211

typedef struct rib_node_tag {
  char func_name[MAX_NAME_LEN + 1];
  int rib_number;
  struct rib_node_tag *next;
} rib_node;

rib_node *rib_name_table[PRIME];

void initialize_rib_name_table(void)
{
  int i;
  
  for(i=0;i<PRIME;i++)
      rib_name_table[i]=NULL;
}

void insert_name(char *name, int rib_number)
{
  int index;
  int i;
  rib_node *ptr;
          
  index=hashpjw(name);
  
  ptr=(rib_node *)malloc(sizeof(rib_node));
  strcpy(ptr->func_name,name);
  ptr->rib_number = rib_number;
  ptr->next=rib_name_table[index];
  rib_name_table[index]=ptr;
  
  return;
}

rib_node *search_name(char *name)        /* returns NULL if not found */
{
  int index;
  rib_node *ptr;
  
  index=hashpjw(name);
  ptr=rib_name_table[index];
  while((ptr!=NULL) && (strcmp(ptr->func_name,name)!=0))
    ptr=ptr->next;
  return(ptr);
}

int hashpjw(char *str)
{
  char *p;
  unsigned h=0,g;
  
  for(p=str;*p != 0;p++)
    {
      h=(h<<24)+(*p);
      if(g=h & 0xf0000000)
        {
	  h = h^(g<<24);
	  h = h^g;
	}
    }
  return (h%PRIME);
}











