/* This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in. */

#include "sysDefns.h"
#include "sockfunc.c"
#include "sockplus.c"
#include "initLocalDesc.c"

void ProcessMsgLC(int LCindex)
{ 
  int sockfd;
  int msgSize,argSize;
  Message msg,failMsg;
  char* buffer;
  char lf[10];
  XDR p;
  int i,k;
  LFnode temp;  
  LFmessage LFreply;
  LFnode loadArr[MAX_NUM_LCS];
  Message tempmsg;
  int numLFreq;
  int min,tmpmsgtype;


  sockfd = LocalDesc[LCindex].sockfd;

  read_on_a_sock(sockfd,(char *) &msg, sizeof msg);
  tmpmsgtype = ntohl(msg.type1);

  switch (tmpmsgtype) {

  case GET_LOAD_FACTOR: 
    if(DEBUG_ON) printf("Got the message ..GET_LOAD_FACTOR from %s\n",LocalDesc[LCindex].mcName);

    /* Find the least 'n' LFs from the LocalDesc array and
       send it to the requesting LC (in i) */

    /* Copy the load info of active LCs to separate arr*/
    for (i=0, k=0; (i< MAX_NUM_LCS) && (k<numActiveLCs) ; i++) 
      if (LocalDesc[i].valid) {
	xdrmem_create(&p,loadArr[k].loadFactor,10,XDR_ENCODE);
	xdr_float(&p,&LocalDesc[i].loadFactor);
	loadArr[k].index = i;
	k++;
      }

    /* Find least 'n' from loadarray*/
    numLFreq = ntohl(msg.numLFreq);
    if (numLFreq > numActiveLCs) numLFreq = numActiveLCs;
    for (i=0; i<numLFreq ; i++) {
      min = i;
      for (k=i+1; k < numActiveLCs ;k++)
	if (loadArr[k].loadFactor < loadArr[min].loadFactor) min = k;

      /* swap i & min*/
      temp = loadArr[i];
      loadArr[i] = loadArr[min];
      loadArr[min] =  temp; 
      LFreply.LFinfo[i].index = htonl(loadArr[i].index);
      for(k = 0;k < 10;k++) LFreply.LFinfo[i].loadFactor[k] = loadArr[i].loadFactor[k];
    }
    LFreply.number = htonl(numLFreq);

    /* Send the  Take_load_factor messg followed by LFmesg*/
    tempmsg.type1 = htonl(TAKE_LOAD_FACTOR);
    tempmsg.srcPid = msg.srcPid;
    tempmsg.destLocal = htonl(LCindex);
    tempmsg.follMsgSize = htonl(sizeof(LFreply)); 
    write_on_a_sock(sockfd, (char *)&tempmsg , sizeof(tempmsg));
    write_on_a_sock(sockfd,(char *)&LFreply,sizeof(LFreply));
    break;

  case TAKE_LOAD_FACTOR:
    if(DEBUG_ON) printf("Got the message ..TAKE_LOAD_FACTOR from %s\n",LocalDesc[LCindex].mcName);

    xdrmem_create(&p,msg.loadFactor,10,XDR_DECODE);
    xdr_float(&p,&LocalDesc[LCindex].loadFactor);
    break;

  case TAKE_WORK:
    if(DEBUG_ON) printf("Got the message ..TAKE_WORK from %s\n",LocalDesc[LCindex].mcName);

    /* Send the message TAKE_WORK to the destination LC */
    argSize = ntohl(msg.follArgSize);
    msgSize = ntohl(msg.follMsgSize);
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,(char *) &msg, sizeof (msg));

    /* Receive & Send the rib name next */
    buffer = (char *) malloc(msgSize);
    read_on_a_sock(sockfd,buffer,msgSize);
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,buffer,msgSize);
    free(buffer);

    /* Send the arguments next */
    buffer = (char *) malloc(argSize);
    read_on_a_sock(sockfd,buffer,argSize);
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,buffer,argSize);
    fflush(stdout);
    free(buffer);
    break;

  case GET_FILE: 
    /* Forward the message to Task Generator */
    write_on_a_sock(LocalDesc[ntohl(msg.srcLocal)].sockfd,(char *)&msg,sizeof(msg));

    /* Read the RIB name next and send it to the Task Generator */
    msgSize = ntohl(msg.follMsgSize);
    buffer = (char *)malloc(msgSize);
    read_on_a_sock(sockfd,buffer,msgSize);

    write_on_a_sock(LocalDesc[ntohl(msg.srcLocal)].sockfd,buffer,msgSize);
    free(buffer);
    break;

  case TAKE_FILE:
    if(DEBUG_ON) printf("Got the message ..TAKE_FILE from %s\n",LocalDesc[LCindex].mcName);

    /* Forward the message to the Task Executor */
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,(char *)&msg,sizeof msg);

    /* Send the filename to the Task Executor */
    msgSize = ntohl(msg.follMsgSize);
    argSize = ntohl(msg.follArgSize);
    buffer = (char *)malloc(msgSize);
    read_on_a_sock(sockfd,buffer,msgSize);
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,buffer,msgSize);
    free(buffer);

    /* Send the actual file next */
    buffer = (char *)malloc(argSize);
    read_on_a_sock(sockfd,buffer,argSize);
    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,buffer,argSize);

    free(buffer);
    break;
    
  case TAKE_RESULTS:
    if(DEBUG_ON) printf("Got the message ..TAKE_RESULTS(seqno = %d) from %s\n",ntohl(msg.srcSeqNo),LocalDesc[LCindex].mcName);

    write_on_a_sock(LocalDesc[ntohl(msg.srcLocal)].sockfd,(char *) &msg, sizeof(msg));
    /* Send the actual results next*/
    msgSize = ntohl(msg.follMsgSize);
    buffer = (char *) malloc(msgSize);
    read_on_a_sock(sockfd,buffer,msgSize);
    write_on_a_sock(LocalDesc[ntohl(msg.srcLocal)].sockfd,buffer,msgSize);
    free(buffer);
    break;

  case GENERATOR_FAILED:
    if(DEBUG_ON) printf("Got the message ..GENERATOR_FAILED from %s\n",LocalDesc[LCindex].mcName);

    write_on_a_sock(LocalDesc[ntohl(msg.destLocal)].sockfd,(char *) &msg, sizeof(msg));
    break;

  case LC_TERMINATING:
    if(DEBUG_ON) printf("Got the message ..LC_TERMINATING from %s\n",LocalDesc[LCindex].mcName);

    /* Mark that LC as invalid and broadcast */
    LocalDesc[LCindex].valid = 0;
    numActiveLCs--;

    failMsg.type1 = htonl(LC_TERMINATING);
    failMsg.srcLocal = htonl(LCindex);

    for (i=0, k=0; (i< MAX_NUM_LCS) && (k<numActiveLCs) ; i++) 
      if (LocalDesc[i].valid) {
	failMsg.destLocal = htonl(i);
	write_on_a_sock(LocalDesc[i].sockfd,(char *)&failMsg, sizeof(failMsg));
	k++;
      }

    break;

  default:
    if(DEBUG_ON) printf("Got UNKNOWN MESSAGE %d FROM %s\n",tmpmsgtype,LocalDesc[LCindex].mcName);

    /* Mark that LC as invalid and broadcast */
    LocalDesc[LCindex].valid = 0;
    numActiveLCs--;
    failMsg.type1 = htonl(LC_TERMINATING);
    failMsg.srcLocal = htonl(LCindex);

    for (i=0, k=0; (i< MAX_NUM_LCS) && (k<numActiveLCs) ; i++) 
      if (LocalDesc[i].valid) {
	if(DEBUG_ON) printf("Sending message LC_TERMINATING to %s\n",LocalDesc[i].mcName);
	failMsg.destLocal = htonl(i);
	write_on_a_sock(LocalDesc[i].sockfd,(char *)&failMsg, sizeof(failMsg));
	k++;
      }

    break;
  }
}

void main()
{
  int connxnSock;
  int tmpSock;
  int ready;			/* if a socket is ready for reading */
  int STATE ;			/* State of the System Coordinator */
  Message msg; /* may be unnecessary */
  struct sockaddr_in connxnSockAddr;
  int i,k;
  int ALL_OVER=0;
  int newLocal,newLocalhs;

  STATE = SYS_INIT;

  do {
    switch (STATE) {
		
    case SYS_INIT: 
      initLocalDescArray();
    
      /* Mark all locals as inactive */
      for (i=0; i<MAX_NUM_LCS; i++)
	LocalDesc[i].valid=FALSE;
		
      numActiveLCs = 0;
      connxnSock = get_server_sock_and_listen ();
      connxnSockAddr = get_sock_addr (connxnSock); 
      STATE = SYS_CHK_MSG;
      break;
		
    case SYS_CHK_MSG:
      /* check in the common channel */
      ready = read_ready_test (connxnSock,0,100);
      if (ready > 0) {
	tmpSock = accept_on_a_sock(connxnSock);
	
	/* Assume that the LC will send its index first the new connxn */
	read_on_a_sock(tmpSock,(char *)&newLocalhs,sizeof(int));	
	newLocal = ntohl(newLocalhs);
	printf("Connecting %s \n",LocalDesc[newLocal].mcName);
	LocalDesc[newLocal].sockfd = tmpSock;
	LocalDesc[newLocal].loadFactor = INFINITE_LOAD;
	LocalDesc[newLocal].valid = TRUE;
	numActiveLCs++;
      }
		
      /* check in the active exclusive channels */
      for (i=0, k=0; (i< MAX_NUM_LCS) && (k<numActiveLCs) ; i++) 
	if (LocalDesc[i].valid) {
	  ready = read_ready_test (LocalDesc[i].sockfd,0,100);
	  k++;
	  if (ready) ProcessMsgLC(i);
	}

      STATE = SYS_CHK_MSG;
      break;

    default:
      printf("SC in unknown STATE. Exiting ...\n");
      exit(1);
    }
  
  }while (!ALL_OVER);
}































