/***************************************************************************/
/*      FILENAME        : sockfunc.c					   */
/*                                                                         */
/*      AUTHOR		: R K JOSHI				2.2.94	   */
/***************************************************************************/
/* This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details.

You should have received a copy of the GNU General Public License 
along with this program; if not, write to Prof. D. Janakiram, 
DOS Lab, Department of CS & E, Indian Institute of Technology Madras, 
Chennai - 600036, India. Alternatively, you can send a mail to 
d.janakiram@cs.iitm.ernet.in.

****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <netdb.h>
#include <rpc/rpc.h>
#ifdef SPARC 
#include <string.h>
#endif

struct hostent *gethostbyname(const char *);

void exit(int);
/* void *memset(void *, int, int); */

/********************sever side:idomain,stream *****************************/

int get_server_sock_and_listen (void);
int accept_on_a_sock(int);
int module_test(void);

/********************client side:idomain,stream ****************************/

int get_client_sock_and_connect(char *, int); /*takes name and port no. */

int make_and_connect_client(struct sockaddr_in);/* takes sockaddr struct */

/********************general:idomain,stream ********************************/

int read_on_a_sock(int, char*, int);
char *read_arbit_data_with_header (int);
int write_on_a_sock(int,char *,int);
int write_arbit_data_on_a_sock(int,char *,int);
int getportno (int);
struct hostent *get_my_hostent(void);
struct sockaddr_in get_sock_addr (int);


/*********************udomain: stream **************************************/

int create_sock_pair (char *, char *, int[2]);
int get_u_server_sock_and_listen (char *);
int get_u_client_sock_and_connect(char *); 
		       	    /* other funcs are same, e.g. accept, read..*/

/***************************************************************************/
/*      Function        : get_server_sock_and_listen().			   */
/*      Purpose         : creates a connection socket and listens to it.   */
/*      Parameters      : none.						   */
/*      Returns	(succ)	: socket id.					   */
/*	Returns (fail)	: performs exit (1).				   */
/*      Modifies        : none.		                                   */
/*      Calls           : socket(), listen(), bind(), printf(),            */
/*			  getsockname(),sizeof().			   */
/*      Called by       :						   */
/*      Bugs / Warnings	: none.					       	   */
/***************************************************************************/

int get_server_sock_and_listen (void)
{
  int sock;
  struct sockaddr_in s;
  int length;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock == 0) { 
    perror ("sock opening in server"); 
    exit (1);
  }

  s.sin_family = AF_INET;
  s.sin_addr.s_addr = INADDR_ANY;
  s.sin_port = htons(SYSPORTNO);

  if (bind (sock, (struct sockaddr *) &s, sizeof s) < 0) {
    perror ("in binding"); 
    exit (1);
  }
  length = sizeof (s);
  if (getsockname (sock,(struct sockaddr *) &s,&length) <0) {
    perror ("in getting sock name"); 
    exit(1);
  }

#ifdef SPARC
  listen (sock, 5);
#else
  listen (sock, 5);
#endif
  return (sock);

}


/********************************************************************************/
/*      Function        : accept_on_a_sock().					*/
/*      Purpose         : performs an accept() call on a socket.		*/
/*      Parameters      : the socket-id on which to perform accept().		*/
/*      Returns (succ)	: the new socket-id which is connected.			*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.		                                        */
/*      Calls           : accept().						*/
/*      Called by       :							*/
/*      Bugs / Warnings : none.							*/
/********************************************************************************/

int accept_on_a_sock(int sock)

{
  int sockforuse;

  sockforuse = accept (sock, (struct sockaddr *) 0, (int *)0);
  if ( sockforuse == -1) { perror ("in accept"); exit (1); }
  return (sockforuse);

}

/********************************************************************************/
/*      Function        : read_on_a_sock().					*/
/*      Purpose         : reads data on a socket.				*/
/*      Parameters      : socket-id, malloced buffer pointer, total data bytes.	*/
/*      Returns (succ)	: number of bytes read.					*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : it first zeros the entire buffer before read.		*/
/*      Calls           : read(), bzero().					*/
/*      Called by       :							*/
/*      Bugs / Warnings : none.							*/
/********************************************************************************/


int read_on_a_sock(int sock, char *buf, int size)
{
  int rval;
  char *remaining_buffer;
  int remaining_size;

#ifdef SPARC
  buf = (char *) memset ((void *) buf, 0, (size_t) size);
#else
  bzero (buf,size);
#endif

  remaining_buffer = buf;
  remaining_size = size;  /* for size > 1024 */

  do {

    rval = read (sock,remaining_buffer,remaining_size);

    if (rval <= 0)  return  (rval);   

    /* 0 bytes will be read if there is no connection - e.g. due to a closed socket */

    remaining_size = remaining_size - rval ;
    remaining_buffer = remaining_buffer + rval;
    
  } while (remaining_size > 0 ); 
  
  return (size); 
}

/********************************************************************************/
/*      Function        : read_arbit_data_with_header().			*/
/*      Purpose         : reads data of unknown length on a socket.		*/
/*      Parameters      : only-the-socket-id.					*/
/*	Assumes		: data on socket first carries data-size-in-bytes 	*/
/*			  and then actual-data					*/
/*	Returns (succ)	: pointer to the data buffer which carries malloc.	*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : mallocs the data buffer forrequired data size.	*/
/*      Calls           : read(), bzero(), malloc().				*/
/*      Called by       :							*/
/*      Bugs / Warnings : 							*/
/********************************************************************************/

char *read_arbit_data_with_header (int sock)
{
  int rval;
  int header;
  int size;
  char *buf;

#ifdef SPARC
  memset (&header, 0, (size_t) sizeof header);
#else
  bzero (&header, sizeof header);
#endif

  if ( (rval = read (sock,(char *)&header,sizeof header)) <=0 ) { 
    perror ("reading error"); 
    exit (1);
  }

  /* 0 bytes will be read if there is no connection - e.g. due to a closed socket */

  size = header;
  buf = (char *) malloc (size);
  if (buf == NULL) {
    perror ("reading error"); 
    exit (1);
  }

#ifdef SPARC
  memset ( buf, 0, (size_t) size);
#else
  bzero (buf,size);
#endif

  if((rval = read (sock,buf,size)) <= 0 ) {
    perror ("reading error"); 
    exit (1);
  }

  return (buf); 
}

/********************************************************************************/
/*      Function        : get_client_sock_and_connect ().			*/
/*      Purpose         : creates a connection client, connects to given server.*/
/*      Parameters      : server's-machine-name, server's-port-number.		*/
/*	Returns (succ)	: client-sock-id					*/
/*	Returns (fail)	: -1							*/
/*      Modifies        : none.							*/
/*      Calls           : socket() gethostbyname() fprintf() bcopy() connect()	*/
/*      Called by       :							*/
/*      Bugs / Warnings :							*/
/********************************************************************************/

int get_client_sock_and_connect(char *name, int port) {
  int sock;
  struct sockaddr_in s;
  struct hostent *hp;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock == 0) { 
    perror ("sock opening in server"); 
    exit (1); 
  }
  hp = gethostbyname (name);
  
  if (hp == 0) { 
    fprintf (stderr, "%s:unknown host", name); 
    exit(2); 
  }
 
#ifdef SPARC
  memcpy ((void *) &s.sin_addr, (void *)hp->h_addr,(size_t) hp->h_length); 
#else
  bcopy ((char *)hp->h_addr, (char *) &s.sin_addr, hp->h_length);
#endif

  s.sin_family = AF_INET;
  s.sin_port = htons(port);

  if ( connect (sock, (struct sockaddr *) &s, sizeof s) < 0) {
    perror ("in connect"); 
    exit (1);
  }

  return (sock);

}

/********************************************************************************/
/*      Function        : make_and_connect_client ().			        */
/*      Purpose         : creates a connection client, connects to given server.*/
/*      Parameters      : server's-sockaddr-structure.		                */
/*	Returns (succ)	: client-sock-id					*/
/*	Returns (fail)	: -1							*/
/*      Modifies        : none.							*/
/*      Calls           : socket() gethostbyname() fprintf() bcopy() connect()	*/
/*      Called by       :							*/
/*      Bugs / Warnings :							*/
/********************************************************************************/

int make_and_connect_client(struct sockaddr_in sock_name)
{
  int sock;
  struct sockaddr_in s;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock == 0) { 
    perror ("sock opening in server"); 
    exit (1);
  }
 
  if ( connect (sock, (struct sockaddr *) &sock_name, sizeof sock_name) < 0) {
    perror ("in connect");
    exit (1); 
  }

  return (sock);

}

/********************************************************************************/
/*      Function        : write_on_a_sock().					*/
/*      Purpose         : writes data on a socket.				*/
/*      Parameters      : socket-id, data buffer pointer, total data bytes.	*/
/*      Returns (succ)	: number of bytes written.				*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.							*/
/*      Calls           : none.					*/
/*      Called by       :							*/
/*      Bugs / Warnings : none.							*/
/********************************************************************************/

int write_on_a_sock(int sock, char *buf, int size)
{
  int wval;
  char *remaining_buffer;
  int remaining_size;

  remaining_buffer = buf;
  remaining_size = size;  /* for size > 1024 */

  do {
    wval = write (sock,remaining_buffer,remaining_size);
    if (wval <= 0)  return  (wval);   

    /* 0 bytes will be written if there is no connection - e.g. due to a closed socket */

    remaining_size = remaining_size - wval ;
    remaining_buffer = remaining_buffer + wval;
  } while (remaining_size > 0 ); 
  
  return (size); 
}

/********************************************************************************/
/*      Function        : write_arbit_data_with_header().			*/
/*      Purpose         : writes data of unknown length on a socket.		*/
/*      Parameters      : socket-id, buffer-pointer, size-of-data		*/
/*	Assumes		: sends on socket: first-the-data-size-in-bytes(header)	*/
/*			  and-then-actual-data					*/
/*	Returns (succ)	: nunber-of-bytes-written-including-the-header.		*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.							*/
/*      Calls           : none.							*/
/*      Called by       :							*/
/*      Bugs / Warnings : 							*/
/********************************************************************************/


int write_arbit_data_on_a_sock(int sock,char *buf,int size)
{
  int wval1, wval2;

  if ( (wval1 = write (sock ,(char *)&size, sizeof size) )<= 0) 
    perror ("write");

  if ( (wval2 = write (sock ,buf,size) )<= 0) 
	perror ("write");
  return (wval1 + wval2);

}

/********************************************************************************/
/*      Function        : getportno().						*/
/*	Purpose		: gives port number associated o a socket.		*/
/*      Parameters      : socket-id.						*/
/*	Returns (succ)	: port-number						*/
/*	Returns (fail)	: exit () values of functons in case of errors.		*/
/*      Modifies        : none.		                                        */
/*      Calls           : 							*/
/*      Called by       :							*/
/*      Bugs / Warnings :  							*/
/********************************************************************************/

int getportno (int sock)
{
  int n;
  struct sockaddr_in s;

  n = sizeof s;
  
  if (getsockname (sock,(struct sockaddr *) &s,&n) <0) { 
    perror ("in getting sock name"); 
    exit(1);
  }
  
  return (ntohs (s.sin_port));

}

/********************************************************************************/
/*      Function        : create_sock_pair().					*/
/*      Purpose         : udomain connected and named sockets.			*/
/*      Parameters      : caller_name: for host socket				*/
/*			  remote_name: given for remote sock			*/
/*      Returns	(succ)	: 0							*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.		                                        */
/*      Calls           : 							*/
/*			  							*/
/*      Called by       :							*/
/*      Bugs / Warnings	: 							*/
/********************************************************************************/

int create_sock_pair(char *caller_name, char *remote_name, int sockvec[2])
{
  int sval;
  struct sockaddr_un caller,remote ;
  
  sval = socketpair (AF_UNIX, SOCK_STREAM,0,sockvec);
  if (sval < 0) { perror ("in sockpair"); exit(0);  }
  caller.sun_family = AF_UNIX;
  strcpy (caller.sun_path, caller_name);

  if (bind (sockvec[0], (struct sockaddr *) &caller,sizeof(struct sockaddr_un)) < 0) { 
    perror ("in bind"); 
    exit(0); 
  }

  remote.sun_family = AF_UNIX;
  strcpy (remote.sun_path, remote_name);
  
  if (bind (sockvec[1], (struct sockaddr *) &remote,sizeof(struct sockaddr_un)) <  0) {
    perror ("in bind"); 
    exit(0); 
  }     

  return (0);
}

/********************************************************************************/
/*      Function        : get_u_server_sock_and_listen().			*/
/*      Purpose         : udomain server socket.				*/
/*      Parameters      : name: for  socket					*/
/*      Returns	(succ)	: 0							*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.		                                        */
/*      Calls           : 							*/
/*			  							*/
/*      Called by       :							*/
/*      Bugs / Warnings	: 							*/
/********************************************************************************/


int get_u_server_sock_and_listen (char *name)
{
  struct sockaddr_un server ;
  int mysock;

  mysock = socket (AF_UNIX, SOCK_STREAM,0);
  if (mysock < 0) { perror ("in server u sock"); exit(0);  }
  
  server.sun_family = AF_UNIX;
  strcpy (server.sun_path, name);
  
  if (bind (mysock, (struct sockaddr *) &server,sizeof(struct sockaddr_un)) < 0) {
    perror ("in bind, server u sock"); 
    exit(0);
  }

  listen (mysock,3);
  return (mysock);
}

/********************************************************************************/
/*      Function        : get_u_client_sock_and_connect().			*/
/*      Purpose         : udomain client sockets.				*/
/*      Parameters      : name: for  socket					*/
/*      Returns	(succ)	: 0							*/
/*	Returns (fail)	: performs exit (1).					*/
/*      Modifies        : none.		                                        */
/*      Calls           : 							*/
/*			  							*/
/*      Called by       :							*/
/*      Bugs / Warnings	: 							*/
/********************************************************************************/

int get_u_client_sock_and_connect(char *server_name)
{
  struct sockaddr_un server ;
  int sock;

  sock = socket (AF_UNIX, SOCK_STREAM,0);
  if (sock < 0) {
    perror ("in client u sock");
    exit(0); 
  }
	
  server.sun_family = AF_UNIX;
  strcpy (server.sun_path, server_name);

  if (connect (sock, (struct sockaddr *) &server,sizeof  (struct sockaddr_un)) < 0) { 
    close (sock);
    perror ("in connect client u sock"); 
    exit(0);
  }

  return (sock);
} 

/********************************************************************************/
struct hostent *get_my_hostent(void)  /* gives callers address etc. */
{
  char hostname[50];
  struct hostent *he;

  gethostname (hostname, 50);
  he = gethostbyname (hostname);
  return (he);
}

/********************************************************************************/

struct sockaddr_in get_sock_addr (int sock)
{
  char hostname[50];
  struct hostent *he;
  struct sockaddr_in socket_name;
  int sock_name_len;
  int flag;
  
  sock_name_len  = sizeof socket_name;
  
  
  if (getsockname (sock, (struct sockaddr *) &socket_name,&sock_name_len) < 0) {
    perror ("in getting name\n"); 
    exit (1); 
  }

  flag = gethostname (hostname, 50);
  if (flag !=0) {
    printf("SC:gethostname failed in get_sock_addr");
    exit(1);
  }

  he = gethostbyname (hostname);
  if (he==NULL) {
    printf("SC:gethostbyname failed in get_sock_addr");
    exit(1);
  }

#ifdef SPARC 
  memcpy ((void *) &socket_name.sin_addr, (void *)he->h_addr,(size_t) he->h_length);
#else
  bcopy ((char *)he->h_addr,(char *) &socket_name.sin_addr,he->h_length);
#endif

  return (socket_name);
}
